Package: minfi
Version: 1.59.0
Title: Analyze Illumina Infinium DNA methylation arrays
Description: Tools to analyze & visualize Illumina Infinium methylation arrays.
Authors@R: c(person(c("Kasper", "Daniel"), "Hansen", role = c("cre", "aut"),
                    email = "kasperdanielhansen@gmail.com"),
             person("Martin", "Aryee", role = "aut"),
             person(c("Rafael", "A."), "Irizarry", role = "aut"),
             person(c("Andrew", "E."), "Jaffe", role = "ctb"),
             person("Jovana", "Maksimovic", role = "ctb"),
             person(c("E.", "Andres"), "Houseman", role = "ctb"),
             person("Jean-Philippe", "Fortin", role = "ctb"),
             person("Tim", "Triche", role = "ctb"),
             person(c("Shan", "V."), "Andrews", role = "ctb"),
             person(c("Peter", "F."), "Hickey", role = "ctb"))
Depends: methods, BiocGenerics (>= 0.15.3), GenomicRanges (>= 1.61.1),
        SummarizedExperiment (>= 1.39.1), Biostrings (>= 2.77.2),
        bumphunter (>= 1.1.9)
Suggests: IlluminaHumanMethylation450kmanifest (>= 0.2.0),
        IlluminaHumanMethylation450kanno.ilmn12.hg19 (>= 0.2.1),
        minfiData (>= 0.18.0), minfiDataEPIC, FlowSorted.Blood.450k (>=
        1.0.1), RUnit, digest, BiocStyle, knitr, rmarkdown, tools
Imports: S4Vectors, Seqinfo, Biobase (>= 2.33.2), IRanges, beanplot,
        RColorBrewer, lattice, nor1mix, siggenes, limma,
        preprocessCore, illuminaio (>= 0.23.2), DelayedMatrixStats (>=
        1.3.4), mclust, genefilter, nlme, reshape, MASS, quadprog,
        data.table, GEOquery, stats, grDevices, graphics, utils,
        DelayedArray (>= 0.15.16), HDF5Array, BiocParallel
Collate: AllGenerics.R MethylSet-class.R RatioSet-class.R
        RGChannelSet-class.R RGChannelSetExtended-class.R
        GenomicMethylSet-class.R GenomicRatioSet-class.R eSet_methods.R
        utils.R IlluminaMethylationManifest-class.R
        IlluminaMethylationAnnotation-class.R minfiQC.R getSex.R
        dmpFinder.R blocks.R plot.R plotBetasByType.R preprocessRaw.R
        preprocessIllumina.R detectionP.R preprocessSwan.R
        preprocessQuantile.R preprocessNoob.R preprocessFunnorm.R qc.R
        read.450k.R read.meth.R read.meth2.R read.geo.R read.manifest.R
        bumphunter.R estimateCellCounts.R gaphunter.R compartments.R
        combineArrays.R DelayedArray_utils.R
VignetteBuilder: knitr
License: Artistic-2.0
URL: https://github.com/hansenlab/minfi
BugReports: https://github.com/hansenlab/minfi/issues
biocViews: ImmunoOncology, DNAMethylation, DifferentialMethylation,
        Epigenetics, Microarray, MethylationArray, MultiChannel,
        TwoChannel, DataImport, Normalization, Preprocessing,
        QualityControl
git_url: https://git.bioconductor.org/packages/minfi
git_branch: devel
git_last_commit: c1d5c0b
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.24
Date/Publication: 2026-04-29
NeedsCompilation: no
Packaged: 2026-04-29 22:56:24 UTC; biocbuild
Author: Kasper Daniel Hansen [cre, aut],
  Martin Aryee [aut],
  Rafael A. Irizarry [aut],
  Andrew E. Jaffe [ctb],
  Jovana Maksimovic [ctb],
  E. Andres Houseman [ctb],
  Jean-Philippe Fortin [ctb],
  Tim Triche [ctb],
  Shan V. Andrews [ctb],
  Peter F. Hickey [ctb]
Maintainer: Kasper Daniel Hansen <kasperdanielhansen@gmail.com>

import(methods)
import(DelayedMatrixStats)
import(BiocGenerics)
importMethodsFrom(Biobase, exprs, fData,
                  featureNames, "featureNames<-",
                  pData, "pData<-",
                  sampleNames, "sampleNames<-")
import(Biostrings, except = c(type))
importMethodsFrom(Seqinfo, seqlevels, "seqlevels<-", genome, "genome<-", seqlengths, "seqlengths<-")
importClassesFrom(Seqinfo, Seqinfo)
import(GenomicRanges)
import(SummarizedExperiment)
import(S4Vectors)
import(IRanges)
import(DelayedArray)
import(BiocParallel)
import(HDF5Array) # TODO: Remove once DelayedArray_utils.R are moved to DelayedArray
importFrom(Biobase, assayDataElementNames, assayDataElementReplace,
           assayDataElement, "assayDataElement<-", validMsg, assayDataValidMembers)

## The following should not be necessary, but Hadley does not use S3export in reshape
importFrom(reshape, melt, melt.array, melt.data.frame, melt.list, melt.matrix, melt.table)
importFrom(mclust, Mclust)
importFrom(RColorBrewer, brewer.pal)
importFrom(beanplot, beanplot)
importFrom(nor1mix, norMix, pnorMix, qnorMix)
importFrom(siggenes, pi0.est, qvalue.cal)
importFrom(limma, squeezeVar, lmFit, eBayes, contrasts.fit, normexp.signal)
importFrom(preprocessCore, normalize.quantiles, normalize.quantiles.use.target)
importFrom(illuminaio, readIDAT)
importFrom(genefilter, rowFtests, rowttests)
importFrom(utils, packageVersion, read.csv, read.table, data)
importFrom(bumphunter, bumphunter, bumphunterEngine, locfitByCluster,
           clusterMaker, boundedClusterMaker)
importMethodsFrom(bumphunter, bumphunter)
importFrom(lattice, panel.abline, panel.stripplot, xyplot)
importFrom(nlme, lme, getVarCov)
importFrom(MASS, huber)
importFrom(quadprog, solve.QP)
importFrom(GEOquery,getGEO)
importFrom(data.table, fread)
importFrom("grDevices", "dev.off")
importFrom("graphics", "abline", "axis", "legend", "lines", "par",
           "plot", "plot.new", "stripchart", "text")
importFrom("stats", "approx", "as.formula", "cmdscale", "contrasts",
           "dist", "fitted", "kmeans", "lm", "lm.fit", "model.matrix",
           "pchisq", "pf", "prcomp", "pt", "rnorm", "runif",
           "vcov")

export(getManifest, getProbeInfo, getManifestInfo,
       IlluminaMethylationManifest, IlluminaMethylationAnnotation,
       getControlAddress,
       getRed, getGreen, getNBeads, getMeth, getUnmeth,
       getBeta, getM, getCN, getMethSignal,
       getOOB, getSnpBeta,
       dropMethylationLoci, getLocations, getAnnotation,
       getAnnotationObject,
       mapToGenome, subsetByLoci,
       getIslandStatus, getProbeType, getSnpInfo, addSnpInfo, dropLociWithSnps,
       normalize.illumina.control, bgcorrect.illumina,
       preprocessRaw, preprocessIllumina,
       preprocessSWAN, preprocessQuantile,
       preprocessFunnorm, preprocessNoob,
       RGChannelSet, RGChannelSetExtended,
       MethylSet, RatioSet,
       GenomicMethylSet, GenomicRatioSet,
       read.450k, read.450k.sheet, read.450k.exp,
       read.metharray, read.metharray.sheet, read.metharray.exp,
       getGenomicRatioSetFromGEO, makeGenomicRatioSetFromMatrix,
       readGEORawFile, readTCGA,
       logit2, ilogit2,
       dmpFinder, mdsPlot, plotCpg, plotBetasByType,
       detectionP, qcReport, controlStripPlot, densityBeanPlot, densityPlot,
       getQC, plotQC, addQC, minfiQC, fixMethOutliers,
       getSex, plotSex, addSex, bumphunter,
       blockFinder, cpgCollapse,
       estimateCellCounts, gaphunter,
       compartments, createCorMatrix, extractAB,
       convertArray, combineArrays)

exportClasses(RGChannelSet, RGChannelSetExtended,
              MethylSet, RatioSet,
              GenomicMethylSet, GenomicRatioSet,
              IlluminaMethylationManifest, IlluminaMethylationAnnotation)

exportMethods(show, initialize, getBeta, getM, getCN, getMeth, getUnmeth,
              getManifest, annotation, "annotation<-", preprocessMethod,
              combine, mapToGenome, ratioConvert, bumphunter,
              pData, "pData<-", sampleNames, "sampleNames<-",
              featureNames, "featureNames<-", coerce)

# R package: minfi
This is the developer version of Bioconductor package [minfi](http://bioconductor.org/packages/devel/bioc/html/minfi.html).  Install in R as:

```r
if (!requireNamespace("BiocManager", quietly=TRUE))
    install.packages("BiocManager")
BiocManager::install('minfi')
```

## Software status

| Resource:     | Bioconductor        | Travis CI     |
| ------------- | ------------------- | ------------- |
| _Platforms:_  | _Multiple_          | _Linux_       |
| R CMD check   | <a href="http://bioconductor.org/checkResults/release/bioc-LATEST/minfi/"><img border="0" src="http://bioconductor.org/shields/build/release/bioc/minfi.svg" alt="Build status"></a> (release)</br><a href="http://bioconductor.org/checkResults/devel/bioc-LATEST/minfi/"><img border="0" src="http://bioconductor.org/shields/build/devel/bioc/minfi.svg" alt="Build status"></a> (devel) | <a href="https://travis-ci.org/kasperdanielhansen/minfi"><img src="https://travis-ci.org/kasperdanielhansen/minfi.svg" alt="Build status"></a> |
| Test coverage |                     | <a href="https://codecov.io/github/kasperdanielhansen/minfi?branch=master"><img src="https://codecov.io/github/kasperdanielhansen/minfi/coverage.svg?branch=master" alt="Coverage Status"/></a>   |                  |

# Exported functions -----------------------------------------------------------

read.450k <- function(basenames, extended = FALSE, verbose = FALSE) {
    .Defunct("read.metharray")
}

read.450k.sheet <- function(base, pattern = "csv$", ignore.case = TRUE,
                            recursive = TRUE, verbose = TRUE) {
    .Defunct("read.metharray.sheet")
}


read.450k.exp <- function(base = NULL, targets = NULL, extended = FALSE,
                          recursive = FALSE, verbose = FALSE) {
    .Defunct("read.metharray.exp")
}

# Exported classes -------------------------------------------------------------

setClass("RGChannelSetExtended", contains = "RGChannelSet")

# Validity methods -------------------------------------------------------------

setValidity("RGChannelSetExtended", function(object) {
    msg <- validMsg(
        NULL,
        .checkAssayNames(
            object, c("Red", "Green", "RedSD", "GreenSD", "NBeads")))
    if (is.null(msg)) TRUE else msg
})

# Exported functions -----------------------------------------------------------

RGChannelSetExtended <- function(Green = new("matrix"), Red = new("matrix"),
                                 GreenSD = new("matrix"), RedSD = new("matrix"),
                                 NBeads = new("matrix"), annotation = "",
                                 ...) {
    # Check rownames, colnames
    assays <- SimpleList(
        Green = Green,
        Red = Red,
        GreenSD = GreenSD,
        RedSD = RedSD,
        NBeads = NBeads)

    new("RGChannelSetExtended",
        SummarizedExperiment(assays = assays, ...),
        annotation = annotation
    )
}

getNBeads <- function(object) {
    if (!is(object, "RGChannelSetExtended")) {
        msg <- "object is of class '%s', but needs to be of class 'RGChannelSetExtended'"
        stop(sprintf(msg, class(object)))
    }
    assay(object, "NBeads")
}

# Exported methods -------------------------------------------------------------

setMethod(
    "updateObject",
    signature(object = "RGChannelSetExtended"),
    function(object, ..., verbose = FALSE) {
        if (verbose) message("updateObject(object = 'RGChannelSetExtended')")
        if ("assayData" %in% names(getObjectSlots(object))) {
            # This is an ExpressionSet based object
            object <- RGChannelSetExtended(
                Green = getObjectSlots(object)[["assayData"]][["Green"]],
                Red = getObjectSlots(object)[["assayData"]][["Red"]],
                GreenSD = getObjectSlots(object)[["assayData"]][["GreenSD"]],
                RedSD = getObjectSlots(object)[["assayData"]][["RedSD"]],
                NBeads = getObjectSlots(object)[["assayData"]][["NBeads"]],
                colData = getObjectSlots(
                    getObjectSlots(object)[["phenoData"]])[["data"]],
                annotation = getObjectSlots(object)[["annotation"]])
        }
        object
    }
)

setMethod(
    "coerce",
    signature(from = "RGChannelSetExtended", to = "RGChannelSet"),
    function(from, to){
        if (nrow(from) > 0 || ncol(from) > 0) {
            assays(from) <- assays(from)[c("Green", "Red")]
        }
        class(from) <- "RGChannelSet"
        from
    }
)

# Exported functions -----------------------------------------------------------

qcReport <- function(rgSet, sampNames = NULL, sampGroups = NULL,
                     pdf = "qcReport.pdf", maxSamplesPerPage = 24,
                     controls = c("BISULFITE CONVERSION I",
                                  "BISULFITE CONVERSION II", "EXTENSION",
                                  "HYBRIDIZATION", "NON-POLYMORPHIC",
                                  "SPECIFICITY I", "SPECIFICITY II",
                                  "TARGET REMOVAL")) {
    .isRGOrStop(rgSet)
    if (is.null(sampNames)) sampNames <- colnames(rgSet)
    n <- ncol(rgSet)
    o <- rev(order(sampNames))
    rgSet <- rgSet[, o]
    sampNames <- sampNames[o]
    if (is.null(sampGroups)) sampGroups <- rep(1, n)
    sampGroups <- sampGroups[o]
    sampGroups <- as.factor(sampGroups)
    numPages <- ceiling(n/maxSamplesPerPage)
    samplesPerPage <- ceiling(n/numPages)
    sampleIdxs <- suppressWarnings(
        split(seq_len(n), rep(seq_len(numPages), each = samplesPerPage)))
    pdf(file = pdf, width = 8, height = 11)
    par(mfrow = c(2, 1))
    densityPlot(rgSet, sampGroups = sampGroups, main = "Beta", xlab = "Beta")
    plot.new()
    par(mfrow = c(1, 1), oma = c(2, 10, 1, 1))
    for (sampleIdx in sampleIdxs) {
        densityBeanPlot(
            dat = rgSet[, sampleIdx],
            sampGroups = sampGroups[sampleIdx],
            sampNames = sampNames[sampleIdx])
    }
    for (controlType in controls) {
        for (sampleIdx in sampleIdxs) {
            controlStripPlot(
                rgSet = rgSet[, sampleIdx],
                sampNames = sampNames[sampleIdx],
                controls = controlType)
        }
    }
    dev.off()
}

controlStripPlot <- function(rgSet,
                             controls = c("BISULFITE CONVERSION I",
                                          "BISULFITE CONVERSION II"),
                             sampNames = NULL, xlim = c(5, 17)) {
    .isRGOrStop(rgSet)

    r <- getRed(rgSet)
    g <- getGreen(rgSet)

    for (controlType in controls) {
        ctrlAddress <- getControlAddress(rgSet, controlType = controlType)

        # Red channel
        ctlWide <- as.matrix(log2(r[ctrlAddress, ,drop = FALSE]))
        if (!is.null(sampNames)) colnames(ctlWide) <- sampNames
        ctlR <- melt(ctlWide, varnames = c("address", "sample"))

        # Green channel
        ctlWide <- as.matrix(log2(g[ctrlAddress, ,drop = FALSE]))
        if (!is.null(sampNames)) colnames(ctlWide) <- sampNames
        ctlG <- melt(ctlWide, varnames = c("address", "sample"))

        # Plot
        ctl <- rbind(
            cbind(channel = "Red", ctlR),
            cbind(channel = "Green", ctlG))
        if (any((ctl$value < xlim[1]) | (ctl$value > xlim[2]))) {
            message("Warning: ", controlType, " probes outside plot range")
        }
        fig <- xyplot(
            x = sample ~ value | channel,
            groups = channel, horizontal = TRUE, pch = 19,
            col = c("darkred", "darkgreen"),
            xlab = "Log2 Intensity",
            xlim = xlim,
            main = paste("Control:", controlType),
            layout = c(2, 1),
            data = ctl,
            panel = function(x, y,...) {
                panel.stripplot(x, y,...)
                panel.abline(h = (as.numeric(y) - 0.5), lty = 3, col = "grey70")
            })
        print(fig)
    }
}

densityBeanPlot <- function(dat, sampGroups = NULL, sampNames = NULL,
                            main = NULL, pal = brewer.pal(8, "Dark2"),
                            numPositions = 10000) {
    if (is(dat, "RGChannelSet") || is(dat, "MethylSet")) {
        b <- getBeta(dat)
    } else if (is(dat, "matrix") || is(dat, "DelayedMatrix")) {
        b <- dat
    } else {
        stop("argument 'dat' must be an 'RGChannelSet', a 'MethylSet', ",
             "a 'matrix', or a 'DelayedMatrix'.")
    }
    n <- ncol(b)
    if (!is.null(sampNames)) colnames(b) <- sampNames
    if (is.null(main)) main <- "Beta"
    if (is.null(sampGroups))  sampGroups <- rep(1, n)
    sampGroups <- as.factor(sampGroups)
    col <- lapply(sampGroups, function(x) rep(pal[x],4))
    if (is.null(numPositions)) {
        idx <- 1:dim(dat)[1]
    } else {
        idx <- sample(nrow(b), numPositions)
    }
    # NOTE: At this point, the data need to be in-memory as an ordinary matrix
    # TODO: Update in response to any suggestions in
    #       https://github.com/Bioconductor/DelayedArray/issues/13
    b_subset <- as.matrix(b[idx, ])
    x <- melt(b_subset, varnames = c("cpg", "sample"))
    o <- order(colnames(b))
    beanplot(
        value ~ sample,
        horizontal = TRUE,
        what = c(0, 1, 1, 0),
        log = "",
        las = 1,
        ylim = c(0, 1),
        xlab = "Beta",
        main = main,
        col = col[o],
        data = x,
        cex.lab = 0.9,
        beanlinewd = 1,
        border = NA)
    abline(h = seq(1, n + 1) - 0.5, lty = 3, col = "grey70")
}


densityPlot <- function(dat, sampGroups = NULL, main = "", xlab = "Beta",
                        pal = brewer.pal(8, "Dark2"),
                        xlim, ylim, add = TRUE, legend = TRUE, ...) {
    if (is(dat, "RGChannelSet") || is(dat, "MethylSet")) {
        b <- getBeta(dat)
    } else if (is(dat, "matrix")) {
        b <- dat
    } else {
        stop("argument 'dat' must be an 'RGChannelSet', a 'MethylSet' or ",
             "matrix.")
    }
    # NOTE: Have to ensure data are in memory as an ordinary vector before
    #       computing density
    d <- apply(b, 2, function(x) density(as.vector(x), na.rm = TRUE))
    if (missing(ylim)) ylim <- range(sapply(d, function(i) range(i$y)))
    if (missing(xlim)) xlim <- range(sapply(d, function(i) range(i$x)))
    if (is.null(sampGroups)) {
        sampGroups <- rep(1, ncol(b))
    } else if (length(sampGroups) == 1) {
        sampGroups <- rep(sampGroups, ncol(b))
    }
    sampGroups <- as.factor(sampGroups)
    if (add) {
        plot(x = 0,
             type = "n",
             ylim = ylim,
             xlim = xlim,
             ylab = "Density",
             xlab = xlab,
             main = main, ...)
        abline(h = 0, col = "grey80")
    }
    for (i in seq_along(d)) {
        lines(d[[i]], col = pal[sampGroups[i]])
    }
    if (legend & length(levels(sampGroups)) > 1) {
        legend("topright", legend = levels(sampGroups), text.col = pal)
    }
}

# Exported classes -------------------------------------------------------------

setClass(
    "IlluminaMethylationAnnotation",
    representation(
        data = "environment",
        annotation = "character",
        defaults = "character"
    )
)

# Validity methods -------------------------------------------------------------

# TODO: Uncomment, fix, or delete this validity method

## setValidity("IlluminaMethylationAnnotation", function(object) {
##     msg <- NULL
##     if(!(all(sapply(object@data, class) == "DataFrame")))
##         msg <- paste(msg,
##                      "All objects in 'objects@data' has to be of class 'DataFrame'", sep = "\n")
##     if (is.null(msg)) TRUE else msg
## })

# Internal functions------------------------------------------------------------

.availableAnnotation <- function(object) {
    object <- getAnnotationObject(object)
    allAnnoNames <- ls(object@data)
    annoClasses <- sub("\\..*", "", allAnnoNames)
    annoClassesChoices <- sub(".*\\.", "", allAnnoNames)
    annoClassesChoices[grep("\\.", allAnnoNames, invert = TRUE)] <- ""
    annoClassesChoices <- split(annoClassesChoices, annoClasses)
    annoClasses <- unique(annoClasses)
    defaults <- object@defaults
    out <- list(
        names = allAnnoNames,
        classes = annoClasses,
        classChoices = annoClassesChoices,
        defaults = defaults)
    out
}

.show.availableAnnotation <- function(object, indent = "  ") {
    available <- .availableAnnotation(object)
    cat("Available annotation\n")
    sapply(available$names, function(xx) {
        cat(sprintf("%s%s\n", indent, xx))
    })
    cat("Defaults\n")
    sapply(available$defaults, function(xx) {
        cat(sprintf("%s%s\n", indent, xx))
    })
}

.annoGet <- function(what, envir) {
    pointer <- get(what, envir = envir)
    if (is(pointer, "DataFrame")) return(pointer)
    get(pointer$what, envir = as.environment(pointer$envir))
}

.getIslandAnnotation <- function(object, islandAnno = NULL) {
    av <- .availableAnnotation(object)
    if (is.null(islandAnno)) {
        islandAnno <- grep(
            pattern = "^Islands\\.",
            x = getAnnotationObject(object)@defaults,
            value = TRUE)
    } else {
        islandAnno <- sub("^Islands\\.", "", islandAnno)
        if (!islandAnno %in% av$annoClassesChoices) {
            stop(sprintf("islandAnno '%s' is not part of the annotation",
                         islandAnno))
        } else {
            islandAnno <- sprintf("Islands.%s", islandAnno)
        }
    }
    getAnnotation(object, what = islandAnno)
}

.doSnpOverlap <- function(map, grSnp) {
    stopifnot(is(map, "GRanges"))
    stopifnot(is(grSnp, "GRanges"))
    stopifnot(all(c("SBE", "probeStart", "probeEnd") %in% names(mcols(map))))
    cat("removing Snps with width != 1\n")
    grSnp <- grSnp[width(grSnp) == 1]
    cpgGR <- GRanges(seqnames(map), IRanges(start(map), width = 2))
    ooCpG <- findOverlaps(cpgGR, grSnp)
    sbeGR <- GRanges(seqnames(map), IRanges(map$SBE, map$SBE))
    ooSbe <- findOverlaps(sbeGR, grSnp)
    ## just match to whole probe then drop CpG overlaps
    probeGR <- GRanges(seqnames(map), IRanges(map$probeStart, map$probeEnd))
    ooProbe <- findOverlaps(probeGR, grSnp, ignore.strand = TRUE)
    ooProbe <- ooProbe[-which(queryHits(ooProbe) %in% queryHits(ooCpG))]
    snpAnno <- DataFrame(matrix(nrow = length(map), ncol = 6))
    colnames(snpAnno) = c("Probe_rs" , "Probe_maf", "CpG_rs",
                          "CpG_maf" ,  "SBE_rs" ,   "SBE_maf")
    rownames(snpAnno) <- names(map)
    snpAnno$Probe_rs[queryHits(ooProbe)] <- names(grSnp)[subjectHits(ooProbe)]
    snpAnno$Probe_maf[queryHits(ooProbe)] <- grSnp$MAF[subjectHits(ooProbe)]
    snpAnno$CpG_rs[queryHits(ooCpG)] <- names(grSnp)[subjectHits(ooCpG)]
    snpAnno$CpG_maf[queryHits(ooCpG)] <- grSnp$MAF[subjectHits(ooCpG)]
    snpAnno$SBE_rs[queryHits(ooSbe)] <- names(grSnp)[subjectHits(ooSbe)]
    snpAnno$SBE_maf[queryHits(ooSbe)] <- grSnp$MAF[subjectHits(ooSbe)]
    snpAnno
}



# Exported functions -----------------------------------------------------------

IlluminaMethylationAnnotation <- function(objectNames, annotation = "",
                                          defaults = "", packageName = "") {
    stopifnot(annotation != "")
    stopifnot(packageName != "")
    stopifnot(all(c("array", "annotation", "genomeBuild") %in%
                      names(annotation)))

    # TODO: Uncomment, fix, or delete the below
    ## stopifnot(all(c("Manifest", "Locations") %in% names(listOfObjects)))
    ## Manifest <- listOfObjects[["Manifest"]]
    ## stopifnot(setequal(names(Manifest),
    ##                    c("Name", "AddressA", "AddressB", "ProbeSeqA",
    ##                      "ProbeSeqB", "Type", "NextBase", "Color")))
    ## stopifnot(all(sapply(listOfObjects, class) %in% c("DataFrame", "data.frame")))
    ## stopifnot(all(nrow(Manifest) == sapply(listOfObjects, nrow)))
    ## stopifnot(all(sapply(listOfObjects, function(obj) {
    ##     all(rownames(obj) == rownames(Manifest))
    ## })))
    ## stopifnot(all(c("chr", "pos") %in% names(listOfObjects[["Locations"]])))
    ## stopifnot(all(listOfObjects[["Locations"]]$chr %in% .seqnames.order.all))
    ## ##available <- .availableAnnotation(listOfObjects)
    stopifnot(all(defaults %in% objectNames))
    ## stopifnot(!anyDuplicated(sub("\\..*", "", defaults)))
    ## TODO: Check column names of any Islands object
    ## Instantiating
    data <- new.env(parent = emptyenv())
    for (nam in objectNames) {
        assign(
            x = nam,
            value = list(
                what = nam,
                envir = sprintf("package:%s", packageName)),
            envir = data)
    }
    lockEnvironment(data, bindings = TRUE)
    anno <- new("IlluminaMethylationAnnotation",
                annotation = annotation, data = data, defaults = defaults)
    anno
}


getAnnotationObject <- function(object) {
    if (is(object, "MethylSet") || is(object, "RatioSet") ||
        is(object, "GenomicMethylSet") || is(object, "GenomicRatioSet") ||
        is(object, "RGChannelSet")) {
        object <- .getAnnotationString(object@annotation)
    }
    if (is.character(object)) {
        if (!require(object, character.only = TRUE)) {
            stop(sprintf("cannot load annotation package %s", object))
        }
        object <- get(object)
    }
    if (!is(object, "IlluminaMethylationAnnotation")) {
        stop("Could not locate annotation object for 'object' of class",
             class(object))
    }
    object
}

getAnnotation <- function(object, what = "everything", lociNames = NULL,
                          orderByLocation = FALSE, dropNonMapping = FALSE) {
    # Processing of arguments and check
    annoObject <- getAnnotationObject(object)
    available <- .availableAnnotation(annoObject)
    if ("everything" %in% what)  what <- available$defaults
    if (!(all(what %in% available$names))) {
        stop("the value of argument 'what' is not part of the annotation ",
             "package or 'everything'")
    }
    n_choices <- vapply(
        available$classes,
        function(cl) length(grep(cl, what)),
        integer(1))
    if (any(n_choices > 1)) stop("only one choice per allowable class")
    if (!any(grepl("^Locations", what)) &&
        (orderByLocation || dropNonMapping)) {
        stop("To use 'orderbyLocation' or 'dropNonMapping' specify Locations ",
             "as part of 'what'")
    }
    # TODO: Ensure same order always
    # Old code for inspiration
    # bestOrder <- c("Locations", "Manifest", "IlluminaSNPs", "Annotation")
    # what <- bestOrder[bestOrder %in% what]

    out <- do.call(cbind, lapply(what, function(wh) {
        updateObject(.annoGet(wh, envir = annoObject@data))
    }))

    if (!is.null(lociNames)) {
        lociNames <- lociNames[lociNames %in% rownames(out)]
    }
    if (is(object, "MethylSet") || is(object, "RatioSet") ||
        is(object, "GenomicMethylSet") || is(object, "GenomicRatioSet")) {
        rNames <- rownames(object)
        if (is.null(lociNames)) {
            lociNames <- rNames[rNames %in% rownames(out)]
        } else {
            lociNames <- rNames[rNames %in% lociNames]
        }
    }
    if (!is.null(lociNames)) out <- out[match(lociNames, rownames(out)),]
    if (dropNonMapping) {
        seqOrder <- .seqnames.order
        wh <- which(out$chr %in% seqOrder)
        out <- out[wh,]
    } else {
        seqOrder <- .seqnames.order.all
    }
    if (orderByLocation) {
        # NOTE: We have a data.frame with chr, pos.  We want to sort accoridng
        #       to chr nested with pos, but with a custom ordering on chr.
        # We will make chr into an integer
        chrInteger <- as.integer(factor(out$chr, levels = seqOrder))
        od <- order(chrInteger, out$pos)
        out <- out[od,]
        # TODO: Uncomment, fix, or delete the below
        # TODO: very memory intensive; improve
        ## sp <- split(out, out$chr)
        ## sp <- sp[seqOrder[seqOrder %in% names(sp)]]
        ## out <- do.call(rbind, lapply(sp, function(df) {
        ##     od <- order(df$pos)
        ##     df[od,]
        ## }))
    }
    out
}

getLocations <- function(object, mergeManifest = FALSE,
                         orderByLocation = FALSE, lociNames = NULL) {
    if (mergeManifest) {
        what <- c("Locations", "Manifest")
    } else {
        what <- "Locations"
    }
    locs <- getAnnotation(
        object = object,
        what = what,
        dropNonMapping = TRUE,
        orderByLocation = orderByLocation,
        lociNames = lociNames)
    gr <- GRanges(seqnames = locs$chr,
                  ranges = IRanges(start = locs$pos, width = 1))
    seqlevels(gr) <- .seqnames.order[.seqnames.order %in% seqlevels(gr)]
    names(gr) <- rownames(locs)
    if (mergeManifest) {
        if ("strand" %in% names(locs)) {
            names(locs)[names(locs) == "strand"] <- "assayStrand"
        }
        mcols(gr) <- locs[, !names(locs) %in% c("chr", "pos", "strand")]
    }
    genome(gr) <- unname(getAnnotationObject(object)@annotation["genomeBuild"])
    gr
}


getIslandStatus <- function(object, islandAnno = NULL) {
    regionType <- .getIslandAnnotation(
        object = object,
        islandAnno = islandAnno)$Relation_to_Island
    regionType <- sub("^[SN]_", "", regionType)
    regionType
}

getProbeType <- function(object, withColor = FALSE) {
    if (withColor) {
        probeType <- paste0(
            getAnnotation(object, what = "Manifest")$Type,
            getAnnotation(object, what = "Manifest")$Color)
    } else {
        probeType <- getAnnotation(object, what = "Manifest")$Type
    }
    probeType
}

getSnpInfo <- function(object, snpAnno = NULL) {
    av <- .availableAnnotation(object)
    if (is.null(snpAnno)) {
        snpAnno <- grep(
            pattern = "^SNPs\\.",
            x = getAnnotationObject(object)@defaults,
            value = TRUE)
    } else {
        snpAnno <- sub("^SNPs\\.", "", snpAnno)
        if (!snpAnno %in% av$annoClassesChoices) {
            stop(sprintf("snpAnno '%s' is not part of the annotation", snpAnno))
        } else {
            snpAnno <- sprintf("SNPs.%s", snpAnno)
        }
    }
    snps <- getAnnotation(object, what = snpAnno)
    snps
}

addSnpInfo <- function(object, snpAnno = NULL) {
    .isGenomicOrStop(object)
    snps <- getSnpInfo(object = object, snpAnno = snpAnno)
    mcolsNames <- names(mcols(granges(object)))
    if (any(mcolsNames %in% names(snps))) {
        cat("Replacing existing snp information\n")
    }
    mcols(object@rowRanges) <- cbind(mcols(granges(object)), snps)
    object
}

dropLociWithSnps <- function(object, snps = c("CpG", "SBE"), maf = 0,
                             snpAnno = NULL){
    .isGenomicOrStop(object)
    maf_cols <- paste0(snps, "_maf")
    snpDF  <- getSnpInfo(object, snpAnno = snpAnno)
    choices <- c("Probe_maf", "CpG_maf", "SBE_maf")
    if (!all(choices %in% colnames(snpDF))) {
        stop("The specificed 'snpAnno' is not supported by this function")
    }
    if (sum(!(maf_cols %in% choices)) > 0) {
        stop("snps vector argument must be a combination of  \"Probe\", ",
             "\"CpG\" and \"SBE\"")
    }
    if (!is.numeric(maf) || maf < 0 || maf > 1) {
        stop("maf argument must be a numeric value between 0 and 1")
    }
    wh <- Reduce(union, lapply(maf_cols, function(xx) {
        which(snpDF[, xx] >= maf)
    }))
    wh <- sort(wh)
    if (length(wh) == 0) return(object)
    object[-wh,]
}

# Exported methods -------------------------------------------------------------

setMethod("show", "IlluminaMethylationAnnotation", function(object) {
    cat("IlluminaMethylationAnnotation object\n")
    .show.annotation(object@annotation)
    .show.availableAnnotation(object)
})

setMethod(
    "getManifest",
    signature(object = "IlluminaMethylationAnnotation"),
    function(object) {
        maniString <- .getManifestString(object@annotation)
        if (!require(maniString, character.only = TRUE)) {
            stop(sprintf("cannot load manifest package %s", maniString))
        }
        get(maniString)
    }
)

library(minfiData)
library(digest)

GRset.quantile <- preprocessQuantile(MsetEx)
digest_preprocessQuantile <- list(M = minfi:::.digestMatrix(getM(GRset.quantile)),
                                  CN = minfi:::.digestMatrix(getCN(GRset.quantile)))
save(digest_preprocessQuantile, file = "../unitTests/digest_preprocessQuantile.rda")

sessionInfo()
rm(list = ls())

# ------------------------------------------------------------------------------
# DelayedArray tests
#

library(DelayedArray)

assays(MsetEx) <- endoapply(assays(MsetEx), DelayedArray)

GRset.quantile <- preprocessQuantile(MsetEx)
digest_preprocessQuantile <- list(M = minfi:::.digestMatrix(getM(GRset.quantile)),
                                  CN = minfi:::.digestMatrix(getCN(GRset.quantile)))
save(digest_preprocessQuantile, file = "../unitTests/digest_preprocessQuantile.DelayedArray.rda")

sessionInfo()
rm(list = ls())

library(minfiData)
library(digest)

Mset.raw <- preprocessRaw(RGsetEx)
digest_preprocessRaw <- list(Meth = minfi:::.digestMatrix(getMeth(Mset.raw)),
                              Unmeth = minfi:::.digestMatrix(getUnmeth(Mset.raw)))
save(digest_preprocessRaw, file = "../unitTests/digest_preprocessRaw.rda")

sessionInfo()
rm(list = ls())

# ------------------------------------------------------------------------------
# DelayedArray tests
#

library(DelayedArray)

assays(RGsetEx) <- endoapply(assays(RGsetEx), DelayedArray)

Mset.raw <- preprocessRaw(RGsetEx)
digest_preprocessRaw <- list(Meth = minfi:::.digestMatrix(getMeth(Mset.raw)),
                             Unmeth = minfi:::.digestMatrix(getUnmeth(Mset.raw)))
save(digest_preprocessRaw, file = "../unitTests/digest_preprocessRaw.DelayedArray.rda")

sessionInfo()
rm(list = ls())

## This script compares the output of Genome Studio to preprocess.Illumina

require(minfi)
require(DelayedMatrixStats)

basepath <- "/thumper2/feinbergLab/core/arrays/illumina/IL002"
sheet <- read.450k.sheet(basepath, pattern = "IL002_v2.csv$")
RGset <- read.450k.exp(targets = sheet, verbose = TRUE)
ISet.raw <- preprocessIllumina(RGset, bg.correct = FALSE, normalize = "no")
ISet.bg <- preprocessIllumina(RGset, bg.correct = TRUE, normalize = "no")
ISet.norm <- preprocessIllumina(RGset, bg.correct = FALSE, normalize = "controls",
                                reference = 8)
ISet.norm.bg <- preprocessIllumina(RGset, bg.correct = TRUE,
                                   normalize = "controls", reference = 8)

od <- c(1,7,8,2,14,9, 3,21,10,15,16,22,
        4,11,23,17,12,24,18,19,5,6,13,20)
basepath <- "/thumper2/feinbergLab/personal/khansen/450k/data"
GS.raw <- minfi:::read.GenomeStudio(file.path(basepath,
                                      "FinalReport_noBack_noNorm.txt"))
odN <- colnames(GS.raw$beta)[od]
GS.raw <- lapply(GS.raw, function(xx) as.matrix(xx[featureNames(ISet.raw), odN]))

GS.bg <- minfi:::read.GenomeStudio(file.path(basepath,
                                     "FinalReport_yesBack_noNorm.txt"))
GS.bg <- lapply(GS.bg, function(xx) as.matrix(xx[featureNames(ISet.bg), odN]))

GS.norm <- minfi:::read.GenomeStudio(file.path(basepath,
                                       "FinalReport_noBack_yesNorm.txt"))
GS.norm <- lapply(GS.norm, function(xx) as.matrix(xx[featureNames(ISet.norm), odN]))

GS.norm.bg <- minfi:::read.GenomeStudio(file.path(basepath,
                                          "FinalReport_yesBack_yesNorm.txt"))
GS.norm.bg <- lapply(GS.norm.bg, function(xx) as.matrix(xx[featureNames(ISet.norm.bg), odN]))

## Now for comparisons

cmax <- colMaxs(abs(getMeth(ISet.raw) - GS.raw$SignalB))
round(cmax, 1)
cmax <- colMaxs(abs(getUnmeth(ISet.raw) - GS.raw$SignalA))
round(cmax, 1)
cmax <- colMaxs(abs(getBeta(ISet.raw, type = "Illumina") - GS.raw$beta), na.rm = TRUE)
round(cmax, 3)

cmax <- colMaxs(abs(getMeth(ISet.bg) - GS.bg$SignalB))
round(cmax, 1)
cmax <- colMaxs(abs(getUnmeth(ISet.bg) - GS.bg$SignalA))
round(cmax, 1)
cmax <- colMaxs(abs(getBeta(ISet.bg, type = "Illumina") - GS.bg$beta), na.rm = TRUE)
round(cmax, 3)

cmax <- colMaxs(abs(getMeth(ISet.norm) - GS.norm$SignalB))
round(cmax, 2)
cmax <- colMaxs(abs(getUnmeth(ISet.norm) - GS.norm$SignalA))
round(cmax, 2)
cmax <- colMaxs(abs(getBeta(ISet.norm, type = "Illumina") - GS.norm$beta), na.rm = TRUE)
round(cmax, 3)

cmax <- colMaxs(abs(getMeth(ISet.norm.bg) - GS.norm.bg$SignalB))
round(cmax, 2)
cmax <- colMaxs(abs(getUnmeth(ISet.norm.bg) - GS.norm.bg$SignalA))
round(cmax, 2)
cmax <- colMaxs(abs(getBeta(ISet.norm.bg, type = "Illumina") - GS.norm.bg$beta), na.rm = TRUE)
round(cmax, 3)

# Example ----------------------------------------------------------------------

# TODO: Move the commented out example to examples/tests or remove entirely

# Author: Jean-Philippe Fortin
# May 6th 2015

# library(minfiData)
# library(lineprof)
# GMset <- mapToGenome(MsetEx)
# prof1 <- lineprof({
# con <- createCorMatrix(GMset, res=500*1000)
# })
# prof2 <- lineprof({
# ab <- extractAB(con)
# })
# prof3 <- lineprof({
# a <- compartments(GMset, resolution=500*1000)
# })

# Internal functions -----------------------------------------------------------

.imputeMatrix <- function(matrix) {
    matrix[is.infinite(matrix) & matrix > 0] <- max(matrix[is.finite(matrix)])
    matrix[is.infinite(matrix) & matrix < 0] <- min(matrix[is.finite(matrix)])

    # Imputation of the missing values:
    missing <- which(is.na(matrix), arr.ind = TRUE)
    if (length(missing) != 0) {
        for (j in seq_len(nrow(missing))) {
            mean <- mean(
                x = matrix[missing[j, 1L], ][
                    is.finite(matrix[missing[j, 1L], ])],
                na.rm = TRUE)
            matrix[missing[j, 1L], missing[j, 2L]] <- mean
        }
    }
    matrix
}

.returnBinnedMatrix <- function(gr.unbinnedCor, resolution){

    bin2D <- function(matrix, ids, n) {
        unique.ids <- sort(unique(ids))
        bin.matrix <- matrix(0, n, n)
        m <- length(unique.ids)
        for (i in seq_len(n)) {
            for (j in seq_len(n)) {
                # TODO: We should be able to speed this one up a lot
                indices1 <- which(ids == unique.ids[i])
                indices2 <- which(ids == unique.ids[j])
                bin.matrix[unique.ids[i],unique.ids[j]] <- median(
                    x = matrix[indices1, indices2],
                    na.rm = TRUE)
            }
        }
        # TODO: This line should not be necessary
        bin.matrix[is.na(bin.matrix)] <- 1
        bin.matrix
    }

    # TODO: Use Bioconductor infrastructure to get chromosome lengths and
    #       don't directly call structure() to create object
    chr.lengths <- structure(
        c(249250621L, 243199373L, 198022430L, 191154276L, 180915260L,
          171115067L, 159138663L, 146364022L, 141213431L, 135534747L,
          135006516L, 133851895L, 115169878L, 107349540L, 102531392L,
          90354753L, 81195210L, 78077248L, 59128983L, 63025520L, 48129895L,
          51304566L, 155270560L, 59373566L),
        .Names = c("chr1", "chr2", "chr3", "chr4", "chr5", "chr6", "chr7",
                   "chr8", "chr9", "chr10", "chr11", "chr12", "chr13", "chr14",
                   "chr15", "chr16", "chr17", "chr18", "chr19", "chr20",
                   "chr21", "chr22", "chrX", "chrY"))
    seqlengths(gr.unbinnedCor) <- chr.lengths[seqlevels(gr.unbinnedCor)]

    stopifnot(length(seqlevels(gr.unbinnedCor)) == 1 &&
                  !is.na(seqlengths(gr.unbinnedCor)))
    stopifnot("cor.matrix" %in% names(mcols(gr.unbinnedCor)))

    gr.binnedCor <- tileGenome(
        seqlengths = seqlengths(gr.unbinnedCor),
        tilewidth = resolution,
        cut.last.tile.in.chrom = TRUE)
    ids <- subjectHits(findOverlaps(gr.unbinnedCor, gr.binnedCor))
    gr.binnedCor$cor.matrix <- bin2D(
        matrix = gr.unbinnedCor$cor.matrix,
        ids = ids,
        n = length(gr.binnedCor))
    gr.binnedCor
}

.removeBadBins <- function(gr) {
    good.bins <- which(!colAlls(gr$cor.matrix, value = 0))
    if (length(good.bins) < nrow(gr$cor.matrix)) {
        gr <- gr[good.bins]
        gr$cor.matrix <- gr$cor.matrix[, good.bins]
    }
    gr
}

.extractOpenClosed <- function(gr, cutoff = 0){
    pc <- gr$pc
    ifelse(pc < cutoff, "open", "closed")
}

.getFirstPC <- function(matrix, method){
    # Centre the matrix
    center <- rowMeans2(matrix, na.rm = TRUE)
    matrix <- sweep(matrix, 1L, center, check.margin = FALSE)
    # TODO: Remove commented code if not needed
    ## if(method == "nipals")
    ##     pc <- mixOmics::nipals(matrix, ncomp = 1)$p[,1]
    .fsvd(matrix, k = 1, method = method)$u
}

.meanSmoother <- function(x, k = 1L, iter = 2L, na.rm = TRUE) {
    meanSmoother.internal <- function(x, k = 1L, na.rm = TRUE) {
        n <- length(x)
        y <- rep(NA_real_, n)

        window.mean <- function(x, j, k, na.rm = na.rm){
            if (k >= 1) {
                return(mean(x[seq(j - (k + 1L), j + k)], na.rm = na.rm))
            } else {
                x[j]
            }
        }

        for (i in (seq(k + 1L, n - k))) {
            y[i] <- window.mean(x, i, k, na.rm)
        }
        for (i in seq_len(k)) {
            y[i] <- window.mean(x, i, i - 1L, na.rm)
        }
        for (i in seq(n - k + 1L, n)) {
            y[i] <- window.mean(x, i, n - i, na.rm)
        }
        y
    }

    for (i in seq_len(iter)) {
        x <- meanSmoother.internal(x, k = k, na.rm = na.rm)
    }
    x
}

.unitarize <- function(x, medianCenter = TRUE) {
    if (medianCenter) x <- x - median(x, na.rm = TRUE)
    bad <- is.na(x)
    x[!bad] <- x[!bad] / sqrt(sum(x[!bad]^2))
    n.bad <- sum(bad)
    if (n.bad > 0) {
        message(
            sprintf("[.unitarize] %i missing values were ignored.\n", n.bad))
    }
    x
}

.fsvd <- function(A, k, i = 1, p = 2, method = c("qr", "svd", "exact")) {
    method <- match.arg(method)
    l <- k + p
    n <- ncol(A)
    m <- nrow(A)
    tall <- TRUE
    # NOTE: In the case a WIDE matrix is provided:
    if (m < n) {
        A <- t(A) # This is potentially expensive
        n <- ncol(A)
        m <- nrow(A)
        tall <- FALSE
    }
    if (l > ncol(A) && method != "exact") {
        stop("(k+p) is greater than the number of columns. Please decrease ",
             "the value of k.")
    }

    # Construct G to be n x l and Gaussian
    G <- matrix(rnorm(n * l, 0, 1), nrow = n, ncol = l)

    if (method == "svd") {
        # Power method:
        H <- A %*% G # m x l matrix
        for (j in seq_len(i)) {
            H <- A %*% (crossprod(A, H))
        }
        # NOTE: We use a SVD to find an othogonal basis Q:
        # TODO: Remove this commented line if not needed
        # H = FF %*% Omega %*% t(S)
        svd <- svd(crossprod(H))
        FF <- svd$u # l x l
        omega <- diag(1/sqrt(svd$d)) # l x l
        S <- H %*% FF %*% omega # m x l
        # Define the orthogonal basis:
        Q <- S[, seq_len(k), drop = FALSE] # m x k
        # TODO: Remove these commented lines if not needed
        # TT <- t(A) %*% Q # n x k
        # TT <- t(TT)
        TT <- crossprod(Q, A)
    } else if (method == "qr") {
        # NOTE: Need to create a list of H matrices
        h.list <- vector("list", i + 1L)
        h.list[[1]] <- A %*% G
        for (j in seq(2, i + 1L)) {
            h.list[[j]] <- A %*% (crossprod(A, h.list[[j - 1L]]))
        }
        H <- do.call("cbind", h.list) # n x [(1+1)l] matrix
        # QR algorithm
        Q <- qr.Q(qr(H, 0))
        # TODO: Remove these commented lines if not needed
        # TT <- t(A)%*%Q # n x [(i+1)l]
        # TT <- t(TT)
        TT <- crossprod(Q, A)
    }
    if (method == "svd" || method == "qr") {
        svd <- svd(TT)
        u <- Q %*% svd$u[,seq_len(k), drop = FALSE]
        v <- svd$v[, seq_len(k), drop = FALSE]
        d <- svd$d[seq_len(k)]
    } else {
        # Exact SVD
        svd <- svd(A)
        u <- svd$u[, seq_len(k), drop = FALSE]
        v <- svd$v[, seq_len(k), drop = FALSE]
        d <- svd$d[seq_len(k)]
    }
    if (!tall) {
        uu <- v
        v <- u
        u <- uu
    }
    list(u = u, v = v, d = d)
}

# Exported functions -----------------------------------------------------------

compartments <- function(object, resolution = 100*1000, what="OpenSea",
                         chr = "chr22", method = c("pearson", "spearman"),
                         keep = TRUE) {

    .isMatrixBackedOrStop(object, "compartments")

    .isGenomicOrStop(object)
    stopifnot(length(chr) == 1 && chr %in% seqlevels(object))
    method <- match.arg(method)
    gr <- createCorMatrix(
        object = object,
        resolution = resolution,
        what = what,
        chr = chr,
        method = method)
    gr <- extractAB(gr, keep = keep)
    gr$compartment <- .extractOpenClosed(gr)
    gr
}

createCorMatrix <- function(object, resolution = 100 * 1000, what = "OpenSea",
                            chr = "chr22", method = c("pearson", "spearman")) {
    .isMatrixBackedOrStop(object, "createCorMatrix")
    .isGenomicOrStop(object)
    stopifnot(length(chr) == 1 && chr %in% seqlevels(object))
    method <- match.arg(method)

    if (is(object, "GenomicMethylSet")) {
        object <- ratioConvert(object, what = "M", keepCN = FALSE)
    }
    assay(object, "M") <- .imputeMatrix(getM(object))

    # Next we subset to a chromosome, keep OpenSea probes and remove SNPs
    seqlevels(object, pruning.mode = "coarse") <- chr
    object <- object[getIslandStatus(object) %in% what,]
    object <- dropLociWithSnps(object, snps = c("CpG", "SBE"), maf = 0.01)

    gr.unbinnedCor <- granges(object)
    gr.unbinnedCor$cor.matrix <- cor(t(getM(object)), method = method)
    gr.cor <- .returnBinnedMatrix(gr.unbinnedCor, resolution = resolution)
    gr.cor <- .removeBadBins(gr.cor)
    gr.cor
}

extractAB <- function(gr, keep = TRUE, svdMethod = "qr"){
    if (!(is(gr, "GRanges") && "cor.matrix" %in% names(mcols(gr)))) {
        stop("'gr' must be an object created by createCorMatrix")
    }
    pc <- .getFirstPC(gr$cor.matrix, method = svdMethod)
    pc <- .meanSmoother(pc)
    pc <- .unitarize(pc)
    # Fix sign of eigenvector
    if (cor(colSums2(gr$cor.matrix), pc) < 0 ) {
        pc <- -pc
    }
    pc <- pc * sqrt(length(pc))
    gr$pc <- pc

    if (!keep) gr$cor.matrix <- NULL
    gr
}

test_compartments <- function() {
    ## stopifnot(require(minfiData))
    ## stopifnot(require(digest))
    ## stopifnot(require(DelayedArray))
    ## load(file.path(path.package("minfi"),
    ##                "unitTests",
    ##                "digest_compartments.rda"))

    ## # Original test
    ## GMsetEx <- mapToGenome(MsetEx)
    ## gr.cor <- createCorMatrix(GMsetEx, res = 500*1000)
    ## checkEquals(digest_compartments$cor.matrix,
    ##             minfi:::.digestMatrix(gr.cor$cor.matrix))
    ## set.seed(456)
    ## gr.ab <- extractAB(gr.cor)
    ## checkEquals(digest_compartments$pc,
    ##             minfi:::.digestVector(gr.ab$pc, digits = 2))

    ## # Testing with DelayedArray-backed objects
    ## MsetEx <- realize(MsetEx)
    ## GMsetEx <- mapToGenome(MsetEx)
    ## checkException(gr.cor <- createCorMatrix(GMsetEx, res = 500*1000),
    ##                silent = TRUE)
    ## # TODO: Uncomment once createCorMatrix() supports DelayedArray-backed
    ## #       minfi objects
    ## # gr.cor <- createCorMatrix(GMsetEx, res = 500*1000)
    ## # checkEquals(digest_compartments$cor.matrix,
    ## #             minfi:::.digestMatrix(gr.cor$cor.matrix))
    ## # set.seed(456)
    ## # gr.ab <- extractAB(gr.cor)
    ## # checkEquals(digest_compartments$pc,
    ## #             minfi:::.digestVector(gr.ab$pc, digits = 2))
    checkTrue(TRUE)
}

# Internal functions -----------------------------------------------------------

.fixMethOutliers <- function(mat, K = -3, verbose = FALSE) {
    # Compute cutoff
    log2_mat <- log2(mat + 0.5)
    mu <- colMedians(log2_mat)
    sd <- colMads(log2_mat)
    cutoff <- 2 ^ (K * sd + mu)

    # NOTE: nFixes is only computed if verbose is TRUE because it's not
    #       otherwise needed
    if (verbose) {
        nFixes <- colSums2(sweep(mat, 2, cutoff, `<`))
        msg <- paste0(
            "[.fixMethOutliers] for sample %s, fixing %s outliers with ",
            "2^cutoff=%s\n")
        for (j in seq_len(ncol(mat))) {
            message(sprintf(msg, j, nFixes[j], round(cutoff[j], 0)))
        }
    }

    # Threshold matrix values so that all values are greater than or equal to
    # the column-specific `cutoff`
    sweep(mat, 2, cutoff, pmax2)
}

# Exported functions -----------------------------------------------------------

fixMethOutliers <- function(object, K = -3, verbose = FALSE) {
    .isMethylOrStop(object)
    if (verbose) message("[fixMethOutliers] fixing Meth channel\n")
    assay(object, "Meth") <- .fixMethOutliers(getMeth(object), K, verbose)
    if (verbose) message("[fixMethOutliers] fixing Unmeth channel\n")
    assay(object, "Unmeth") <- .fixMethOutliers(getUnmeth(object), K, verbose)
    return(object)
}

addQC <- function(object, qc) {
    .pDataAdd(object, qc)
}

plotQC <- function(qc, badSampleCutoff = 10.5) {
    meds <- (qc$mMed + qc$uMed)/2
    whichBad <- which((meds < badSampleCutoff))
    plot(qc$mMed, qc$uMed,
         xlim = c(8,14), ylim = c(8,14), xaxt = "n", yaxt = "n",
         xlab = "Meth median intensity (log2)",
         ylab = "Unmeth median intensity (log2)",
         col = ifelse(1:nrow(qc) %in% whichBad, "red", "black"))
    axis(side = 1, at = c(9,11,13))
    axis(side = 2, at = c(9,11,13))
    ## abline(h = badSampleCutoff, lty = 2)
    ## abline(v = badSampleCutoff, lty = 2)
    abline(badSampleCutoff * 2 , -1, lty = 2)
    if (length(whichBad) > 0) {
        text(qc$mMed[whichBad], qc$uMed[whichBad] - 0.25,
             labels = whichBad, col = "red")
    }
    legend("topleft", legend = c("good", "bad, with sample index"), pch = 1,
           col = c("black", "red"), bty = "n")
    invisible(NULL)
}

getQC <- function(object) {
    .isMethylOrStop(object)
    U.medians <- log2(colMedians(getUnmeth(object)))
    M.medians <- log2(colMedians(getMeth(object)))
    df <- DataFrame(mMed = M.medians, uMed = U.medians)
    rownames(df) <- colnames(object)
    df
}

minfiQC <- function(object, fixOutliers=TRUE, verbose = FALSE){
    .isMethylOrStop(object)
    subverbose <- max(as.integer(verbose) - 1L, 0L)
    if (fixOutliers) {
        if (verbose) message("[minfiQC] fixing outliers\n")
        fixMethOutliers(object, verbose = subverbose)
    }
    qc <- getQC(object)
    object <- addQC(object, qc = qc)
    if (!is(object, "GenomicMethylSet")) {
        sex <- getSex(mapToGenome(object))
    } else {
        sex <- getSex(object)
    }
    object <- addSex(object, sex = sex)
    return(list(object = object, qc = cbind(qc, sex)))
}

## # A helper function to check whether two SummarizedExperiment objects are
## # equivalent
## checkEquivalentSummarizedExperiments <- function(SE1, SE2) {
##     stopifnot(require(SummarizedExperiment))
##     stopifnot(is(SE1, "SummarizedExperiment"),
##               is(SE2, "SummarizedExperiment"))
##     assays(SE1) <- endoapply(assays(SE1), as.matrix)
##     assays(SE2) <- endoapply(assays(SE2), as.matrix)
##     if (isTRUE(all.equal(SE1, SE2))) {
##         return(all.equal(assays(SE1), assays(SE2)))
##     }
##     FALSE
## }

## # TODO: Uncomment once bumphunter() supports DelayedArray-backed minfi
## #       objects
## test_bumphunter_examples <- function() {

##     stopifnot(require(minfiData))
##     stopifnot(require(DelayedArray))
##     # Original example
##     # gmSet <- preprocessQuantile(MsetEx)
##     # design <- model.matrix(~ gmSet$status)
##     # bumps <- bumphunter(
##     #     object = gmSet,
##     #     design = design,
##     #     B = 0,
##     #     type = "Beta",
##     #     cutoff = 0.25)
##     # bumps0 <- bumps

##     # DelayedArray-backed version
##     MsetEx <- realize(MsetEx)
##     checkException(gmSet <- preprocessQuantile(MsetEx),
##                    silent = TRUE)
##     # design <- model.matrix(~ gmSet$status)
##     # checkException(bumps <- bumphunter(
##     #     object = gmSet,
##     #     design = design,
##     #     B = 0,
##     #     type = "Beta",
##     #     cutoff = 0.25))

##     # Check equivalency
##     # checkIdentical(bumps0, bumps)
## }

## test_combineArrays_examples <- function() {
##     stopifnot(require(minfiData))
##     stopifnot(require(minfiDataEPIC))
##     stopifnot(require(DelayedArray))

##     # Original example
##     data(RGsetEx.sub)
##     data(RGsetEPIC)
##     rgSet0 <- combineArrays(RGsetEPIC, RGsetEx.sub)

##     # DelayedArray-backed version
##     RGsetEx.sub <- realize(RGsetEx.sub)
##     RGsetEPIC <- realize(RGsetEPIC)
##     rgSet <- combineArrays(RGsetEPIC, RGsetEx.sub)

##     # Check equivalency
##     checkEquivalentSummarizedExperiments(rgSet0, rgSet)
## }

## # TODO: Uncomment once compartments() supports DelayedArray-backed minfi objects
## test_compartments_examples <- function() {
##     stopifnot(require(minfiData))
##     stopifnot(require(DelayedArray))

##     # Original example
##     # GMset <- mapToGenome(MsetEx)
##     # comps <- compartments(GMset, res = 10^6)
##     # comps0 <- comps

##     # DelayedArray-backed version
##     MsetEx <- realize(MsetEx)
##     GMset <- mapToGenome(MsetEx)
##     checkException(comps <- compartments(GMset, res = 10^6), silent = TRUE)

##     # Check equivalency
##     # checkEquals(comps0, comps)
## }

## test_convertArray_examples <- function() {
##     stopifnot(require(minfiData))
##     stopifnot(require(DelayedArray))

##     # Original example
##     data(RGsetEx.sub)
##     rgSet <- convertArray(RGsetEx.sub, outType = "IlluminaHumanMethylationEPIC")
##     rgSet0 <- rgSet

##     # DelayedArray-backed version
##     RGsetEx.sub <- realize(RGsetEx.sub)
##     rgSet <- convertArray(RGsetEx.sub, outType = "IlluminaHumanMethylationEPIC")

##     # Check equivalency
##     checkEquivalentSummarizedExperiments(rgSet0, rgSet)
## }

## test_detectionP_examples <- function() {
##     stopifnot(require(minfiData))
##     stopifnot(require(DelayedArray))

##     # Original example
##     detP <- detectionP(RGsetEx.sub)
##     failed0 <- detP > 0.01

##     # DelayedArray-backed version
##     RGsetEx.sub <- realize(RGsetEx.sub)
##     detP <- detectionP(RGsetEx.sub)
##     failed <- detP > 0.01

##     # Check equivalency
##     # NOTE: detectionP() returns a DelayedMatrix when applied to a
##     #       DelayedArray-backed object, hence need for as.matrix()
##     checkIdentical(failed0, as.matrix(failed))
## }

## # TODO: Uncomment once dmpFinder() supports DelayedArray-backed minfi objects
## test_dmpFinder_examples <- function() {
##     stopifnot(require(minfiData))
##     stopifnot(require(DelayedArray))

##     # Original example
##     MsetExSmall <- MsetEx[seq_len(1e4), ]
##     # grp <- pData(MsetEx)$Sample_Group
##     # M <- getM(MsetExSmall, type = "beta", betaThreshold = 0.001)
##     # dmp0 <- dmpFinder(M, pheno = grp, type = "categorical")

##     # DelayedArray-backed version
##     MsetExSmall <- realize(MsetExSmall)
##     M <- getM(MsetExSmall, type = "beta", betaThreshold = 0.001)
##     checkException(dmp <- dmpFinder(M, pheno = grp, type = "categorical"),
##                    silent = TRUE)

##     # Check equivalency
##     # checkIdentical(dmp0, dmp)
## }

## # TODO: Uncomment once estimateCellCounts() supports DelayedArray-backed minfi
## #       objects
## test_estimateCellCounts_examples <- function() {
##     stopifnot(require(FlowSorted.Blood.450k))
##     stopifnot(require(DelayedArray))

##     # Original example
##     wh.WBC <- which(FlowSorted.Blood.450k$CellType == "WBC")
##     wh.PBMC <- which(FlowSorted.Blood.450k$CellType == "PBMC")
##     RGset <- FlowSorted.Blood.450k[, c(wh.WBC, wh.PBMC)]
##     # The following line is purely to work around an issue with repeated
##     # sampleNames and Biobase::combine()
##     sampleNames(RGset) <- paste(
##         RGset$CellType,
##         c(seq(along = wh.WBC), seq(along = wh.PBMC)),
##         sep = "_")
##     # counts0 <- estimateCellCounts(RGset, meanPlot = FALSE)

##     # DelayedArray-backed version
##     RGset <- realize(RGset)
##     checkException(counts <- estimateCellCounts(RGset, meanPlot = FALSE),
##                    silent = TRUE)

##     # Check equivalency
##     # checkIdentical(counts0, counts)
## }

## test_fixMethOutliers_examples <- function() {
##     stopifnot(require(minfiData))
##     stopifnot(require(DelayedArray))

##     # Original example
##     MsetEx0 <- fixMethOutliers(MsetEx)

##     # DelayedArray-backed version
##     MsetEx <- realize(MsetEx)
##     MsetEx <- fixMethOutliers(MsetEx)

##     # Check equivalency
##     checkEquivalentSummarizedExperiments(MsetEx0, MsetEx)
## }

## # TODO: Uncomment once gaphunter() supports DelayedArray-backed minfi objects
## test_gaphunter_examples <- function() {
##     stopifnot(require(minfiData))
##     stopifnot(require(DelayedArray))

##     # Original example
##     # gapres0 <- gaphunter(MsetEx.sub, threshold = 0.3, keepOutliers = TRUE)

##     # DelayedArray-backed version
##     MsetEx.sub <- realize(MsetEx.sub)
##     checkException(
##         gapres <- gaphunter(MsetEx.sub, threshold = 0.3, keepOutliers = TRUE),
##         silent = TRUE)

##     # Check equivalency
##     # checkEquivalentSummarizedExperiments(gapres0, gapres)
## }

## test_getAnnotation_examples <- function() {
##     stopifnot(require(minfiData))
##     stopifnot(require(DelayedArray))

##     # Original example
##     anno0 <- getAnnotation(MsetEx, what = "Manifest")

##     # DelayedArray-backed version
##     MsetEx <- realize(MsetEx)
##     anno <- getAnnotation(MsetEx, what = "Manifest")

##     # Check equivalency
##     checkIdentical(anno0, anno)
## }

## test_getQC_examples <- function() {
##     stopifnot(require(minfiData))
##     stopifnot(require(DelayedArray))

##     # Original example
##     qc <- getQC(MsetEx)
##     MsetEx0 <- addQC(MsetEx, qc = qc)

##     # DelayedArray-backed version
##     MsetEx <- realize(MsetEx)
##     qc <- getQC(MsetEx)
##     MsetEx <- addQC(MsetEx, qc = qc)

##     # Check equivalency
##     checkEquivalentSummarizedExperiments(MsetEx0, MsetEx)
## }

## test_getSex_examples <- function() {
##     stopifnot(require(minfiData))
##     stopifnot(require(DelayedArray))

##     # Original example
##     GMsetEx <- mapToGenome(MsetEx)
##     estSex <- getSex(GMsetEx)
##     GMsetEx0 <- addSex(GMsetEx, sex = estSex)

##     # DelayedArray-backed version
##     MsetEx <- realize(MsetEx)
##     GMsetEx <- mapToGenome(MsetEx)
##     estSex <- getSex(GMsetEx)
##     GMsetEx <- addSex(GMsetEx, sex = estSex)

##     # Check equivalency
##     checkEquivalentSummarizedExperiments(GMsetEx0, GMsetEx)
## }

## test_logit2_examples <- function() {
##     stopifnot(require(DelayedArray))

##     # Original example
##     x <- matrix(c(0.25, 0.5, 0.75))
##     val0 <- logit2(x)

##     # DelayedArray-backed version
##     x <- DelayedArray(x)
##     val <- logit2(x)

##     # Check equivalency
##     # NOTE: logi2() returns a DelayedMatrix when applied to a DelayedArra,
##     #       hence need for as.matrix()
##     checkIdentical(val0, as.matrix(val))
## }

## # TODO: Uncomment once makeGenomicRatioSetFromMatrix() supports DelayedArray
## test_makeGenomicRatioSetFromMatrix_examples <- function() {
##     stopifnot(require(DelayedArray))

##     # Original example
##     mat <- matrix(10, 5, 2)
##     rownames(mat) <- c(
##         "cg13869341", "cg14008030","cg12045430", "cg20826792","cg00381604")
##     # grset0 <- makeGenomicRatioSetFromMatrix(mat)

##     # DelayedArray-backed version
##     mat <- realize(mat)
##     checkException(grset <- makeGenomicRatioSetFromMatrix(mat), silent = TRUE)

##     # Check equivalency
##     # checkEquivalentSummarizedExperiments(grset0, grset)
## }

## test_mapToGenome_examples <- function() {
##     stopifnot(require(minfiData))
##     stopifnot(require(DelayedArray))

##     # Original example
##     GMsetEx.sub0 <- mapToGenome(MsetEx.sub)

##     # DelayedArray-backed version
##     MsetEx.sub <- realize(MsetEx.sub)
##     GMsetEx.sub <- mapToGenome(MsetEx.sub)

##     checkEquivalentSummarizedExperiments(GMsetEx.sub0, GMsetEx.sub)
## }

## test_minfiQC_examples <- function() {
##     stopifnot(require(minfiData))
##     stopifnot(require(DelayedArray))

##     # Original example
##     out0 <- minfiQC(MsetEx)

##     # DelayedArray-backed version
##     MsetEx <- realize(MsetEx)
##     out <- minfiQC(MsetEx)

##     # Check equivalency
##     checkEquivalentSummarizedExperiments(out0$object, out$object)
##     checkIdentical(out0$qc, out$qc)
## }

## # TODO: Uncomment once preprocessFunnorm() supports DelayedArray-backed minfi
## #       objects
## test_preprocessFunnorm_examples <- function() {
##     stopifnot(require(minfiData))
##     stopifnot(require(DelayedArray))

##     # Original example
##     # Mset.sub.funnorm0 <- preprocessFunnorm(RGsetEx.sub)

##     # DelayedArray-backed version
##     RGsetEx.sub <- realize(RGsetEx.sub)
##     checkException(Mset.sub.funnorm <- preprocessFunnorm(RGsetEx.sub),
##                    silent = TRUE)

##     # Check equivalency
##     # checkEquivalentSummarizedExperiments(Mset.sub.funnorm0, Mset.sub.funnorm)
## }

## test_preprocessIllumina_examples <- function() {
##     stopifnot(require(minfiData))
##     stopifnot(require(DelayedArray))

##     # Original example
##     dat0 <- preprocessIllumina(
##         rgSet = RGsetEx,
##         bg.correct = FALSE,
##         normalize = "controls")

##     # DelayedArray-backed version
##     RGsetEx <- realize(RGsetEx)
##     dat <- preprocessIllumina(
##         rgSet = RGsetEx,
##         bg.correct = FALSE,
##         normalize = "controls")

##     # Check equality
##     checkEquivalentSummarizedExperiments(dat0, dat)
## }

## test_preprocessNoob_examples <- function() {
##     stopifnot(require(minfiData))
##     stopifnot(require(DelayedArray))

##     # Original example
##     MsetEx.sub.noob <- preprocessNoob(RGsetEx.sub)
##     dyeMethods <- c(ssNoob = "single", refNoob = "reference")
##     GRsets0 <- lapply(
##         X = dyeMethods,
##         FUN = function(m) preprocessNoob(RGsetEx, dyeMethod = m))

##     # DelayedArray-backed version
##     RGsetEx.sub <- realize(RGsetEx.sub)
##     MsetEx.sub.noob <- preprocessNoob(RGsetEx.sub)
##     GRsets <- lapply(
##         X = dyeMethods,
##         FUN = function(m) preprocessNoob(RGsetEx, dyeMethod = m))

##     # Check equivalency
##     Map(checkEquivalentSummarizedExperiments, GRsets0, GRsets)
## }

## # TODO: Uncomment once preprocessQuantile() supports DelayedArray-backed minfi
## #       objects
## test_preprocessQuantile_examples <- function() {
##     stopifnot(require(minfiData))
##     stopifnot(require(DelayedArray))

##     # Original example
##     # GMset.sub.quantile0 <- preprocessQuantile(RGsetEx.sub)
##     # GMset0 <- preprocessQuantile(RGsetEx)

##     # DelayedArray-backed version
##     RGsetEx.sub <- realize(RGsetEx.sub)
##     checkException(GMset.sub.quantile <- preprocessQuantile(RGsetEx.sub),
##                    silent = TRUE)
##     RGsetEx <- realize(RGsetEx)
##     checkException(GMset <- preprocessQuantile(RGsetEx),
##                    silent = TRUE)

##     # Check equivalency
##     # checkEquivalentSummarizedExperiments(
##     #     GMset.sub.quantile0,
##     #     GMset.sub.quantile)
##     # checkEquivalentSummarizedExperiments(GMset0, GMset)
## }

## test_preprocessRaw_examples <- function() {
##     stopifnot(require(minfiData))
##     stopifnot(require(DelayedArray))

##     # Original example
##     dat0 <- preprocessRaw(RGsetEx)

##     # DelayedArray-backed version
##     RGsetEx <- realize(RGsetEx)
##     dat <- preprocessRaw(RGsetEx)

##     # Check equivalency
##     checkEquivalentSummarizedExperiments(dat0, dat)
## }

## test_preprocessSwan_examples <- function() {
##     stopifnot(require(minfiData))
##     stopifnot(require(DelayedArray))

##     # Original example
##     set.seed(666)
##     MsetEx.sub.swan0 <- preprocessSWAN(RGsetEx.sub)
##     dat0 <- preprocessRaw(RGsetEx)
##     set.seed(777)
##     datSwan0 <- preprocessSWAN(RGsetEx, mSet = dat0)
##     datIlmn0 <- preprocessIllumina(RGsetEx)
##     set.seed(888)
##     datIlmnSwan0 <- preprocessSWAN(RGsetEx, mSet = datIlmn0)

##     # DelayedArray-backed version
##     RGsetEx.sub <- realize(RGsetEx.sub)
##     RGsetEx <- realize(RGsetEx)
##     set.seed(666)
##     MsetEx.sub.swan <- preprocessSWAN(RGsetEx.sub)
##     dat <- preprocessRaw(RGsetEx)
##     set.seed(777)
##     datSwan <- preprocessSWAN(RGsetEx, mSet = dat)
##     datIlmn <- preprocessIllumina(RGsetEx)
##     set.seed(888)
##     datIlmnSwan <- preprocessSWAN(RGsetEx, mSet = datIlmn)

##     # Check equivalency
##     checkEquivalentSummarizedExperiments(dat0, dat)
##     checkEquivalentSummarizedExperiments(datSwan0, datSwan)
##     checkEquivalentSummarizedExperiments(datIlmn0, datIlmn)
##     checkEquivalentSummarizedExperiments(datIlmnSwan0, datIlmnSwan)
## }

## test_ratioConvert_examples <- function() {
##     stopifnot(require(minfiData))
##     stopifnot(require(DelayedArray))

##     # Original examples
##     RsetEx.sub0 <- ratioConvert(MsetEx.sub, keepCN = TRUE)

##     # DelayedArray-backed version
##     MsetEx.sub <- realize(MsetEx.sub)
##     RsetEx.sub <- ratioConvert(MsetEx.sub, keepCN = TRUE)

##     # Check equivalency
##     checkEquivalentSummarizedExperiments(RsetEx.sub0, RsetEx.sub)
## }

## test_subsetByLoci_examples <- function() {
##     stopifnot(require(minfiData))
##     stopifnot(require(DelayedArray))

##     # Original examples
##     loci <- c("cg00050873", "cg00212031", "cg00213748", "cg00214611")
##     a0 <- subsetByLoci(RGsetEx.sub, includeLoci = loci)
##     b0 <- subsetByLoci(RGsetEx.sub, excludeLoci = loci)

##     # DelayedArray-backed version
##     RGsetEx.sub <- realize(RGsetEx.sub)
##     a <- subsetByLoci(RGsetEx.sub, includeLoci = loci)
##     b <- subsetByLoci(RGsetEx.sub, excludeLoci = loci)

##     # Check equivalency
##     checkEquivalentSummarizedExperiments(a0, a)
##     checkEquivalentSummarizedExperiments(b0, b)
## }

## test_utils_examples <- function() {
##     stopifnot(require(minfiData))
##     stopifnot(require(DelayedArray))

##     # Original examples
##     a0 <- head(getMethSignal(MsetEx, what = "Beta"))

##     # DelayedArray-backed version
##     MsetEx <- realize(MsetEx)
##     a <- head(getMethSignal(MsetEx, what = "Beta"))

##     # Check equivalency
##     # NOTE: getMethSignal() returns a DelayedMatrix when applied to a
##     #       DelayedArray-backed object, hence need for as.matrix()
##     checkEquals(a0, as.matrix(a))
## }

# Internal functions -----------------------------------------------------------

pickCompProbes <- function(mSet, cellTypes = NULL, numProbes = 50,
                           compositeCellType = compositeCellType,
                           probeSelect = probeSelect) {
    .isMatrixBackedOrStop(mSet)
    splitit <- function(x) {
        split(seq_along(x), x)
    }

    p <- getBeta(mSet)
    pd <- as.data.frame(colData(mSet))
    if (!is.null(cellTypes)) {
        if (!all(cellTypes %in% pd$CellType))
            stop("elements of argument 'cellTypes' is not part of ",
                 "'mSet$CellType'")
        keep <- which(pd$CellType %in% cellTypes)
        pd <- pd[keep,]
        p <- p[,keep]
    }
    # NOTE: Make cell type a factor
    pd$CellType <- factor(pd$CellType, levels = cellTypes)
    ffComp <- rowFtests(p, pd$CellType)
    prof <- vapply(
        X = splitit(pd$CellType),
        FUN = function(j) rowMeans2(p, cols = j),
        FUN.VALUE = numeric(nrow(p)))
    r <- rowRanges(p)
    compTable <- cbind(ffComp, prof, r, abs(r[, 1] - r[, 2]))
    names(compTable)[1] <- "Fstat"
    names(compTable)[c(-2, -1, 0) + ncol(compTable)] <-
        c("low", "high", "range")
    tIndexes <- splitit(pd$CellType)
    tstatList <- lapply(tIndexes, function(i) {
        x <- rep(0,ncol(p))
        x[i] <- 1
        return(rowttests(p, factor(x)))
    })

    if (probeSelect == "any") {
        probeList <- lapply(tstatList, function(x) {
            y <- x[x[, "p.value"] < 1e-8, ]
            yAny <- y[order(abs(y[, "dm"]), decreasing = TRUE), ]
            c(rownames(yAny)[seq(numProbes * 2)])
        })
    } else {
        probeList <- lapply(tstatList, function(x) {
            y <- x[x[, "p.value"] < 1e-8, ]
            yUp <- y[order(y[, "dm"], decreasing = TRUE), ]
            yDown <- y[order(y[, "dm"], decreasing = FALSE), ]
            c(rownames(yUp)[seq_len(numProbes)],
              rownames(yDown)[seq_len(numProbes)])
        })
    }

    trainingProbes <- unique(unlist(probeList))
    p <- p[trainingProbes,]

    pMeans <- colMeans2(p)
    names(pMeans) <- pd$CellType

    form <- as.formula(
        sprintf("y ~ %s - 1", paste(levels(pd$CellType), collapse = "+")))
    phenoDF <- as.data.frame(model.matrix(~ pd$CellType - 1))
    colnames(phenoDF) <- sub("^pd\\$CellType", "", colnames(phenoDF))
    if (ncol(phenoDF) == 2) {
        # Two group solution
        X <- as.matrix(phenoDF)
        coefEsts <- t(solve(t(X) %*% X) %*% t(X) %*% t(p))
    } else {
        # > 2 groups solution
        tmp <- validationCellType(Y = p, pheno = phenoDF, modelFix = form)
        coefEsts <- tmp$coefEsts
    }

    list(
        coefEsts = coefEsts,
        compTable = compTable,
        sampleMeans = pMeans)
}



projectCellType <- function(Y, coefCellType, contrastCellType = NULL,
                            nonnegative = TRUE, lessThanOne = FALSE) {
    if (is.null(contrastCellType)) {
        Xmat <- coefCellType
    } else {
        Xmat <- tcrossprod(coefCellType, contrastCellType)
    }

    nCol <- dim(Xmat)[2]
    if (nCol == 2) {
        Dmat <- crossprod(Xmat)
        mixCoef <- t(
            apply(Y, 2, function(x) solve(Dmat, crossprod(Xmat, x))))
        colnames(mixCoef) <- colnames(Xmat)
        return(mixCoef)
    } else {
        nSubj <- dim(Y)[2]

        mixCoef <- matrix(0, nSubj, nCol)
        rownames(mixCoef) <- colnames(Y)
        colnames(mixCoef) <- colnames(Xmat)

        if (nonnegative) {
            if (lessThanOne) {
                Amat <- cbind(rep(-1, nCol), diag(nCol))
                b0vec <- c(-1, rep(0, nCol))
            } else {
                Amat <- diag(nCol)
                b0vec <- rep(0, nCol)
            }
            for (i in seq_len(nSubj)) {
                obs <- which(!is.na(Y[,i]))
                Dmat <- crossprod(Xmat[obs,])
                mixCoef[i,] <- solve.QP(
                    Dmat = Dmat,
                    dvec = crossprod(Xmat[obs,], Y[obs,i]),
                    Amat = Amat,
                    bvec = b0vec)$sol
            }
        } else {
            for (i in seq_len(nSubj)) {
                obs <- which(!is.na(Y[,i]))
                Dmat <- crossprod(Xmat[obs,])
                mixCoef[i,] <- solve(Dmat, t(Xmat[obs,]) %*% Y[obs,i])
            }
        }
        mixCoef
    }
}

validationCellType <- function(Y, pheno, modelFix, modelBatch=NULL,
                               L.forFstat = NULL, verbose = FALSE){
    N <- dim(pheno)[1]
    pheno$y <- rep(0, N)
    xTest <- model.matrix(modelFix, pheno)
    sizeModel <- dim(xTest)[2]
    M <- dim(Y)[1]

    if (is.null(L.forFstat)) {
        # NOTE: All non-intercept coefficients
        L.forFstat <- diag(sizeModel)[-1,]
        colnames(L.forFstat) <- colnames(xTest)
        rownames(L.forFstat) <- colnames(xTest)[-1]
    }

    # Initialize various containers
    sigmaResid <- sigmaIcept <- nObserved <- nClusters <- Fstat <- rep(NA, M)
    coefEsts <- matrix(NA, M, sizeModel)
    coefVcovs <- list()

    if (verbose) cat("[validationCellType] ")
    # Loop over each CpG
    for (j in seq_len(M)) {
        # Remove missing methylation values
        ii <- !is.na(Y[j, ])
        nObserved[j] <- sum(ii)
        pheno$y <- Y[j,]

        if (j %% round(M / 10) == 0 && verbose) cat(".") # Report progress

        # Try to fit a mixed model to adjust for plate
        try({
            if (!is.null(modelBatch)) {
                fit <- try(
                    lme(modelFix, random = modelBatch, data = pheno[ii, ]))
                # NOTE: If LME can't be fit, just use OLS
                OLS <- inherits(fit, "try-error")
            } else {
                OLS <- TRUE
            }

            if (OLS) {
                fit <- lm(modelFix, data = pheno[ii, ])
                fitCoef <- fit$coef
                sigmaResid[j] <- summary(fit)$sigma
                sigmaIcept[j] <- 0
                nClusters[j] <- 0
            } else {
                fitCoef <- fit$coef$fixed
                sigmaResid[j] <- fit$sigma
                sigmaIcept[j] <- sqrt(getVarCov(fit)[1])
                nClusters[j] <- length(fit$coef$random[[1]])
            }
            coefEsts[j,] <- fitCoef
            coefVcovs[[j]] <- vcov(fit)

            useCoef <- L.forFstat %*% fitCoef
            useV <- L.forFstat %*% coefVcovs[[j]] %*% t(L.forFstat)
            Fstat[j] <- (t(useCoef) %*% solve(useV, useCoef)) / sizeModel
        })
    }
    if (verbose) cat(" done\n")

    # Name the rows so that they can be easily matched to the target data set
    rownames(coefEsts) <- rownames(Y)
    colnames(coefEsts) <- names(fitCoef)
    degFree <- nObserved - nClusters - sizeModel + 1

    # Get P values corresponding to F statistics
    Pval <- 1 - pf(Fstat, sizeModel, degFree)

    list(
        coefEsts = coefEsts,
        coefVcovs = coefVcovs,
        modelFix = modelFix,
        modelBatch = modelBatch,
        sigmaIcept = sigmaIcept,
        sigmaResid = sigmaResid,
        L.forFstat = L.forFstat,
        Pval = Pval,
        orderFstat = order(-Fstat),
        Fstat = Fstat,
        nClusters = nClusters,
        nObserved = nObserved,
        degFree = degFree)
}

# Exported functions -----------------------------------------------------------

estimateCellCounts <- function(rgSet, compositeCellType = "Blood",
                               processMethod = "auto", probeSelect = "auto",
                               cellTypes = c("CD8T", "CD4T", "NK", "Bcell",
                                             "Mono", "Gran"),
                               referencePlatform = c(
                                   "IlluminaHumanMethylation450k",
                                   "IlluminaHumanMethylationEPIC",
                                   "IlluminaHumanMethylation27k"),
                               returnAll = FALSE, meanPlot = FALSE,
                               verbose = TRUE, ...) {

    # Check inputs
    .isMatrixBackedOrStop(rgSet, "estimateCellCounts")
    .isRGOrStop(rgSet)
    rgSet <- as(rgSet, "RGChannelSet")
    referencePlatform <- match.arg(referencePlatform)
    rgPlatform <- sub(
        "IlluminaHumanMethylation",
        "",
        annotation(rgSet)[which(names(annotation(rgSet)) == "array")])
    platform <- sub("IlluminaHumanMethylation", "", referencePlatform)
    if ((compositeCellType == "CordBlood") && (!"nRBC" %in% cellTypes)) {
        message("[estimateCellCounts] Consider including 'nRBC' in argument 'cellTypes' for cord blood estimation.\n")
    }
    referencePkg <- sprintf("FlowSorted.%s.%s", compositeCellType, platform)
    subverbose <- max(as.integer(verbose) - 1L, 0L)
    if (!require(referencePkg, character.only = TRUE)) {
        stop(sprintf("Could not find reference data package for compositeCellType '%s' and referencePlatform '%s' (inferred package name is '%s')",
                     compositeCellType, platform, referencePkg))
    }
    data(list = referencePkg)
    referenceRGset <- get(referencePkg)
    if (rgPlatform != platform) {
        rgSet <- convertArray(
            object = rgSet,
            outType = referencePlatform,
            verbose = subverbose)
    }
    if (!"CellType" %in% names(colData(referenceRGset))) {
        stop(sprintf("the reference sorted dataset (in this case '%s') needs to have a phenoData column called 'CellType'"),
             names(referencePkg))
    }
    if (sum(colnames(rgSet) %in% colnames(referenceRGset)) > 0) {
        stop("the sample/column names in the user set must not be in the ",
             "reference data ")
    }
    if (!all(cellTypes %in% referenceRGset$CellType)) {
        stop(sprintf("all elements of argument 'cellTypes' needs to be part of the reference phenoData columns 'CellType' (containg the following elements: '%s')",
                     paste(unique(referenceRGset$cellType), collapse = "', '")))
    }
    if (length(unique(cellTypes)) < 2) {
        stop("At least 2 cell types must be provided.")
    }
    if ((processMethod == "auto") &&
        (compositeCellType %in% c("Blood", "DLPFC"))) {
        processMethod <- "preprocessQuantile"
    }
    if ((processMethod == "auto") &&
        (!compositeCellType %in% c("Blood", "DLPFC"))) {
        processMethod <- "preprocessNoob"
    }
    processMethod <- get(processMethod)
    if ((probeSelect == "auto") && (compositeCellType == "CordBlood")) {
        probeSelect <- "any"
    }
    if ((probeSelect == "auto") && (compositeCellType != "CordBlood")) {
        probeSelect <- "both"
    }

    if (verbose) {
        message("[estimateCellCounts] Combining user data with reference ",
                "(flow sorted) data.\n")
    }
    newpd <- DataFrame(
        sampleNames = c(colnames(rgSet), colnames(referenceRGset)),
        studyIndex = rep(
            x = c("user", "reference"),
            times = c(ncol(rgSet), ncol(referenceRGset))),
        stringsAsFactors = FALSE)
    referencePd <- colData(referenceRGset)
    combinedRGset <- combineArrays(
        object1 = rgSet,
        object2 = referenceRGset,
        outType = "IlluminaHumanMethylation450k")
    colData(combinedRGset) <- newpd
    colnames(combinedRGset) <- newpd$sampleNames
    rm(referenceRGset)

    if (verbose) {
        message("[estimateCellCounts] Processing user and reference data ",
                "together.\n")
    }
    if (compositeCellType == "CordBlood") {
        # NOTE: Here Shan wants to discard probes that they have decided
        #       shouldn't be used, for example multi-mapping probes. This is
        #       done by only using probes with names in the comptable.
        #       This is kind of ugly, and dataset dependent.
        combinedMset <- processMethod(combinedRGset, verbose = subverbose)
        compTable <- get(paste0(referencePkg, ".compTable"))
        combinedMset <- combinedMset[
            which(rownames(combinedMset) %in% rownames(compTable)),]
    } else {
        combinedMset <- processMethod(combinedRGset)
    }
    rm(combinedRGset)

    # Extract normalized reference data
    referenceMset <- combinedMset[, combinedMset$studyIndex == "reference"]
    colData(referenceMset) <- as(referencePd, "DataFrame")
    mSet <- combinedMset[, combinedMset$studyIndex == "user"]
    colData(mSet) <- as(colData(rgSet), "DataFrame")
    rm(combinedMset)

    if (verbose) {
        message("[estimateCellCounts] Picking probes for composition ",
                "estimation.\n")
    }
    compData <- pickCompProbes(
        mSet = referenceMset,
        cellTypes = cellTypes,
        compositeCellType = compositeCellType,
        probeSelect = probeSelect)
    coefs <- compData$coefEsts
    # TODO: Shouldn't be necessary to rm() anything
    rm(referenceMset)

    if (verbose) message("[estimateCellCounts] Estimating composition.\n")
    counts <- projectCellType(getBeta(mSet)[rownames(coefs), ], coefs)
    rownames(counts) <- colnames(rgSet)

    if (meanPlot) {
        smeans <- compData$sampleMeans
        smeans <- smeans[order(names(smeans))]
        sampleMeans <- c(
            colMeans2(
                x = getBeta(mSet),
                rows = match(rownames(coefs), rownames(mSet))),
            smeans)
        sampleColors <- c(
            rep(1, ncol(mSet)),
            1 + as.numeric(factor(names(smeans))))
        plot(sampleMeans, pch = 21, bg = sampleColors)
        legend("bottomleft",
               c("blood", levels(factor(names(smeans)))),
               col = 1:7,
               pch = 15)
    }
    if (returnAll) {
        return(list(
            counts = counts,
            compTable = compData$compTable,

            normalizedData = mSet))
    } else {
        counts
    }
}

library(minfiData)
library(digest)

GMsetEx <- mapToGenome(MsetEx)
gr.cor <- createCorMatrix(GMsetEx, res=500*1000)
set.seed(456)
gr.ab <- extractAB(gr.cor)

digest_compartments <- list(cor.matrix = minfi:::.digestMatrix(gr.cor$cor.matrix),
                           pc = minfi:::.digestVector(gr.ab$pc, digits = 2))
save(digest_compartments, file = "../unitTests/digest_compartments.rda")

sessionInfo()
rm(list = ls())

# ------------------------------------------------------------------------------
# DelayedArray tests
#

library(DelayedArray)

assays(MsetEx) <- endoapply(assays(MsetEx), DelayedArray)

GMsetEx <- mapToGenome(MsetEx)
gr.cor <- createCorMatrix(GMsetEx, res=500*1000)
set.seed(456)
gr.ab <- extractAB(gr.cor)

digest_compartments <- list(cor.matrix = minfi:::.digestMatrix(gr.cor$cor.matrix),
                            pc = minfi:::.digestVector(gr.ab$pc, digits = 2))
save(digest_compartments, file = "../unitTests/digest_compartments.DelayedArray.rda")

sessionInfo()
rm(list = ls())

# Internal functions -----------------------------------------------------------

.qnormStratifiedHelper <- function(mat, probeType, regionType) {
    # Check inputs
    if (ncol(mat) == 1) return(mat)
    if (length(probeType) != length(regionType)) {
        stop("length of 'probeType' and 'regionType' needs to be the same.")
    }
    if (nrow(mat) != length(probeType)) {
        stop("'mat' needs to have as many rows as entries in 'probeType'")
    }

    # Quantile normalize probes in each region type
    regionTypes <- unique(regionType)
    for (i in seq_along(regionTypes)) {
        inRegion <- (regionType == regionTypes[i])
        Index1 <- which(inRegion & probeType == "I")
        Index2 <- which(inRegion & probeType == "II")
        mat[Index2,] <- normalize.quantiles(mat[Index2, ])
        # NOTE: The following code is easy to understand, but we are not using
        #       it because we want the results to stay stable. It gives almost
        #       the same results as the code below:
        # mat[Index1,] <- normalize.quantiles.use.target(
        #     x = mat[Index1, , drop = FALSE],
        #     target = mat[Index2, 1, drop = TRUE])
        target <- approx(
            x = seq(along = Index2),
            y = sort(mat[Index2, 1]),
            xout = seq(1, length(Index2), length.out = length(Index1)))$y
        mat[Index1, ] <- normalize.quantiles.use.target(
            x = mat[Index1, , drop = FALSE],
            target = target)
    }

    mat
}

.qnormStratified <- function(mat, auIndex, xIndex, yIndex, sex = NULL,
                             probeType, regionType) {
    # Normalize probes on the autosomes
    mat[auIndex, ] <- .qnormStratifiedHelper(
        mat = mat[auIndex, ],
        probeType = probeType[auIndex],
        regionType = regionType[auIndex])

    # Normalize probes on the sex chromosomes
    if (is.null(sex)) {
        # NOTE: If sex if not given, we will assume all samples have same sex
        sexIndexes <- list(seq_len(ncol(mat)))
    } else {
        sexIndexes <- split(seq_len(ncol(mat)), sex)

    }
    sexIndexes <- sexIndexes[lengths(sexIndexes) > 1]
    for (idxes in sexIndexes) {
        mat[c(xIndex, yIndex), idxes] <- .qnormStratifiedHelper(
            mat = mat[c(xIndex, yIndex), idxes, drop = FALSE],
            probeType = probeType[c(xIndex, yIndex)],
            regionType = regionType[c(xIndex, yIndex)])
    }

    mat
}

# Quantile normalize but do chrX and chrY separately by sex
.qnormNotStratified <- function(mat, auIndex, xIndex, yIndex, sex = NULL) {
    # Normalize probes on the autosomes
    mat[auIndex,] <- normalize.quantiles(mat[auIndex, ])

    # Normalize probes on the sex chromosomes
    if (is.null(sex)) {
        sexIndexes <- list(U = seq_len(ncol(mat)))
    } else {
        sexIndexes <- split(seq_len(ncol(mat)), sex)
    }
    for (i in seq_along(sexIndexes)) {
        Index <- sexIndexes[[i]]
        if (length(Index) > 1) {
            mat[c(xIndex, yIndex), Index] <- normalize.quantiles(
                x = mat[c(xIndex, yIndex), Index])
        } else {
            warning(
                sprintf("Only one sample of sex: %s. Not normalizing the sex chromosomes for that sample.",
                        names(sexIndexes)[i]))
        }
    }

    mat
}

# Exported functions -----------------------------------------------------------

preprocessQuantile <- function(object, fixOutliers = TRUE,
                               removeBadSamples = FALSE, badSampleCutoff = 10.5,
                               quantileNormalize = TRUE, stratified = TRUE,
                               mergeManifest = FALSE, sex = NULL,
                               verbose = TRUE) {
    # Check inputs
    .isMatrixBackedOrStop(object, "preprocessQuantile")
    # NOTE (Kasper): We could use [Genomic]MethylSet if the object has been
    #                processed with preprocessRaw()
    # TODO: Add the above support?
    if (!(is(object, "RGChannelSet") ||
          is(object, "MethylSet") ||
          is(object, "GenomicMethylSet"))) {
        stop("object must be of class 'RGChannelSet' or '[Genomic]MethylSet'")
    }
    if ((is(object, "MethylSet") || is(object, "GenomicMethylSet")) &&
        (is.na(preprocessMethod(object)["rg.norm"]) ||
         preprocessMethod(object)["rg.norm"] !=
         "Raw (no normalization or bg correction)")) {
        warning("preprocessQuantile has only been tested with 'preprocessRaw'")
    }
    if (.is27k(object) && stratified) {
        stratified <- FALSE
        warning("The stratification option is not available for 27k arrays.")
    }

    # Map to genome
    if (verbose) message("[preprocessQuantile] Mapping to genome.")
    object <- mapToGenome(object, mergeManifest = mergeManifest)

    # Get sex
    if (is.null(sex)) {
        object <- addSex(object)
        sex <- colData(object)$predictedSex
    } else {
        sex <- .checkSex(sex)
    }

    # Fix outliers
    if (fixOutliers) {
        if (verbose) message("[preprocessQuantile] Fixing outliers.")
        object <- fixMethOutliers(object)
    }

    # Run QC
    qc <- getQC(object)
    meds <- (qc$uMed + qc$mMed) / 2
    keepIndex <- which(meds > badSampleCutoff)
    if (length(keepIndex) == 0 && removeBadSamples) {
        stop("All samples found to be bad")
    }
    if (length(keepIndex) < ncol(object) && removeBadSamples) {
        if (verbose) {
            message(
                sprintf("[preprocessQuantile] Found and removed %s bad samples",
                        ncol(object) - length(keepIndex)))
        }
        object <- object[, keepIndex]
    }

    xIndex <- which(seqnames(object) == "chrX")
    yIndex <- which(seqnames(object) == "chrY")
    auIndex <- which(seqnames(object) %in% paste0("chr", 1:22))

    # Quantile normalize Meth and Unmeth
    U <- getUnmeth(object)
    M <- getMeth(object)
    if (quantileNormalize) {
        if (verbose) message("[preprocessQuantile] Quantile normalizing.")
        if (!stratified) {
            U <- .qnormNotStratified(
                mat = U,
                auIndex = auIndex,
                xIndex = xIndex,
                yIndex = yIndex,
                sex = sex)
            if (!is.null(getAutoRealizationBackend())) {
                U <- realize(U)
            }
            M <- .qnormNotStratified(
                mat = M,
                auIndex = auIndex,
                xIndex = xIndex,
                yIndex = yIndex,
                sex = sex)
            if (!is.null(getAutoRealizationBackend())) {
                M <- realize(M)
            }
        } else {
            probeType <- getProbeType(object)
            regionType <- getIslandStatus(object)
            regionType[regionType %in% c("Shelf", "OpenSea")] <- "Far"
            U <- .qnormStratified(
                mat = U,
                auIndex = auIndex,
                xIndex = xIndex,
                yIndex = yIndex,
                sex = sex,
                probeType = probeType,
                regionType = regionType)
            if (!is.null(getAutoRealizationBackend())) {
                U <- realize(U)
            }
            M <- .qnormStratified(
                mat = M,
                auIndex = auIndex,
                xIndex = xIndex,
                yIndex = yIndex,
                sex = sex,
                probeType = probeType,
                regionType = regionType)
            if (!is.null(getAutoRealizationBackend())) {
                M <- realize(M)
            }
        }
    }

    # Construct output GenomicRatioSet
    preprocessMethod <- c(
        mu.norm = "preprocessQuantile",
        preprocessMethod(object))
    GenomicRatioSet(
        gr = granges(object),
        Beta = NULL,
        M = log2(M / U),
        CN = log2(U + M),
        colData = colData(object),
        annotation = annotation(object),
        preprocessMethod = preprocessMethod)
}

test_subsetRGChannelSet <- function() {
##     stopifnot(require(minfiData))
##     stopifnot(require(DelayedArray))
##     exclude <- c("cg02004872", #  I-grn
##                  "cg00050873", # I-red
##                  "cg00035864") # II

##     # Original tests
##     rgSetSub <- subsetByLoci(RGsetEx, excludeLoci = exclude)
##     checkEquals(dim(preprocessRaw(rgSetSub)),
##                 dim(MsetEx[!rownames(MsetEx) %in% exclude, ]))
##     checkEquals(dim(preprocessNoob(rgSetSub)),
##                 dim(MsetEx[!rownames(MsetEx) %in% exclude, ]))
##     # TODO: Why as.numeric()?
##     checkEquals(as.numeric(dim(preprocessFunnorm(rgSetSub))),
##                 as.numeric(dim(MsetEx[!rownames(MsetEx) %in% exclude, ])))

##     # Testing with DelayedArray-backed objects
##     RGsetEx <- realize(RGsetEx)
##     MsetEx <- realize(MsetEx)
##     rgSetSub <- subsetByLoci(RGsetEx, excludeLoci = exclude)
##     checkEquals(dim(preprocessRaw(rgSetSub)),
##                 dim(MsetEx[!rownames(MsetEx) %in% exclude, ]))
##     checkEquals(dim(preprocessNoob(rgSetSub)),
##                 dim(MsetEx[!rownames(MsetEx) %in% exclude, ]))
##     # TODO: Uncomment once preprocessFunnorm() supports DelayedArray-backed
##     #       minfi objects
##     checkException(preprocessFunnorm(rgSetSub), silent = TRUE)
##     # checkEquals(as.numeric(dim(preprocessFunnorm(rgSetSub))),
##     #             as.numeric(dim(MsetEx[!rownames(MsetEx) %in% exclude, ])))
    checkTrue(TRUE)
}

test_preprocessRaw <- function() {
    stopifnot(require(minfiData))
    stopifnot(require(digest))
    stopifnot(require(DelayedArray))
    load(file.path(
        path.package("minfi"),
        "unitTests",
        "digest_preprocessRaw.rda"))

    # Original tests
    Mset <- preprocessRaw(RGsetEx)
    checkEquals(digest_preprocessRaw$Meth,
                minfi:::.digestMatrix(getMeth(Mset)))
    checkEquals(digest_preprocessRaw$Unmeth,
                minfi:::.digestMatrix(getUnmeth(Mset)))

    # Testing with DelayedArray-backed objects
    RGsetEx <- realize(RGsetEx)
    Mset <- preprocessRaw(RGsetEx)
    checkEquals(digest_preprocessRaw$Meth,
                minfi:::.digestMatrix(as.matrix(getMeth(Mset))))
    checkEquals(digest_preprocessRaw$Unmeth,
                minfi:::.digestMatrix(as.matrix(getUnmeth(Mset))))
}

test_preprocessIllumina <- function() {
    ## stopifnot(require(minfiData))
    ## stopifnot(require(digest))
    ## stopifnot(require(DelayedArray))
    ## load(file.path(path.package("minfi"),
    ##                "unitTests",
    ##                "digest_preprocessIllumina.rda"))

    ## # Original tests
    ## Mset <- preprocessIllumina(RGsetEx)
    ## checkEquals(digest_preprocessIllumina$Meth,
    ##             minfi:::.digestMatrix(getMeth(Mset)))
    ## checkEquals(digest_preprocessIllumina$Unmeth,
    ##             minfi:::.digestMatrix(getUnmeth(Mset)))

    ## # Testing with DelayedArray-backed objects
    ## RGsetEx <- realize(RGsetEx)
    ## Mset <- preprocessIllumina(RGsetEx)
    ## checkEquals(digest_preprocessIllumina$Meth,
    ##             minfi:::.digestMatrix(as.matrix(getMeth(Mset))))
    ## checkEquals(digest_preprocessIllumina$Unmeth,
    ##             minfi:::.digestMatrix(as.matrix(getUnmeth(Mset))))
    checkTrue(TRUE)
}

test_preprocessSWAN <- function() {
    ## stopifnot(require(minfiData))
    ## stopifnot(require(digest))
    ## stopifnot(require(DelayedArray))
    ## load(file.path(path.package("minfi"),
    ##                "unitTests",
    ##                "digest_preprocessSWAN.rda"))

    ## # Original tests
    ## set.seed(456)
    ## Mset <- preprocessSWAN(RGsetEx)
    ## checkEquals(digest_preprocessSWAN$Meth,
    ##             minfi:::.digestMatrix(getMeth(Mset)))
    ## checkEquals(digest_preprocessSWAN$Unmeth,
    ##             minfi:::.digestMatrix(getUnmeth(Mset)))

    ## # Testing with DelayedArray-backed objects
    ## RGsetEx <- realize(RGsetEx)
    ## set.seed(456)
    ## Mset <- preprocessSWAN(RGsetEx)
    ## checkEquals(digest_preprocessSWAN$Meth,
    ##             minfi:::.digestMatrix(as.matrix(getMeth(Mset))))
    ## checkEquals(digest_preprocessSWAN$Unmeth,
    ##             minfi:::.digestMatrix(as.matrix(getUnmeth(Mset))))
    checkTrue(TRUE)
}

test_preprocessQuantile <- function() {
    ## stopifnot(require(minfiData))
    ## stopifnot(require(digest))
    ## stopifnot(require(DelayedArray))
    ## load(file.path(path.package("minfi"),
    ##                "unitTests",
    ##                "digest_preprocessQuantile.rda"))

    ## # Original tests
    ## GRset <- preprocessQuantile(MsetEx)
    ## checkEquals(digest_preprocessQuantile$M,
    ##             minfi:::.digestMatrix(getM(GRset)))
    ## checkEquals(digest_preprocessQuantile$CN,
    ##             minfi:::.digestMatrix(getCN(GRset)))

    ## # Testing with DelayedArray-backed objects
    ## MsetEx <- realize(MsetEx)
    ## checkException(preprocessQuantile(MsetEx), silent = TRUE)
    ## # TODO: Uncomment once preprocessQuantile() supports DelayedArray-backed
    ## #       minfi objects
    ## # checkEquals(digest_preprocessQuantile$M,
    ## #             minfi:::.digestMatrix(as.matrix(getM(GRset))))
    ## # checkEquals(digest_preprocessQuantile$CN,
    ## #             minfi:::.digestMatrix(as.matrix(getCN(GRset))))
    checkTrue(TRUE)
}

test_preprocessNoob <- function() {
    ## stopifnot(require(minfiData))
    ## stopifnot(require(digest))
    ## stopifnot(require(DelayedArray))
    ## load(file.path(path.package("minfi"),
    ##                "unitTests",
    ##                "digest_preprocessNoob.rda"))

    ## # Original tests
    ## Mset <- preprocessNoob(RGsetEx)
    ## checkEquals(digest_preprocessNoob$Meth,
    ##             minfi:::.digestMatrix(getMeth(Mset)))
    ## checkEquals(digest_preprocessNoob$Unmeth,
    ##             minfi:::.digestMatrix(getUnmeth(Mset)))

    ## # Testing with DelayedArray-backed objects
    ## RGsetEx <- realize(RGsetEx)
    ## Mset <- preprocessNoob(RGsetEx)
    ## checkEquals(digest_preprocessNoob$Meth,
    ##             minfi:::.digestMatrix(as.matrix(getMeth(Mset))))
    ## checkEquals(digest_preprocessNoob$Unmeth,
    ##             minfi:::.digestMatrix(as.matrix(getUnmeth(Mset))))
    checkTrue(TRUE)
}

test_preprocessFunnorm <- function() {
    ## stopifnot(require(minfiData))
    ## stopifnot(require(digest))
    ## stopifnot(require(DelayedArray))
    ## load(file.path(path.package("minfi"),
    ##                "unitTests",
    ##                "digest_preprocessFunnorm.rda"))

    ## # Original tests
    ## GRset <- preprocessFunnorm(RGsetEx)
    ## checkEquals(digest_preprocessFunnorm$M,
    ##             minfi:::.digestMatrix(getM(GRset)))
    ## checkEquals(digest_preprocessFunnorm$CN,
    ##             minfi:::.digestMatrix(getCN(GRset)))

    ## # Testing with DelayedArray-backed objects
    ## RGsetEx <- realize(RGsetEx)
    ## checkException(GRset <- preprocessFunnorm(RGsetEx), silent = TRUE)

    ## # TODO: Uncomment once preprocessFunnorm() supports DelayedArray-backed
    ## #       minfi objects
    ## # checkEquals(digest_preprocessFunnorm$M,
    ## #             minfi:::.digestMatrix(getM(GRset)))
    ## # checkEquals(digest_preprocessFunnorm$CN,
    ## #             minfi:::.digestMatrix(getCN(GRset)))
    checkTrue(TRUE)
}

plotBetasByType <- function(data, probeTypes = NULL, legendPos = "top",
                            colors = c("black", "red", "blue"), main = "",
                            lwd = 3, cex.legend = 1) {
    # Check inputs
    if (!(is(data, "MethylSet") || is.matrix(data) || is.vector(data) ||
          is(data, "DelayedMatrix"))) {
        stop("'data' needs to be a 'MethylSet' or a matrix, 'DelayedMatrix', ",
             "or vector of beta values")
    }
    if (!is.vector(data)) {
        if (ncol(data) > 1) stop("'data' must only contain one sample")
    }
    if (!is(data, "MethylSet")) {
        r <- range(data, na.rm = TRUE)
        if (!(r[1] >= 0 && r[2] <= 1)) {
            stop("'data' should be beta values in the range [0, 1]")
        }
    }
    if (is(data, "MethylSet")) {
        if (is.null(probeTypes)) {
            typeI <- getProbeInfo(data, type = "I")[, c("Name", "nCpG")]
            typeII <- getProbeInfo(data, type = "II")[, c("Name", "nCpG")]
            probeTypes <- rbind(typeI, typeII)
            probeTypes$Type <- rep(
                x = c("I", "II"),
                times = c(nrow(typeI), nrow(typeII)))
        }
    }
    if (!all(c("Name", "Type") %in% colnames(probeTypes))) {
        stop("'probeTypes' must be a data.frame with a column 'Name' of probe ",
             "IDs and a column 'Type' indicating their design type")
    }

    # Construct 1-column matrix of beta values
    if (is(data, "MethylSet")) {
        betas <- as.matrix(getBeta(data))
    } else if (is.vector(data)) {
        betas <- matrix(
            data = data,
            ncol = 1,
            dimnames = list(names(data), NULL))
    } else {
        betas <- as.matrix(data)
    }

    # Compute densities
    betas <- betas[!is.na(betas), , drop = FALSE]
    n_probes <- nrow(betas)
    n_type1 <- sum(probeTypes$Type == "I" &
                       probeTypes$Name %in% rownames(betas))
    n_type2 <- sum(probeTypes$Type == "II" &
                       probeTypes$Name %in% rownames(betas))
    betas_density <- density(betas)
    typeI_density <- suppressWarnings(
        density(
            x = betas[rownames(betas) %in%
                          probeTypes$Name[probeTypes$Type == "I"]],
            weights = rep(1 / n_probes, n_type1)))
    typeII_density <- suppressWarnings(
        density(
            x = betas[rownames(betas) %in%
                          probeTypes$Name[probeTypes$Type == "II"]],
            weights = rep(1 / n_probes, n_type2)))

    # Plot densities
    plot(
        x = betas_density,
        main = main,
        xlab = "Beta values",
        ylim = c(0, max(betas_density$y)),
        lwd = lwd,
        col = colors[1])
    lines(
        x = typeI_density,
        col = colors[2],
        lty = 2,
        lwd = lwd)
    lines(
        x = typeII_density,
        col = colors[3],
        lty = 2,
        lwd = lwd)
    legend(
        x = legendPos,
        y = c("All probes", "Infinium I", "Infinium II"),
        col = colors,
        lwd = lwd,
        bg = "white",
        cex = cex.legend,
        lty = c(1, 2, 2))
}

# Exported classes -------------------------------------------------------------

setClass(
    "MethylSet",
    representation(preprocessMethod = "character", annotation = "character"),
    contains = "SummarizedExperiment"
)

# Validity methods -------------------------------------------------------------

setValidity("MethylSet", function(object) {
    msg <- validMsg(NULL, .checkAssayNames(object, c("Meth", "Unmeth")))
    if (is.null(msg)) TRUE else msg
})


# Exported functions -----------------------------------------------------------

MethylSet <- function(Meth = new("matrix"), Unmeth = new("matrix"),
                      annotation = "", preprocessMethod = "", ...) {
    # Check rownames, colnames
    assays <- SimpleList(Meth = Meth, Unmeth = Unmeth)
    new("MethylSet",
        SummarizedExperiment(assays = assays, ...),
        annotation = annotation,
        preprocessMethod = preprocessMethod
    )
}

dropMethylationLoci <- function(object, dropRS = TRUE, dropCH = TRUE) {
    stopifnot(class(object) %in% c(
        "MethylSet", "GenomicMethylSet", "RatioSet", "GenomicRatioSet"))
    dropRegEx <- ""
    if (dropRS) dropRegEx <- c(dropRegEx, "^rs")
    if (dropCH) dropRegEx <- c(dropRegEx, "^ch\\.")
    dropRegEx <- dropRegEx[nchar(dropRegEx) > 0]
    if (length(dropRegEx) == 0) return(object)
    dropRegEx <- sprintf("(%s)", paste(dropRegEx, collapse = "|"))
    whDrop <- grep(dropRegEx, rownames(object))
    if (length(whDrop) == 0) return(object)
    object[-whDrop, ]
}


# Exported methods -------------------------------------------------------------

setMethod("show", signature(object = "MethylSet"), function(object) {
    callNextMethod()
    .show.annotation(annotation(object))
    .show.preprocessMethod(preprocessMethod(object))
})

setMethod("getMeth", signature(object = "MethylSet"), function(object) {
    assay(object, "Meth")
})

setMethod("getUnmeth", signature(object = "MethylSet"), function(object) {
    assay(object, "Unmeth")
})

setMethod(
    "getBeta",
    signature(object = "MethylSet"),
    function(object, type = "", offset = 0, betaThreshold = 0) {
        if (type == "Illumina") offset <- 100
        .betaFromMethUnmeth(
            Meth = getMeth(object),
            Unmeth = getUnmeth(object),
            offset = offset,
            betaThreshold = betaThreshold)
    }
)

setMethod(
    "getM",
    signature(object = "MethylSet"),
    function(object, type = "", ...) {
        if (type == "") return(log2(getMeth(object) / getUnmeth(object)))
        if (type == "beta" || type == "Beta") logit2(getBeta(object, ...))
    }
)

setMethod("getCN", signature(object = "MethylSet"), function(object, ...) {
    log2(getMeth(object) + getUnmeth(object))
})

setMethod(
    "preprocessMethod",
    signature(object = "MethylSet"),
    function(object) object@preprocessMethod
)

setMethod("annotation", signature(object = "MethylSet"), function(object) {
    object@annotation
})

setReplaceMethod(
    "annotation",
    signature(object = "MethylSet"),
    function(object, value) {
        object@annotation <- value
        object
    }
)

setMethod(
    "getManifest",
    signature(object = "MethylSet"),
    function(object) {
        maniString <- .getManifestString(object@annotation)
        if (!require(maniString, character.only = TRUE))
            stop(sprintf("cannot load manifest package %s", maniString))
        get(maniString)
    }
)

setMethod(
    "mapToGenome",
    signature(object = "MethylSet"),
    function(object, mergeManifest = FALSE) {
        gr <- getLocations(
            object = object,
            mergeManifest = mergeManifest,
            orderByLocation = TRUE,
            lociNames = rownames(object))
        object <- object[names(gr), ]
        GenomicMethylSet(
            gr = gr,
            Meth = getMeth(object),
            Unmeth = getUnmeth(object),
            colData = colData(object),
            preprocessMethod = preprocessMethod(object),
            annotation = annotation(object),
            metadata = metadata(object))
    }
)

setMethod(
    "updateObject",
    signature(object = "MethylSet"),
    function(object, ..., verbose=FALSE) {
        if (verbose) message("updateObject(object = 'MethylSet')")
        if ("assayData" %in% names(getObjectSlots(object))) {
            # This is an ExpressionSet based object
            object <- MethylSet(
                Meth = getObjectSlots(object)[["assayData"]][["Meth"]],
                Unmeth = getObjectSlots(object)[["assayData"]][["Unmeth"]],
                colData = getObjectSlots(
                    getObjectSlots(object)[["phenoData"]])[["data"]],
                annotation = getObjectSlots(object)[["annotation"]],
                preprocessMethod = getObjectSlots(object)[["preprocessMethod"]])
        } else {
            object <- callNextMethod()
        }
        object
    }
)


setMethod(
    "ratioConvert",
    signature(object = "MethylSet"),
    function(object, what = c("beta", "M", "both"), keepCN = TRUE, ...) {
        what <- match.arg(what)
        if (what == "beta") {
            Beta <- getBeta(object, ...)
            M <- NULL
        }
        if (what == "M") {
            Beta <- NULL
            M <- getM(object, ...)
        }
        if (what == "both") {
            Beta <- getBeta(object, ...)
            M <- getM(object, ...)
        }
        CN <- if (keepCN) getCN(object) else NULL
        RatioSet(
            Beta = Beta,
            M = M,
            CN = CN,
            colData = colData(object),
            annotation = annotation(object),
            preprocessMethod = preprocessMethod(object),
            rowData = rowData(object),
            metadata = metadata(object))
    }
)

setMethod(
    "combine",
    signature(x = "MethylSet", y = "MethylSet"),
    function(x, y, ...) {
        colDataFix <- .harmonizeDataFrames(
            .pDataFix(colData(x)),
            .pDataFix(colData(y)))
        colData(x) <- colDataFix$x
        colData(y) <- colDataFix$y
        cbind(x, y)
    }
)

# Internal functions -----------------------------------------------------------

read.manifest.Mammal <- function(file, type = 2) {
    # NOTE: As is, requires grep
    control.line <- system(
        sprintf("grep -n \\\\[Controls\\\\] %s", file), intern = TRUE)
    control.line <- as.integer(sub(":.*", "", control.line))
    stopifnot(length(control.line) == 1 &&
                  is.integer(control.line) &&
                  !is.na(control.line))
    if(type == 1) {
        assay.line <- system(
            sprintf("grep -n \\\\[Assay\\\\] %s", file), intern = TRUE)
        assay.line <- as.integer(sub(":.*", "", assay.line))
        stopifnot(length(assay.line) == 1 &&
                  is.integer(assay.line) &&
                  !is.na(assay.line))
    } else {
        assay.line  <- 0
    }
    colNames <- readLines(file, n = assay.line + 1L)[assay.line + 1L]
    colNames <- strsplit(colNames, ",")[[1]]
    colClasses <- rep("character", length(colNames))
    names(colClasses) <- colNames
    names(colClasses) <- make.names(names(colClasses))
    colClasses[c("MAPINFO")] <- "integer"
    if(type == 1) {
        colClasses <- c(colClasses, drop = "logical")
    }
    manifest <- read.table(
        file = file,
        header = FALSE,
        col.names = names(colClasses),
        sep = ",",
        comment.char = "",
        quote = "",
        skip = assay.line + 1L,
        colClasses = colClasses,
        nrows = control.line - assay.line - 2L)
    manifest$drop <- NULL
    if(type == 2) {
        names(manifest)[c(1,2,30)] <- c("Name", "Internal_Name", "Internal_Name2")
    }
    
    manifest$AddressA_ID <- gsub("^0*", "", manifest$AddressA_ID)
    manifest$AddressB_ID <- gsub("^0*", "", manifest$AddressB_ID)
    
    TypeI <- manifest[
        manifest$Infinium_Design_Type == "I",
        c("Name", "AddressA_ID", "AddressB_ID", "Color_Channel", "Next_Base",
          "AlleleA_ProbeSeq", "AlleleB_ProbeSeq")]
    names(TypeI)[c(2, 3, 4, 5, 6, 7)] <- c(
        "AddressA", "AddressB", "Color", "NextBase", "ProbeSeqA", "ProbeSeqB")
    
    TypeI <- as(TypeI, "DataFrame")
    TypeI$ProbeSeqA <- DNAStringSet(TypeI$ProbeSeqA)
    TypeI$ProbeSeqB <- DNAStringSet(TypeI$ProbeSeqB)
    TypeI$NextBase <- DNAStringSet(TypeI$NextBase)
    TypeI$nCpG <- as.integer(
        oligonucleotideFrequency(TypeI$ProbeSeqB, width = 2)[, "CG"] - 1L)
    TypeI$nCpG[TypeI$nCpG < 0] <- 0L
    TypeSnpI <- TypeI[grep("^rs", TypeI$Name), ]
    TypeI <- TypeI[-grep("^rs", TypeI$Name), ]

    TypeII <- manifest[
        manifest$Infinium_Design_Type == "II",
        c("Name", "AddressA_ID", "AlleleA_ProbeSeq")]
    names(TypeII)[c(2,3)] <- c("AddressA", "ProbeSeqA")
    TypeII <- as(TypeII, "DataFrame")
    TypeII$ProbeSeqA <- DNAStringSet(TypeII$ProbeSeqA)
    TypeII$nCpG <- as.integer(letterFrequency(TypeII$ProbeSeqA, letters = "R"))
    TypeII$nCpG[TypeII$nCpG < 0] <- 0L
    TypeSnpII <- TypeII[grep("^rs", TypeII$Name), ]
    TypeII <- TypeII[-grep("^rs", TypeII$Name), ]

    if(type == 1) {
        controls <- read.table(
            file = file,
            skip = control.line,
            sep = ",",
            comment.char = "",
            quote = "",
            colClasses = c(rep("character", 5)))[, 1:5]
        TypeControl <- controls[, 1:4]
        names(TypeControl) <- c("Address", "Type", "Color", "ExtendedType")
        TypeControl <- as(TypeControl, "DataFrame")
    } else {
        controls <- read.table(
            file = file,
            skip = control.line,
            sep = ",",
            comment.char = "",
            quote = "",
            colClasses = c(rep("character", 4)))[, 1:4]
        TypeControl <- controls[, 1:4]
        names(TypeControl) <- c("Address", "Type", "Color", "ExtendedType")
        TypeControl$Type  <- toupper(TypeControl$Type)
        TypeControl <- as(TypeControl, "DataFrame")
    }
        
    
    list(
        manifestList = list(
            TypeI = TypeI,
            TypeII = TypeII,
            TypeControl = TypeControl,
            TypeSnpI = TypeSnpI,
            TypeSnpII = TypeSnpII),
        manifest = manifest,
        controls = controls)
}

read.manifest.Allergy <- function(file) {
    # NOTE: As is, requires grep
    control.line <- system(
        sprintf("grep -n \\\\[Controls\\\\] %s", file), intern = TRUE)
    control.line <- as.integer(sub(":.*", "", control.line))
    stopifnot(length(control.line) == 1 &&
                  is.integer(control.line) &&
                  !is.na(control.line))
    assay.line <- system(
        sprintf("grep -n \\\\[Assay\\\\] %s", file), intern = TRUE)
    assay.line <- as.integer(sub(":.*", "", assay.line))
    stopifnot(length(assay.line) == 1 &&
                  is.integer(assay.line) &&
                  !is.na(assay.line))
    colNames <- readLines(file, n = assay.line + 1L)[assay.line + 1L]
    colNames <- strsplit(colNames, ",")[[1]]
    colClasses <- rep("character", length(colNames))
    names(colClasses) <- colNames
    names(colClasses) <- make.names(names(colClasses))
    colClasses[c("MAPINFO")] <- "integer"
    manifest <- read.table(
        file = file,
        header = FALSE,
        col.names = names(colClasses),
        sep = ",",
        comment.char = "",
        quote = "",
        skip = assay.line + 1L,
        colClasses = colClasses,
        nrows = control.line - assay.line - 2L)
    manifest$AddressA_ID <- gsub("^0*", "", manifest$AddressA_ID)
    manifest$AddressB_ID <- gsub("^0*", "", manifest$AddressB_ID)
    TypeI <- manifest[
        manifest$Infinium_Design_Type == "I",
        c("IlmnID", "AddressA_ID", "AddressB_ID", "Color_Channel", "Next_Base",
          "AlleleA_ProbeSeq", "AlleleB_ProbeSeq")]
    names(TypeI)[c(1, 2, 3, 4, 5, 6, 7)] <- c("Name",
        "AddressA", "AddressB", "Color", "NextBase", "ProbeSeqA", "ProbeSeqB")
    TypeI <- as(TypeI, "DataFrame")
    TypeI$ProbeSeqA <- DNAStringSet(TypeI$ProbeSeqA)
    TypeI$ProbeSeqB <- DNAStringSet(TypeI$ProbeSeqB)
    TypeI$NextBase <- DNAStringSet(TypeI$NextBase)
    TypeI$nCpG <- as.integer(
        oligonucleotideFrequency(TypeI$ProbeSeqB, width = 2)[, "CG"] - 1L)
    TypeI$nCpG[TypeI$nCpG < 0] <- 0L
    TypeSnpI <- TypeI[grep("^rs", TypeI$Name), ]
    TypeI <- TypeI[grep("^rs", TypeI$Name, invert=TRUE), ]

    TypeII <- manifest[
        manifest$Infinium_Design_Type == "II",
        c("IlmnID", "AddressA_ID", "AlleleA_ProbeSeq")]
    names(TypeII)[c(1,2,3)] <- c("Name", "AddressA", "ProbeSeqA")
    TypeII <- as(TypeII, "DataFrame")
    TypeII$ProbeSeqA <- DNAStringSet(TypeII$ProbeSeqA)
    TypeII$nCpG <- as.integer(letterFrequency(TypeII$ProbeSeqA, letters = "R"))
    TypeII$nCpG[TypeII$nCpG < 0] <- 0L
    TypeSnpII <- TypeII[grep("^rs", TypeII$Name), ]
    TypeII <- TypeII[grep("^rs", TypeII$Name, invert=TRUE), ]

    controls <- read.table(
        file = file,
        skip = control.line,
        sep = ",",
        comment.char = "",
        quote = "",
        colClasses = c(rep("character", 5)))[, 1:5]
    TypeControl <- controls[, 1:4]
    names(TypeControl) <- c("Address", "Type", "Color", "ExtendedType")
    TypeControl <- as(TypeControl, "DataFrame")

    list(
        manifestList = list(
            TypeI = TypeI,
            TypeII = TypeII,
            TypeControl = TypeControl,
            TypeSnpI = TypeSnpI,
            TypeSnpII = TypeSnpII),
        manifest = manifest,
        controls = controls)
}


read.manifest.sesame.Allergy <- function(file) {
    ID_column <- "IlmnID"
    colNames <- readLines(file, n = 1)
    colNames <- strsplit(colNames, ",")[[1]]
    colClasses <- rep("character", length(colNames))
    names(colClasses) <- colNames
    names(colClasses) <- make.names(names(colClasses))
    colClasses[c("MAPINFO")] <- "integer"
    manifest <- read.table(
        file = file,
        header = TRUE,
        sep = ",",
        comment.char = "",
        colClasses = colClasses,
        quote = "")
    names(manifest)[names(manifest) == "Probe_ID"] <- "IlmnID"
    names(manifest)[names(manifest) == "U"] <- "AddressA_ID"
    names(manifest)[names(manifest) == "M"] <- "AddressB_ID"
    manifest$Infinium_Design_Type <- sub("1", "I", manifest$Infinium_Design_Type)
    manifest$Infinium_Design_Type <- sub("2", "II", manifest$Infinium_Design_Type)
    manifest$Color_Channel <- sub("Both", "", manifest$Color_Channel)
    controls.idx <- grep("^ctl_", manifest$IlmnID)
    nocontrols.idx <- grep("^ctl_", manifest$IlmnID, invert = TRUE)
    controls <- manifest[controls.idx,]
    manifest <- manifest[nocontrols.idx,]
    dupNames <- unique(manifest$Name[duplicated(manifest$Name)])
    manifest <- manifest[! manifest$Name %in% dupNames,]
    TypeI <- manifest[
        manifest$Infinium_Design_Type == "I",
        c(ID_column, "AddressA_ID", "AddressB_ID", "Color_Channel", "Next_Base",
          "AlleleA_ProbeSeq", "AlleleB_ProbeSeq")]
    names(TypeI) <- c("Name", "AddressA", "AddressB", "Color", "NextBase", "ProbeSeqA", "ProbeSeqB")
    TypeI$NextBase[is.na(TypeI$NextBase)] <- ""
    TypeI$Color[is.na(TypeI$Color)] <- ""
    TypeI <- as(TypeI, "DataFrame")
    TypeI$ProbeSeqA <- DNAStringSet(TypeI$ProbeSeqA)
    TypeI$ProbeSeqB <- DNAStringSet(TypeI$ProbeSeqB)
    TypeI$NextBase <- DNAStringSet(TypeI$NextBase)
    TypeI$nCpG <- as.integer(
        oligonucleotideFrequency(TypeI$ProbeSeqB, width = 2)[, "CG"] - 1L)
    TypeI$nCpG[TypeI$nCpG < 0] <- 0L
    TypeSnpI <- TypeI[grep("^rs", TypeI$Name), ]
    TypeI <- TypeI[grep("^rs", TypeI$Name, invert = TRUE), ]

    TypeII <- manifest[
        manifest$Infinium_Design_Type == "II",
        c(ID_column, "AddressA_ID", "AlleleA_ProbeSeq")]
    names(TypeII) <- c("Name", "AddressA", "ProbeSeqA")
    TypeII <- as(TypeII, "DataFrame")
    TypeII$ProbeSeqA <- DNAStringSet(TypeII$ProbeSeqA)
    TypeII$nCpG <- as.integer(letterFrequency(TypeII$ProbeSeqA, letters = "R"))
    TypeII$nCpG[TypeII$nCpG < 0] <- 0L
    TypeSnpII <- TypeII[grep("^rs", TypeII$Name), ]
    TypeII <- TypeII[grep("^rs", TypeII$Name, invert=TRUE), ]

    TypeControl <- as(controls[, c("AddressA_ID", "Probe_Type", "IlmnID")], "DataFrame")
    names(TypeControl) <- c("Address", "Type", "ExtendedType")
    TypeControl$Type <- sub("BISULFITE_CONVERSION_", "BISULFITE CONVERSION ", TypeControl$Type)
    TypeControl$Type <- sub("SPECIFICITY_", "SPECIFICITY ", TypeControl$Type)
    TypeControl$Type <- sub("TARGET_REMOVAL", "TARGET REMOVAL", TypeControl$Type)
    TypeControl$ExtendedType <- sub("^ctl_", "", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("Biotin_5K", "Biotin(5K)", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("Biotin_Bkg", "Biotin (Bkg)", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("Biotin_High", "Biotin (High)", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("BS_Conversion_", "BS Conversion ", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("I_", "I-", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("DNP_20K", "DNP(20K)", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("DNP_Bkg", "DNP (Bkg)", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("DNP_High", "DNP (High)", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("Extension_A", "Extension (A)", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("Extension_C", "Extension (C)", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("Extension_G", "Extension (G)", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("Extension_T", "Extension (T)", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("GT_Mismatch_", "GT Mismatch ", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("_MM", " (MM)", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("_PM", " (PM)", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("Hyb_High", "Hyb (High)", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("Hyb_Low", "Hyb (Low)", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("Hyb_Medium", "Hyb (Medium)", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("NP_A", "NP (A)", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("NP_C", "NP (C)", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("NP_G_1", "NP (G) 1", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("NP_G_2", "NP (G) 2", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("NP_G_3", "NP (G) 3", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("NP_G_4", "NP (G) 4", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("NP_G_5", "NP (G) 5", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("NP_G", "NP (G)", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("NP_T", "NP (T)", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("Negative_", "Negative ", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("Specificity_", "Specificity ", TypeControl$ExtendedType)
    TypeControl$ExtendedType <- sub("Target_Removal_", "Target Removal ", TypeControl$ExtendedType)
    
    return(list(manifestList = list(
                    TypeI = TypeI,
                    TypeII = TypeII,
                    TypeControl = TypeControl,
                    TypeSnpI = TypeSnpI,
                    TypeSnpII = TypeSnpII),
                manifest = manifest, controls = controls))
}


read.manifest.EPIC <- function(file) {
    # NOTE: As is, requires grep
    control.line <- system(
        sprintf("grep -n \\\\[Controls\\\\] %s", file), intern = TRUE)
    control.line <- as.integer(sub(":.*", "", control.line))
    stopifnot(length(control.line) == 1 &&
                  is.integer(control.line) &&
                  !is.na(control.line))
    assay.line <- system(
        sprintf("grep -n \\\\[Assay\\\\] %s", file), intern = TRUE)
    assay.line <- as.integer(sub(":.*", "", assay.line))
    stopifnot(length(assay.line) == 1 &&
                  is.integer(assay.line) &&
                  !is.na(assay.line))
    colNames <- readLines(file, n = assay.line + 1L)[assay.line + 1L]
    colNames <- strsplit(colNames, ",")[[1]]
    colClasses <- rep("character", length(colNames))
    names(colClasses) <- colNames
    names(colClasses) <- make.names(names(colClasses))
    colClasses[c("MAPINFO")] <- "integer"
    manifest <- read.table(
        file = file,
        header = FALSE,
        col.names = names(colClasses),
        sep = ",",
        comment.char = "",
        quote = "",
        skip = assay.line + 1L,
        colClasses = colClasses,
        nrows = control.line - assay.line - 2L)
    manifest$AddressA_ID <- gsub("^0*", "", manifest$AddressA_ID)
    manifest$AddressB_ID <- gsub("^0*", "", manifest$AddressB_ID)
    TypeI <- manifest[
        manifest$Infinium_Design_Type == "I",
        c("Name", "AddressA_ID", "AddressB_ID", "Color_Channel", "Next_Base",
          "AlleleA_ProbeSeq", "AlleleB_ProbeSeq")]
    names(TypeI)[c(2, 3, 4, 5, 6, 7)] <- c(
        "AddressA", "AddressB", "Color", "NextBase", "ProbeSeqA", "ProbeSeqB")
    TypeI <- as(TypeI, "DataFrame")
    TypeI$ProbeSeqA <- DNAStringSet(TypeI$ProbeSeqA)
    TypeI$ProbeSeqB <- DNAStringSet(TypeI$ProbeSeqB)
    TypeI$NextBase <- DNAStringSet(TypeI$NextBase)
    TypeI$nCpG <- as.integer(
        oligonucleotideFrequency(TypeI$ProbeSeqB, width = 2)[, "CG"] - 1L)
    TypeI$nCpG[TypeI$nCpG < 0] <- 0L
    TypeSnpI <- TypeI[grep("^rs", TypeI$Name), ]
    TypeI <- TypeI[-grep("^rs", TypeI$Name), ]

    TypeII <- manifest[
        manifest$Infinium_Design_Type == "II",
        c("Name", "AddressA_ID", "AlleleA_ProbeSeq")]
    names(TypeII)[c(2,3)] <- c("AddressA", "ProbeSeqA")
    TypeII <- as(TypeII, "DataFrame")
    TypeII$ProbeSeqA <- DNAStringSet(TypeII$ProbeSeqA)
    TypeII$nCpG <- as.integer(letterFrequency(TypeII$ProbeSeqA, letters = "R"))
    TypeII$nCpG[TypeII$nCpG < 0] <- 0L
    TypeSnpII <- TypeII[grep("^rs", TypeII$Name), ]
    TypeII <- TypeII[-grep("^rs", TypeII$Name), ]

    controls <- read.table(
        file = file,
        skip = control.line,
        sep = ",",
        comment.char = "",
        quote = "",
        colClasses = c(rep("character", 5)))[, 1:5]
    TypeControl <- controls[, 1:4]
    names(TypeControl) <- c("Address", "Type", "Color", "ExtendedType")
    TypeControl <- as(TypeControl, "DataFrame")

    list(
        manifestList = list(
            TypeI = TypeI,
            TypeII = TypeII,
            TypeControl = TypeControl,
            TypeSnpI = TypeSnpI,
            TypeSnpII = TypeSnpII),
        manifest = manifest,
        controls = controls)
}

read.manifest.450k <- function(file) {
    # NOTE: As is, requires grep
    control.line <- system(
        sprintf("grep -n \\\\[Controls\\\\] %s", file), intern = TRUE)
    control.line <- as.integer(sub(":.*", "", control.line))
    stopifnot(length(control.line) == 1 &&
                  is.integer(control.line) &&
                  !is.na(control.line))
    assay.line <- system(
        sprintf("grep -n \\\\[Assay\\\\] %s", file), intern = TRUE)
    assay.line <- as.integer(sub(":.*", "", assay.line))
    stopifnot(length(assay.line) == 1 &&
                  is.integer(assay.line) &&
                  !is.na(assay.line))

    # NOTE: Column headers is in line 8, hardcoded
    colNames <- readLines(file, n = assay.line + 1L)[assay.line + 1L]
    colNames <- strsplit(colNames, ",")[[1]]
    colClasses <- rep("character", length(colNames))
    names(colClasses) <- colNames
    colClasses[c("MAPINFO")] <- "integer"
    manifest <- read.table(
        file = file,
        header = TRUE,
        sep = ",",
        comment.char = "",
        quote = "",
        skip = 7,
        colClasses = colClasses,
        nrows = control.line - 9)
    TypeI <- manifest[
        manifest$Infinium_Design_Type == "I",
        c("Name", "AddressA_ID", "AddressB_ID", "Color_Channel", "Next_Base",
          "AlleleA_ProbeSeq", "AlleleB_ProbeSeq")]
    names(TypeI)[c(2, 3, 4, 5, 6 , 7)] <-
        c("AddressA", "AddressB", "Color", "NextBase", "ProbeSeqA", "ProbeSeqB")
    TypeI <- as(TypeI, "DataFrame")
    TypeI$ProbeSeqA <- DNAStringSet(TypeI$ProbeSeqA)
    TypeI$ProbeSeqB <- DNAStringSet(TypeI$ProbeSeqB)
    TypeI$NextBase <- DNAStringSet(TypeI$NextBase)
    TypeI$nCpG <- as.integer(
        oligonucleotideFrequency(TypeI$ProbeSeqB, width = 2)[, "CG"] - 1L)
    TypeI$nCpG[TypeI$nCpG < 0] <- 0L
    TypeSnpI <- TypeI[grep("^rs", TypeI$Name), ]
    TypeI <- TypeI[-grep("^rs", TypeI$Name), ]

    TypeII <- manifest[
        manifest$Infinium_Design_Type == "II",
        c("Name", "AddressA_ID", "AlleleA_ProbeSeq")]
    names(TypeII)[c(2, 3)] <- c("AddressA", "ProbeSeqA")
    TypeII <- as(TypeII, "DataFrame")
    TypeII$ProbeSeqA <- DNAStringSet(TypeII$ProbeSeqA)
    TypeII$nCpG <- as.integer(letterFrequency(TypeII$ProbeSeqA, letters = "R"))
    TypeII$nCpG[TypeII$nCpG < 0] <- 0L
    TypeSnpII <- TypeII[grep("^rs", TypeII$Name), ]
    TypeII <- TypeII[-grep("^rs", TypeII$Name), ]

    controls <- read.table(
        file = file,
        skip = control.line,
        sep = ",",
        comment.char = "",
        quote = "",
        colClasses = c(rep("character", 5)))[, 1:5]
    TypeControl <- controls[, 1:4]
    names(TypeControl) <- c("Address", "Type", "Color", "ExtendedType")
    TypeControl <- as(TypeControl, "DataFrame")

    list(
        manifestList = list(
            TypeI = TypeI,
            TypeII = TypeII,
            TypeControl = TypeControl,
            TypeSnpI = TypeSnpI,
            TypeSnpII = TypeSnpII),
        manifest = manifest,
        controls = controls)
}


read.manifest.27k <- function(file) {
    # NOTE: As is, requires grep
    control.line <- system(
        sprintf("grep -a -n \\\\[Controls\\\\] %s", file), intern = TRUE)
    control.line <- as.integer(sub(":.*", "", control.line))
    assay.line <- system(
        sprintf("grep -a -n \\\\[Assay\\\\] %s", file), intern = TRUE)
    assay.line <- as.integer(sub(":.*", "", assay.line))

    # NOTE: Column headers is in line 8, hardcoded
    colNames <- tail(readLines(file, n = assay.line + 1), n = 1)
    colNames <- strsplit(colNames, ",")[[1]]
    colClasses <- rep("character", length(colNames))
    names(colClasses) = colNames
    colClasses[c("MAPINFO")] <- "integer"
    manifest <- read.table(
        file = file,
        header = TRUE,
        sep = ",",
        comment.char = "",
        quote = "",
        skip = assay.line,
        colClasses = colClasses,
        nrows = control.line - (assay.line + 1),
        fill = TRUE)
    TypeI <- manifest[
        c("Name", "AddressA_ID", "AddressB_ID", "Color_Channel", "Next_Base",
          "AlleleA_ProbeSeq", "AlleleB_ProbeSeq")]
    TypeI <- TypeI[TypeI$Name != "", ]
    names(TypeI)[c(2, 3, 4, 5, 6, 7)] <- c(
        "AddressA", "AddressB", "Color", "NextBase", "ProbeSeqA", "ProbeSeqB")
    TypeI <- as(TypeI, "DataFrame")
    TypeI$ProbeSeqA <- DNAStringSet(TypeI$ProbeSeqA)
    TypeI$ProbeSeqB <- DNAStringSet(TypeI$ProbeSeqB)
    TypeI$NextBase <- DNAStringSet(TypeI$NextBase)
    TypeI$nCpG <- as.integer(
        oligonucleotideFrequency(TypeI$ProbeSeqB, width = 2)[, "CG"] - 1L)
    TypeI$nCpG[TypeI$nCpG < 0] <- 0L

    TypeII <- manifest[
        manifest$Infinium_Design_Type == "II",
        c("Name", "AddressA_ID", "AlleleA_ProbeSeq")]
    names(TypeII)[c(2, 3)] <- c("AddressA", "ProbeSeqA")
    TypeII <- as(TypeII, "DataFrame")
    TypeII$ProbeSeqA <- BStringSet(TypeII$ProbeSeqA)
    TypeII$nCpG <- as.integer(letterFrequency(TypeII$ProbeSeq, letters = "R"))

    controls <- read.table(
        file = file,
        skip = control.line,
        sep = ",",
        comment.char = "",
        quote = "",
        colClasses = c(rep("character", 5)))

    TypeControl <- controls[, 1:4]
    names(TypeControl) <- c("Address", "Type", "Color", "ExtendedType")
    TypeControl <- as(TypeControl, "DataFrame")

    snps <- TypeControl[TypeControl$Type == "Genotyping",]
    TypeControl <- TypeControl[TypeControl$Type != "Genotyping",]
    rsname <- sub("_[AB]", "", snps$ExtendedType)
    snps.sp <- split(snps, rsname)
    snps.sp <- lapply(names(snps.sp), function(rs) {
        snp <- snps.sp[[rs]]
        DataFrame(
            Name = rs,
            AddressA = snp[grep("_A", snp$ExtendedType), "Address"],
            AddressB = snp[grep("_B", snp$ExtendedType), "Address"],
            Color = "Unknown")
    })
    TypeSnpI <- do.call(rbind, snps.sp)
    TypeSnpII <- TypeSnpI[0, ]

    list(manifestList =
             list(TypeI = TypeI,
                  TypeII = TypeII,
                  TypeControl = TypeControl,
                  TypeSnpI = TypeSnpI,
                  TypeSnpII = TypeSnpII),
         manifest = manifest,
         controls = controls)
}

# TODOs ------------------------------------------------------------------------

# TODO: Lots of duplicated code; DRY

# Exported functions -----------------------------------------------------------

dmpFinder <- function(dat, pheno, type = c("categorical", "continuous"),
                      qCutoff = 1, shrinkVar = FALSE) {

    # Check inputs
    type <- match.arg(type)
    if (is(dat, "MethylSet")) {
        .isMatrixBackedOrStop(dat, "dmpFinder")
        M <- getM(dat)
    } else if (is(dat, "DelayedMatrix")) {
        stop("This function does not yet support DelayedArray objects")
    } else {
        stopifnot(is.numeric(dat))
        M <- dat
        if (is.vector(M)) M <- matrix(M, nrow = 1)
    }
    n <- length(pheno)
    if (n != ncol(M)) stop("length of pheno does not equal number of samples")

    if (type == "categorical") {
        pheno <- factor(as.character(pheno))
        design <- model.matrix(~pheno)
        fit <- lmFit(M, design)
        if (shrinkVar) {
            fit <- contrasts.fit(fit, contrasts(pheno))
            fit <- eBayes(fit)
            tab <- data.frame(
                intercept = fit$coefficients[, 1],
                f = fit[["F"]],
                pval = fit[["F.p.value"]])
        } else {
            fit1 <- lmFit(M)
            RSS1 <- rowSums2((M - fitted(fit1))^2)
            RSS <- rowSums2((M - fitted(fit))^2)
            df1 <- length(levels(pheno)) - 1
            df2 <- n - length(levels(pheno))
            Fstat <- ((RSS1 - RSS)/df1)/(RSS/df2)
            if (df2 > 1e+06) {
                F.p.value <- pchisq(df1 * Fstat, df1, lower.tail = FALSE)
            }
            else {
                F.p.value <- pf(Fstat, df1, df2, lower.tail = FALSE)
            }
            tab <- data.frame(
                intercept = fit$coefficients[, 1],
                f = Fstat,
                pval = F.p.value)
        }
    }
    else if (type == "continuous") {
        design <- model.matrix(~pheno)
        fit <- lmFit(M, design)
        if (shrinkVar) {
            fit <- eBayes(fit)
            sigma <- sqrt(fit$s2.post)
            df <- fit$df.prior + fit$df.residual
        } else {
            sigma <- fit$sigma
            df <- fit$df.residual
        }
        coef <- fit$coefficients
        if (is.vector(coef)) coef <- matrix(coef, ncol = 2)
        stdev <- fit$stdev.unscaled
        if (is.vector(stdev)) stdev <- matrix(stdev, ncol = 2)
        t <- coef[, 2] / (stdev[, 2] * sigma)
        pval <- 2 * pt(abs(t), df = df, lower.tail = FALSE)
        tab <- data.frame(
            intercept = coef[, 1],
            beta = coef[, 2],
            t = t,
            pval = pval)
    }
    p0 <- pi0.est(tab$pval[!is.na(tab$pval)])$p0
    tab$qval <- qvalue.cal(tab$pval, p0)
    if (qCutoff < 1) tab <- tab[tab$qval <= qCutoff,]
    o <- order(tab$pval)
    tab <- tab[o, ]
    ## tab <- cbind(tab, annotate(rownames(tab)))
    if (nrow(tab) == 0) {
        message(sprintf("No significant DMPs at FDR cutoff of %s.", qCutoff))
    }
    tab
}

gaphunter <- function(object, threshold = 0.05, keepOutliers = FALSE,
                      outCutoff = 0.01, verbose = TRUE) {
    # Check inputs
    if ((threshold <= 0) || (threshold >= 1)) {
        stop("[gaphunter] 'threshold' must be between 0 and 1.")
    }
    if ((outCutoff <= 0) || (outCutoff >= 0.5)) {
        stop("[gaphunter] 'outCutoff' must be between 0 and 0.5.")
    }
    if (is(object, "GenomicRatioSet") || is(object, "GenomicMethylSet") ||
        is(object, "MethylSet") || is(object,"RatioSet")) {
        .isMatrixBackedOrStop(object, "gaphunter")
        if (verbose) message("[gaphunter] Calculating beta matrix.")
        Beta <- getBeta(object)
    } else {
        if (is(object, "matrix")) {
            test <- rowRanges(object)
            if (sum(test[, 1] < 0, na.rm = TRUE) == 0 &&
                sum(test[, 2] > 1, na.rm = TRUE) == 0) {
                Beta <- object
            } else {
                stop("[gaphunter] Matrix must be of Beta values with range ",
                     "from 0 to 1") }
        } else {
            stop("[gaphunter] Object must be one of (Genomic)RatioSet, ",
                 "(Genomic)MethylSet, or matrix")
        }
    }
    nacheck <- rowCounts(Beta, value = NA_real_)
    if (sum(nacheck > 0) > 0) {
        if (verbose) {
            message("[gaphunter] Removing probes containing missing beta ",
                    "values.")
        }
        Beta <- Beta[which(nacheck == 0),]
    }

    if (verbose) {
        message(
            "[gaphunter] Using ",
            prettyNum(nrow(Beta), big.mark = ",", scientific = FALSE),
            " probes and ",
            prettyNum(ncol(Beta), big.mark = ",", scientific = FALSE),
            " samples.")
        message("[gaphunter] Searching for gap signals.")
    }

    sorting <- t(
        apply(Beta, 1, sort.int, method = "quick", index.return = TRUE))
    sortedindices <- lapply(sorting, function(n) n$ix)
    sortedbeta <- do.call("rbind", lapply(sorting, function(n) n$x))
    rownames(sortedbeta) <- rownames(Beta)
    diffs <- rowDiffs(sortedbeta)
    gapind <- rowSums2(diffs > threshold)
    sortedindices <- lapply(which(gapind > 0), function(h) sortedindices[[h]])
    sortedbeta <- sortedbeta[which(gapind > 0),]
    diffs <- diffs[which(gapind > 0),]
    gapind <- gapind[which(gapind > 0)]

    breakpoints <- apply(diffs, 1, function(x) which(x > threshold))

    returngroups <- function(x, y) {
        template <- rep(1, ncol(sortedbeta))
        count <- 2
        for (j in x) {
            template[y[seq(j + 1L, length(y))]] <- count
            count <- count + 1
        }
        template
    }

    groupanno <- t(mapply(returngroups, breakpoints, sortedindices))
    maxGroups <- max(groupanno)
    gapanno <- matrix(0, nrow = nrow(groupanno), ncol = maxGroups + 1)
    gapanno[,1] <- rowMaxs(groupanno)
    for (ii in 1:maxGroups) {
        gapanno[, ii + 1] <- rowCounts(groupanno, value = ii)
    }
    # TODO: Remove commented chunk if not needed
    ## gapanno <- apply(groupanno, 1, function(k) {
    ##     tempgap <- rep(0, max(gapind) + 2)
    ##     kl <- length(unique(k))
    ##     tempgap[1] <- kl
    ##     tempgap[1 + 1:kl] <- table(k)
    ##     return(tempgap)
    ## })
    ## gapanno <- t(gapanno)
    rownames(gapanno) <- rownames(sortedbeta)
    rownames(groupanno) <- rownames(sortedbeta)

    if (verbose) {
        message(
            "[gaphunter] Found ",
            prettyNum(nrow(gapanno), big.mark = ",", scientific = FALSE),
            " gap signals.")
    }

    if (keepOutliers == FALSE) {
        if (verbose) {
            message("[gaphunter] Filtering out gap signals driven by outliers.")
        }
        markme <- unlist(lapply(seq_len(nrow(gapanno)), function(blah) {
            analyze <- gapanno[blah, -1]
            maxgroup <- which(analyze == max(analyze, na.rm = TRUE))
            if (length(maxgroup) == 1) {
                if (sum(analyze[-maxgroup], na.rm = TRUE) <
                    outCutoff * ncol(Beta)) {
                    return(blah)
                }
            }
        }))
        if (length(markme) > 0) {
            gapanno <- gapanno[-markme, ]
            groupanno <- groupanno[-markme, ]
        }
        if (verbose) {
            message(
                "[gaphunter] Removed ",
                prettyNum(length(markme), big.mark = ",", scientific = FALSE),
                " gap signals driven by outliers from results.")
        }
    }
    gapanno <- data.frame(gapanno)

    colnames(gapanno) <- c("Groups", paste0("Group",seq_len(ncol(gapanno) - 1)))

    algorithm <- list(
        threshold = threshold,
        outCutoff = outCutoff,
        keepOutliers = keepOutliers)

    list(
        proberesults = gapanno,
        sampleresults = groupanno,
        algorithm = algorithm)
}

library(minfiData)
library(digest)

GRset.funnorm <- preprocessFunnorm(RGsetEx)
digest_preprocessFunnorm <- list(M = minfi:::.digestMatrix(getM(GRset.funnorm)),
                                 CN = minfi:::.digestMatrix(getCN(GRset.funnorm)))
save(digest_preprocessFunnorm, file = "../unitTests/digest_preprocessFunnorm.rda")

sessionInfo()
rm(list = ls())

# ------------------------------------------------------------------------------
# DelayedArray tests
#

library(DelayedArray)

assays(RGsetEx) <- endoapply(assays(RGsetEx), DelayedArray)

GRset.funnorm <- preprocessFunnorm(RGsetEx)
digest_preprocessFunnorm <- list(M = minfi:::.digestMatrix(getM(GRset.funnorm)),
                                 CN = minfi:::.digestMatrix(getCN(GRset.funnorm)))
save(digest_preprocessFunnorm, file = "../unitTests/digest_preprocessFunnorm.DelayedArray.rda")

sessionInfo()
rm(list = ls())

######################################################
## Functional normalization of the 450k array
## Jean-Philippe Fortin
## Sept 24 2013
#####################################################

## Return a normalized beta matrix
##
## need stuff from gmodels.  Where?
preprocessFunnorm <- function(rgSet, nPCs=2, sex = NULL, bgCorr = TRUE, dyeCorr = TRUE, keepCN = TRUE, ratioConvert = TRUE, verbose = TRUE) {

    .isMatrixBackedOrStop(rgSet, "preprocessFunnorm")

    .isRGOrStop(rgSet)
    rgSet <- updateObject(rgSet) ## FIXM: might not KDH: technically, this should not be needed, but might be nice

    # Background correction and dye bias normalization:
    if (bgCorr){
        if(verbose && dyeCorr) {
            message("[preprocessFunnorm] Background and dye bias correction with noob")
        } else {
            message("[preprocessFunnorm] Background correction with noob")
        }
        gmSet <- preprocessNoob(rgSet, dyeCorr = dyeCorr)
        if(verbose) message("[preprocessFunnorm] Mapping to genome")
        gmSet <- mapToGenome(gmSet)
    } else {
        if(verbose) message("[preprocessFunnorm] Mapping to genome")
        gmSet <- mapToGenome(rgSet)
    }

    subverbose <- max(as.integer(verbose) - 1L, 0)

    if(verbose) message("[preprocessFunnorm] Quantile extraction")
    extractedData <- .extractFromRGSet450k(rgSet)
    rm(rgSet)

    if (is.null(sex)) {
        gmSet <- addSex(gmSet, getSex(gmSet, cutoff = -3))
        sex <- rep(1L, length(gmSet$predictedSex))
        sex[gmSet$predictedSex == "F"] <- 2L
    }
    if(verbose) message("[preprocessFunnorm] Normalization")
    if(keepCN) {
        CN <- getCN(gmSet)
    }
    gmSet <- .normalizeFunnorm450k(object = gmSet, extractedData = extractedData,
                                   sex = sex, nPCs = nPCs, verbose = subverbose)
    preprocessMethod <- c(preprocessMethod(gmSet),
                          mu.norm = sprintf("Funnorm, nPCs=%s", nPCs))
    if(ratioConvert) {
        grSet <- ratioConvert(gmSet, type = "Illumina", keepCN = keepCN)
        if(keepCN) {
            assay(grSet, "CN") <- CN
        }
        grSet@preprocessMethod <- preprocessMethod
        return(grSet)
    } else {
        gmSet@preprocessMethod <- preprocessMethod
        return(gmSet)
    }
 }

 .getFunnormIndices <- function(object) {
     .isGenomicOrStop(object)
     probeType <- getProbeType(object, withColor = TRUE)
     autosomal <- (seqnames(object) %in% paste0("chr", 1:22))
     indices <- list(IGrn = which(probeType == "IGrn" & autosomal),
                     IRed = which(probeType == "IRed" & autosomal),
                     II = which(probeType == "II" & autosomal),
                     X = which(seqnames(object) == "chrX"),
                     Y = which(seqnames(object) == "chrY"))
     indices
 }

 .normalizeFunnorm450k <- function(object, extractedData, nPCs, sex, verbose = TRUE) {
     normalizeQuantiles <- function(matrix, indices, sex = NULL) {
         matrix <- matrix[indices,,drop=FALSE]
         ## uses probs, model.matrix, nPCS, through scoping)
         oldQuantiles <- t(colQuantiles(matrix, probs = probs))
         if(is.null(sex)) {
             newQuantiles <- .returnFit(controlMatrix = model.matrix, quantiles = oldQuantiles, nPCs = nPCs)
         } else {
             newQuantiles <- .returnFitBySex(controlMatrix = model.matrix, quantiles = oldQuantiles, nPCs = nPCs, sex = sex)
         }
         .normalizeMatrix(matrix, newQuantiles)
     }

    indicesList <- .getFunnormIndices(object)
    model.matrix <- .buildControlMatrix450k(extractedData)
    probs <- seq(from = 0, to = 1, length.out = 500)
    Meth <- getMeth(object)
    Unmeth <- getUnmeth(object)
    if (nPCs > 0){
        for (type in c("IGrn", "IRed", "II")) {
            indices <- indicesList[[type]]
            if(length(indices) > 0) {
                if(verbose) message(sprintf("[normalizeFunnorm450k] Normalization of the %s probes", type))
                Unmeth[indices,] <- normalizeQuantiles(Unmeth, indices = indices, sex = NULL)
                Meth[indices,] <- normalizeQuantiles(Meth, indices = indices, sex = NULL)
            }
        }

        indices <- indicesList[["X"]]
        if(length(indices) > 0) {
            if(verbose) message("[normalizeFunnorm450k] Normalization of the X-chromosome")
            Unmeth[indices,] <- normalizeQuantiles(Unmeth, indices = indices, sex = sex)
            Meth[indices,] <- normalizeQuantiles(Meth, indices = indices, sex = sex)
        }
    }

    indices <- indicesList[["Y"]]
    if(length(indices) > 0) {
        if(verbose) message("[normalizeFunnorm450k] Normalization of the Y-chromosome")
        sex <- as.character(sex)
        levels <- unique(sex)
        nSexes <- length(levels)
        if (nSexes == 2) {
            level1 <- levels[1]
            level2 <- levels[2]
        }
        if (nSexes == 2) {
            if (sum(sex == level1)>1) {
                Meth[indices, sex==level1]   <- preprocessCore::normalize.quantiles(Meth[indices, sex == level1, drop=FALSE])
                Unmeth[indices, sex==level1] <- preprocessCore::normalize.quantiles(Unmeth[indices, sex == level1,drop=FALSE])
            }
            if (sum(sex == level2)>1) {
                Meth[indices, sex==level2]   <- preprocessCore::normalize.quantiles(Meth[indices, sex == level2,drop=FALSE])
                Unmeth[indices, sex==level2] <- preprocessCore::normalize.quantiles(Unmeth[indices, sex == level2,drop=FALSE])
            }
        } else {
            Meth[indices,] <- preprocessCore::normalize.quantiles(Meth[indices,])
            Unmeth[indices,] <- preprocessCore::normalize.quantiles(Unmeth[indices,])
        }
    }
    assay(object, "Meth") <- Meth
    assay(object, "Unmeth") <- Unmeth
    return(object)
}



### To extract quantiles and control probes from rgSet
.extractFromRGSet450k <- function(rgSet) {
    rgSet <- updateObject(rgSet)
    controlType <- c("BISULFITE CONVERSION I",
                     "BISULFITE CONVERSION II",
                     "EXTENSION",
                     "HYBRIDIZATION",
                     "NEGATIVE",
                     "NON-POLYMORPHIC",
                     "NORM_A",
                     "NORM_C",
                     "NORM_G",
                     "NORM_T",
                     "SPECIFICITY I",
                     "SPECIFICITY II",
                     "TARGET REMOVAL",
                     "STAINING")

    array <- annotation(rgSet)[["array"]]
    ## controlAddr <- getControlAddress(rgSet, controlType = controlType, asList = TRUE)
    ctrls <- getProbeInfo(rgSet, type = "Control")
    if(!all(controlType %in% ctrls$Type))
        stop("The `rgSet` does not contain all necessary control probes")
    ctrlsList <- split(ctrls, ctrls$Type)[controlType]

    redControls <- getRed(rgSet)[ctrls$Address,,drop=FALSE]
    redControls <- lapply(ctrlsList, function(ctl) redControls[ctl$Address,,drop=FALSE])
    greenControls <- getGreen(rgSet)[ctrls$Address,,drop=FALSE]
    greenControls <- lapply(ctrlsList, function(ctl) greenControls[ctl$Address,,drop=FALSE])

    ## Extraction of the undefined negative control probes
    oobRaw <- getOOB(rgSet)
    probs <- c(0.01, 0.50, 0.99)
    greenOOB <- t(colQuantiles(oobRaw$Grn, na.rm = TRUE, probs = probs))
    redOOB   <- t(colQuantiles(oobRaw$Red, na.rm=TRUE,  probs = probs))
    oob      <- list(greenOOB = greenOOB, redOOB = redOOB)

    return(list(
        greenControls = greenControls,
        redControls = redControls,
        oob = oob, ctrlsList = ctrlsList,
        array = array))
}


## Extraction of the Control matrix
.buildControlMatrix450k <- function(extractedData) {
    getCtrlsAddr <- function(exType, index) {
        ctrls <- ctrlsList[[index]]
        addr <- ctrls$Address
        names(addr) <- ctrls$ExtendedType
        na.omit(addr[exType])
    }

    array <- extractedData$array
    greenControls <- extractedData$greenControls
    redControls <- extractedData$redControls
    controlNames <- names(greenControls)
    ctrlsList <- extractedData$ctrlsList

    ## Bisulfite conversion extraction for probe type II:
    index <- match("BISULFITE CONVERSION II", controlNames)
    redControls.current <- redControls[[ index ]]
    bisulfite2 <- colMeans2(redControls.current, na.rm = TRUE)

    ## Bisulfite conversion extraction for probe type I:
    index <- match("BISULFITE CONVERSION I", controlNames)
    if (array=="IlluminaHumanMethylation450k"){
        addr <- getCtrlsAddr(exType = sprintf("BS Conversion I%sC%s", c(" ", "-", "-"), 1:3), index = index)
    } else {
        addr <- getCtrlsAddr(exType = sprintf("BS Conversion I%sC%s", c("-", "-"), 1:2), index = index)
    }
    greenControls.current <- greenControls[[ index ]][addr,,drop=FALSE]
    if (array=="IlluminaHumanMethylation450k"){
        addr <- getCtrlsAddr(exType = sprintf("BS Conversion I-C%s", 4:6), index = index)
    } else {
        addr <- getCtrlsAddr(exType = sprintf("BS Conversion I-C%s", 3:5), index = index)
    }
    redControls.current <- redControls[[ index ]][addr,, drop=FALSE]
    if (nrow(redControls.current)==nrow(greenControls.current)){
        bisulfite1 <- colMeans2(redControls.current + greenControls.current, na.rm = TRUE)
    } else {
        bisulfite1 <- colMeans2(redControls.current, na.rm=TRUE) + colMeans2(greenControls.current, na.rm = TRUE)
    }


    ## Staining
    index <- match("STAINING", controlNames)
    addr <- getCtrlsAddr(exType = "Biotin (High)", index = index)
    stain.green <- t(greenControls[[ index ]][addr,,drop=FALSE])
    addr <- getCtrlsAddr(exType = "DNP (High)", index = index)
    stain.red <- t(redControls[[ index ]][addr,, drop=FALSE ])

    ## Extension
    index <-    match("EXTENSION", controlNames)
    addr <- getCtrlsAddr(exType = sprintf("Extension (%s)", c("A", "T")), index = index)
    extension.red <- t(redControls[[index]][addr,,drop=FALSE])
    colnames(extension.red) <- paste0("extRed", 1:ncol(extension.red))
    addr <- getCtrlsAddr(exType = sprintf("Extension (%s)", c("C", "G")), index = index)
    extension.green <- t(greenControls[[index]][addr,,drop=FALSE])
    colnames(extension.green) <- paste0("extGrn", 1:ncol(extension.green))

    ## Hybridization should be monitored only in the green channel
    index <- match("HYBRIDIZATION", controlNames)
    hybe <- t(greenControls[[index]])
    colnames(hybe) <- paste0("hybe", 1:ncol(hybe))

    ## Target removal should be low compared to hybridization probes
    index <- match("TARGET REMOVAL", controlNames)
    targetrem <- t(greenControls[[index]])
    colnames(targetrem) <- paste0("targetrem", 1:ncol(targetrem))

    ## Non-polymorphic probes
    index <- match("NON-POLYMORPHIC", controlNames)
    addr <- getCtrlsAddr(exType = sprintf("NP (%s)", c("A", "T")), index = index)
    nonpoly.red <- t(redControls[[index]][addr, ,drop=FALSE])
    colnames(nonpoly.red) <- paste0("nonpolyRed", 1:ncol(nonpoly.red))
    addr <- getCtrlsAddr(exType = sprintf("NP (%s)", c("C", "G")), index = index)
    nonpoly.green <- t(greenControls[[index]][addr, ,drop=FALSE])
    colnames(nonpoly.green) <- paste0("nonpolyGrn", 1:ncol(nonpoly.green))

    ## Specificity II
    index <- match("SPECIFICITY II", controlNames)
    greenControls.current <- greenControls[[index]]
    redControls.current <- redControls[[index]]
    spec2.green <- t(greenControls.current)
    colnames(spec2.green) <- paste0("spec2Grn", 1:ncol(spec2.green))
    spec2.red <- t(redControls.current)
    colnames(spec2.red) <- paste0("spec2Red", 1:ncol(spec2.red))
    spec2.ratio <- colMeans2(greenControls.current, na.rm = TRUE) /
        colMeans2(redControls.current, na.rm = TRUE)

    ## Specificity I
    index <- match("SPECIFICITY I", controlNames)
    addr <- getCtrlsAddr(exType = sprintf("GT Mismatch %s (PM)", 1:3), index = index)
    greenControls.current <- greenControls[[index]][addr,,drop=FALSE]
    redControls.current <- redControls[[index]][addr,,drop=FALSE]
    spec1.green <- t(greenControls.current)
    colnames(spec1.green) <- paste0("spec1Grn", 1:ncol(spec1.green))
    spec1.ratio1 <- colMeans2(redControls.current, na.rm = TRUE) /
        colMeans2(greenControls.current, na.rm = TRUE)

    index <- match("SPECIFICITY I", controlNames) # Added that line
    addr <- getCtrlsAddr(exType = sprintf("GT Mismatch %s (PM)", 4:6), index = index)
    greenControls.current <- greenControls[[index]][addr,,drop=FALSE]
    redControls.current <- redControls[[index]][addr,,drop=FALSE]
    spec1.red <- t(redControls.current)
    colnames(spec1.red) <- paste0("spec1Red", 1:ncol(spec1.red))
    spec1.ratio2 <- colMeans2(greenControls.current, na.rm = TRUE) /
        colMeans2(redControls.current, na.rm = TRUE)
    spec1.ratio <- (spec1.ratio1 + spec1.ratio2) / 2

    ## Normalization probes:
    index <- match(c("NORM_A"), controlNames)
    normA <- colMeans2(redControls[[index]], na.rm = TRUE)
    index <- match(c("NORM_T"), controlNames)
    normT <- colMeans2(redControls[[index]], na.rm = TRUE)
    index <- match(c("NORM_C"), controlNames)
    normC <- colMeans2(greenControls[[index]], na.rm = TRUE)
    index <- match(c("NORM_G"), controlNames)
    normG <- colMeans2(greenControls[[index]], na.rm = TRUE)

    dyebias <- (normC + normG)/(normA + normT)

    oobG <- extractedData$oob$greenOOB
    oobR <- extractedData$oob$redOOB
    oob.ratio <- oobG[2,]/oobR[2,]
    oobG <- t(oobG)
    colnames(oobG) <- paste0("oob", c(1,50,99))

    model.matrix <- cbind(
        bisulfite1, bisulfite2, extension.green, extension.red, hybe,
        stain.green, stain.red, nonpoly.green, nonpoly.red,
        targetrem, spec1.green, spec1.red, spec2.green, spec2.red, spec1.ratio1,
        spec1.ratio, spec2.ratio, spec1.ratio2, normA, normC, normT, normG, dyebias,
        oobG, oob.ratio)


    ## Imputation
    for (colindex in 1:ncol(model.matrix)) {
        if(any(is.na(model.matrix[,colindex]))) {
            column <- model.matrix[,colindex]
            column[is.na(column)]    <- mean(column, na.rm = TRUE)
            model.matrix[ , colindex] <- column
        }
    }

    ## Scaling
    model.matrix <- scale(model.matrix)

    ## Fixing outliers
    model.matrix[model.matrix > 3] <- 3
    model.matrix[model.matrix < (-3)] <- -3

    ## Rescaling
    model.matrix <- scale(model.matrix)

    return(model.matrix)
}


### Return the normalized quantile functions
.returnFit <- function(controlMatrix, quantiles, nPCs) {
    stopifnot(is.matrix(quantiles))
    stopifnot(is.matrix(controlMatrix))
    stopifnot(ncol(quantiles) == nrow(controlMatrix))
    ## Fixing potential problems with extreme quantiles
    quantiles[1,] <- 0
    quantiles[nrow(quantiles),] <- quantiles[nrow(quantiles) - 1,] + 1000
    meanFunction <- rowMeans2(quantiles)
    res <- quantiles - meanFunction
    controlPCs <- prcomp(controlMatrix)$x[,1:nPCs,drop=FALSE]
    design <- model.matrix(~controlPCs)
    fits <- lm.fit(x = design, y = t(res))
    newQuantiles <- meanFunction + t(fits$residuals)
    newQuantiles <- .regularizeQuantiles(newQuantiles)
    return(newQuantiles)
}

.returnFitBySex <- function(controlMatrix, quantiles, nPCs, sex) {
    stopifnot(is.matrix(quantiles))
    stopifnot(is.matrix(controlMatrix))
    stopifnot(ncol(quantiles) == nrow(controlMatrix))
    sex    <- as.character(sex)
    levels <- unique(sex)
    nSexes <- length(levels)
    if (nSexes == 2) {
        sex1 <- sum(sex == levels[1])
        sex2 <- sum(sex == levels[2])

    } else {
        sex1 <- sum(sex == levels[1])
        sex2 <- 0
    }

    ## When normalization should not be performed by sex separately:
    if ((sex1 <= 10) | (sex2 <= 10)) {
        newQuantiles <- .returnFit(controlMatrix = controlMatrix,
                                  quantiles = quantiles,
                                  nPCs = nPCs)
    } else {
        quantiles1 <- quantiles[, sex == levels[1]]
        controlMatrix1 <- controlMatrix[sex == levels[1], ]

        newQuantiles1 <- .returnFit(controlMatrix = controlMatrix1,
                                   quantiles = quantiles1,
                                   nPCs = nPCs)

        quantiles2 <- quantiles[, sex == levels[2]]
        controlMatrix2 <- controlMatrix[sex == levels[2], ]

        newQuantiles2 <- .returnFit(controlMatrix = controlMatrix2,
                                   quantiles = quantiles2,
                                   nPCs = nPCs)

        newQuantiles <- quantiles
        newQuantiles[, sex == levels[1]] <- newQuantiles1
        newQuantiles[, sex == levels[2]] <- newQuantiles2
    }

    return(newQuantiles)
}

### Normalize a matrix of intensities
.normalizeMatrix <- function(intMatrix, newQuantiles) {
    ## normMatrix <- matrix(NA, nrow(intMatrix), ncol(intMatrix))
    n <- nrow(newQuantiles)
    normMatrix <- sapply(1:ncol(intMatrix), function(i) {
        crtColumn <- intMatrix[ , i]
        crtColumn.reduced <- crtColumn[!is.na(crtColumn)]
        ## Generation of the corrected intensities:
        target <- sapply(1:(n-1), function(j) {
            start <- newQuantiles[j,i]
            end <- newQuantiles[j+1,i]
            if (!isTRUE(all.equal(start,end))){
                sequence <- seq(start, end,( end-start)/n)[-(n+1)]
            } else {
                sequence <- rep(start, n)
            }
            return(sequence)
        })
        target <- as.vector(target)
        result <- preprocessCore::normalize.quantiles.use.target(matrix(crtColumn.reduced), target)
        return(result)
    })
    return(normMatrix)
}

# To ensure a monotonically increasing and non-negative quantile function
# Necessary for pathological cases
.regularizeQuantiles <- function(x){
    x[x<0] <- 0
    colCummaxs(x)
}

# Global variables -------------------------------------------------------------

utils::globalVariables(c("channel"))

# Other hardcoded variables ----------------------------------------------------

.default.27k.annotation  <- "ilmn12.hg19"
.default.450k.annotation <- "ilmn12.hg19"
.default.epic.annotation <- "ilm10b4.hg19"
.default.allergy.annotation <- "ilm12.hg19"
.default.epicv2.annotation <- "20a1.hg38"
.metharray.types <- c("IlluminaHumanMethylation450k",
                      "IlluminaHumanMethylationEPIC",
		      "IlluminaHumanMethylationEPICv2",
                      "IlluminaHumanMethylation27k",
                      "IlluminaHumanMethylationAllergy",
                      "HorvathMammalMethylChip40")
.seqnames.order.all <- c(paste0("chr", c(1:22, "X", "Y")), "multi", "unmapped")
.seqnames.order <- paste0("chr", c(1:22, "X", "Y"))

# Internal functions -----------------------------------------------------------

logit2 <- function(x) log2(x) - log2(1 - x)

ilogit2 <- function(x) 2^x / (1 + 2^x)

.show.annotation <- function(annotation, indent = "  ") {
    cat("Annotation\n")
    if (length(annotation) == 1) {
        cat(sprintf("%sarray: %s\n", indent, annotation))
    } else {
        sapply(seq(along = annotation), function(ii) {
            cat(sprintf("%s%s: %s\n",
                        indent,
                        names(annotation)[ii],
                        annotation[ii]))
        })
    }
}

.show.preprocessMethod <- function(preprocessMethod) {
    if (length(preprocessMethod) == 3 && is.null(names(preprocessMethod))) {
        names(preprocessMethod) <- c("rg.norm", "minfi", "manifest")
    }
    if (length(preprocessMethod) == 0) {
        preprocessMethod <- c(
            rg.norm = "unknown",
            minfi = "unknown",
            manifest = "unknown")
    }
    cat("Preprocessing\n")
    cat(sprintf("  Method: %s\n  minfi version: %s\n  Manifest version: %s\n",
                preprocessMethod["rg.norm"],
                preprocessMethod["minfi"],
                preprocessMethod["manifest"]))
}

.getManifestString <- function(annotation) {
    if (length(annotation) == 1) {
        if (annotation == "Unknown") {
            stop("Cannot get Manifest object for an 'Unknown' array")
        }
        return(paste0(annotation, "manifest"))
    }
    if ("array" %in% names(annotation)) {
        if (annotation["array"] == "Unknown") {
            stop("Cannot get Manifest object for an 'Unknown' array")
        }
        return(paste0(annotation["array"], "manifest"))
    }
    stop("unable to get the manifest string for this object")
}

.getAnnotationString <- function(annotation) {
    if (length(annotation) == 1) {
        if (annotation == "Unknown") {
            stop("Cannot get Annotation object for an 'Unknown' array")
        }
        return(sprintf("%sanno", annotation))
    }
    if (all(c("array", "annotation") %in% names(annotation))) {
        if (annotation["array"] == "Unknown") {
            stop("Cannot get Annotation object for an 'Unknown' array")
        }
        return(
            sprintf("%sanno.%s", annotation["array"], annotation["annotation"]))
    }
    stop("unable to get the annotation string for this object")
}

.betaFromMethUnmeth <- function(Meth, Unmeth, object, offset = 0,
                                betaThreshold = 0, minZero = TRUE) {
    stopifnot(offset >= 0)
    stopifnot(betaThreshold >= 0 & betaThreshold <= 0.5)
    if (minZero) {
        Meth <- pmax2(Meth, 0)
        Unmeth <- pmax2(Unmeth, 0)
    }
    beta <- Meth / (Meth + Unmeth + offset)
    if (betaThreshold > 0) {
        beta <- pmin2(pmax2(beta, betaThreshold), 1 - betaThreshold)
    }
    beta
}

.checkAssayNames <- function(object, names) {
    nms <- names(assays(object, withDimnames = FALSE))
    if (!all(names %in% nms)) {
        return(sprintf(
            "object of class '%s' needs to have assay slots with names '%s'",
            class(object),
            paste0(names, collapse = ", ")))
    } else {
        NULL
    }
}

.digestMatrix <- function(mat, digits = 6) {
    content <- sprintf(paste0("%.", digits, "f"), mat)
    # NOTE: Handling signed zero as per IEEE specs
    zero <- paste(c("0.", rep("0", digits)), collapse = "")
    content[content == paste0("-", zero)] <- zero
    digest::digest(c(content, rownames(mat), colnames(mat)))
}

.digestVector <- function(vec, digits = 6) {
    content <- sprintf(paste0("%.", digits, "f"), vec)
    # NOTE: Handling signed zero as per IEEE specs
    zero <- paste(c("0.", rep("0", digits)), collapse = "")
    content[content == paste0("-", zero)] <- zero
    digest::digest(content)
}

.isGenomicOrStop <- function(object) {
    if (!is(object, "GenomicMethylSet") && !is(object, "GenomicRatioSet")) {
        stop("object is of class '", class(object), "', but needs to be of ",
             "class 'GenomicMethylSet' or 'GenomicRatioSet'")
    }
}

.isMethylOrStop <- function(object) {
    if (!is(object, "MethylSet") && !is(object, "GenomicMethylSet")) {
        stop("object is of class '", class(object), "', but needs to be of ",
             "class 'MethylSet' or 'GenomicMethylSet'")
    }
}

.isMethylOrRatio <- function(object) {
    is(object, "MethylSet") ||
        is(object, "GenomicMethylSet") ||
        is(object, "RatioSet") ||
        is(object, "GenomicRatioSet")
}

.isRGOrStop <- function(object) {
    if (!is(object, "RGChannelSet")) {
        stop("object is of class '", class(object), "', but needs to be of ",
             "class 'RGChannelSet' or 'RGChannelSetExtended'")
    }
}

.isMatrixBacked <- function(object) {
    stopifnot(is(object, "SummarizedExperiment"))
    all(vapply(assays(object), is.matrix, logical(1L)))
}

.isDelayedArrayBacked <- function(object) {
    stopifnot(is(object, "SummarizedExperiment"))
    all(vapply(assays(object), is, logical(1L), "DelayedArray"))
}

.isMatrixBackedOrWarning <- function(object, FUN) {
    if (.isDelayedArrayBacked(object)) {
        warning("Memory usage may be high because '", FUN, "()' is not yet ",
                "optimized for use with DelayedArray-backed minfi objects.",
                call. = FALSE,
                immediate. = TRUE)
    } else if (!.isMatrixBacked(object)) {
        warning("Memory usage may be high because '", FUN, "()' is not yet ",
                "optimized for use with non-matrix-backed minfi objects.",
                call. = FALSE,
                immediate. = TRUE)
    }
}

.isMatrixBackedOrStop <- function(object, FUN) {
    if (!.isMatrixBacked(object)) {
        stop("'", FUN, "()' only supports matrix-backed minfi objects.",
             call. = FALSE)
    }
}

.is27k <- function(object) {
    annotation(object)["array"] == "IlluminaHumanMethylation27k"
}

.is450k <- function(object) {
    annotation(object)["array"] == "IlluminaHumanMethylation450k"
}

.isEPIC <- function(object) {
    annotation(object)["array"] == "IlluminaHumanMethylationEPIC"
}

.isEPICv2 <- function(object) {
    annotation(object)["array"] == "IlluminaHumanMethylationEPICv2"
}



.isAllergy <- function(object) {
    annotation(object)["array"] == "IlluminaHumanMethylationAllergy"
}

.harmonizeSex <- function(vector) {
    # TODO: This function is not yet implemented
    stop("function not done")
    validMale <- c("M", "MALE")
    validFemale <- c("F", "FEMALE")
    # validUnknown <- c("U", "Unknown")
    if (is.factor(vector)) vector <- as.character(vector)
    if (!is.character(vector)) {
        stop("[.harmonizeSet] argument 'vector' needs to be either a ",
             "character or a factor")
    }
    vector <- toupper(vector)
    vector[vector %in% validMale] <- "M"
    vector[vector %in% validFemale] <- "F"
    if (any(!vector %in% c("M", "F"))) {
        stop("[.harmonizeSet] could not harmonize the vector argument to be ",
             "either 'M' or 'F'")
    }
    vector
}

.harmonizeDataFrames <- function(x, y) {
    stopifnot(is(x, "DataFrame"))
    stopifnot(is(y, "DataFrame"))
    x.only <- setdiff(names(x), names(y))
    y.only <- setdiff(names(y), names(x))
    if (length(x.only) > 0) {
        df.add <- x[1, x.only, drop = FALSE]
        is.na(df.add[1, ]) <- TRUE
        y <- cbind(y, df.add)
    }
    if (length(y.only) > 0) {
        df.add <- y[1, y.only, drop = FALSE]
        is.na(df.add[1, ]) <- TRUE
        x <- cbind(x, df.add)
    }
    list(x = x, y = y[, names(x)])
}

.pDataAdd <- function(object, df) {
    stopifnot(is(df, "data.frame") || is(df, "DataFrame"))
    pD <- colData(object)
    if (any(names(df) %in% names(pD))) {
        alreadyPresent <- intersect(names(df), names(pD))
        warning(sprintf(
            "replacing the following columns in colData(object): %s",
            paste(alreadyPresent, collapse = ", ")))
        pD[, alreadyPresent] <- df[, alreadyPresent]
        df <- df[, !names(df) %in% alreadyPresent]
    }
    if (ncol(df) > 0) {
        # NOTE: Work around for bug in cbind(DataFrame, DataFrame)
        rownam <- rownames(pD)
        pD <- cbind(pD, df)
        rownames(pD) <- rownam
    }
    colData(object) <- pD
    object
}

.pDataFix <- function(df) {
    characterColumns <- c(
        "Slide", "Array", "Sample_Name", "Basename", "SampleID")
    for (col in characterColumns) {
        if (col %in% names(df))
            df[[col]] <- as.character(df[[col]])
    }
    df
}

.NA_type <- function(type) {
    c(vector(type), NA)
}

# Exported functions -----------------------------------------------------------

getMethSignal <- function(object, what = c("Beta", "M"), ...) {
    what <- match.arg(what)
    switch(what,
           "Beta" = getBeta(object, ...),
           "M" = getM(object, ...)
    )
}

# Internal functions -----------------------------------------------------------

read.GenomeStudio <- function(filename) {
    colnames <- strsplit(readLines(filename, n = 9)[9], "\t")[[1]]
    colClasses <- rep("NULL", length(colnames))
    names(colClasses) <- colnames
    colClasses["TargetID"] <- "character"
    colClasses[grep("AVG_Beta", colnames)] <- "numeric"
    colClasses[grep("Signal_A", colnames)] <- "numeric"
    colClasses[grep("Signal_B", colnames)] <- "numeric"
    # NOTE: File has trailing tab...
    colClasses <- c(colClasses, dummy = "NULL")
    mat <- read.table(
        file = filename,
        header = TRUE,
        sep = "\t",
        comment.char = "",
        quote = "",
        skip = 8,
        colClasses = colClasses)
    sampleNames <- sub(
        "\\.AVG_Beta", "", grep("AVG_Beta", colnames(mat), value = TRUE))
    beta <- mat[, grep("AVG_Beta", colnames(mat))]
    colnames(beta) <- sampleNames
    rownames(beta) <- mat$TargetID
    SignalA <- mat[, grep("Signal_A", colnames(mat))]
    colnames(SignalA) <- sampleNames
    rownames(SignalA) <- mat$TargetID
    SignalB <- mat[, grep("Signal_B", colnames(mat))]
    colnames(SignalB) <- sampleNames
    rownames(SignalB) <- mat$TargetID
    list(beta = beta, SignalA = SignalA, SignalB = SignalB)
}

# Exported functions -----------------------------------------------------------

makeGenomicRatioSetFromMatrix <- function(mat,rownames = NULL, pData = NULL,
                                          array = "IlluminaHumanMethylation450k",
                                          annotation = .default.450k.annotation,
                                          mergeManifest = FALSE,
                                          what = c("Beta","M")) {

    what <- match.arg(what)

    if (!is.matrix(mat)) {
        stop(sprintf("'mat' must be a matrix. It is a %s.", class(mat)))
    }

    if (is.null(rownames)) {
        rownames <- rownames(mat)
    } else {
        if (length(rownames) != nrow(mat)) {
            stop("Number of rows of mat and length of rownames must match.")
        }
        rownames(mat) <- rownames
    }

    if (is.data.frame(pData)) pData <- as(pData,"DataFrame")
    if (is.null(colnames(mat))) colnames(mat) <- seq_len(ncol(mat))
    if (is.null(pData)) {
        pData <- DataFrame(X1 = seq_len(ncol(mat)), row.names = colnames(mat))
    }
    if (!is(pData, "DataFrame")) {
        stop(sprintf(
            "'pData' must be DataFrame or data.frame. It is a %s.",
            class(pData)))
    }

    # Create GRanges
    ann <- .getAnnotationString(c(array = array, annotation = annotation))
    if (!require(ann, character.only = TRUE)) {
        stop(sprintf("cannot load annotation package %s", ann))
    }
    object <- get(ann)

    gr <- getLocations(object, mergeManifest = mergeManifest,
                       orderByLocation = TRUE)

    locusNames <- names(gr)

    # NOTE:Tthis might return NAs but it's ok
    # TODO: Fix this. return only what is sent
    common <- intersect(locusNames, rownames(mat))
    if (length(common) == 0) {
        stop("No rowname matches. 'rownames' need to match ",
             "IlluminaHumanMethylation450k probe names.")
    }

    # NOTE: we give no warning if some of the rownames have no match.
    ind1 <- match(common,rownames(mat))
    ind2 <- match(common,locusNames)

    preprocessing <- c(
        rg.norm = "Matrix converted with makeGenomicRatioSetFromMatrix")

    if (what == "Beta") {
        out <- GenomicRatioSet(
            gr = gr[ind2, ],
            Beta = mat[ind1, ,drop = FALSE],
            M = NULL,
            CN = NULL,
            colData = pData,
            annotation = c(array = array, annotation = annotation),
            preprocessMethod = preprocessing)
    } else {
        out <- GenomicRatioSet(
            gr = gr[ind2, ],
            Beta = NULL,
            M = mat[ind1, ,drop = FALSE],
            CN = NULL,
            colData = pData,
            annotation = c(array = array, annotation = annotation),
            preprocessMethod = preprocessing)
    }
    out
}


getGenomicRatioSetFromGEO <- function(GSE = NULL, path = NULL,
                                      array = "IlluminaHumanMethylation450k",
                                      annotation = .default.450k.annotation,
                                      what = c("Beta","M"),
                                      mergeManifest = FALSE,
                                      i = 1) {

    what <- match.arg(what)

    if (is.null(GSE) && is.null(path)) {
        stop("Either GSE or path must be supplied.")
    }

    # Read the GEO Main Files Information
    if (!is.null(GSE)) {
        gset <- getGEO(GSE)
    } else {
        gset <- getGEO(file.path(path, list.files(path, pattern = ".soft")))
    }

    if (length(gset) == 0) stop("Empty list retrieved from GEO.")
    if (length(gset) > 1) {
        warning(
            "More than one ExpressionSet found:\n",
            names(gset),
            "\nUsing entry ",
            i)
        gset <- gset[[i]]
    } else gset <- gset[[1]]
    platform <- annotation(gset)

    if (platform != "GPL13534") {
        warning(
            sprintf("%s is not the platform ID associated with IlluminaHumanMethylation450k. Should be GPL13534.",
                    platform))
    }
    if (what ==" Beta" && (min(exprs(gset)[, 1], na.rm = TRUE) < 0 ||
                           max(exprs(gset)[, 1], na.rm = TRUE) > 1 )) {
        warning("Values outside [0,1] detected. 'what' argument should not ",
                "be Beta.")
    }

    # GEO data is read, now ready to create a minfi object
    ann <- .getAnnotationString(c(array = array, annotation = annotation))
    if (!require(ann, character.only = TRUE)) {
        stop(sprintf("cannot load annotation package %s", ann))
    }
    object <- get(ann)

    gr <- getLocations(
        object = object,
        mergeManifest = mergeManifest,
        orderByLocation = TRUE)

    locusNames <- names(gr)

    sampleNames(gset) <- gset$title

    # TODO: We could call makeGenomicRatioSetFromMatrix() but rewrite to avoid
    # a copy of exprs(gset)
    common <- intersect(locusNames, fData(gset)$Name)
    if (length(common) == 0) {
        stop("No rowname matches. 'rownames' need to match ",
             "IlluminaHumanMethylation450k probe names.")
    }

    # NOTE: We give no warning if some rownames have no match
    ind1 <- match(common,fData(gset)$Name)
    ind2 <- match(common,locusNames)

    preprocessing <- c(rg.norm = paste0("See GEO ", GSE, " for details"))

    if (what == "Beta") {
        out <- GenomicRatioSet(
            gr = gr[ind2, ],
            Beta = exprs(gset)[ind1, ,drop = FALSE],
            M = NULL,
            CN = NULL,
            colData = as(pData(gset), "DataFrame"),
            annotation = c(array = array, annotation = annotation),
            preprocessMethod = preprocessing)
    } else {
        out <- GenomicRatioSet(
            gr = gr[ind2, ],
            Beta = NULL,
            M = exprs(gset)[ind1, ,drop = FALSE],
            CN = NULL,
            colData = as(pData(gset), "DataFrame"),
            annotation = c(array = array, annotation = annotation),
            preprocessMethod = preprocessing)
    }

    out
}

readTCGA <- function(filename, sep = "\t", keyName = "Composite Element REF",
                     Betaname = "Beta_value", pData = NULL,
                     array = "IlluminaHumanMethylation450k",
                     annotation = .default.450k.annotation,
                     mergeManifest = FALSE,
                     showProgress = TRUE) {
    # NOTE: We assume first column are sample names and second column are the
    #       value identifiers
    colnames <- strsplit(readLines(filename, n = 2), sep)

    select <- sort(
        c(grep(keyName, colnames[[2]]), grep(Betaname,colnames[[2]])))

    mat <- fread(
        input = filename,
        header = FALSE,
        sep = sep,
        select = select,
        showProgress = showProgress,
        skip = 2)
    rowNames <- as.matrix(mat[, 1, with = FALSE])
    mat <- as.matrix(mat[, -1, with = FALSE])
    rownames(mat) <- rowNames
    colnames(mat) <- colnames[[1]][select][-1]
    # TODO: Shouldn't be necessary to rm() anything
    rm(rowNames,colnames)

    makeGenomicRatioSetFromMatrix(
        mat = mat,
        pData = pData,
        array = array,
        annotation = annotation,
        mergeManifest = mergeManifest, what = "Beta")
}


readGEORawFile <- function(filename, sep = ",", Uname = "Unmethylated signal",
                           Mname = "Methylated signal", row.names = 1,
                           pData = NULL, array = "IlluminaHumanMethylation450k",
                           annotation = .default.450k.annotation,
                           mergeManifest = FALSE, showProgress = TRUE, ...) {
    colnames <- strsplit(readLines(filename, n = 1), sep)[[1]]

    if (all(!grepl(Uname, colnames))) {
        stop("No columns contain Uname. Use readLines or look at file header ",
             "to see column names.")
    }

    if (all(!grepl(Mname, colnames))) {
        stop("No columns contain Mname. Use readLines or look at file header ",
             "to see column names.")
    }

    select <- sort(c(row.names, grep(Uname,colnames), grep(Mname,colnames)))

    mat <- fread(
        input = filename,
        header = TRUE,
        sep = sep,
        select = select,
        showProgress = showProgress,
        ...)

    rowNames <- as.matrix(mat[, 1, with = FALSE])
    mat <- as.matrix(mat[,-1, with = FALSE])
    rownames(mat) <- rowNames
    rm(rowNames)

    uindex <- grep(Uname,colnames(mat))
    mindex <- grep(Mname,colnames(mat))

    trim <- function(x){
        x <- gsub("^\\s+|\\s+$", "", x)
        x <- gsub("^\\.+|\\.+$", "", x)
        x <- gsub("^\\_+|\\_$", "", x)
        return(x)
    }

    UsampleNames <- trim(sub(Uname, "", colnames(mat)[uindex]))
    MsampleNames <- trim(sub(Mname, "", colnames(mat)[mindex]))

    index <- match(UsampleNames,MsampleNames)
    MsampleNames <- MsampleNames[index]
    mindex <- mindex[index]

    if (!identical(UsampleNames,MsampleNames)) {
        stop("Sample names do not match for Meth and Unmeth channels.")
    }

    if (is.data.frame(pData)) pData <- as(pData,"DataFrame")

    if (is.null(pData)) {
        pData <- DataFrame(
            X1 = seq_along(UsampleNames),
            row.names = UsampleNames)
    }

    ann <- .getAnnotationString(c(array = array, annotation = annotation))
    if (!require(ann, character.only = TRUE)) {
        stop(sprintf("cannot load annotation package %s", ann))
    }
    object <- get(ann)

    gr <- getLocations(
        object = object,
        mergeManifest = mergeManifest,
        orderByLocation = TRUE)

    locusNames <- names(gr)

    # NOTE: This might return NAs but it's ok
    # TODO: Fix this. return only what is sent
    common <- intersect(locusNames,rownames(mat))
    if (length(common) == 0) {
        stop("No rowname matches. 'rownames' need to match ",
             "IlluminaHumanMethylation450k probe names.")
    }

    # NOTE: We give no warning if some of the rownames have no match.
    ind1 <- match(common,rownames(mat))
    ind2 <- match(common,locusNames)

    preprocessing <- c(rg.norm = paste0("Data read from file ", filename, "."))
    colnames(mat) <- NULL
    GenomicMethylSet(
        gr =  gr[ind2, ],
        Meth = mat[ind1, mindex],
        Unmeth = mat[ind1, uindex],
        colData = pData,
        preprocessMethod = preprocessing,
        annotation = c(array = array, annotation = annotation))
}

# TODOs ------------------------------------------------------------------------

# TODO: Lots of duplicated code; DRY

# Missing methods --------------------------------------------------------------

# TODO: DelayedArray::type() for all RealizationSink subclasses
setMethod("type", "HDF5RealizationSink", function(x) {
    x@type
})
setMethod("type", "arrayRealizationSink", function(x) {
    DelayedArray::type(x@result_envir$result)
})
setMethod("type", "RleRealizationSink", function(x) {
    x@type
})
# TODO: dimnames() for all RealizationSink subclasses
setMethod("dimnames", "arrayRealizationSink", function(x) {
    dimnames(x@result_envir$result)
})

# Get the 'highest' DelayedArray::type() of array-like objects -----------------
.highestType <- function(...) {
    dots <- list(...)
    types <- vapply(dots, DelayedArray::type, character(1L))
    vector <- do.call(c, lapply(types, vector))  # guaranteed to be atomic
    typeof(vector)
}

# Advanced block processing routines -------------------------------------------

# NOTE: DelayedArray::blockApply() with the option to write the blocks to
#       'sink'. Useful, for example, to apply a function across column-blocks
#       of a DelayedMatrix, write these results to disk, and then wrap
#       these in a DelayedMatrix.
# TODO: See https://github.com/Bioconductor/DelayedArray/issues/10
blockApplyWithRealization <- function(x, FUN, ..., sink = NULL, x_grid = NULL,
                                      sink_grid = NULL, BPREDO = list(),
                                      BPPARAM = bpparam()) {
    FUN <- match.fun(FUN)

    # Check conformable dots_grids and sinks_grids
    x_grid <- DelayedArray:::normarg_grid(x_grid, x)
    sink_grid <- DelayedArray:::normarg_grid(sink_grid, sink)
    if (!identical(dim(x_grid), dim(sink_grid))) {
        stop("non-conformable 'x_grid' and 'sink_grid'")
    }

    # Loop over blocks of `x` and write to `sink`
    nblock <- length(x_grid)
    bplapply(seq_len(nblock), function(b) {
        if (DelayedArray:::get_verbose_block_processing()) {
            message("Processing block ", b, "/", nblock, " ... ",
                    appendLF = FALSE)
        }
        x_viewport <- x_grid[[b]]
        sink_viewport <- sink_grid[[b]]
        block <- read_block(x, x_viewport)
        attr(block, "from_grid") <- x_grid
        attr(block, "block_id") <- b
        block_ans <- FUN(block, ...)
        # NOTE: This is the only part different from DelayedArray::blockApply()
        if (!is.null(sink)) {
            write_block(sink, sink_viewport, block_ans)
            block_ans <- NULL
        }
        if (DelayedArray:::get_verbose_block_processing()) {
            message("OK")
        }
    },
    BPREDO = BPREDO,
    BPPARAM = BPPARAM)
}

# NOTE: A mapply()-like function for conformable arrays.
# NOTE: Different from DelayedArray:::block_Mapply(); designed to have an API
#       more like DelayedArray::blockArray()
# TODO: See https://github.com/Bioconductor/DelayedArray/issues/11
blockMapply <- function(FUN, ..., MoreArgs = NULL, grids = NULL,
                        BPREDO = list(), BPPARAM = bpparam()) {
    FUN <- match.fun(FUN)
    dots <- unname(list(...))
    # Check conformable grids
    if (is.null(grids)) {
        grids <- replicate(length(dots), NULL)
    }
    grids <- mapply(
        FUN = DelayedArray:::normarg_grid,
        grids,
        dots,
        SIMPLIFY = FALSE,
        USE.NAMES = FALSE)
    grids_dims <- lapply(grids, dim)
    all_same_grids_dims <- all(
        vapply(X = grids_dims,
               FUN = function(dim) all(dim == grids_dims[[1L]]),
               FUN.VALUE = logical(1L)))
    if (!all_same_grids_dims) {
        stop("non-conformable grids")
    }
    stopifnot(length(dots) == length(grids))
    nblock <- length(grids[[1]])
    bplapply(seq_len(nblock), function(b) {
        if (DelayedArray:::get_verbose_block_processing()) {
            message("Processing block ", b, "/", nblock, " ... ",
                    appendLF = FALSE)
        }
        viewports <- lapply(grids, function(grid) grid[[b]])
        blocks <- mapply(
            FUN = function(x, grid, viewport) {
                block <- read_block(x, viewport)
                attr(block, "from_grid") <- grid
                attr(block, "block_id") <- b
                block
            },
            x = dots,
            grid = grids,
            viewport = viewports,
            SIMPLIFY = FALSE,
            USE.NAMES = FALSE)
        block_ans <- do.call(FUN, c(blocks, MoreArgs))
        if (DelayedArray:::get_verbose_block_processing()) {
            message("OK")
        }
        block_ans
    },
    BPREDO = BPREDO,
    BPPARAM = BPPARAM)
}

# NOTE: blockMapply() with the option to write the blocks to multiple 'sinks'.
#       Useful, for example, to apply a function across column-blocks of
#       multiple DelayedMatrix objects, write these results to disk, and then
#       wrap these in a DelayedMatrix.
# NOTE: `dots_grids`, `sinks_grids`, and `sinks` should all be lists
# TODO: See https://github.com/Bioconductor/DelayedArray/issues/11
blockMapplyWithRealization <- function(FUN, ..., MoreArgs = NULL, sinks = NULL,
                                       dots_grids = NULL, sinks_grids = NULL,
                                       BPREDO = list(), BPPARAM = bpparam()) {
    FUN <- match.fun(FUN)
    dots <- unname(list(...))
    # Check valid `sinks`
    stopifnot(is.null(sinks) || is.list(sinks))
    # Check conformable dots_grids and sinks_grids
    if (is.null(dots_grids)) {
        dots_grids <- replicate(length(dots), NULL)
    } else {
        stopifnot(is.list(dots_grids))
    }
    dots_grids <- mapply(
        FUN = DelayedArray:::normarg_grid,
        dots_grids,
        dots,
        SIMPLIFY = FALSE,
        USE.NAMES = FALSE)
    if (is.null(sinks_grids)) {
        sinks_grids <- replicate(length(sinks), NULL)
    } else {
        stopifnot(is.list(sinks_grids))
    }
    sinks_grids <- mapply(
        FUN = DelayedArray:::normarg_grid,
        sinks_grids,
        sinks,
        SIMPLIFY = FALSE,
        USE.NAMES = FALSE)
    grids_dims <- lapply(c(dots_grids, sinks_grids), dim)
    all_same_grids_dims <- all(
        vapply(X = grids_dims,
               FUN = function(dim) all(dim == grids_dims[[1L]]),
               FUN.VALUE = logical(1L)))
    if (!all_same_grids_dims) {
        stop("non-conformable 'dots_grids' and 'sinks_grids'")
    }
    stopifnot(length(dots) == length(dots_grids),
              length(sinks) == length(sinks_grids))

    # Loop over blocks of `dots` and write to `sinks`
    nblock <- length(dots_grids[[1]])
    bplapply(seq_len(nblock), function(b) {
        if (DelayedArray:::get_verbose_block_processing()) {
            message("Processing block ", b, "/", nblock, " ... ",
                    appendLF = FALSE)
        }
        input_viewports <- lapply(dots_grids, function(grid) grid[[b]])
        output_viewports <- lapply(sinks_grids, function(grid) grid[[b]])
        blocks <- mapply(
            FUN = function(x, grid, viewport) {
                block <- read_block(x, viewport)
                attr(block, "from_grid") <- grid
                attr(block, "block_id") <- b
                block
            },
            x = dots,
            grid = dots_grids,
            viewport = input_viewports,
            SIMPLIFY = FALSE,
            USE.NAMES = FALSE)
        block_ans <- do.call(FUN, c(blocks, MoreArgs))
        if (!is.list(block_ans)) {
            block_ans <- list(block_ans)
        }
        # NOTE: This is the only part different from blockMapply()
        if (!is.null(sinks)) {
            mapply(function(ans, sink, viewport) {
                write_block(sink, viewport, ans)
            }, ans = block_ans, sink = sinks, viewport = output_viewports,
            SIMPLIFY = FALSE, USE.NAMES = FALSE)
            block_ans <- NULL
        }
        if (DelayedArray:::get_verbose_block_processing()) {
            message("OK")
        }
        block_ans
    },
    BPREDO = BPREDO,
    BPPARAM = BPPARAM)
}

# Internal functions -----------------------------------------------------------

.getSex <- function(CN = NULL, xIndex = NULL, yIndex = NULL, cutoff = -2) {
    if (is.null(CN) || is.null(xIndex) || is.null(yIndex)) {
        stop("must provide CN, xIndex, and yIndex")
    }
    # TODO: This does not handle only females or only males. This ought to be
    #       handled by the 'centers' (see below) being too close together
    xMed <- colMedians(CN, rows = xIndex, na.rm = TRUE)
    yMed <- colMedians(CN, rows = yIndex, na.rm = TRUE)
    dd <- yMed - xMed
    k <- kmeans(dd, centers = c(min(dd), max(dd)))

    sex0 <- ifelse(dd < cutoff, "F", "M")
    sex0 <- .checkSex(sex0)
    sex1 <- ifelse(k$cluster == which.min(k$centers), "F", "M")
    sex1 <- .checkSex(sex1)

    if (!identical(sex0, sex1)) {
        warning("An inconsistency was encountered while determining sex. One ",
                "possibility is that only one sex is present. We recommend ",
                "further checks, for example with the plotSex function.")
    }
    df <- DataFrame(xMed = xMed, yMed = yMed, predictedSex = sex0)
    rownames(df) <- colnames(CN)
    df
}

.checkSex <- function(sex) {
    if (!(is.character(sex) && !any(is.na(sex)) &&
          all(sex %in% c("M", "F")))) {
        stop("'sex' seems wrong (needs to be a character, without ",
             "missing values, of 'M' and 'F'")
    }
    sex
}

# Exported functions -----------------------------------------------------------

getSex <- function(object = NULL, cutoff = -2){
    .isGenomicOrStop(object)
    if (is(object, "GenomicMethylSet")) CN <- getCN(object)
    if (is(object, "GenomicRatioSet")) CN <- getCN(object)
    # TODO: Add test for logarithmic scale or non-log scale
    xIndex <- which(seqnames(object) == "chrX")
    yIndex <- which(seqnames(object) == "chrY")
    .getSex(
        CN = CN,
        xIndex = xIndex,
        yIndex = yIndex,
        cutoff = cutoff)
}

addSex <- function(object, sex = NULL) {
    if (is.null(sex)) {
        sex <- getSex(object)
    }
    if (is(sex, "DataFrame")) {
        stopifnot(all(c("predictedSex", "xMed", "yMed") %in% colnames(sex)))
    }
    sex$predictedSex <- .checkSex(sex$predictedSex)
    .pDataAdd(object, sex)
}

plotSex <- function(object, id = NULL) {
    stopifnot(all(c("predictedSex", "xMed", "yMed") %in%
                      colnames(colData(object))))
    if (is.null(id)) id <- seq_along(object$predictedSex)
    if (length(id) != length(object$predictedSex)) {
        stop("id length must match number of samples.")
    }
    plot(
        x = object$xMed,
        y = object$yMed,
        type = "n",
        xlab = "X chr, median total intensity (log2)",
        ylab = "Y chr, median total intensity (log2)")
    text(
        x = object$xMed,
        y = object$yMed,
        labels = id,
        col = ifelse(object$predictedSex == "M", "deepskyblue", "deeppink3"))
    legend(
        "bottomleft",
        c("M", "F"),
        col = c("deepskyblue", "deeppink3"),
        pch = 16)
}

test_sex <- function() {
    ## stopifnot(require(minfiData))
    ## stopifnot(require(DelayedArray))
    ## data(MsetEx)

    ## # Original tests
    ## gmSet <- mapToGenome(MsetEx)
    ## gmSet <- addSex(gmSet)
    ## checkEquals(gmSet$sex, gmSet$predictedSex)

    ## # Testing with DelayedArray-backed objects
    ## MsetEx <- realize(MsetEx)
    ## gmSet <- mapToGenome(MsetEx)
    ## gmSet <- addSex(gmSet)
    ## checkEquals(gmSet$sex, gmSet$predictedSex)
    checkTrue(TRUE)
}

library(minfiData)
library(digest)

Mset.noob <- preprocessNoob(RGsetEx)
digest_preprocessNoob <- list(Meth = minfi:::.digestMatrix(getMeth(Mset.noob)),
                              Unmeth = minfi:::.digestMatrix(getUnmeth(Mset.noob)))
save(digest_preprocessNoob, file = "../unitTests/digest_preprocessNoob.rda")

sessionInfo()
rm(list = ls())

# ------------------------------------------------------------------------------
# DelayedArray tests
#

library(DelayedArray)

assays(RGsetEx) <- endoapply(assays(RGsetEx), DelayedArray)
Mset.noob <- preprocessNoob(RGsetEx)
digest_preprocessNoob <- list(Meth = minfi:::.digestMatrix(getMeth(Mset.noob)),
                              Unmeth = minfi:::.digestMatrix(getUnmeth(Mset.noob)))
save(digest_preprocessNoob, file = "../unitTests/digest_preprocessNoob.DelayedArray.rda")

sessionInfo()
rm(list = ls())

# Internal functions -----------------------------------------------------------

.guessArrayTypes <- function(nProbes) {
    if (nProbes >= 622000 && nProbes <= 623000) {
        arrayAnnotation <- c(
            array = "IlluminaHumanMethylation450k",
            annotation = .default.450k.annotation)
    } else if (nProbes >= 1050000 && nProbes <= 1053000) {
        # NOTE: "Current EPIC scan type"
        arrayAnnotation <- c(
            array = "IlluminaHumanMethylationEPIC",
            annotation = .default.epic.annotation)
    } else if (nProbes >= 1032000 && nProbes <= 1033000) {
        # NOTE: "Old EPIC scan type"
        arrayAnnotation <- c(
            array = "IlluminaHumanMethylationEPIC",
            annotation = .default.epic.annotation)
    } else if (nProbes >= 1105000 && nProbes <= 1105300) {
        arrayAnnotation <- c(
            array = "IlluminaHumanMethylationEPICv2",
            annotation = .default.epicv2.annotation)
    } else if (nProbes >= 55200 && nProbes <= 55400) {
        arrayAnnotation <- c(
            array = "IlluminaHumanMethylation27k",
            annotation = .default.27k.annotation)
    } else if (nProbes >= 54700 && nProbes <= 54800) {
        arrayAnnotation <- c(
            array = "IlluminaHumanMethylationAllergy",
            annotation = .default.allergy.annotation)
    }  else if (nProbes >= 41000 & nProbes <= 41100) {
        arrayAnnotation <- c(
            array = "HorvathMammalMethylChip40",
            annotation = "test.unknown")
    } else if (nProbes >= 43650 & nProbes <= 43680) {
        arrayAnnotation <- c(
            array = "IlluminaHumanMethylationAllergy",
            annotation = .default.allergy.annotation)
    } else {
        arrayAnnotation <- c(array = "Unknown", annotation = "Unknown")
    }
    arrayAnnotation
}

# Exported functions -----------------------------------------------------------

read.metharray <- function(basenames, extended = FALSE, verbose = FALSE,
                           force = FALSE) {

    # TODO: Need to think about API. Currently, if `BACKEND` is NULL then the
    #       RGSet is matrix-backed. However, within the DelayedArray package,
    #       BACKEND = NULL means to create a DelayedArray with an ordinary
    #       array as the seed. This is a source of slight tension. However, in
    #       general there is no advantage to using a DelayedMatrix with an
    #       ordinary matrix as the seed over directly using the ordinary matrix
    #       (at least within minfi).
    BACKEND <- getAutoRealizationBackend()
    BPREDO <- list()
    BPPARAM <- SerialParam()

    # Check files exist
    basenames <- sub("_Grn\\.idat.*", "", basenames)
    basenames <- sub("_Red\\.idat.*", "", basenames)
    stopifnot(!anyDuplicated(basenames))
    G.files <- paste(basenames, "_Grn.idat", sep = "")
    names(G.files) <- basename(basenames)
    these.dont.exists <- !file.exists(G.files)
    if (any(these.dont.exists)) {
        G.files[these.dont.exists] <- paste0(G.files[these.dont.exists], ".gz")
    }
    if (!all(file.exists(G.files))) {
        noexits <- sub("\\.gz", "", G.files[!file.exists(G.files)])
        stop("The following specified files do not exist:",
             paste(noexits, collapse = ", "))
    }
    R.files <- paste(basenames, "_Red.idat", sep = "")
    names(R.files) <- basename(basenames)
    these.dont.exists <- !file.exists(R.files)
    if (any(these.dont.exists)) {
        R.files[these.dont.exists] <- paste0(R.files[these.dont.exists], ".gz")
    }
    if (!all(file.exists(R.files))) {
        noexits <- R.files[!file.exists(G.files)]
        stop("The following specified files do not exist:",
             paste(noexits, collapse = ", "))
    }

    # Load 'Quants' from IDAT file, including 'SD' and 'NBeads' if extend is
    #       TRUE
    # TODO: Perform fewer realizations. E.g., if the block size is set large
    #       enough so that k arrays' data can be kept in memory then should
    #       only perform a realization every k-th array.
    stime <- system.time({
        G.Quants <- bplapply(G.files, function(xx) {
            if (verbose) message("[read.metharray] Reading ", basename(xx))
            Quants <- readIDAT(xx)[["Quants"]]
            if (!extended) {
                Quants <- Quants[, "Mean", drop = FALSE]
            }
            if (!is.null(BACKEND)) {
                Quants <- realize(Quants, BACKEND = BACKEND)
            }
            Quants
        }, BPREDO = BPREDO, BPPARAM = BPPARAM)
        R.Quants <- bplapply(R.files, function(xx) {
            if (verbose) message("[read.metharray] Reading ", basename(xx))
            Quants <- readIDAT(xx)[["Quants"]]
            if (!extended) {
                Quants <- Quants[, "Mean", drop = FALSE]
            }
            if (!is.null(BACKEND)) {
                Quants <- realize(Quants, BACKEND = BACKEND)
            }
            Quants
        }, BPREDO = BPREDO, BPPARAM = BPPARAM)
    })[3]
    if (verbose) {
        message(sprintf("[read.metharray] Read idat files in %.1f seconds",
                        stime))
    }
    if (verbose) {
        message("[read.metharray] Creating data matrices ... ",
                appendLF = FALSE)
    }
    ptime1 <- proc.time()
    allNProbes <- vapply(G.Quants, nrow, integer(1L))
    arrayTypes <- cbind(do.call(rbind, lapply(allNProbes, .guessArrayTypes)),
                        size = allNProbes)
    sameLength <- (length(unique(arrayTypes[, "size"])) == 1)
    sameArray <- (length(unique(arrayTypes[, "array"])) == 1)

    if (!sameLength && !sameArray) {
        cat("[read.metharray] Trying to parse IDAT files from different arrays.\n")
        cat("  Inferred Array sizes and types:\n")
        print(arrayTypes[, c("array", "size")])
        stop("[read.metharray] Trying to parse different IDAT files, of ",
             "different size and type.")
    }
    if (!sameLength && sameArray && !force) {
        stop("[read.metharray] Trying to parse IDAT files with different ",
             "array size but seemingly all of the same type.\n  You can force ",
             "this by 'force=TRUE', see the man page ?read.metharray")
    }
    commonAddresses <- as.character(
        Reduce("intersect", lapply(G.Quants, rownames)))
    # NOTE: Must manually set colnames because it is not safe to assume these
    #       will be correctly deparsed
    GreenMean <- do.call(
        cbind,
        lapply(G.Quants, function(xx) xx[commonAddresses, "Mean"]))
    colnames(GreenMean) <- names(G.Quants)
    RedMean <- do.call(
        cbind,
        lapply(R.Quants, function(xx) xx[commonAddresses, "Mean"]))
    colnames(RedMean) <- names(R.Quants)
    if (extended) {
        GreenSD <- do.call(
            cbind,
            lapply(G.Quants, function(xx) xx[commonAddresses, "SD"]))
        colnames(GreenSD) <- names(G.Quants)
        RedSD <- do.call(
            cbind,
            lapply(R.Quants, function(xx) xx[commonAddresses, "SD"]))
        colnames(RedSD) <- names(R.Quants)
        NBeads <- do.call(
            cbind,
            lapply(G.Quants, function(xx) xx[commonAddresses, "NBeads"]))
        colnames(NBeads) <- names(G.Quants)
    }
    ptime2 <- proc.time()
    stime <- (ptime2 - ptime1)[3]
    if (verbose) {
        message(sprintf("done in %.1f seconds", stime))
    }
    if (verbose) {
        message("[read.metharray] Instantiating final object ... ",
                appendLF = FALSE)
    }
    ptime1 <- proc.time()
    if (extended) {
        out <- RGChannelSetExtended(
            Red = RedMean,
            Green = GreenMean,
            RedSD = RedSD,
            GreenSD = GreenSD,
            NBeads = NBeads)
    } else {
        out <- RGChannelSet(Red = RedMean, Green = GreenMean)
    }
    rownames(out) <- commonAddresses
    out@annotation <- c(array = arrayTypes[1, 1], annotation = arrayTypes[1, 2])
    ptime2 <- proc.time()
    stime <- (ptime2 - ptime1)[3]
    if (verbose) {
        message(sprintf("done in %.1f seconds", stime))
    }
    out
}

read.metharray.sheet <- function(base, pattern = "csv$", ignore.case = TRUE,
                                 recursive = TRUE, verbose = TRUE) {
    readSheet <- function(file) {
        dataheader <- grep("^\\[DATA\\]", readLines(file), ignore.case = TRUE)
        if (length(dataheader) == 0) dataheader <- 0
        df <- read.csv(file, stringsAsFactor = FALSE, skip = dataheader)
        nam <- grep(
            pattern = "Sentrix_Position",
            x = names(df),
            ignore.case = TRUE,
            value = TRUE)
        if (length(nam) == 1) {
            df$Array <- as.character(df[, nam])
            df[, nam] <- NULL
        }
        nam <- grep(
            pattern = "Array[\\._]ID",
            x = names(df),
            ignore.case = TRUE,
            value = TRUE)
        if (length(nam) == 1) {
            df$Array <- as.character(df[, nam])
            df[, nam] <- NULL
        }
        if (!"Array" %in% names(df)) {
            warning(sprintf("Could not infer array name for file: %s", file))
        }
        nam <- grep("Sentrix_ID", names(df), ignore.case = TRUE, value = TRUE)
        if (length(nam) == 1) {
            df$Slide <- as.character(df[, nam])
            df[, nam] <- NULL
        }
        nam <- grep(
            pattern = "Slide[\\._]ID",
            x = names(df),
            ignore.case = TRUE,
            value = TRUE)
        if (length(nam) == 1) {
            df$Slide <- as.character(df[, nam])
            df[, nam] <- NULL
        }
        if (!"Slide" %in% names(df)) {
            warning(sprintf("Could not infer slide name for file: %s", file))
        } else {
            df[, "Slide"] <- as.character(df[, "Slide"])
        }
        nam <- grep(
            pattern = "Plate[\\._]ID",
            x = names(df),
            ignore.case = TRUE,
            value = TRUE)
        if (length(nam) == 1) {
            df$Plate <- as.character(df[, nam])
            df[, nam] <- NULL
        }
        for (nam in c("Pool_ID", "Sample_Plate", "Sample_Well")) {
            if (nam %in% names(df)) {
                df[[nam]] <- as.character(df[[nam]])
            }
        }

        if (!is.null(df$Array)) {
            patterns <- sprintf("%s_%s_Grn.idat", df$Slide, df$Array)
            allfiles <- list.files(
                path = dirname(file),
                recursive = recursive,
                full.names = TRUE)
            # TODO: Switch sapply() to vapply()
            basenames <- sapply(
                X = patterns,
                FUN = function(xx) grep(xx, allfiles, value = TRUE))
            names(basenames) <- NULL
            basenames <- sub("_Grn\\.idat.*", "", basenames, ignore.case = TRUE)
            df$Basename <- basenames
        }
        df
    }
    if (!all(file.exists(base))) stop("'base' does not exists")
    info <- file.info(base)
    if (!all(info$isdir) && !all(!info$isdir)) {
        stop("'base needs to be either directories or files")
    }
    if (all(info$isdir)) {
        csvfiles <- list.files(
            path = base,
            recursive = recursive,
            pattern = pattern,
            ignore.case = ignore.case,
            full.names = TRUE)
        if (verbose) {
            message("[read.metharray.sheet] Found the following CSV files:")
            print(csvfiles)
        }
    } else {
        csvfiles <- list.files(base, full.names = TRUE)
    }
    dfs <- lapply(csvfiles, readSheet)
    namesUnion <- Reduce(union, lapply(dfs, names))
    df <- do.call(
        what = rbind,
        args = lapply(dfs, function(df) {
            newnames <- setdiff(namesUnion, names(df))
            newdf <- matrix(
                data = NA,
                ncol = length(newnames),
                nrow = nrow(df),
                dimnames = list(NULL, newnames))
            cbind(df, as.data.frame(newdf))
        }))
    df
}

read.metharray.exp <- function(base = NULL, targets = NULL, extended = FALSE,
                               recursive = FALSE, verbose = FALSE,
                               force = FALSE) {
    if (!is.null(targets)) {
        if (!"Basename" %in% names(targets)) {
            stop("Need 'Basename' amongst the column names of 'targets'")
        }
        if (!is.null(base)) {
            files <- file.path(base, basename(targets$Basename))
        } else {
            files <- targets$Basename
        }
        rgSet <- read.metharray(
            basenames = files,
            extended = extended,
            verbose = verbose,
            force = force)
        pD <- targets
        pD$filenames <- files
        rownames(pD) <- colnames(rgSet)
        colData(rgSet) <- as(pD, "DataFrame")
        return(rgSet)
    }
    # Now we just read all files in the directory
    Grn.files <- list.files(
        path = base,
        pattern = "_Grn.idat$",
        recursive = recursive,
        ignore.case = TRUE,
        full.names = TRUE)
    Red.files <- list.files(
        path = base,
        pattern = "_Red.idat$",
        recursive = recursive,
        ignore.case = TRUE,
        full.names = TRUE)
    if (length(Grn.files) == 0 || length(Red.files) == 0) {
        stop("No IDAT files were found")
    }
    commonFiles <- intersect(
        x = sub("_Grn.idat$", "", Grn.files),
        y = sub("_Red.idat$", "", Red.files))
    if (length(commonFiles) == 0) {
        stop("No IDAT files with both Red and Green channel were found")
    }
    commonFiles.Grn <- paste(commonFiles, "_Grn.idat", sep = "")
    commonFiles.Red <- paste(commonFiles, "_Red.idat", sep = "")
    if (!setequal(commonFiles.Grn, Grn.files)) {
        warning(
            sprintf("the following files only exists for the green channel: %s",
                    paste(
                        setdiff(Grn.files, commonFiles.Grn), collapse = ", ")))
    }
    if (!setequal(commonFiles.Red, Red.files)) {
        warning(
            sprintf("the following files only exists for the red channel: %s",
                    paste(
                        setdiff(Red.files, commonFiles.Red), collapse = ", ")))
    }
    read.metharray(
        basenames = commonFiles,
        extended = extended,
        verbose = verbose,
        force = force)
}

# Internal functions -----------------------------------------------------------

clusterMaker4Blocks <- function(gr, relationToIsland, islandName, maxGap,
                                maxClusterWidth) {
    # Get middle position of each island
    if (!is.character(relationToIsland) ||
        !all(unique(relationToIsland) %in%
             c("OpenSea", "Island", "Shelf", "N_Shelf", "S_Shelf", "Shore",
               "N_Shore", "S_Shore")) ||
        length(gr) != length(relationToIsland)) {
        stop("argument 'relationToIsland' is either not a character or seems ",
             "to have wrong values")
    }
    if (!is.character(islandName) || length(gr) != length(islandName)) {
        stop("argument 'islandName' is not a character or it has the wrong ",
             "length")
    }

    fullName <- paste0(islandName, relationToIsland)
    isNotSea <- (fullName != "OpenSea")

    pos <- start(gr)
    # NOTE: These are the "average positions" on shelfs, shores, and islands
    ipos <- round(tapply(pos[isNotSea], fullName[isNotSea], mean))
    # Make non-island/shore/shelf positions their own island
    islFactor <- ifelse(isNotSea, fullName, seq_along(pos))
    # Check which of the new positions correspond to islands/shores/shelfs,
    # i.e. open sea. Assign probes in islands/shores/shelves just one position
    pos[isNotSea] <- ipos[islFactor[isNotSea]]
    # make first pass clusters with new positions
    pns <- boundedClusterMaker(
        chr = as.numeric(seqnames(gr)),
        pos = pos,
        assumeSorted = TRUE,
        maxGap = maxGap,
        maxClusterWidth = maxClusterWidth)
    # But islands must be on their own so we change those beginning and ends of
    # each genomic region type: island, shore, shelf
    types <- unique(relationToIsland)
    # Take out open sea
    types <- types[types != "OpenSea"]
    for (i in types) {
        ind <- relationToIsland == i
        StartEnd <- abs(diff(c(0,ind) != 0))
        add2pns <- cumsum(StartEnd)
        pns <- pns + add2pns
    }
    # where are the islands?
    pns <- as.numeric(as.factor(pns))

    # annotation
    type <- sub("^[NS]_", "", relationToIsland)

    data.frame(pns = pns, type = type)
}


cpgCollapseAnnotation <- function(gr, relationToIsland, islandName,
                                  maxGap = 500, maxClusterWidth = 1500,
                                  blockMaxGap = 2.5*10^5, verbose = TRUE) {

    # Check inputs
    if (is.unsorted(order(gr))) stop("object has to be ordered.")
    # NOTE: This function aggregates probes in islands and the rest are
    #       aggregated into clusters. Then annotation is created for the new
    #       aggregated data the indexes of the original probes along with the
    #       group ids are returned indexes and pns respectively. Note that
    #       these two are redudant but they save me work
    # Make sure chr is a factor.
    # TODO (Kasper): Should this be done from get-go?

    if (verbose) {
        message("[cpgCollapseAnnotation] Clustering islands and clusters of ",
                "probes.\n")
    }

    # Block tab has the block group ids
    blocktab <- clusterMaker4Blocks(
        gr = gr,
        relationToIsland = relationToIsland,
        islandName = islandName,
        maxGap = maxGap,
        maxClusterWidth = maxClusterWidth)

    # Split rows by group id
    groupIndexes <- split(seq_along(blocktab$pns), blocktab$pns)

    # make an object with results
    if (verbose) message("[cpgCollapseAnnotation] Computing new annotation.\n")
    tmpRanges <- t(sapply(groupIndexes, function(ind) range(start(gr)[ind])))

    anno <- GRanges(
        seqnames = Rle(
            tapply(as.vector(seqnames(gr)), blocktab$pns, function(x) x[1])),
        ranges = IRanges(start = tmpRanges[,1], end = tmpRanges[,2]),
        id = as.numeric(names(groupIndexes)),
        type = as.vector(
            tapply(
                as.character(blocktab$type), blocktab$pns, function(x) x[1])))
    res <- list(anno = anno, indexes = groupIndexes)
    seql <- seqlevels(res$anno)
    seqlevels(res$anno, pruning.mode = "coarse") <-
        .seqnames.order[.seqnames.order %in% seql]

    if (verbose) message("[cpgCollapseAnnotation] Defining blocks.\n")
    ind <- (res$anno$type == "OpenSea")
    pns <- rep(NA, length(res$anno))
    pns[ind] <- clusterMaker(
        chr = as.numeric(seqnames(res$anno[ind,])),
        pos = start(res$anno[ind,]),
        maxGap = blockMaxGap)
    res$anno$blockgroup <- pns
    res$pns <- blocktab$pns
    res
}

# Internal generics ------------------------------------------------------------

# `x` is either `meth_signal` or `cn`
# `...` are additional arguments passed to methods.
setGeneric(
    ".cpgCollapse",
    function(x, Indexes, dataSummary, na.rm, verbose, ...)
        standardGeneric(".cpgCollapse"),
    signature = "x")

# Internal methods -------------------------------------------------------------

setMethod(
    ".cpgCollapse",
    "matrix",
    function(x, Indexes, dataSummary, na.rm, verbose) {

        # Set up return matrix
        n_clusters <- length(Indexes)
        n_cpgs_per_cluster <- lengths(Indexes)
        stopifnot(all(n_cpgs_per_cluster) > 0)
        # NOTE: .cpgCollapse() can return non-integer values, so fill a numeric
        #        matrix
        collapsed_x <- matrix(NA_real_, nrow = n_clusters, ncol = ncol(x))

        # Process clusters with a single CpG
        cluster_idx <- n_cpgs_per_cluster == 1L
        cpg_idx <- unlist(Indexes[cluster_idx], use.names = FALSE)
        collapsed_x[cluster_idx, ] <- x[cpg_idx, , drop = FALSE]

        # Process clusters with 0 or > 1 CpG
        cluster_idx <- which(n_cpgs_per_cluster != 1L)
        for (i in cluster_idx) {
            if (verbose) if (runif(1) < 0.0001) cat(".")
            # TODO: If willing to assume dataSummary comes from
            #       DelayedMatrixStats, could use `rows` rather than explicit
            #       subsetting.
            collapsed_x[i, ] <- dataSummary(
                x[Indexes[[i]], , drop = FALSE],
                na.rm = na.rm)
        }
        if (verbose) cat("\n")
        collapsed_x
    })

setMethod(
    ".cpgCollapse",
    "DelayedMatrix",
    function(x, Indexes, dataSummary, na.rm, verbose, BPREDO = list(),
             BPPARAM = SerialParam()) {

        # Set up intermediate RealizationSink object of appropriate dimensions
        # and type
        n_clusters <- length(Indexes)
        n_cpgs_per_cluster <- lengths(Indexes)
        # NOTE: This is ultimately coerced to the output DelayedMatrix object
        # NOTE: .cpgCollapse() can return non-integer values, so fill a "double"
        #       sink
        ans_type <- "double"
        sink <- DelayedArray::AutoRealizationSink(
            dim = c(n_clusters, ncol(x)),
            type = ans_type)
        on.exit(close(sink))

        # Set up ArrayGrid instances over `x` as well as "parallel" ArrayGrid
        # instance over `sink`
        x_grid <- colAutoGrid(x)
        sink_grid <- RegularArrayGrid(
            refdim = dim(sink),
            spacings = c(nrow(sink), ncol(sink) / length(x_grid)))
        # Sanity check ArrayGrid objects have the same dim
        stopifnot(dim(x_grid) == dim(sink_grid))

        # Loop over column-blocks of `x`, perform SWAN normalization, and write
        # to `normalized_x_sink`
        blockApplyWithRealization(
            x = x,
            FUN = .cpgCollapse,
            Indexes = Indexes,
            dataSummary = dataSummary,
            na.rm = na.rm,
            verbose = verbose,
            sink = sink,
            x_grid = x_grid,
            sink_grid = sink_grid,
            BPREDO = BPREDO,
            BPPARAM = BPPARAM)

        # Return as DelayedMatrix
        as(sink, "DelayedArray")
    }
)

# Exported functions -----------------------------------------------------------

# NOTE: Collapses a minfi object into islands, shores, and defines block regions
cpgCollapse <- function(object, what = c("Beta", "M"), maxGap = 500,
                        blockMaxGap = 2.5 * 10^5, maxClusterWidth = 1500,
                        dataSummary = colMeans, na.rm = FALSE,
                        returnBlockInfo = TRUE, islandAnno = NULL,
                        verbose = TRUE, ...) {

    # Check inputs
    # TODO: ?cpgCollapse suggests `object` needn't be a
    #       Genomic[MethlSet|RatioSet] but the code assumes `granges(object)`
    #       works
    what <- match.arg(what)

    # Construct annotation
    if (verbose) message("[cpgCollapse] Creating annotation.\n")
    islands <- .getIslandAnnotation(object = object, islandAnno = islandAnno)
    relationToIsland <- islands$Relation_to_Island
    islandName <- islands$Islands_Name
    gr <- granges(object)
    anno <- cpgCollapseAnnotation(
        gr = gr,
        relationToIsland = relationToIsland,
        islandName = islandName,
        maxGap = maxGap,
        blockMaxGap = blockMaxGap,
        maxClusterWidth = maxClusterWidth,
        verbose = verbose)
    Indexes <- split(seq_along(anno$pns), anno$pns)

    # Collapse data
    if (verbose) message("[cpgCollapse] Collapsing data")
    meth_signal <- getMethSignal(object, what = what, ...)
    collapsed_meth_signal <- .cpgCollapse(
        x = meth_signal,
        Indexes = Indexes,
        dataSummary = dataSummary,
        na.rm = na.rm,
        verbose = verbose)
    cn <- getCN(object,...)
    if (!is.null(cn)) {
        collapsed_cn <- .cpgCollapse(
            x = cn,
            Indexes = Indexes,
            dataSummary = dataSummary,
            na.rm = na.rm,
            verbose = verbose)
    } else {
      collapsed_cn <- NULL 
    }
    
    # Construct output
    preproc <- c(collapse = "cpgCollapse", preprocessMethod(object))
    if (what == "M") {
        ret <- GenomicRatioSet(
            gr = anno$anno,
            Beta = NULL,
            M = collapsed_meth_signal,
            CN = collapsed_cn,
            colData = colData(object),
            annotation = annotation(object),
            preprocessMethod = preproc)
    }
    else {
        ret <- GenomicRatioSet(
            gr = anno$anno,
            Beta = collapsed_meth_signal,
            M = NULL,
            CN = collapsed_cn,
            colData = colData(object),
            annotation = annotation(object),
            preprocessMethod = preproc)
    }
    # NOTE: Take out annotation as we already kept it
    anno <- anno[2:3]
    if (returnBlockInfo) {
        return(list(object = ret, blockInfo = anno))
    }
    ret
}

# NOTE: blockFinder() just uses a cluster object or granges()$blockgroup
#       where clusters are constructed by
#           cpgCollpase() which calls
#               cpgCollapseAnnotation() which calls
#                   clusterMaker4Blocks() which calls
#                       boundedClusterMaker()
blockFinder <- function(object, design, coef = 2, what = c("Beta", "M"),
                        cluster = NULL, cutoff = NULL,
                        pickCutoff = FALSE, pickCutoffQ = 0.99,
                        nullMethod = c("permutation","bootstrap"),
                        smooth = TRUE, smoothFunction = locfitByCluster,
                        B = ncol(permutations), permutations = NULL,
                        verbose = TRUE, bpSpan = 2.5*10^5, ...) {

    # Check inputs
    .isMatrixBackedOrStop(object)
    if (!is(object,"GenomicRatioSet")) stop("object must be 'GenomicRatioSet'")
    if (is.null(cluster)) cluster <- granges(object)$blockgroup
    if (is.null(cluster)) stop("need 'cluster'")
    what <- match.arg(what)
    nullMethod <- match.arg(nullMethod)
    idx <- which(granges(object)$type == "OpenSea")
    if (length(idx) == 0) stop("need OpenSea types in granges(object)")


    pos <- start(granges(object)) / 2 + end(granges(object)) / 2
    res <- bumphunterEngine(
        mat = getMethSignal(object, what)[idx,],
        design = design,
        coef = coef,
        chr = as.character(seqnames(object))[idx],
        pos = pos[idx], cluster = cluster[idx],
        cutoff = cutoff,  pickCutoff = pickCutoff,
        pickCutoffQ = pickCutoffQ,
        nullMethod = nullMethod,
        smooth = smooth,
        smoothFunction = smoothFunction,
        B = B,
        permutations = permutations,
        verbose = verbose,
        bpSpan = bpSpan,...)

    ## FIXME: reindex like below
    res$coef <- bumphunter:::.getEstimate(getMethSignal(object, what), design, coef = coef)

    ## Re-indexing because we only fit the model on the idx indexes
    res$table$indexStart <- idx[res$table$indexStart]
    res$table$indexEnd <- idx[res$table$indexEnd]
    fitted <- rep(NA, length(granges(object)))
    fitted[idx] <- res$fitted
    res$fitted <- fitted
    pvaluesMarginal <- rep(NA,length(granges(object)))
    pvaluesMarginal[idx] <- res$pvaluesMarginal
    res$pvaluesMarginal <- pvaluesMarginal

    return(res)
}

# Exported classes -------------------------------------------------------------

setClass(
    "IlluminaMethylationManifest",
    representation(data = "environment", annotation = "character"))

# Validity methods -------------------------------------------------------------

setValidity("IlluminaMethylationManifest", function(object) {
    msg <- NULL
    if (!"TypeI" %in% ls(object@data) ||
        !is(object@data[["TypeI"]], "DataFrame")) {
        msg <- paste(
            msg,
            "'data' slot must contain a DataFrame with TypeI probes",
            sep = "\n")
    }
    if (!"TypeII" %in% ls(object@data) ||
        !is(object@data[["TypeII"]], "DataFrame")) {
        msg <- paste(
            msg,
            "'data' slot must contain a DataFrame with TypeII probes",
            sep = "\n")
    }
    if (!"TypeControl" %in% ls(object@data) ||
        !is(object@data[["TypeControl"]], "DataFrame")) {
        msg <- paste(
            msg,
            "'data' slot must contain a DataFrame with Control probes",
            sep = "\n")
    }
    if (!"TypeSnpI" %in% ls(object@data) ||
        !is(object@data[["TypeSnpI"]], "DataFrame")) {
        msg <- paste(
            msg,
            "'data' slot must contain a DataFrame with Snp I probes",
            sep = "\n")
    }
    if (!"TypeSnpII" %in% ls(object@data) ||
        !is(object@data[["TypeSnpII"]], "DataFrame")) {
        msg <- paste(
            msg,
            "'data' slot must contain a DataFrame with Snp II probes",
            sep = "\n")
    }

    # Check Names
    if (!all(c("Name", "AddressA", "AddressB", "Color", "nCpG") %in%
             colnames(object@data[["TypeI"]]))) {
        msg <- paste(
            msg,
            "'TypeI' has to have column names 'Name', 'AddressA', 'AddressB', 'Color', 'nCpG'")
    }
    if (!is.character(object@data[["TypeI"]]$Name) ||
        !is.character(object@data[["TypeI"]]$AddressA) ||
        !is.character(object@data[["TypeI"]]$AddressB) ||
        !is.character(object@data[["TypeI"]]$Color) ||
        !is.integer(object@data[["TypeI"]]$nCpG)) {
        msg <- paste(msg, "'TypeI' columns has wrong classes")
    }

    if (!all(c("Name", "AddressA", "nCpG") %in%
             colnames(object@data[["TypeII"]]))) {
        msg <- paste(
            msg, "'TypeII' has to have column names 'Name', 'AddressA', 'nCpG'")
    }
    if (!is.character(object@data[["TypeII"]]$Name) ||
        !is.character(object@data[["TypeII"]]$AddressA) ||
        !is.integer(object@data[["TypeII"]]$nCpG)) {
        msg <- paste(msg, "'TypeII' columns has wrong classes")
    }

    if (!all(c("Type", "Address") %in%
             colnames(object@data[["TypeControl"]]))) {
        msg <- paste(
            msg, "'TypeControl' has to have column names 'Type', 'Address'")
    }
    if (!is.character(object@data[["TypeControl"]]$Type) ||
        !is.character(object@data[["TypeControl"]]$Address)) {
        msg <- paste(msg, "'TypeControl' columns has wrong classes")
    }

    if (!all(c("Name", "AddressA", "AddressB", "Color") %in%
             colnames(object@data[["TypeSnpI"]]))) {
        msg <- paste(
            msg,
            "'TypeSnpI' has to have column names 'Name', 'AddressA', 'AddressB', 'Color'")
    }
    if (!is.character(object@data[["TypeSnpI"]]$Name) ||
        !is.character(object@data[["TypeSnpI"]]$AddressA) ||
        !is.character(object@data[["TypeSnpI"]]$AddressB) ||
        !is.character(object@data[["TypeSnpI"]]$Color)) {
        msg <- paste(msg, "'TypeSnpI' columns has wrong classes")
    }

    if (!all(c("Name", "AddressA") %in% colnames(object@data[["TypeSnpII"]]))) {
        msg <- paste(
            msg, "'TypeSnpII' has to have column names 'Name', 'AddressA'")
    }
    if (!is.character(object@data[["TypeSnpII"]]$Name) ||
        !is.character(object@data[["TypeSnpII"]]$AddressA)) {
        msg <- paste(msg, "'TypeSnpII' columns has wrong classes")
    }

    if (is.null(msg)) TRUE else msg
})

# updateObject() ---------------------------------------------------------------

setMethod(
    "updateObject",
    signature(object = "IlluminaMethylationManifest"),
    function(object, ..., verbose=FALSE) {
        if (verbose) message("updateObject(object = 'IlluminaMethylationManifest')")
        for (name in ls(object@data, all.names=TRUE)) {
            x <- get(name, envir=object@data, inherits=FALSE)
            x <- updateObject(x, ..., verbose=verbose)
            unlockBinding(name, env=object@data)
            assign(name, x, envir=object@data)
            lockBinding(name, env=object@data)
        }
        object
    }
)

# Internal functions -----------------------------------------------------------

.getProbePositionsDetailed <- function(map) {
    # NOTE: map is GRanges with metadata columns strand and type
    stopifnot(is(map, "GRanges"))
    stopifnot(c("Strand", "Type") %in% names(mcols(map)))

    probeStart <- rep(NA, length(map))
    wh.II.F <- which(map$Type == "II" & map$Strand == "+")
    wh.II.R <- which(map$Type == "II" & map$Strand == "-")
    wh.I.F <- which(map$Type == "I" & map$Strand == "+")
    wh.I.R <- which(map$Type == "I" & map$Strand == "-")

    probeStart[wh.II.F] <- start(map)[wh.II.F]
    probeStart[wh.II.R] <- start(map)[wh.II.R] - 50
    probeStart[wh.I.F] <- start(map)[wh.I.F] - 1
    probeStart[wh.I.R] <- start(map)[wh.I.R] - 49
    map$probeStart <- probeStart

    probeEnd <- rep(NA, length(map))
    probeEnd[wh.II.F] <- start(map)[wh.II.F] + 50
    probeEnd[wh.II.R] <- start(map)[wh.II.R]
    probeEnd[wh.I.F] <- start(map)[wh.I.F] + 49
    probeEnd[wh.I.R] <- start(map)[wh.I.R] + 1
    map$probeEnd <- probeEnd

    sbe <- rep(NA, length(map))
    sbe[wh.II.F] <- start(map)[wh.II.F]
    sbe[wh.II.R] <- start(map)[wh.II.R] + 1
    sbe[wh.I.F] <- start(map)[wh.I.F] - 1
    sbe[wh.I.R] <- start(map)[wh.I.R] + 2
    map$SBE <- sbe

    map
}

.getAddressesNotInManifest <- function(rgSet) {
    .isRGOrStop(rgSet)
    addressesInManifest <- c(
        getProbeInfo(rgSet, type = "I")$AddressA,
        getProbeInfo(rgSet, type = "I")$AddressB,
        getProbeInfo(rgSet, type = "II")$AddressA,
        getProbeInfo(rgSet, type = "Control")$Address,
        getProbeInfo(rgSet, type = "SnpI")$AddressA,
        getProbeInfo(rgSet, type = "SnpI")$AddressB,
        getProbeInfo(rgSet, type = "SnpII")$AddressA)
    setdiff(rownames(rgSet), addressesInManifest)
}

# Exported functions -----------------------------------------------------------

IlluminaMethylationManifest <- function(TypeI = DataFrame(),
                                        TypeII = DataFrame(),
                                        TypeControl = DataFrame(),
                                        TypeSnpI = DataFrame(),
                                        TypeSnpII = DataFrame(),
                                        annotation = "") {
    data <- new.env(parent = emptyenv())
    data[["TypeI"]] <- updateObject(TypeI)
    data[["TypeII"]] <- updateObject(TypeII)
    data[["TypeControl"]] <- updateObject(TypeControl)
    data[["TypeSnpI"]] <- updateObject(TypeSnpI)
    data[["TypeSnpII"]] <- updateObject(TypeSnpII)
    lockEnvironment(data, bindings = TRUE)
    new("IlluminaMethylationManifest",
        annotation = annotation,
        data = data)
}

getProbeInfo <- function(object,
                         type = c("I", "II", "Control", "I-Green", "I-Red",
                                  "SnpI", "SnpII")) {
    type <- match.arg(type)
    if (type %in% c("I", "II", "Control", "SnpI", "SnpII")) {
        out <- getManifest(object)@data[[paste("Type", type, sep = "")]]
    }
    if (type == "I-Green") {
        out <- getManifest(object)@data[["TypeI"]]
        out <- out[out$Color == "Grn",]
    }
    if (type == "I-Red") {
        out <- getManifest(object)@data[["TypeI"]]
        out <- out[out$Color == "Red",]
    }
    if (is(object, "RGChannelSet")) {
        if ("Address" %in% names(out)) {
            out <- out[out$Address %in% rownames(object), ]
        }
        if ("AddressA" %in% names(out)) {
            out <- out[out$AddressA %in% rownames(object), ]
        }
        if ("AddressB" %in% names(out)) {
            out <- out[out$AddressB %in% rownames(object), ]
        }
    }
    if (.isMethylOrRatio(object)) {
        out <- out[out$Name %in% rownames(object),]
    }
    updateObject(out)
}

getManifestInfo <- function(object, type = c("nLoci", "locusNames")) {
    type <- match.arg(type)
    switch(type,
           "nLoci" = {
               nrow(getProbeInfo(object, type = "I")) +
                   nrow(getProbeInfo(object, type = "II"))
           },
           "locusNames" = {
               c(getProbeInfo(object, type = "I")$Name,
                 getProbeInfo(object, type = "II")$Name)
           })
}

getControlAddress <- function(object,
                              controlType = c("NORM_A", "NORM_C", "NORM_G",
                                              "NORM_T"),
                              asList = FALSE) {
    if (asList) {
        ctrls <- getProbeInfo(object, type = "Control")
        out <- split(ctrls$Address, ctrls$Type)
        return(out[tolower(names(out)) %in% tolower(controlType)])
    } else {
        ctrls <- getProbeInfo(object, type = "Control")
        out <- ctrls[tolower(ctrls$Type) %in% tolower(controlType), "Address"]
    }
    out
}

# Exported methods -------------------------------------------------------------

setMethod("show", "IlluminaMethylationManifest", function(object) {
    cat("IlluminaMethylationManifest object\n")
    object <- updateObject(object)
    .show.annotation(object@annotation)
    cat("Number of type I probes:", nrow(object@data[["TypeI"]]), "\n")
    cat("Number of type II probes:", nrow(object@data[["TypeII"]]), "\n")
    cat("Number of control probes:", nrow(object@data[["TypeControl"]]), "\n")
    cat("Number of SNP type I probes:", nrow(object@data[["TypeSnpI"]]), "\n")
    cat("Number of SNP type II probes:", nrow(object@data[["TypeSnpII"]]), "\n")
})

setMethod(
    "getManifest",
    signature(object = "IlluminaMethylationManifest"),
          function(object) object # removed updateObject(object)
)

setMethod("getManifest", signature(object = "character"), function(object) {
    maniString <- .getManifestString(object)
    if (!require(maniString, character.only = TRUE))
        stop(sprintf("cannot load manifest package %s", maniString))
    ## updateObject(get(maniString))
    get(maniString)
})

## ToDo
##  - handle non-HDF5 backends; ie. both DelayedArray and non-DelayedArray
##  - code sharing between DelayedArray and non-DelayedArray
##  - DONE dealing with the case of arrays with unequal dimension 
##  - SEMI-DONE set chunkdim somehow in arguments, perhaps compression level as well
##  - DONE include output dir in arguments
##  - return with right annotation
##  - POSTPONED TO WAIT ON PETE return a fully saved SE
##  - POSTPONED TO WAIT ON PETE progress bar
##  - DONE verbosity



read.metharray2 <- function(basenames,
                            extended = FALSE,
                            verbose = FALSE,
                            dir = "my_h5_se",
                            nArrays = 1,
                            BPPARAM = bpparam("SerialParam"),
                            nArrays.chunk = 1) {
    ## This function assumes that all IDATs are identical.
    ## This is not a safe assumption; will be relaxed later.
    ## At least I check for it.
    
    ## Setting up file names
    basenames <- sub("_Grn\\.idat.*", "", basenames)
    basenames <- sub("_Red\\.idat.*", "", basenames)
    stopifnot(!anyDuplicated(basenames))
    G.files <- paste(basenames, "_Grn.idat", sep = "")
    names(G.files) <- basename(basenames)
    these.dont.exists <- !file.exists(G.files)
    if (any(these.dont.exists)) {
        G.files[these.dont.exists] <- paste0(G.files[these.dont.exists], ".gz")
    }
    if (!all(file.exists(G.files))) {
        noexits <- sub("\\.gz", "", G.files[!file.exists(G.files)])
        stop("The following specified files do not exist:",
             paste(noexits, collapse = ", "))
    }
    R.files <- paste(basenames, "_Red.idat", sep = "")
    names(R.files) <- basename(basenames)
    these.dont.exists <- !file.exists(R.files)
    if (any(these.dont.exists)) {
        R.files[these.dont.exists] <- paste0(R.files[these.dont.exists], ".gz")
    }
    if (!all(file.exists(R.files))) {
        noexits <- R.files[!file.exists(G.files)]
        stop("The following specified files do not exist:",
             paste(noexits, collapse = ", "))
    }
    ans_ncol <- length(G.files)

    ## First we check whether these are all the same array type and has the
    ## same number of probes

    nProbes <- sapply(G.files, readIDAT, what = "nSNPsRead") # FIXME: potentially parallize this
    arrayTypes <- cbind(do.call(rbind, lapply(nProbes, minfi:::.guessArrayTypes)),
                        size = nProbes)
    sameLength <- (length(unique(arrayTypes[, "size"])) == 1)
    sameArray <- (length(unique(arrayTypes[, "array"])) == 1)
    if (!sameArray) {
        cat("[read.metharray2] Trying to parse IDAT files from different arrays.\n")
        cat("  Inferred Array sizes and types:\n")
        print(arrayTypes[, c("array", "size")])
        stop("[read.metharray2] Trying to parse different IDAT files, of ",
             "different size and type." )
    }
    if (sameArray && !sameLength && !force) {
        stop("[read.metharray] Trying to parse IDAT files with different ",
             "array size but seemingly all of the same type.\n  You can force ",
             "this by 'force=TRUE', see the man page ?read.metharray")
    }
    if (sameArray && !sameLength && force) {
        ## get the rownames as bpiterate
        iterator <- function(files, nArrays) {
            grid <- RegularArrayGrid(length(files), nArrays)
            gridIdx <- lapply(seq_along(grid), function(xx) {
                ir <- ranges(grid[[xx]])
                seq(from = start(ir), to = end(ir))
            })
            b <- 0L
            function() {
                if(b == length(grid)) return(NULL)
                b <<- b + 1L
                gridIdx[[b]]
            }
        }
        computer <- function(files) {
            Reduce("intersect", lapply(files, readIDAT, what = "IlluminaIDs"))
        }
        reducer <- function(x, y) {
            intersect(x,y)
        }
        rownames.out <- bpiterate(iterator(G.files, nArrays = nArrays),
                                  FUN = computer,
                                  REDUCE = reducer,
                                  BPPARAM = BPPARAM)
    }
    if (sameArray && sameLength) {
        first.file <- readIDAT(G.files[1])
        rownames.out <- rownames(first.file$Quants)
    }
    ans_nrow <- length(rownames.out)
    grid <- RegularArrayGrid(c(ans_nrow, ans_ncol),
                             c(ans_nrow, nArrays)) 
    readAndCheck <- function(b, files, grid, sink, sink_lock, rownames.out) {
        ## Before calling this function, we ensure that rownames.out
        ## is a subset of the rownames in the file.
        cat(".")
        viewport <- grid[[b]]
        col_range <- ranges(viewport)[2L]
        col_indices <- seq(from = start(col_range), to = end(col_range))
        Quants <- lapply(files[col_indices], function(xx) {
            readIDAT(xx)[["Quants"]]
        })
        Mean <- do.call(cbind, lapply(Quants, function(xx) xx[rownames.out, "Mean", drop=FALSE]))
        write_block(sink, viewport = viewport, block = Mean)
    }
    Gmean_sink <- HDF5RealizationSink(
        dim = c(ans_nrow, ans_ncol),
        filepath = file.path(dir, "assays.h5"),
        dimnames = NULL, # NOTE: Never allow dimnames.
        type = "integer",
        name = "Green",
        chunkdim = c(ans_nrow, nArrays.chunk))
    on.exit(close(Gmean_sink), add = TRUE)
    sink_lock <- ipcid()
    on.exit(ipcremove(sink_lock), add = TRUE)
    if(verbose)
        message(sprintf("[read.metharray2]: reading %i Green files\n", length(G.files)))
    Gmean <- bplapply(seq_along(grid),
                      FUN = readAndCheck,
                      files = G.files,
                      grid = grid, 
                      sink = Gmean_sink,
                      sink_lock = sink_lock,
                      rownames.out = rownames.out,
                      BPPARAM = BPPARAM)
    ipcremove(sink_lock)
    Gmean <- as(Gmean_sink, "DelayedArray")
    Rmean_sink <- HDF5RealizationSink(
        dim = c(ans_nrow, ans_ncol),
        filepath = file.path(dir, "assays.h5"),
        dimnames = NULL, # NOTE: Never allow dimnames.
        type = "integer",
        name = "Red",
        chunkdim = c(ans_nrow, nArrays.chunk))
    on.exit(close(Rmean_sink), add = TRUE)
    if(verbose)
        message(sprintf("[read.metharray2]: reading %i Red files\n", length(R.files)))
    Rmean <- bplapply(seq_along(grid),
                      FUN = readAndCheck,
                      files = R.files,
                      grid = grid, 
                      sink = Rmean_sink,
                      sink_lock = sink_lock,
                      rownames.out = rownames.out,
                      BPPARAM = BPPARAM)
    Rmean <- as(Rmean_sink, "DelayedArray")
    out <- RGChannelSet(Red = Rmean, Green = Gmean)
    rownames(out) <- rownames.out
    colnames(out) <- names(G.files)
    colData(out)$filename
    ## TODO: set annotation slot
    out
}

# Internal generics ------------------------------------------------------------

# NOTE: `...` are additional arguments passed to methods.
setGeneric(
    ".detectionP",
    function(Red, Green, locusNames, controlIdx, TypeI.Red, TypeI.Green, TypeII,
             ...)
        standardGeneric(".detectionP"),
    signature = c("Red", "Green"))

# Internal methods -------------------------------------------------------------

# NOTE: `...` is ignored
setMethod(
    ".detectionP",
    c("matrix", "matrix"),
    function(Red, Green, locusNames, controlIdx, TypeI.Red, TypeI.Green, TypeII,
             ...) {
        # Set up output matrix with appropriate dimensions and type
        detP <- matrix(NA_real_,
                       nrow = length(locusNames),
                       ncol = ncol(Red),
                       dimnames = list(locusNames, colnames(Red)))

        # Compute summary statistics needed for calculations
        rBg <- Red[controlIdx, , drop = FALSE]
        rMu <- colMedians(rBg)
        rSd <- colMads(rBg)
        gBg <- Green[controlIdx, , drop = FALSE]
        gMu <- colMedians(gBg)
        gSd <- colMads(gBg)

        # Fill output matrix
        for (j in seq_len(ncol(detP))) {
            # Type I Red
            intensity <- Red[TypeI.Red$AddressA, j] + Red[TypeI.Red$AddressB, j]
            detP[TypeI.Red$Name, j] <- pnorm(
                q = intensity,
                mean = 2 * rMu[j],
                sd = 2 * rSd[j],
                lower.tail = FALSE)
            # Type I Green
            intensity <- Green[TypeI.Green$AddressA, j] +
                Green[TypeI.Green$AddressB, j]
            detP[TypeI.Green$Name, j] <- pnorm(
                q = intensity,
                mean = 2 * gMu[j],
                sd = 2 * gSd[j],
                lower.tail = FALSE)
            # Type II
            intensity <- Red[TypeII$AddressA, j] + Green[TypeII$AddressA, j]
            detP[TypeII$Name, j] <- pnorm(
                q = intensity,
                mean = rMu[j] + gMu[j],
                sd = rSd[j] + gSd[j],
                lower.tail = FALSE)
        }

        # Return output matrix
        detP
    }
)

setMethod(
    ".detectionP",
    c("DelayedMatrix", "DelayedMatrix"),
    function(Red, Green, locusNames, controlIdx, TypeI.Red, TypeI.Green, TypeII,
             BPREDO = list(), BPPARAM = SerialParam()) {
        # Set up intermediate RealizationSink object of appropriate dimensions
        # and type
        # NOTE: This is ultimately coerced to the output DelayedMatrix objects,
        #       `detP`
        detP_sink <- DelayedArray::AutoRealizationSink(
            dim = c(length(locusNames), ncol(Red)),
            dimnames = list(locusNames, colnames(Red)),
            type = "double")
        on.exit(close(detP_sink))

        # Set up ArrayGrid instances over `Red` and `Green` as well as
        # "parallel" ArrayGrid instance over `detP_sink`.
        Red_grid <- colAutoGrid(Red)
        Green_grid <- colAutoGrid(Green)
        detP_sink_grid <- RegularArrayGrid(
            refdim = dim(detP_sink),
            spacings = c(nrow(detP_sink), ncol(detP_sink) / length(Red_grid)))
        # Sanity check ArrayGrid objects have the same dim
        stopifnot(dim(Red_grid) == dim(Green_grid),
                  dim(Red_grid) == dim(detP_sink_grid))

        # Loop over blocks of `Red` and `Green` and write to `detP_sink`
        blockMapplyWithRealization(
            FUN = .detectionP,
            Red = Red,
            Green = Green,
            MoreArgs = list(
                locusNames = locusNames,
                controlIdx = controlIdx,
                TypeI.Red = TypeI.Red,
                TypeI.Green = TypeI.Green,
                TypeII = TypeII),
            sinks = list(detP_sink),
            dots_grids = list(Red_grid, Green_grid),
            sinks_grids = list(detP_sink_grid),
            BPREDO = BPREDO,
            BPPARAM = BPPARAM)

        # Return as DelayedMatrix
        as(detP_sink, "DelayedArray")
    }
)

# Exported functions -----------------------------------------------------------

# TODO: Document: because we simultaneously walk over column-blocks of
#       `Red` and `Green`, the number of elements loaded into memory is
#       doubled. If running into memory issues, try halving
#       getOption("DelayedArray.block.size")
detectionP <- function(rgSet, type = "m+u") {
    .isRGOrStop(rgSet)

    # Extract data to pass to low-level function that constructs `detP`
    locusNames <- getManifestInfo(rgSet, "locusNames")
    controlIdx <- getControlAddress(rgSet, controlType = "NEGATIVE")
    Red <- getRed(rgSet)
    Green <- getGreen(rgSet)
    TypeI.Red <- getProbeInfo(rgSet, type = "I-Red")
    TypeI.Green <- getProbeInfo(rgSet, type = "I-Green")
    TypeII <- getProbeInfo(rgSet, type = "II")

    # Construct `detP`
    detP <- .detectionP(
        Red = Red,
        Green = Green,
        locusNames = locusNames,
        controlIdx = controlIdx,
        TypeI.Red = TypeI.Red,
        TypeI.Green = TypeI.Green,
        TypeII = TypeII)

    detP
}

# Global variables -------------------------------------------------------------

utils::globalVariables("oob")

# Internal functions -----------------------------------------------------------

# TODO: Profile
normexp.get.xs <- function(xf, controls, offset = 50, verbose = FALSE) {
    if (verbose) {
        message(
            "[normexp.get.xs] Background mean & SD estimated from ",
            nrow(controls),
            " probes")
    }
    mu <- sigma <- alpha <- rep(NA_real_, ncol(xf))
    for (i in seq_len(ncol(xf))) {
        ests <- huber(controls[, i])
        mu[i] <- ests$mu
        sigma[i] <- ests$s
        alpha[i] <- max(huber(xf[, i])$mu - mu[i], 10)
    }
    pars <- data.frame(mu = mu, lsigma = log(sigma), lalpha = log(alpha))
    for (i in seq_len(ncol(xf))) {
        xf[, i] <- normexp.signal(as.numeric(pars[i, ]), xf[, i])
    }
    list(
        xs = xf + offset,
        params = data.frame(
            mu = mu, sigma = sigma, alpha = alpha, offset = offset),
        meta = c("background mean", "background SD", "signal mean", "offset"))
}

# TODO: Profile
normexp.get.xcs <- function(xcf, params) {
    stopifnot(any(grepl("mu", names(params))),
              any(grepl("sigma", names(params))),
              any(grepl("alpha", names(params))),
              any(grepl("offset", names(params))))
    pars <- data.frame(
        mu = params[[grep("mu", names(params), value = TRUE)]],
        sigma = log(params[[grep("sigma", names(params), value = TRUE)]]),
        alpha = log(params[[grep("alpha", names(params), value = TRUE)]]))
    for (i in seq_len(ncol(xcf))) {
        xcf[, i] <- normexp.signal(as.numeric(pars[i, ]), xcf[, i])
    }

    xcf + params[[grep("offset", names(params), value = TRUE)]][1]
}

# From TJT (2016-06-16)
# Performing dye bias normalization
#
# "single" = just reciprocate out the dye bias, don't use a reference. (similar
#            to, but implemented differently from, unmaintained "asmn" package
#            by Decker et al., doi:10.4161/epi.26037)
# "reference" = use the least-worst sample in the batch (previous default)
#
# "single" is now the default: it provides single-sample preprocessing and
# betas/M-values produced by this method are identical to those from the
# "reference" version used in (e.g.) the TCGA data processing pipeline.
dyeCorrection <- function(Meth, Unmeth, Red, Green, control_probes,
                          Green_probes, Red_probes, d2.probes, estimates,
                          array_type, dyeMethod, verbose) {

    # Background correct the Illumina normalization controls
    redControls <- Red[control_probes$Address, , drop = FALSE]
    greenControls <- Green[control_probes$Address, ,drop = FALSE]
    rownames(redControls) <- rownames(greenControls) <-
        control_probes$Type
    internal.controls <- list(Green = greenControls, Red = redControls)
    xcs <- lapply(names(internal.controls), function(nch) {
        xcf <- as.matrix(internal.controls[[nch]])
        normexp.get.xcs(xcf = xcf, params = estimates[[nch]][["params"]])
    })
    names(xcs) <- names(internal.controls)
    internal.controls[["Green"]] <- xcs[["Green"]]
    internal.controls[["Red"]] <- xcs[["Red"]]

    if (array_type %in% c("IlluminaHumanMethylation27k")) {
        CG.controls <- which(
            rownames(internal.controls[[1]]) %in% c("Normalization-Green"))
        AT.controls <- which(
            rownames(internal.controls[[1]]) %in% c("Normalization-Red"))
    } else {
        CG.controls <- which(
            rownames(internal.controls[[1]]) %in% c("NORM_C", "NORM_G"))
        AT.controls <- which(
            rownames(internal.controls[[1]]) %in% c("NORM_A", "NORM_T"))
    }

    # Dye bias normalization with the corrected Illumina control probes
    Green.avg <- colMeans2(x = internal.controls[["Green"]], rows = CG.controls)
    Red.avg <- colMeans2(x = internal.controls[["Red"]], rows = AT.controls)
    R.G.ratio <- Red.avg / Green.avg

    if (dyeMethod == "single") {
        if (verbose) {
            message(
                "[dyeCorrection] Applying R/G ratio flip to fix dye bias")
        }
        Red.factor <- 1 / R.G.ratio
        Green.factor <- 1
    } else if (dyeMethod == "reference") {
        reference <- which.min(abs(R.G.ratio - 1))
        if (verbose) {
            message(
                "[dyeCorrection] Using sample number ",
                reference,
                " as reference level")
        }
        ref <- (Green.avg + Red.avg)[reference] / 2
        if (is.na(ref)) {
            stop("'reference' refers to an array that is not present")
        }
        Green.factor <- ref / Green.avg
        Red.factor <- ref / Red.avg
    }

    Green <- list(
        M = Meth[Green_probes,, drop=FALSE],
        U = Unmeth[Green_probes,, drop=FALSE],
        D2 = Meth[d2.probes,, drop=FALSE])
    Red <- list(
        M = Meth[Red_probes,, drop=FALSE],
        U = Unmeth[Red_probes,, drop=FALSE],
        D2 = Unmeth[d2.probes,, drop=FALSE ])

    # NOTE: Adjust Red regardless of reference or equalization approach
    #       but only adjust Green if using older reference method
    Red <- lapply(
        X = Red,
        FUN = function(x) {
            sweep(x, MARGIN = 2, STATS = Red.factor, FUN = "*")
        })
    Meth[Red_probes, ] <- Red$M
    Unmeth[Red_probes, ] <- Red$U
    Unmeth[d2.probes, ] <- Red$D2
    if (dyeMethod == "reference") {
        Green <- lapply(
            X = Green,
            FUN = function(x) {
                sweep(x, MARGIN = 2, STATS = Green.factor, FUN = "*")
            })
        Meth[Green_probes, ] <- Green$M
        Unmeth[Green_probes, ] <- Green$U
        Meth[d2.probes, ] <- Green$D2
    }

    list(Meth = Meth, Unmeth = Unmeth)
}

# Internal generics ------------------------------------------------------------

# `...` are additional arguments passed to methods.
setGeneric(
    ".preprocessNoob",
    function(Meth, Unmeth, GreenOOB, RedOOB, Green_probes, Red_probes,
             d2.probes, offset, dyeCorr, Red, Green, control_probes, array_type,
             dyeMethod, verbose, ...)
        standardGeneric(".preprocessNoob"),
    signature = c("Meth", "Unmeth")
)

# Internal methods -------------------------------------------------------------

setMethod(
    ".preprocessNoob",
    c("matrix", "matrix"),
    function(Meth, Unmeth, GreenOOB, RedOOB, Green_probes, Red_probes,
             d2.probes, offset, dyeCorr, Red, Green, control_probes, array_type,
             dyeMethod, verbose) {
        subverbose <- max(as.integer(verbose) - 1L, 0)

        # Threshold Meth and Unmeth to be positive
        Meth[Meth <= 0] <- 1L
        Unmeth[Unmeth <= 0] <- 1L
        
        # NormExp estimates for Green and Red
        dat <- list(
            Green = list(
                M =  Meth[Green_probes, , drop = FALSE],
                U =  Unmeth[Green_probes, , drop = FALSE],
                D2 = Meth[d2.probes, , drop = FALSE]),
            Red = list(
                M =  Meth[Red_probes, , drop = FALSE],
                U =  Unmeth[Red_probes, , drop = FALSE],
                D2 = Unmeth[d2.probes, , drop = FALSE]))
        estimates <- lapply(names(dat), function(nch, oob) {
            xf <- rbind(
                dat[[nch]][["M"]],
                dat[[nch]][["U"]],
                dat[[nch]][["D2"]])
            xs <- normexp.get.xs(
                xf = xf,
                controls = oob[[nch]],
                offset = offset,
                verbose = subverbose)
            names(xs[["params"]]) <- paste(
                names(xs[["params"]]), nch, sep = ".")
            names(xs[["meta"]]) <- paste(names(xs[["meta"]]), nch, sep = ".")
            xs
        }, oob = list(Green = GreenOOB, Red = RedOOB))
        names(estimates) <- names(dat)

        # Correct for Green and Red in Meth and Unmeth
        rows <- lapply(dat, function(x) vapply(x, nrow, integer(1L)))
        last <- lapply(rows, cumsum)
        first <- Map(function(last, rows) last - rows + 1, last, rows)
        if (length(Green_probes) > 0) {
            Green.M <- seq(first[["Green"]][["M"]], last[["Green"]][["M"]])
            Meth[Green_probes, ] <- estimates[["Green"]][["xs"]][Green.M, ]
            Green.U <- seq(first[["Green"]][["U"]], last[["Green"]][["U"]])
            Unmeth[Green_probes, ] <- estimates[["Green"]][["xs"]][Green.U, ]
        }
        if (length(Red_probes) > 0) {
            Red.M <- seq(first[["Red"]][["M"]], last[["Red"]][["M"]])
            Meth[Red_probes, ] <- estimates[["Red"]][["xs"]][Red.M, ]
            Red.U <- seq(first[["Red"]][["U"]], last[["Red"]][["U"]])
            Unmeth[Red_probes, ] <- estimates[["Red"]][["xs"]][Red.U, ]
        }
        if (length(d2.probes) > 0) {
            d2.M <- seq(first[["Green"]][["D2"]], last[["Green"]][["D2"]])
            d2.U <- seq(first[["Red"]][["D2"]], last[["Red"]][["D2"]])
            Meth[d2.probes, ] <- estimates[["Green"]][["xs"]][d2.M, ]
            Unmeth[d2.probes, ] <- estimates[["Red"]][["xs"]][d2.U, ]
        }

        if (!dyeCorr) {
            return(list(Meth = Meth, Unmeth = Unmeth))
        }
        dyeCorrection(
            Meth = Meth,
            Unmeth = Unmeth,
            Red = Red,
            Green = Green,
            control_probes = control_probes,
            Green_probes = Green_probes,
            Red_probes = Red_probes,
            d2.probes = d2.probes,
            estimates = estimates,
            array_type = array_type,
            dyeMethod = dyeMethod,
            verbose = verbose)
    }
)

setMethod(
    ".preprocessNoob",
    c("DelayedMatrix", "DelayedMatrix"),
    function(Meth, Unmeth, GreenOOB, RedOOB, Green_probes, Red_probes,
             d2.probes, offset, dyeCorr, Red, Green, control_probes,
             array_type, dyeMethod, verbose, BPREDO = list(),
             BPPARAM = SerialParam()) {
        # Set up intermediate RealizationSink objects of appropriate dimensions
        # and type
        # NOTE: These are ultimately coerced to the output DelayedMatrix
        #       objects, `Meth` and `Unmeth`
        # NOTE: Noob can return non-integer values, so write to a numeric sink
        # NOTE: Don't do `Unmeth_sink <- Meth__sink` or else these will
        #       reference the same object and clobber each other when written
        #       to!
        ans_type <- "double"
        M_sink <- DelayedArray::AutoRealizationSink(
            dim = dim(Meth),
            dimnames = dimnames(Meth),
            type = ans_type)
        on.exit(close(M_sink))
        U_sink <- DelayedArray::AutoRealizationSink(
            dim = dim(Unmeth),
            dimnames = dimnames(Unmeth),
            type = ans_type)
        on.exit(close(U_sink), add = TRUE)

        # Set up ArrayGrid instances over `Meth`, `Unmeth`, `M_sink`, and
        # `U_sink`. Also set up "parallel" ArrayGrid instances over `GreenOOB`
        # and `RedOOB`, as well as `Red` and `Green` if performing dye
        # correction.
        Meth_grid <- colAutoGrid(Meth)
        Unmeth_grid <- colAutoGrid(Unmeth)
        M_sink_grid <- RegularArrayGrid(
            refdim = dim(M_sink),
            spacings = c(nrow(M_sink), ncol(M_sink) / length(Meth_grid)))
        U_sink_grid <- M_sink_grid
        GreenOOB_grid <- RegularArrayGrid(
            refdim = dim(GreenOOB),
            spacings = c(nrow(GreenOOB), ncol(GreenOOB) / length(Meth_grid)))
        RedOOB_grid <- RegularArrayGrid(
            refdim = dim(RedOOB),
            spacings = c(nrow(RedOOB), ncol(RedOOB) / length(Meth_grid)))
        # Sanity check ArrayGrid objects have the same dim
        stopifnot(identical(dim(Meth_grid), dim(Unmeth_grid)),
                  identical(dim(Meth_grid), dim(M_sink_grid)),
                  identical(dim(Meth_grid), dim(GreenOOB_grid)),
                  identical(dim(Meth_grid), dim(RedOOB_grid)))

        # Loop over blocks of `Meth`, `Unmeth`, `GreenOOB`, and `RedOOB`, as
        # well as `Red` and `Green` if doing dye correction, and write to
        # `M_sink` and `U_sink`.
        if (dyeCorr) {
            Red_grid <- RegularArrayGrid(
                refdim = dim(Red),
                spacings = c(nrow(Red), ncol(Red) / length(Meth_grid)))
            Green_grid <- RegularArrayGrid(
                refdim = dim(Green),
                spacings = c(nrow(Green), ncol(Green) / length(Meth_grid)))
            # Sanity check ArrayGrid objects have the same dim
            stopifnot(dim(Red_grid) == dim(Green_grid),
                      dim(Red_grid) == dim(Meth_grid))
            blockMapplyWithRealization(
                FUN = .preprocessNoob,
                Meth = Meth,
                Unmeth = Unmeth,
                GreenOOB = GreenOOB,
                RedOOB = RedOOB,
                Red = Red,
                Green = Green,
                MoreArgs = list(
                    Green_probes = Green_probes,
                    Red_probes = Red_probes,
                    d2.probes = d2.probes,
                    offset = offset,
                    dyeCorr = dyeCorr,
                    control_probes = control_probes,
                    array_type = array_type,
                    dyeMethod = dyeMethod,
                    verbose = verbose),
                sinks = list(M_sink, U_sink),
                dots_grids = list(Meth_grid, Unmeth_grid,
                                  GreenOOB_grid, RedOOB_grid,
                                  Red_grid, Green_grid),
                sinks_grids = list(M_sink_grid, U_sink_grid),
                BPREDO = BPREDO,
                BPPARAM = BPPARAM)
        } else {
            blockMapplyWithRealization(
                FUN = .preprocessNoob,
                Meth = Meth,
                Unmeth = Unmeth,
                GreenOOB = GreenOOB,
                RedOOB = RedOOB,
                MoreArgs = list(
                    oob = oob,
                    Green_probes = Green_probes,
                    Red_probes = Red_probes,
                    d2.probes = d2.probes,
                    offset = offset,
                    dyeCorr = dyeCorr,
                    control_probes = control_probes,
                    array_type = array_type,
                    dyeMethod = dyeMethod,
                    verbose = verbose),
                sinks = list(M_sink, U_sink),
                dots_grids = list(Meth_grid, Unmeth_grid,
                                  GreenOOB_grid, RedOOB_grid),
                sinks_grids = list(M_sink_grid, U_sink_grid),
                BPREDO = BPREDO,
                BPPARAM = BPPARAM)
        }

        # Return as DelayedMatrix objects
        M <- as(M_sink, "DelayedArray")
        U <- as(U_sink, "DelayedArray")

        list(Meth = M, Unmeth = U)
    }
)

# Exported functions -----------------------------------------------------------

# TODO: Document: because we simultaneously walk over column-blocks of `Meth`,
#       and `Unmeth`, as well as `Red` and `Green` if doing dye correction, the
#       number of elements loaded into memory is doubled or quadrupled. If
#       running into memory issues, try halving/quartering
#       getOption("DelayedArray.block.size")
preprocessNoob <- function(rgSet, offset = 15, dyeCorr = TRUE, verbose = FALSE,
                           dyeMethod = c("single", "reference")) {

    # Check inputs
    .isRGOrStop(rgSet)
    dyeMethod <- match.arg(dyeMethod)

    # Extract data to pass to low-level functions that construct `M` and `U`
    oob <- getOOB(rgSet)
    GreenOOB <- oob[["Grn"]]
    RedOOB <- oob[["Red"]]
    MSet <- preprocessRaw(rgSet)
    anno <- getAnnotation(MSet)
    probe.type <- paste0(anno$Type, anno$Color)
    Green_probes <- intersect(rownames(MSet), anno$Name[probe.type == "IGrn"])
    Red_probes <- intersect(rownames(MSet), anno$Name[probe.type == "IRed"])
    d2.probes <- intersect(rownames(MSet), anno$Name[probe.type == "II"])
    Meth <- getMeth(MSet)
    Unmeth <- getUnmeth(MSet)

    if (dyeCorr) {
        control_probes <- getProbeInfo(rgSet, type = "Control")
        control_probes <- control_probes[
            control_probes$Address %in% rownames(rgSet), ]
        Red <- getRed(rgSet)
        Green <- getGreen(rgSet)
        array_type <- rgSet@annotation[["array"]]
    } else {
        control_probes <- NULL
        Red <- NULL
        Green <- NULL
        array_type <- NULL
    }
    M_and_U <- .preprocessNoob(
        Meth = Meth,
        Unmeth = Unmeth,
        GreenOOB = GreenOOB,
        RedOOB = RedOOB,
        Green_probes = Green_probes,
        Red_probes = Red_probes,
        d2.probes = d2.probes,
        offset = offset,
        dyeCorr = dyeCorr,
        Red = Red,
        Green = Green,
        control_probes = control_probes,
        array_type = array_type,
        dyeMethod = dyeMethod,
        verbose = verbose)

    # Construct MethylSet
    assay(MSet, "Meth") <- M_and_U[["Meth"]]
    assay(MSet, "Unmeth") <- M_and_U[["Unmeth"]]
    # TODO: Is this the correct way to add a preprocessing method? It is
    #       shown as NA by show,MethylSet-method
    MSet@preprocessMethod <- c(
        mu.norm = sprintf("Noob, dyeCorr=%s, dyeMethod=%s", dyeCorr, dyeMethod))
    MSet
}

# TODO: Switch from `Meth` to `M` and `Unmeth` to `U`
# TODO: Rationalise argument names and order across functions, set defaults
#       assuming dyeCorr is FALSE

# Exported classes -------------------------------------------------------------

setClass(
    "RatioSet",
    representation(preprocessMethod = "character", annotation = "character"),
    contains = "SummarizedExperiment")

# Validity methods -------------------------------------------------------------

setValidity("RatioSet", function(object) {
    msg <- validMsg(NULL, NULL)
    # NOTE: Do not check for the presence of 'CN' because it is optional
    msgBeta <- .checkAssayNames(object, c("Beta"))
    msgM <- .checkAssayNames(object, c("M"))
    if (!is.null(msgBeta) && !is.null(msgM)) {
        msg <- validMsg(
            msg,
            sprintf("objects of class '%s needs to have assays slots either 'Beta' or 'M' or both",
                    class(object)))
    }
    if (is.null(msg)) TRUE else msg
})

# Exported functions -----------------------------------------------------------

RatioSet <- function(Beta = NULL, M = NULL, CN = NULL, annotation = "",
                     preprocessMethod = "", ...) {
    msg <- "Need either 'Beta' or 'M' or both"
    if (is.null(Beta) && is.null(M)) stop(msg)
    assays <- SimpleList(Beta = Beta, M = M, CN = CN)
    assays <- assays[!vapply(assays, is.null, logical(1L))]
    new("RatioSet",
        SummarizedExperiment(assays = assays, ...),
        annotation = annotation,
        preprocessMethod = preprocessMethod
    )
}

# Exported methods -------------------------------------------------------------

setMethod("show", signature(object = "RatioSet"), function(object) {
    callNextMethod()
    .show.annotation(annotation(object))
    .show.preprocessMethod(preprocessMethod(object))
})

setMethod("preprocessMethod", signature(object = "RatioSet"), function(object) {
    object@preprocessMethod
})

setMethod("annotation", signature(object = "RatioSet"), function(object) {
    object@annotation
})

setReplaceMethod(
    "annotation",
    signature(object = "RatioSet"),
    function(object, value) {
        object@annotation <- value
        object
    }
)

setMethod("getBeta", signature(object = "RatioSet"), function(object) {
    nms <- names(assays(object, withDimnames = FALSE))
    if ("Beta" %in% nms) return(assay(object, "Beta"))
    if ("M" %in% nms) return(ilogit2(assay(object, "M")))
    stop("object does not contain either 'M' nor 'Beta' amongst assay slots")
})

setMethod("getM", signature(object = "RatioSet"), function(object) {
    nms <- names(assays(object, withDimnames = FALSE))
    if ("M" %in% nms) return(assay(object, "M"))
    if ("Beta" %in% nms) return(logit2(assay(object, "Beta")))
    stop("object does not contain either 'M' nor 'Beta' amongst assay slots")
})

setMethod("getCN", signature(object = "RatioSet"), function(object) {
    nms <- names(assays(object, withDimnames = FALSE))
    if ("CN" %in% nms) return(assay(object, "CN"))
    NULL
})

setMethod(
    "mapToGenome",
    signature(object = "RatioSet"),
    function(object, drop = TRUE, mergeManifest = FALSE) {
        gr <- getLocations(
            object = object,
            mergeManifest = mergeManifest,
            orderByLocation = TRUE)
        object <- object[names(gr),]
        GenomicRatioSet(
            gr = gr,
            Beta = getBeta(object),
            M = getM(object),
            CN = getCN(object),
            colData = colData(object),
            preprocessMethod = preprocessMethod(object),
            annotation = annotation(object))
    }
)

setMethod(
    "updateObject",
    signature(object = "RatioSet"),
    function(object, ..., verbose=FALSE) {
        if (verbose) message("updateObject(object = 'RatioSet')")
        if ("assayData" %in% names(getObjectSlots(object))) {
            # This is an ExpressionSet based object
            object <- RatioSet(
                Beta = getObjectSlots(object)[["assayData"]][["Beta"]],
                M = getObjectSlots(object)[["assayData"]][["M"]],
                CN = getObjectSlots(object)[["assayData"]][["CN"]],
                colData = getObjectSlots(
                    getObjectSlots(object)[["phenoData"]])[["data"]],
                annotation = getObjectSlots(object)[["annotation"]],
                preprocessMethod = getObjectSlots(object)[["preprocessMethod"]])
        }
        object
    }
)

setMethod(
    "combine",
    signature(x = "RatioSet", y = "RatioSet"),
    function(x, y, ...) {
        colDataFix <- .harmonizeDataFrames(.pDataFix(colData(x)),
                                           .pDataFix(colData(y)))
        colData(x) <- colDataFix$x
        colData(y) <- colDataFix$y
        cbind(x, y)
    }
)

# Internal functions -----------------------------------------------------------

.getAnnotationFromOutType <- function(outType = c(
    "IlluminaHumanMethylation450k",
    "IlluminaHumanMethylationEPIC",
    "IlluminaHumanMethylation27k")) {
    outType <- match.arg(outType)
    if (outType == "IlluminaHumanMethylation450k") {
        anno <-  .default.450k.annotation
    } else if (outType == "IlluminaHumanMethylation27k") {
        anno <- .default.27k.annotation
    } else {
        anno <- .default.epic.annotation
    }
    c(array = outType, annotation = anno)
}

.getLociFromOutType <- function(outType = c("IlluminaHumanMethylation450k",
                                            "IlluminaHumanMethylationEPIC",
                                            "IlluminaHumanMethylation27k")) {
    outType <- match.arg(outType)
    manifest <- getManifest(outType)
    probesI <- getProbeInfo(manifest, type = "I")$Name
    probesII <- getProbeInfo(manifest, type = "II")$Name
    probes <- c(probesI, probesII)
    probes
}

# Convert the rgSet into the outType array.
.convertArray_450k_epic <- function(rgSet,
                                    outType = c("IlluminaHumanMethylation450k",
                                                "IlluminaHumanMethylationEPIC"),
                                    verbose = verbose) {

    outType <- match.arg(outType)
    .isRGOrStop(rgSet)
    stopifnot(.is450k(rgSet) || .isEPIC(rgSet))

    array <- annotation(rgSet)["array"]
    if (array == outType) stop("'rgSet' already in the 'outType' array type.")
    manifest1 <- getManifest(outType)
    manifest2 <- getManifest(rgSet)

    keepAddresses <- list(
        I = NULL,
        II = NULL,
        SnpI = NULL,
        SnpII = NULL,
        Control = NULL)

    # Probes of Type I
    probes1 <- getProbeInfo(manifest1, type = "I")
    probes2 <- getProbeInfo(manifest2, type = "I")
    commonNames <- intersect(probes1$Name, probes2$Name)
    probes1 <- probes1[match(commonNames, probes1$Name), ]
    probes2 <- probes2[match(commonNames, probes2$Name), ]
    stopifnot(all(probes1$Color == probes2$Color))
    stopifnot(all(probes1$ProbeSeqA == probes2$ProbeSeqA))
    stopifnot(all(probes1$ProbeSeqB == probes2$ProbeSeqB))
    # Translating rgSet2 addresses to rgSet1 addresses
    translate <- c(probes1$AddressA, probes1$AddressB)
    names(translate) <- c(probes2$AddressA, probes2$AddressB)
    wh <- which(rownames(rgSet) %in% names(translate))
    rownames(rgSet)[wh] <- translate[rownames(rgSet)[wh]]
    keepAddresses$I <- unname(translate)

    # Probes of Type II
    probes1 <- getProbeInfo(manifest1, type = "II")
    probes2 <- getProbeInfo(manifest2, type = "II")
    commonNames <- intersect(probes1$Name, probes2$Name)
    probes1 <- probes1[match(commonNames, probes1$Name),]
    probes2 <- probes2[match(commonNames, probes2$Name),]
    stopifnot(all(probes1$ProbeSeqA == probes2$ProbeSeqA))
    # Translating rgSet2 addresses to rgSet1 addresses
    translate <- probes1$AddressA
    names(translate) <- probes2$AddressA
    wh <- which(rownames(rgSet) %in% names(translate))
    rownames(rgSet)[wh] <- translate[rownames(rgSet)[wh]]
    keepAddresses$II <- unname(translate)

    # Probes of Type SnpI
    probes1 <- getProbeInfo(manifest1, type = "SnpI")
    probes2 <- getProbeInfo(manifest2, type = "SnpI")
    commonNames <- intersect(probes1$Name, probes2$Name)
    probes1 <- probes1[match(commonNames, probes1$Name),]
    probes2 <- probes2[match(commonNames, probes2$Name),]
    stopifnot(all(probes1$ProbeSeqA == probes2$ProbeSeqB))
    stopifnot(all(probes1$ProbeSeqB == probes2$ProbeSeqA))
    # Translating rgSet2 addresses to rgSet1 addresses
    translate <- c(probes1$AddressA, probes1$AddressB)
    names(translate) <- c(probes2$AddressA, probes2$AddressB)
    wh <- which(rownames(rgSet) %in% names(translate))
    rownames(rgSet)[wh] <- translate[rownames(rgSet)[wh]]
    keepAddresses$SnpI <- unname(translate)

    # Probes of Type SnpII
    probes1 <- getProbeInfo(manifest1, type = "SnpII")
    probes2 <- getProbeInfo(manifest2, type = "SnpII")
    commonNames <- intersect(probes1$Name, probes2$Name)
    probes1 <- probes1[match(commonNames, probes1$Name),]
    probes2 <- probes2[match(commonNames, probes2$Name),]
    stopifnot(all(probes1$ProbeSeqA == probes2$ProbeSeqA))
    # Translating rgSet2 addresses to rgSet1 addresses
    translate <- probes1$AddressA
    names(translate) <- probes2$AddressA
    wh <- which(rownames(rgSet) %in% names(translate))
    rownames(rgSet)[wh] <- translate[rownames(rgSet)[wh]]
    keepAddresses$SnpII <- unname(translate)

    # Probes of Type Control
    probes1 <- getProbeInfo(manifest1, type = "Control")
    probes2 <- getProbeInfo(manifest2, type = "Control")
    commonAddress <- intersect(probes1$Address, probes2$Address)
    probes1 <- probes1[match(commonAddress, probes1$Address),]
    probes2 <- probes2[match(commonAddress, probes2$Address),]
    keepAddresses$Control <- unname(probes1$Address)

    # Update rgSet
    keepAddresses <- do.call("c", keepAddresses)
    keepAddresses <- keepAddresses[keepAddresses %in% rownames(rgSet)]
    rgSet  <- rgSet[keepAddresses, ]
    annotation(rgSet) <- .getAnnotationFromOutType(outType)
    rgSet
}

# Exported methods -------------------------------------------------------------

setMethod(
    "combineArrays",
    signature(object1 = "RGChannelSet", object2 = "RGChannelSet"),
    function(object1, object2,
             outType = c("IlluminaHumanMethylation450k",
                         "IlluminaHumanMethylationEPIC"),
             verbose = TRUE) {

        outType <- match.arg(outType)
        array1 <- annotation(object1)[["array"]]
        array2 <- annotation(object2)[["array"]]
        if (array1 == array2) outType <- array1
        if (array1 == "IlluminaHumanMethylation27k" ||
            array2 == "IlluminaHumanMethylation27k") {
            stop("27k arrays cannot be combined at the RGChannelSet level.")
        }
        object1 <- convertArray(object1, outType = outType, verbose = verbose)
        object2 <- convertArray(object2, outType = outType, verbose = verbose)
        features1 <- rownames(object1)
        features2 <- rownames(object2)
        features  <- intersect(features1, features2)
        object1  <- object1[features, ]
        object2  <- object2[features, ]
        rgSet <- combine(object1, object2)
        rgSet$ArrayTypes <- rep(
            x = c(array1, array2),
            times = c(ncol(object1), ncol(object2)))
        rgSet
    }
)

setMethod("combineArrays",
          signature(object1 = "MethylSet", object2 = "MethylSet"),
          function(object1, object2,
                   outType = c("IlluminaHumanMethylation450k",
                               "IlluminaHumanMethylationEPIC",
                               "IlluminaHumanMethylation27k"),
                   verbose = TRUE) {

              outType <- match.arg(outType)
              array1 <- annotation(object1)["array"]
              array2 <- annotation(object2)["array"]
              if (array1 == array2) outType <- array1
              object1 <- convertArray(
                  object = object1,
                  outType = outType,
                  verbose = verbose)
              object2 <- convertArray(
                  object = object2,
                  outType = outType,
                  verbose = verbose)
              common.features <- intersect(rownames(object1), rownames(object2))
              object1 <- object1[common.features, ]
              object2 <- object2[common.features, ]
              Mset <- combine(object1, object2)
              Mset$ArrayTypes <- rep(
                  x = c(array1, array2),
                  times = c(ncol(object1), ncol(object2)))
              Mset
          }
)

setMethod("combineArrays",
          signature(object1 = "RatioSet", object2 = "RatioSet"),
          function(object1, object2,
                   outType = c("IlluminaHumanMethylation450k",
                               "IlluminaHumanMethylationEPIC",
                               "IlluminaHumanMethylation27k"),
                   verbose = TRUE) {

              outType <- match.arg(outType)
              array1 <- annotation(object1)["array"]
              array2 <- annotation(object2)["array"]
              if (array1 == array2) outType <- array1
              object1 <- convertArray(
                  object = object1,
                  outType = outType,
                  verbose = verbose)
              object2 <- convertArray(
                  object = object2,
                  outType = outType,
                  verbose = verbose)
              common.features <- intersect(rownames(object1), rownames(object2))
              object1 <- object1[common.features, ]
              object2 <- object2[common.features, ]
              Rset <- combine(object1, object2)
              Rset$ArrayTypes <- rep(
                  x = c(array1, array2),
                  times = c(ncol(object1), ncol(object2)))
              Rset
          }
)

setMethod(
    "combineArrays",
    signature(object1 = "GenomicRatioSet", object2 = "GenomicRatioSet"),
    function(object1, object2,
             outType = c("IlluminaHumanMethylation450k",
                         "IlluminaHumanMethylationEPIC",
                         "IlluminaHumanMethylation27k"),
             verbose = TRUE) {

        outType <- match.arg(outType)
        array1 <- annotation(object1)["array"]
        array2 <- annotation(object2)["array"]
        if (array1 == array2) outType <- array1
        object1 <- convertArray(object1, outType = outType, verbose = verbose)
        object2 <- convertArray(object2, outType = outType, verbose = verbose)
        colData1 <- colData(object1)
        colData2 <- colData(object2)
        colData1$ArrayTypes <- array1
        colData2$ArrayTypes <- array2
        colData1 <- colData(object1)
        colData2 <- colData(object2)
        by <- c("row.names", intersect(names(colData1), names(colData2)))
        colData.merged <- merge(colData1, colData2, all = TRUE, by = by)
        colData(object1) <- colData.merged[match(
            x = colnames(object1),
            table = colData.merged[, "Row.names"]), ]
        colData(object2) <- colData.merged[match(
            x = colnames(object2),
            table = colData.merged[, "Row.names"]), ]
        gr.common <- intersect(granges(object1), granges(object2))
        object1 <- sort(subsetByOverlaps(object1, gr.common))
        object2 <- sort(subsetByOverlaps(object2, gr.common))
        GRset <- cbind(object1, object2)
        colnames(GRset) <- GRset$Row.names
        GRset$Row.names <- NULL
        GRset
    }
)

setMethod(
    "combineArrays",
    signature(object1 = "GenomicMethylSet", object2 = "GenomicMethylSet"),
    function(object1, object2,
             outType = c("IlluminaHumanMethylation450k",
                         "IlluminaHumanMethylationEPIC",
                         "IlluminaHumanMethylation27k"),
             verbose = TRUE) {

        outType <- match.arg(outType)
        array1 <- annotation(object1)["array"]
        array2 <- annotation(object2)["array"]
        if (array1 == array2) outType <- array1
        object1 <- convertArray(object1, outType = outType, verbose = verbose)
        object2 <- convertArray(object2, outType = outType, verbose = verbose)
        colData1 <- colData(object1)
        colData2 <- colData(object2)
        colData1$ArrayTypes <- array1
        colData2$ArrayTypes <- array2
        by <- c("row.names", intersect(names(colData1), names(colData2)))
        colData.merged <- merge(colData1, colData2, all = TRUE, by = by)
        colData(object1) <- colData.merged[match(
            x = colnames(object1),
            table = colData.merged[, "Row.names"]), ]
        colData(object2) <- colData.merged[match(
            x = colnames(object2),
            table = colData.merged[, "Row.names"]), ]
        gr.common <- intersect(granges(object1), granges(object2))
        object1 <- sort(subsetByOverlaps(object1, gr.common))
        object2 <- sort(subsetByOverlaps(object2, gr.common))
        GMset <- cbind(object1, object2)
        colnames(GMset) <- GMset$Row.names
        GMset$Row.names <- NULL
        GMset
    }
)

setMethod(
    "convertArray",
    signature(object = "RGChannelSet"),
    function(object,
             outType = c("IlluminaHumanMethylation450k",
                         "IlluminaHumanMethylationEPIC"),
             verbose = TRUE) {

        outType <- match.arg(outType)
        array <- annotation(object)[["array"]]
        if (array == outType) return(object)
        if (verbose) message(sprintf("[convertArray] Casting as %s", outType))
        .convertArray_450k_epic(
            rgSet = object,
            outType = outType,
            verbose = verbose)
    }
)

setMethod(
    "convertArray",
    signature(object = "MethylSet"),
    function(object,
             outType = c("IlluminaHumanMethylation450k",
                         "IlluminaHumanMethylationEPIC",
                         "IlluminaHumanMethylation27k"),
             verbose = TRUE) {

        outType <- match.arg(outType)
        array <- annotation(object)[["array"]]
        if (array == outType) return(object)
        if (verbose) message(sprintf("[convertArray] Casting as %s", outType))
        common.features <- intersect(
            x = rownames(object),
            y = .getLociFromOutType(outType))
        object <- object[common.features, ]
        annotation(object) <- .getAnnotationFromOutType(outType)
        object
    }
)

setMethod(
    "convertArray",
    signature(object = "RatioSet"),
    function(object,
             outType = c("IlluminaHumanMethylation450k",
                         "IlluminaHumanMethylationEPIC",
                         "IlluminaHumanMethylation27k"),
             verbose = TRUE) {

        outType <- match.arg(outType)
        array <- annotation(object)[["array"]]
        if (array == outType) return(object)
        if (verbose) message(sprintf("[convertArray] Casting as %s", outType))
        common.features <- intersect(
            x = rownames(object),
            y = .getLociFromOutType(outType))
        object <- object[common.features, ]
        annotation(object) <- .getAnnotationFromOutType(outType)
        object
    }
)

setMethod(
    "convertArray",
    signature(object = "GenomicMethylSet"),
    function(object,
             outType = c("IlluminaHumanMethylation450k",
                         "IlluminaHumanMethylationEPIC",
                         "IlluminaHumanMethylation27k"),
             verbose = TRUE) {


        outType <- match.arg(outType)
        array <- annotation(object)[["array"]]
        if (array == outType) return(object)
        if (verbose) message(sprintf("[convertArray] Casting as %s", outType))
        common.features <- intersect(
            x = rownames(object),
            y = .getLociFromOutType(outType))
        object <- object[common.features, ]
        object@annotation <- .getAnnotationFromOutType(outType)
        object
    }
)

setMethod(
    "convertArray",
    signature(object = "GenomicRatioSet"),
    function(object,
             outType = c("IlluminaHumanMethylation450k",
                         "IlluminaHumanMethylationEPIC",
                         "IlluminaHumanMethylation27k"),
             verbose = TRUE) {

        outType <- match.arg(outType)
        array <- annotation(object)[["array"]]
        if (array == outType) return(object)
        if (verbose) message(sprintf("[convertArray] Casting as %s", outType))
        common.features <- intersect(
            x = rownames(object),
            y = .getLociFromOutType(outType))
        object <- object[common.features, ]
        object@annotation <- .getAnnotationFromOutType(outType)
        object
    }
)

# TODOs ------------------------------------------------------------------------

# TODO: Lots of duplicated code; DRY

# Internal generics ------------------------------------------------------------

# `...` are additional arguments passed to methods.
setGeneric(
    ".preprocessRawMeth",
    function(Red, Green, locusNames, TypeI.Red, TypeI.Green, TypeII, ...)
        standardGeneric(".preprocessRawMeth"),
    signature = c("Red", "Green"))

# `...` are additional arguments passed to methods.
setGeneric(
    ".preprocessRawUnmeth",
    function(Red, Green, locusNames, TypeI.Red, TypeI.Green, TypeII, ...)
        standardGeneric(".preprocessRawUnmeth"),
    signature = c("Red", "Green"))

# `...` are additional arguments passed to methods.
setGeneric(
    ".preprocessRaw",
    function(Red, Green, locusNames, TypeI.Red, TypeI.Green, TypeII, ...)
        standardGeneric(".preprocessRaw"),
    signature = c("Red", "Green"))

# Internal methods -------------------------------------------------------------

setMethod(
    ".preprocessRawMeth",
    c("matrix", "matrix"),
    function(Red, Green, locusNames, TypeI.Red, TypeI.Green, TypeII) {
        # Set up output matrix with appropriate dimension and type
        type <- .highestType(Red, Green)
        M <- matrix(.NA_type(type),
                    nrow = length(locusNames),
                    ncol = ncol(Red),
                    dimnames = list(locusNames, colnames(Red)))

        # Fill output matrix
        M[TypeI.Red$Name, ] <- Red[TypeI.Red$AddressB, ]
        M[TypeI.Green$Name, ] <- Green[TypeI.Green$AddressB, ]
        M[TypeII$Name, ] <- Green[TypeII$AddressA, ]

        # Return output matrix
        M
    }
)

setMethod(
    ".preprocessRawUnmeth",
    c("matrix", "matrix"),
    function(Red, Green, locusNames, TypeI.Red, TypeI.Green, TypeII, ...) {
        # Set up output matrix with appropriate dimension and type
        type <- .highestType(Red, Green)
        U <- matrix(.NA_type(type),
                    nrow = length(locusNames),
                    ncol = ncol(Red),
                    dimnames = list(locusNames, colnames(Red)))

        # Fill output matrix
        U[TypeI.Red$Name, ] <- Red[TypeI.Red$AddressA, ]
        U[TypeI.Green$Name, ] <- Green[TypeI.Green$AddressA, ]
        U[TypeII$Name, ] <- Red[TypeII$AddressA, ]

        # Return output matrix
        U
    }
)

setMethod(
    ".preprocessRaw",
    c("matrix", "matrix"),
    function(Red, Green, locusNames, TypeI.Red, TypeI.Green, TypeII, ...) {
        list(
            M = .preprocessRawMeth(
                Red, Green, locusNames, TypeI.Red, TypeI.Green, TypeII),
            U = .preprocessRawUnmeth(
                Red, Green, locusNames, TypeI.Red, TypeI.Green, TypeII))
    }
)

setMethod(
    ".preprocessRaw",
    c("DelayedMatrix", "DelayedMatrix"),
    function(Red, Green, locusNames, TypeI.Red, TypeI.Green, TypeII,
             BPREDO = list(), BPPARAM = SerialParam()) {
        # Set up intermediate RealizationSink objects of appropriate dimensions
        # and type
        # NOTE: These are ultimately coerced to the output DelayedMatrix
        #       objects, `M` and `U`
        # NOTE: Don't do `U_sink <- M_sink` or else these will reference the
        #       same object and clobber each other when written to!
        ans_type <- .highestType(Red, Green)
        M_sink <- DelayedArray::AutoRealizationSink(
            dim = c(length(locusNames), ncol(Red)),
            dimnames = list(locusNames, colnames(Red)),
            type = ans_type)
        on.exit(close(M_sink))
        U_sink <- DelayedArray::AutoRealizationSink(
            dim = c(length(locusNames), ncol(Red)),
            dimnames = list(locusNames, colnames(Red)),
            type = ans_type)
        on.exit(close(U_sink), add = TRUE)

        # Set up ArrayGrid instances over `Red` and `Green` as well as
        # "parallel" ArrayGrid instances over `M_sink` and `U_sink`.
        Red_grid <- colAutoGrid(Red)
        Green_grid <- colAutoGrid(Green)
        M_sink_grid <- RegularArrayGrid(
            refdim = dim(M_sink),
            spacings = c(nrow(M_sink), Red_grid@spacings[2]))
        U_sink_grid <- M_sink_grid
        # Sanity check ArrayGrid objects have the same dim
        stopifnot(dim(Red_grid) == dim(Green_grid),
                  dim(Red_grid) == dim(M_sink_grid))

        # Loop over blocks of `Red` and `Green` and write to `M_sink` and
        # `U_sink`
        blockMapplyWithRealization(
            FUN = .preprocessRaw,
            Red = Red,
            Green = Green,
            MoreArgs = list(
                locusNames = locusNames,
                TypeI.Red = TypeI.Red,
                TypeI.Green = TypeI.Green,
                TypeII = TypeII),
            sinks = list(M_sink, U_sink),
            dots_grids = list(Red_grid, Green_grid),
            sinks_grids = list(M_sink_grid, U_sink_grid),
            BPREDO = BPREDO,
            BPPARAM = BPPARAM)

        # Return as DelayedMatrix objects
        M <- as(M_sink, "DelayedArray")
        U <- as(U_sink, "DelayedArray")

        list(M = M, U = U)
    }
)

# Exported functions -----------------------------------------------------------

# TODO: Document: because we simultaneously walk over column-blocks of
#       `Red` and `Green`, the number of elements loaded into memory is
#       doubled. If running into memory issues, try halving
#       getOption("DelayedArray.block.size")
preprocessRaw <- function(rgSet) {
    .isRGOrStop(rgSet)

    # Extract data to pass to low-level functions that construct `M` and `U`
    locusNames <- getManifestInfo(rgSet, "locusNames")
    Red <- getRed(rgSet)
    Green <- getGreen(rgSet)
    TypeI.Red <- getProbeInfo(rgSet, type = "I-Red")
    TypeI.Green <- getProbeInfo(rgSet, type = "I-Green")
    TypeII <- getProbeInfo(rgSet, type = "II")

    # Construct `M` and `U`
    M_and_U <- .preprocessRaw(
        Red = Red,
        Green = Green,
        locusNames = locusNames,
        TypeI.Red = TypeI.Red,
        TypeI.Green = TypeI.Green,
        TypeII = TypeII)
    M <- M_and_U[["M"]]
    U <- M_and_U[["U"]]

    # Construct MethylSet
    out <- MethylSet(
        Meth = M,
        Unmeth = U,
        colData = colData(rgSet),
        annotation = annotation(rgSet),
        metadata = metadata(rgSet))

    # TODO: The manifest package version is currently not updated since
    #       `packageVersion(getManifest(rgSet))` fails. `packageVersion()
    #       expects a string
    out@preprocessMethod <- c(
        rg.norm = "Raw (no normalization or bg correction)",
        minfi = as.character(packageVersion("minfi")),
        manifest = as.character(
            packageVersion(.getManifestString(rgSet@annotation))))
    #packageVersion(getManifest(rgSet)))
    out
}

# Exported generics ------------------------------------------------------------

setGeneric("getMeth", function(object) standardGeneric("getMeth"))

setGeneric("getUnmeth", function(object) standardGeneric("getUnmeth"))

setGeneric("getBeta", function(object, ...) standardGeneric("getBeta"))

setGeneric("getM", function(object, ...) standardGeneric("getM"))

setGeneric("getCN", function(object, ...) standardGeneric("getCN"))

setGeneric("mapToGenome", function(object, ...) standardGeneric("mapToGenome"))

setGeneric("getManifest", function(object) standardGeneric("getManifest"))

setGeneric(
    "ratioConvert",
    function(object, ...) standardGeneric("ratioConvert"))

setGeneric(
    "preprocessMethod",
    function(object) standardGeneric("preprocessMethod"))


setGeneric(
    "convertArray",
    function(object, ...) standardGeneric("convertArray"))

setGeneric(
    "combineArrays",
    function(object1, object2, ...) standardGeneric("combineArrays"))

# Internal functions -----------------------------------------------------------

normalize.illumina.control <- function(rgSet, reference = 1) {
    # This function returns an rgset, not a methylset
    # code duplication
    Green <- getGreen(rgSet)
    Red <- getRed(rgSet)

    if (.is450k(rgSet) || .isEPIC(rgSet) || .isAllergy(rgSet) || .isEPICv2(rgSet)) {
        AT.controls <- getControlAddress(
            object = rgSet,
            controlType = c("NORM_A", "NORM_T"))
        CG.controls <- getControlAddress(
            object = rgSet,
            controlType = c("NORM_C", "NORM_G"))
    }
    if (.is27k(rgSet)) {
        AT.controls <- getControlAddress(
            object = rgSet,
            controlType = "Normalization-Red")
        CG.controls <- getControlAddress(
            object = rgSet,
            controlType = "Normalization-Green")
    }

    Green.avg <- colMeans2(Green, rows = match(CG.controls, rownames(Green)))
    Red.avg <- colMeans2(Red, rows = match(AT.controls, rownames(Red)))
    ref <- (Green.avg + Red.avg)[reference] / 2
    if (is.na(ref)) {
        stop("perhaps 'reference' refer to an array that is not present.")
    }
    Green.factor <- ref / Green.avg
    Red.factor <- ref / Red.avg
    Green <- sweep(Green, 2, FUN = "*", Green.factor)
    Red <- sweep(Red, 2, FUN = "*", Red.factor)
    assay(rgSet, "Green") <- Green
    assay(rgSet, "Red") <- Red
    rgSet
}

bgcorrect.illumina <- function(rgSet) {
    .isRGOrStop(rgSet)
    Green <- getGreen(rgSet)
    Red <- getRed(rgSet)
    if (.is450k(rgSet) || .isEPIC(rgSet) || .isAllergy(rgSet) || .isEPICv2(rgSet)) {
        NegControls <- getControlAddress(rgSet, controlType = "NEGATIVE")
    }
    if (.is27k(rgSet)) {
        NegControls <- getControlAddress(rgSet, controlType = "Negative")
    }
    Green.bg <- apply(Green[NegControls, , drop = FALSE], 2, function(xx) {
        sort(as.vector(xx))[31]
    })
    Red.bg <- apply(Red[NegControls, , drop = FALSE], 2, function(xx) {
        sort(as.vector(xx))[31]
    })
    Green <- pmax2(sweep(Green, 2, Green.bg), 0)
    Red <- pmax2(sweep(Red, 2, Red.bg), 0)
    assay(rgSet, "Green") <- Green
    assay(rgSet, "Red") <- Red
    rgSet
}

# Exported functions -----------------------------------------------------------

# TODO: Document: This does not realize the result for a DelayedMatrix-backed
#       RGChannelSet{Extended}
preprocessIllumina <- function(rgSet, bg.correct = TRUE,
                               normalize = c("controls", "no"), reference = 1) {
    .isRGOrStop(rgSet)
    normalize <- match.arg(normalize)

    if (normalize == "controls") {
        rgSet <- normalize.illumina.control(rgSet, reference = reference)
    }
    if (bg.correct) {
        rgSet <- bgcorrect.illumina(rgSet)
    }
    out <- preprocessRaw(rgSet)
    preprocess <- sprintf(
        "Illumina, bg.correct = %s, normalize = %s, reference = %d",
        bg.correct, normalize, reference)

    # TODO: The manifest package version is currently not updated since
    #       `packageVersion(getManifest(rgSet))` fails. `packageVersion()`
    #       expects a string
    out@preprocessMethod <- c(
        rg.norm = preprocess,
        minfi = as.character(packageVersion("minfi")),
        manifest = as.character(
            packageVersion(.getManifestString(rgSet@annotation))))
    #packageVersion(getManifest(rgSet)))
    out
}

# Exported functions -----------------------------------------------------------

mdsPlot <- function(dat, numPositions = 1000, sampNames = NULL,
                    sampGroups = NULL, xlim, ylim, pch = 1,
                    pal = brewer.pal(8, "Dark2"), legendPos = "bottomleft",
                    legendNCol, main = NULL) {
    # Check inputs
    if (is(dat, "MethylSet") || is(dat, "RGChannelSet")) {
        b <- getBeta(dat)
    } else if (is(dat, "matrix")) {
        b <- dat
    } else {
        stop("dat must be an 'MethylSet', 'RGChannelSet', or 'matrix'.")
    }
    if (is.null(main)) {
        main <- sprintf(
            "Beta MDS\n%d most variable positions",
            numPositions)
    }

    o <- order(rowVars(b), decreasing = TRUE)[seq_len(numPositions)]
    d <- dist(t(b[o, ]))
    fit <- cmdscale(d)
    if (missing(xlim)) xlim <- range(fit[, 1]) * 1.2
    if (missing(ylim)) ylim <- range(fit[, 2]) * 1.2
    if (is.null(sampGroups)) sampGroups <- rep(1, numPositions)
    sampGroups <- as.factor(sampGroups)
    col <- pal[sampGroups]
    if (is.null(sampNames)) {
        plot(
            x = fit[, 1],
            y = fit[, 2],
            col = col,
            pch = pch,
            xlim = xlim,
            ylim = ylim,
            xlab = "",
            ylab = "",
            main = main)
    } else {
        plot(
            x = 0,
            y = 0,
            type = "n",
            xlim = xlim,
            ylim = ylim,
            xlab = "",
            ylab = "",
            main = main)
        text(x = fit[, 1], y = fit[, 2], sampNames, col = col)
    }
    numGroups <- length(levels(sampGroups))
    if (missing(legendNCol)) legendNCol <- numGroups
    if (numGroups > 1) {
        legend(
            x = legendPos,
            legend = levels(sampGroups),
            ncol = legendNCol,
            text.col = pal[seq_len(numGroups)])
    }
}

plotCpg <- function(dat, cpg, pheno, type = c("categorical", "continuous"),
                    measure = c("beta", "M"), ylim = NULL, ylab = NULL,
                    xlab = "", fitLine = TRUE, mainPrefix = NULL,
                    mainSuffix = NULL) {
    if (is.numeric(cpg)) cpg <- rownames(dat)[cpg]
    type <- match.arg(type)
    measure <- match.arg(measure)
    if (is(dat, "MethylSet") || is(dat, "RGChannelSet")) {
        if (measure == "beta") {
            # NOTE: as.matrix() necessary in case 'dat' is a
            #       DelayedArray-backed minfi object
            x <- as.matrix(getBeta(dat[cpg, ]))
            if (is.null(ylab)) ylab <- "Beta"
            if (is.null(ylim)) ylim <- c(0, 1)
        } else if (measure == "M") {
            x <- getM(dat[cpg, ])
            if (is.null(ylab)) ylab <- "M"
            if (is.null(ylim)) ylim <- range(x)
        }
    } else {
        if (is.null(ylab)) ylab <- "unknown"
        x <- dat
        if (is.vector(x)) {
            x <- matrix(x, ncol = 1)
        } else {
            x <- as.matrix(x)
        }
    }
    main <- paste(mainPrefix, cpg, mainSuffix)
    names(main) <- cpg
    for (i in cpg) {
        if (type == "categorical") {
            stripchart(
                x = x[i, ] ~ pheno,
                vertical = TRUE,
                method = "jitter",
                jitter = 0.15,
                ylab = ylab,
                ylim = ylim,
                main = main[i])
        } else if (type == "continuous") {
            plot(
                x = pheno,
                y = x[i,],
                ylab = ylab,
                xlab = xlab,
                ylim = ylim,
                main = main[i])
            if (fitLine) abline(lm(x[i, ] ~ pheno), col = "blue")
        }
    }
}

setMethod("sampleNames", signature(object = "RGChannelSet"),
          function(object) {
    colnames(object)
})
setMethod("sampleNames", signature(object = "MethylSet"),
          function(object) {
    colnames(object)
})
setMethod("sampleNames", signature(object = "RatioSet"),
          function(object) {
    colnames(object)
})
setMethod("sampleNames", signature(object = "GenomicMethylSet"),
          function(object) {
    colnames(object)
})
setMethod("sampleNames", signature(object = "GenomicRatioSet"),
          function(object) {
    colnames(object)
})

setReplaceMethod("sampleNames", signature(object = "RGChannelSet"),
          function(object, value) {
    colnames(object) <- value
    object
})
setReplaceMethod("sampleNames", signature(object = "MethylSet"),
          function(object, value) {
    colnames(object) <- value
    object
})
setReplaceMethod("sampleNames", signature(object = "RatioSet"),
          function(object, value) {
    colnames(object) <- value
    object
})
setReplaceMethod("sampleNames", signature(object = "GenomicMethylSet"),
          function(object, value) {
    colnames(object) <- value
    object
})
setReplaceMethod("sampleNames", signature(object = "GenomicRatioSet"),
          function(object, value) {
    colnames(object) <- value
    object
})

setMethod("featureNames", signature(object = "RGChannelSet"),
          function(object) {
    rownames(object)
})
setMethod("featureNames", signature(object = "MethylSet"),
          function(object) {
    rownames(object)
})
setMethod("featureNames", signature(object = "RatioSet"),
          function(object) {
    rownames(object)
})
setMethod("featureNames", signature(object = "GenomicMethylSet"),
          function(object) {
    rownames(object)
})
setMethod("featureNames", signature(object = "GenomicRatioSet"),
          function(object) {
    rownames(object)
})

setReplaceMethod("featureNames", signature(object = "RGChannelSet"),
          function(object, value) {
    rownames(object) <- value
    object
})
setReplaceMethod("featureNames", signature(object = "MethylSet"),
          function(object, value) {
    rownames(object) <- value
    object
})
setReplaceMethod("featureNames", signature(object = "RatioSet"),
          function(object, value) {
    rownames(object) <- value
    object
})
setReplaceMethod("featureNames", signature(object = "GenomicMethylSet"),
          function(object, value) {
    rownames(object) <- value
    object
})
setReplaceMethod("featureNames", signature(object = "GenomicRatioSet"),
          function(object, value) {
    rownames(object) <- value
    object
})

setMethod("pData", signature(object = "RGChannelSet"),
          function(object) {
    colData(object)
})
setMethod("pData", signature(object = "MethylSet"),
          function(object) {
    colData(object)
})
setMethod("pData", signature(object = "RatioSet"),
          function(object) {
    colData(object)
})
setMethod("pData", signature(object = "GenomicMethylSet"),
          function(object) {
    colData(object)
})
setMethod("pData", signature(object = "GenomicRatioSet"),
          function(object) {
    colData(object)
})

setReplaceMethod("pData", signature(object = "RGChannelSet", value = "DataFrame"),
                 function(object, value) {
    colData(object) <- value
    object
})
setReplaceMethod("pData", signature(object = "MethylSet", value = "DataFrame"),
                 function(object, value) {
    colData(object) <- value
    object
})
setReplaceMethod("pData", signature(object = "RatioSet", value = "DataFrame"),
                 function(object, value) {
    colData(object) <- value
    object
})
setReplaceMethod("pData", signature(object = "GenomicMethylSet", value = "DataFrame"),
                 function(object, value) {
    colData(object) <- value
    object
})
setReplaceMethod("pData", signature(object = "GenomicRatioSet", value = "DataFrame"),
                 function(object, value) {
    colData(object) <- value
    object
})

## setReplaceMethod("pData", signature(object = "GenomicMethylSet", value = "DataFrame"),
##                  function(object, value) {
##     object <- BiocGenerics:::replaceSlots(object, colData=value)
##     msg <- SummarizedExperiment:::.valid.SummarizedExperiment.assays_ncol(object)
##     if (!is.null(msg))
##         stop(msg)
##     object
## })

# Exported classes -------------------------------------------------------------

setClass(
    "RGChannelSet",
    representation(annotation = "character"),
    contains = "SummarizedExperiment"
)

# Validity methods -------------------------------------------------------------

setValidity("RGChannelSet", function(object) {
    msg <- validMsg(NULL, .checkAssayNames(object, c("Red", "Green")))
    if (is.null(msg)) TRUE else msg
})

# Internal generics ------------------------------------------------------------

# `x` and `probe_info` are either `Red` and `IGrn` or `Green` and `IRed`,
# respectively.
# `...` are additional arguments passed to methods.
setGeneric(
    ".getOOB",
    function(x, probe_info, ...) standardGeneric(".getOOB"),
    signature = c("x")
)

# Internal methods -------------------------------------------------------------

setMethod(".getOOB", "matrix", function(x, probe_info, ...) {
    rbind(
        x[probe_info$AddressA, , drop = FALSE],
        x[probe_info$AddressB, , drop = FALSE])
})

setMethod(
    ".getOOB",
    "DelayedMatrix",
    function(x, probe_info, BPREDO = list(), BPPARAM = SerialParam()) {
        # Set up intermediate RealizationSink object of appropriate dimensions
        # and type.
        # NOTE: This is ultimately coerced to the output DelayedMatrix object.
        ans_type <- type(x)
        sink <- DelayedArray::AutoRealizationSink(
            dim = c(2L * nrow(probe_info), ncol(x)),
            dimnames = list(
                c(probe_info$AddressA, probe_info$AddressB), colnames(x)),
            type = ans_type)
        on.exit(close(sink))

        # Set up ArrayGrid instances over `x` as well as "parallel" ArrayGrid
        # instance over `sink`.
        x_grid <- colAutoGrid(x)
        sink_grid <- RegularArrayGrid(
            refdim = dim(sink),
            spacings = c(nrow(sink), ncol(sink) / length(x_grid)))
        # Sanity check ArrayGrid objects have the same dim
        stopifnot(dim(x_grid) == dim(sink_grid))

        # Loop over blocks of `Green` and write to `Grn_sink`
        blockApplyWithRealization(
            FUN = .getOOB,
            x = x,
            probe_info = probe_info,
            sink = sink,
            x_grid = x_grid,
            sink_grid = sink_grid,
            BPREDO = BPREDO,
            BPPARAM = BPPARAM)

        # Return as DelayedMatrix
        as(sink, "DelayedArray")
    }
)

# Exported functions -----------------------------------------------------------

RGChannelSet <- function(Green = new("matrix"), Red = new("matrix"),
                         annotation = "", ...) {
    ## Check rownames, colnames
    assays <- SimpleList(Green = Green, Red = Red)
    new("RGChannelSet",
        SummarizedExperiment(assays = assays, ...),
        annotation = annotation
    )
}

getRed <- function(object) {
    .isRGOrStop(object)
    assay(object, "Red")
}

getGreen <- function(object) {
    .isRGOrStop(object)
    assay(object, "Green")
}

getOOB <- function(object) {
    # Check input
    .isRGOrStop(object)

    # Extract data to pass to low-level functions that construct `Red` and `Grn`
    Red <- getRed(object)
    Green <- getGreen(object)
    IRed <- getProbeInfo(object, type = "I-Red")
    IGreen <- getProbeInfo(object, type = "I-Green")

    # Extract OOB probes
    list(Grn = .getOOB(Green, IRed), Red = .getOOB(Red, IGreen))
}

getSnpBeta <- function(object) {
    # Check input
    .isRGOrStop(object)

    # Extract and processSNP probes
    snpProbesI <- getProbeInfo(object, type = "SnpI")
    snpProbesII <- getProbeInfo(object, type = "SnpII")
    snpProbesI.Green <- snpProbesI[snpProbesI$Color == "Grn", , drop = FALSE]
    snpProbesI.Red <- snpProbesI[snpProbesI$Color == "Red", , drop = FALSE]

    # Extract Red and Green channels
    Green <- getGreen(object)
    Red <- getRed(object)

    # Construct M and U
    M.II <- Green[snpProbesII$AddressA, , drop = FALSE]
    U.II <- Red[snpProbesII$AddressA, , drop = FALSE]
    M.I.Red <- Red[snpProbesI.Red$AddressB, , drop = FALSE]
    U.I.Red <- Red[snpProbesI.Red$AddressA, , drop = FALSE]
    M.I.Green <- Green[snpProbesI.Green$AddressB, , drop = FALSE]
    U.I.Green <- Green[snpProbesI.Green$AddressA, , drop = FALSE]
    M <- rbind(M.II, M.I.Red, M.I.Green)
    U <- rbind(U.II, U.I.Red, U.I.Green)
    rownames(M) <- rownames(U) <- c(
        snpProbesII$Name,
        snpProbesI.Red$Name,
        snpProbesI.Green$Name)

    # Compute beta
    M / (U + M + 100)
}

subsetByLoci <- function(rgSet, includeLoci = NULL, excludeLoci = NULL,
                         keepControls = TRUE, keepSnps = TRUE) {
    # Check input
    .isRGOrStop(rgSet)

    # Get probe info
    TypeII <- getProbeInfo(rgSet, type = "II")
    TypeI.Red <- getProbeInfo(rgSet, type = "I-Red")
    TypeI.Green <- getProbeInfo(rgSet, type = "I-Green")

    # Select probes to retain
    if (!is.null(includeLoci)) {
        TypeII <- TypeII[TypeII$Name %in% includeLoci, ]
        TypeI.Red <- TypeI.Red[TypeI.Red$Name %in% includeLoci, ]
        TypeI.Green <- TypeI.Green[TypeI.Green$Name %in% includeLoci, ]
    }
    if (!is.null(excludeLoci)) {
        TypeII <- TypeII[!TypeII$Name %in% excludeLoci, ]
        TypeI.Red <- TypeI.Red[!TypeI.Red$Name %in% excludeLoci, ]
        TypeI.Green <- TypeI.Green[!TypeI.Green$Name %in% excludeLoci, ]
    }
    addresses <- c(
        TypeII$AddressA,
        TypeI.Red$AddressA, TypeI.Red$AddressB,
        TypeI.Green$AddressA, TypeI.Green$AddressB)
    if (keepControls) {
        addresses <- c(addresses, getProbeInfo(rgSet, type = "Control")$Address)
    }
    if (keepSnps) {
        addresses <- c(
            addresses,
            getProbeInfo(rgSet, type = "SnpI")$AddressA,
            getProbeInfo(rgSet, type = "SnpI")$AddressB,
            getProbeInfo(rgSet, type = "SnpII")$AddressA)
    }
    indices <- which(rownames(rgSet) %in% addresses)

    # Filter RGChannelSet to retain only the selected probes
    rgSet[indices, ]
}

# Exported methods -------------------------------------------------------------

setMethod("show", signature(object = "RGChannelSet"), function(object) {
    callNextMethod()
    .show.annotation(annotation(object))
})

setMethod("annotation", signature(object = "RGChannelSet"), function(object) {
    object@annotation
})

setReplaceMethod(
    "annotation",
    signature(object = "RGChannelSet"),
    function(object, value) {
        object@annotation <- value
        object
    }
)

setMethod(
    "updateObject",
    signature(object = "RGChannelSet"),
    function(object, ..., verbose = FALSE) {
        if (verbose) message("updateObject(object = 'RGChannelSet')")
        if ("assayData" %in% names(getObjectSlots(object))) {
            # This is an ExpressionSet based object
            object <- RGChannelSet(
                Green = getObjectSlots(object)[["assayData"]][["Green"]],
                Red = getObjectSlots(object)[["assayData"]][["Red"]],
                colData = getObjectSlots(
                    getObjectSlots(object)[["phenoData"]])[["data"]],
                annotation = getObjectSlots(object)[["annotation"]])
        } else {
            object <- callNextMethod()
        }
        object
    }
)

setMethod("getManifest", signature(object = "RGChannelSet"), function(object) {
    maniString <- .getManifestString(object@annotation)
    if (!require(maniString, character.only = TRUE)) {
        stop(sprintf("cannot load manifest package %s", maniString))
    }
    updateObject(get(maniString))
})

setMethod("getBeta", signature(object = "RGChannelSet"), function(object, ...) {
    object <- preprocessRaw(object)
    callGeneric(object, ...)
})

setMethod(
    "mapToGenome",
    signature(object = "RGChannelSet"),
    function(object, ...) {
        object <- preprocessRaw(object)
        callGeneric(object, ...)
    }
)

setMethod(
    "combine",
    signature(x = "RGChannelSet", y = "RGChannelSet"),
    function(x, y, ...) {
        colDataFix <- .harmonizeDataFrames(
            .pDataFix(colData(x)),
            .pDataFix(colData(y)))
        colData(x) <- colDataFix$x
        colData(y) <- colDataFix$y
        cbind(x, y)
    }
)

# Exported methods -------------------------------------------------------------

setMethod(
    "bumphunter",
    signature(object = "GenomicRatioSet"),
    function(object, design, cluster = NULL, coef = 2, cutoff = NULL,
             pickCutoff = FALSE, pickCutoffQ = 0.99, maxGap = 500,
             nullMethod = c("permutation", "bootstrap"), smooth = FALSE,
             smoothFunction = locfitByCluster, useWeights = FALSE,
             B = ncol(permutations), permutations = NULL, verbose = TRUE,
             type = c("Beta","M"), ...) {

        .isMatrixBackedOrStop(object, "bumphunter")

        type <- match.arg(type)
        bumphunterEngine(
            mat = getMethSignal(object, type),
            design = design,
            chr = as.factor(seqnames(object)),
            pos = start(object),
            cluster = cluster,
            coef = coef,
            cutoff = cutoff,
            pickCutoff = pickCutoff,
            pickCutoffQ = pickCutoffQ,
            maxGap = maxGap,
            nullMethod = nullMethod,
            smooth = smooth,
            smoothFunction = smoothFunction,
            useWeights = useWeights,
            B = B,
            permutations = permutations,
            verbose = verbose,
            ...)
    }
)

library(minfiData)
library(digest)

set.seed(456)
Mset.swan <- preprocessSWAN(RGsetEx)
digest_preprocessSWAN <- list(Meth = minfi:::.digestMatrix(getMeth(Mset.swan)),
                                  Unmeth = minfi:::.digestMatrix(getUnmeth(Mset.swan)))
save(digest_preprocessSWAN, file = "../unitTests/digest_preprocessSWAN.rda")

sessionInfo()
rm(list = ls())

# ------------------------------------------------------------------------------
# DelayedArray tests
#

library(DelayedArray)

assays(RGsetEx) <- endoapply(assays(RGsetEx), DelayedArray)

set.seed(456)
Mset.swan <- preprocessSWAN(RGsetEx)
digest_preprocessSWAN <- list(Meth = minfi:::.digestMatrix(getMeth(Mset.swan)),
                              Unmeth = minfi:::.digestMatrix(getUnmeth(Mset.swan)))
save(digest_preprocessSWAN, file = "../unitTests/digest_preprocessSWAN.DelayedArray.rda")

sessionInfo()
rm(list = ls())

# Exported classes -------------------------------------------------------------

setClass(
    "GenomicRatioSet",
    representation(preprocessMethod = "character", annotation = "character"),
    contains = "RangedSummarizedExperiment")

# Validity methods -------------------------------------------------------------

setValidity("GenomicRatioSet", function(object) {
    msg <- validMsg(NULL, NULL)
    # NOTE: Do not check for the presence of 'CN' because it is optional
    msgBeta <- .checkAssayNames(object, "Beta")
    msgM <- .checkAssayNames(object, "M")
    if (!is.null(msgBeta) && !is.null(msgM)) {
        msg <- validMsg(
            msg,
            sprintf("objects of class '%s needs to have assays slots either 'Beta' or 'M' or both",
                    class(object)))
    }
    if (!is(rowRanges(object), "GRanges")) {
        msg <- validMsg(
            msg,
            sprintf("object of class '%s' needs to have a 'GRanges' in slot 'rowRanges'",
                    class(object)))
    }
    if (is.null(msg)) TRUE else msg
})

# Exported functions -----------------------------------------------------------

GenomicRatioSet <- function(gr = GRanges(), Beta = NULL, M = NULL, CN = NULL,
                            annotation = "", preprocessMethod = "", ...) {
    msg <- "Need either 'Beta' or 'M' or both"
    if (is.null(Beta) && is.null(M)) stop(msg)
    assays <- SimpleList(Beta = Beta, M = M, CN = CN)
    assays <- assays[!vapply(assays, is.null, logical(1L))]
    new("GenomicRatioSet",
        SummarizedExperiment(
            assays = assays,
            rowRanges = as(gr, "GRanges"),
            ...),
        annotation = annotation,
        preprocessMethod = preprocessMethod
    )
}

# Exported methods -------------------------------------------------------------

setMethod("show", signature(object = "GenomicRatioSet"), function(object) {
    callNextMethod()
    .show.annotation(annotation(object))
    .show.preprocessMethod(preprocessMethod(object))
})

setMethod("getBeta", signature(object = "GenomicRatioSet"), function(object) {
    nms <- names(assays(object, withDimnames = FALSE))
    if ("Beta" %in% nms) return(assay(object, "Beta"))
    if ("M" %in% nms) return(ilogit2(assay(object, "M")))
    stop("object does not contain either 'M' nor 'Beta' amongst assay slots")
})

setMethod("getM", signature(object = "GenomicRatioSet"), function(object) {
    nms <- names(assays(object, withDimnames = FALSE))
    if ("M" %in% nms) return(assay(object, "M"))
    if ("Beta" %in% nms) return(logit2(assay(object, "Beta")))
    stop("object does not contain either 'M' nor 'Beta' amongst assay slots")
})

setMethod("getCN", signature(object = "GenomicRatioSet"), function(object) {
    nms <- names(assays(object, withDimnames = FALSE))
    if ("CN" %in% nms) return(assay(object, "CN"))
    NULL
})

setMethod(
    "mapToGenome",
    signature(object = "GenomicRatioSet"),
    function(object, ...) object
)

setMethod(
    "preprocessMethod",
    signature(object = "GenomicRatioSet"),
    function(object) object@preprocessMethod
)

setMethod(
    "annotation",
    signature(object = "GenomicRatioSet"),
    function(object) object@annotation
)

setReplaceMethod(
    "annotation",
    signature(object = "GenomicRatioSet"),
    function(object, value) {
        object@annotation <- value
        object
    }
)

setMethod(
    "updateObject",
    signature(object = "GenomicRatioSet"),
    function(object, ..., verbose = FALSE) {
        if (object@annotation["annotation"] == "ilmn.v1.2") {
            object@annotation["annotation"] <- .default.450k.annotation
        }
        object
    }
)

setMethod(
    "combine",
    signature(x = "GenomicRatioSet", y = "GenomicRatioSet"),
    function(x, y, ...) {
        colDataFix <- .harmonizeDataFrames(.pDataFix(colData(x)),
                                           .pDataFix(colData(y)))
        colData(x) <- colDataFix$x
        colData(y) <- colDataFix$y
        cbind(x,y)
    }
)

library(minfiData)
library(digest)

Mset.illumina <- preprocessIllumina(RGsetEx)
digest_preprocessIllumina <- list(Meth = minfi:::.digestMatrix(getMeth(Mset.illumina)),
                                      Unmeth = minfi:::.digestMatrix(getUnmeth(Mset.illumina)))
save(digest_preprocessIllumina, file = "../unitTests/digest_preprocessIllumina.rda")

sessionInfo()
rm(list = ls())

# ------------------------------------------------------------------------------
# DelayedArray tests
#

library(DelayedArray)

assays(RGsetEx) <- endoapply(assays(RGsetEx), DelayedArray)

Mset.illumina <- preprocessIllumina(RGsetEx)
digest_preprocessIllumina <- list(Meth = minfi:::.digestMatrix(getMeth(Mset.illumina)),
                                  Unmeth = minfi:::.digestMatrix(getUnmeth(Mset.illumina)))
save(digest_preprocessIllumina, file = "../unitTests/digest_preprocessIllumina.DelayedArray.rda")

sessionInfo()
rm(list = ls())

# Internal functions -----------------------------------------------------------

getSubset <- function(counts, subset){
    x <- integer(0)
    for (i in 1:3) {
        x <- c(x, sample(seq.int(1, length(counts))[counts == i], subset))
    }
    seq.int(1, length(counts)) %in% x
}

bgIntensitySwan <- function(rgSet) {
    rows <- match(getControlAddress(rgSet, controlType = "NEGATIVE"),
                  rownames(rgSet))
    grnMed <- colMedians(
        x = getGreen(rgSet),
        rows = rows)
    redMed <- colMedians(
        x = getRed(rgSet),
        rows = rows)
    rowMeans2(cbind(grnMed, redMed))
}

# TODO: Profile and tidy
normaliseChannel <- function(intensityI, intensityII, xNormSet, bg) {
    xTarget <- aveQuantile(
        list(intensityI[xNormSet[[1]]], intensityII[xNormSet[[2]]]))
    xNorm <- unlist(
        subsetQuantileNorm(
            list(intensityI, intensityII), xNormSet, xTarget, bg))
    names(xNorm) <- c(names(intensityI), names(intensityII))
    xNorm
}

# TODO: Profile and tidy
aveQuantile <- function(X) {
    nbrOfChannels <- length(X)
    if (nbrOfChannels == 1) {
        return(X)
    }
    nbrOfObservations <- unlist(lapply(X, FUN = length), use.names = FALSE)
    maxNbrOfObservations <- max(nbrOfObservations)
    if (maxNbrOfObservations == 1) {
        return(X)
    }
    ## nbrOfFiniteObservations <- rep(maxNbrOfObservations, times = nbrOfChannels)
    quantiles <- (0:(maxNbrOfObservations - 1))/(maxNbrOfObservations - 1)
    xTarget <- vector("double", maxNbrOfObservations)
    for (cc in 1:nbrOfChannels) {
        Xcc <- X[[cc]]
        Scc <- sort(Xcc)
        nobs <- length(Scc)
        if (nobs < maxNbrOfObservations) {
            ## tt <- !is.na(Xcc)
            bins <- (0:(nobs - 1))/(nobs - 1)
            Scc <- approx(x = bins, y = Scc, xout = quantiles,ties = "ordered")$y
        }
        xTarget <- xTarget + Scc
    }
    rm(Scc, Xcc)
    xTarget <- xTarget/nbrOfChannels
    xTarget
}

# TODO: Profile and tidy
subsetQuantileNorm <- function(x, xNormSet, xTarget, bg) {
    for(i in 1:length(x)){
        n <- length(x[[i]])
        nTarget <- length(xTarget)
        nNormSet <- sum(xNormSet[[i]])

        if(nNormSet != nTarget){
            targetQuantiles <- (0:(nTarget - 1))/(nTarget - 1)
            r <- rank(x[xNormSet[,i], i])
            xNew <-(r - 1)/(nNormSet - 1)
            xNew <- xNew[order(xNew)]
            xNorm <- approx(x = targetQuantiles, y = xTarget, xout = xNew, ties = "ordered", rule = 2)$y
        } else {
            xNorm<-xTarget
        }

        r <- rank(x[[i]])
        xNew <-(r - 1)/(n - 1)
        quantiles <- xNew[xNormSet[[i]]]
        quantiles <- quantiles[order(quantiles)]
        xmin <- min(x[[i]][xNormSet[[i]]]) #get min value from subset
        xmax <- max(x[[i]][xNormSet[[i]]]) #get max value from subset
        kmax <- which(xNew > max(quantiles))
        kmin<- which(xNew < min(quantiles))
        offsets.max <- x[[i]][kmax]-xmax
        offsets.min <- x[[i]][kmin]-xmin
        x[[i]] <- approx(x = quantiles, y = xNorm, xout = xNew, ties = "ordered")$y #interpolate
        x[[i]][kmax]<- max(xNorm) + offsets.max
        x[[i]][kmin]<- min(xNorm) + offsets.min
        x[[i]] = ifelse(x[[i]] <= 0, bg, x[[i]])
    }
    x
}

# Internal generics ------------------------------------------------------------

# `x` is either `Meth` or `Unmeth`
# `...` are additional arguments passed to methods.
setGeneric(
    ".preprocessSWAN",
    function(x, xNormSet, counts, bg, ...) standardGeneric(".preprocessSWAN"),
    signature = "x")

# Internal methods -------------------------------------------------------------

setMethod(".preprocessSWAN", "matrix", function(x, xNormSet, counts, bg) {
    # NOTE: SWAN can return non-integer values, so fill a numeric matrix
    normalized_x <- matrix(NA_real_,
                           ncol = ncol(x),
                           nrow = nrow(x),
                           dimnames = dimnames(x))
    typeI_idx <- rownames(x) %in% counts$Name[counts$Type == "I"]
    typeII_idx <- rownames(x) %in% counts$Name[counts$Type == "II"]
    for (j in seq_len(ncol(x))) {
        normalized_x[, j] <- normaliseChannel(
            intensityI = x[typeI_idx, j],
            intensityII = x[typeII_idx, j],
            xNormSet = xNormSet,
            bg = bg[j])
    }
    normalized_x
})

setMethod(
    ".preprocessSWAN",
    "DelayedMatrix",
    function(x, xNormSet, counts, bg, BPREDO = list(),
             BPPARAM = SerialParam()) {
        # Set up intermediate RealizationSink object of appropriate dimensions
        # and type
        # NOTE: This is ultimately coerced to the output DelayedMatrix object
        # NOTE: SWAN can return non-integer values, so fill a "double" sink
        ans_type <- "double"
        sink <- DelayedArray::AutoRealizationSink(
            dim = c(nrow(x), ncol(x)),
            dimnames = dimnames(x),
            type = ans_type)
        on.exit(close(sink))

        # Coerce `bg` to a row vector
        # NOTE: This is required in order to iterate in "parallel" over `x`,
        #       `bg`, and `sink`.
        bg <- matrix(bg, ncol = length(bg))

        # Set up ArrayGrid instances over `x` as well as "parallel" ArrayGrid
        # instances over `bg` and `sink`.
        x_grid <- colAutoGrid(x)
        bg_grid <- RegularArrayGrid(
            refdim = dim(bg),
            spacings = c(nrow(bg), ncol(bg) / length(x_grid)))
        sink_grid <- RegularArrayGrid(
            refdim = dim(sink),
            spacings = c(nrow(sink), ncol(sink) / length(x_grid)))
        # Sanity check ArrayGrid objects have the same dim
        stopifnot(dim(x_grid) == dim(sink_grid))

        # Loop over column-blocks of `x`, perform SWAN normalization, and write
        # to `normalized_x_sink`
        blockMapplyWithRealization(
            FUN = .preprocessSWAN,
            x = x,
            bg = bg,
            MoreArgs = list(
                xNormSet = xNormSet,
                counts = counts),
            sinks = list(sink),
            dots_grids = list(x_grid, bg_grid),
            sinks_grids = list(sink_grid),
            BPREDO = BPREDO,
            BPPARAM = BPPARAM)

        # Return as DelayedMatrix
        as(sink, "DelayedArray")
    })

# Exported functions -----------------------------------------------------------

preprocessSWAN <- function(rgSet, mSet = NULL, verbose = FALSE) {
    .isRGOrStop(rgSet)

    # Extract data to pass to low-level functions that construct normalized `M`
    # and `U`
    if (is.null(mSet)) {
        mSet <- preprocessRaw(rgSet)
    } else {
        .isMethylOrStop(mSet)
    }
    typeI <- getProbeInfo(rgSet, type = "I")[, c("Name", "nCpG")]
    typeII <- getProbeInfo(rgSet, type = "II")[, c("Name", "nCpG")]
    # TODO This next part should be fixed so it becomes more elegant
    CpG.counts <- rbind(typeI, typeII)
    # TODO: Unclear why this is necessary
    CpG.counts$Name <- as.character(CpG.counts$Name)
    CpG.counts$Type <- rep(c("I", "II"), times = c(nrow(typeI), nrow(typeII)))
    counts <- CpG.counts[CpG.counts$Name %in% rownames(mSet), ]
    subset <- min(
        table(counts$nCpG[counts$Type == "I" & counts$nCpG %in% 1:3]),
        table(counts$nCpG[counts$Type == "II" & counts$nCpG %in% 1:3]))
    bg <- bgIntensitySwan(rgSet)
    Meth <- getMeth(mSet)
    Unmeth <- getUnmeth(mSet)
    xNormSet <- lapply(c("I", "II"), function(type) {
        getSubset(counts$nCpG[counts$Type == type], subset)
    })

    # Construct normalized data
    M <- .preprocessSWAN(
        x = Meth,
        xNormSet = xNormSet,
        counts = counts,
        bg = bg)
    U <- .preprocessSWAN(
        x = Unmeth,
        xNormSet = xNormSet,
        counts = counts,
        bg = bg)

    # Construct MethylSet
    assay(mSet, "Meth") <- M
    assay(mSet, "Unmeth") <- U
    mSet@preprocessMethod <- c(
        rg.norm = sprintf("SWAN (based on a MethylSet preprocesses as '%s'",
                          preprocessMethod(mSet)[1]),
        minfi = as.character(packageVersion("minfi")),
        manifest = as.character(
            packageVersion(.getManifestString(annotation(rgSet)))))
    mSet
}

# TODOs ------------------------------------------------------------------------

# TODO: Choose more informative names for variables

# Exported classes -------------------------------------------------------------

setClass(
    "GenomicMethylSet",
    representation(preprocessMethod = "character", annotation = "character"),
    contains = "RangedSummarizedExperiment"
)

# Validity methods -------------------------------------------------------------

setValidity("GenomicMethylSet", function(object) {
    msg <- validMsg(NULL, .checkAssayNames(object, c("Meth", "Unmeth")))
    if (class(rowRanges(object)) != "GRanges") {
        msg <- validMsg(
            msg,
            sprintf("object of class '%s' needs to have a 'GRanges' in slot 'rowRanges'",
                    class(object)))
    }
    if (is.null(msg)) TRUE else msg
})

# Exported functions -----------------------------------------------------------

GenomicMethylSet <- function(gr = GRanges(), Meth = new("matrix"),
                             Unmeth = new("matrix"), annotation = "",
                             preprocessMethod = "", ...) {
    assays <- SimpleList(Meth = Meth, Unmeth = Unmeth)
    new("GenomicMethylSet",
        SummarizedExperiment(
            assays = assays,
            rowRanges = as(gr, "GRanges"),
            ...),
        annotation = annotation,
        preprocessMethod = preprocessMethod
    )
}

# Exported methods -------------------------------------------------------------

setMethod("show", signature(object = "GenomicMethylSet"), function(object) {
    callNextMethod()
    .show.annotation(annotation(object))
    .show.preprocessMethod(preprocessMethod(object))
})

setMethod("getMeth", signature(object = "GenomicMethylSet"), function(object) {
    assay(object, "Meth")
})

setMethod(
    "getUnmeth",
    signature(object = "GenomicMethylSet"),
    function(object) assay(object, "Unmeth")
)

setMethod(
    "getBeta",
    signature(object = "GenomicMethylSet"),
    function(object, type = "", offset = 0, betaThreshold = 0) {
        if (type == "Illumina") {
            offset <- 100
        }
        .betaFromMethUnmeth(
            Meth = getMeth(object),
            Unmeth = getUnmeth(object),
            offset = offset,
            betaThreshold = betaThreshold)
    }
)

setMethod(
    "getM",
    signature(object = "GenomicMethylSet"),
    function(object, type = "", ...) {
        if (type == "") return(log2(getMeth(object) / getUnmeth(object)))
        if (type == "beta" || type == "Beta") logit2(getBeta(object, ...))
    }
)

setMethod(
    "getCN",
    signature(object = "GenomicMethylSet"),
    function(object, ...) log2(getMeth(object) + getUnmeth(object))
)

setMethod(
    "mapToGenome",
    signature(object = "GenomicMethylSet"),
    function(object, ...) object
)

setMethod(
    "preprocessMethod",
    signature(object = "GenomicMethylSet"),
    function(object) object@preprocessMethod
)

setMethod(
    "annotation",
    signature(object = "GenomicMethylSet"),
    function(object) object@annotation
)

setReplaceMethod(
    "annotation",
    signature(object = "GenomicMethylSet"),
    function(object, value) {
        object@annotation <- value
        object
    }
)

setMethod(
    "ratioConvert",
    signature(object = "GenomicMethylSet"),
    function(object, what = c("beta", "M", "both"), keepCN = TRUE, ...) {
        what <- match.arg(what)
        if (what == "beta") {
            Beta <- getBeta(object, ...)
            M <- NULL
        }
        if (what == "M") {
            Beta <- NULL
            M <- getM(object, ...)
        }
        if (what == "both") {
            Beta <- getBeta(object, ...)
            M <- getM(object, ...)
        }
        if (keepCN) {
            CN <- getCN(object)
        } else {
            CN <- NULL
        }
        GenomicRatioSet(
            Beta = Beta,
            M = M,
            CN = CN,
            gr = granges(object),
            colData = colData(object),
            annotation = annotation(object),
            preprocessMethod = preprocessMethod(object),
            metadata = metadata(object))
    }
)

setMethod(
    "updateObject",
    signature(object = "GenomicMethylSet"),
    function(object, ..., verbose = FALSE) {
        if (object@annotation["annotation"] == "ilmn.v1.2") {
            object@annotation["annotation"] <- .default.450k.annotation
        }
        object
    }
)

setMethod(
    "combine",
    signature(x = "GenomicMethylSet", y = "GenomicMethylSet"),
    function(x, y, ...) {
        colDataFix <- .harmonizeDataFrames(.pDataFix(colData(x)),
                                           .pDataFix(colData(y)))
        colData(x) <- colDataFix$x
        colData(y) <- colDataFix$y
        cbind(x,y)
    }
)

---
title: "The minfi User's Guide"
shorttitle: "minfi guide"
author: 
  - Kasper D. Hansen
  - Jean-Phillipe Fortin
package: minfi
bibliography: minfi.bib
abstract: >
  A comprehensive guide to using the minfi package for analyzing
  DNA methylation microarrays from Illumina.vignettes.
vignette: >
  %\VignetteIndexEntry{minfi User's Guide}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
output: 
  BiocStyle::html_document:
    toc_float: true
---

# Introduction

The `r Biocpkg("minfi")` package provides tools for analyzing Illumina's Methylation arrays, specifically the 450k and EPIC (also known as the 850k) arrays.  We have partial support for the older 27k array.

The tasks addressed in this package include preprocessing, QC assessments, identification of interesting methylation loci and plotting functionality.  Analyzing these types of arrays is ongoing research in ours and other groups.

The input data to this package are IDAT files, representing two different color channels prior to normalization.  This is the most complete data type, because IDAT files includes measurements on control probes.  It is possible to use Genome Studio files together with the data structures contained in this package, but only some functionality is available because Genome Studio output does not contain control probe information.  In addition, usually Genome Studio output is normalized using the methods implemented in Genome Studio and these are generally considered inferior.

## Citing the minfi package

The MINFI package contains methods which are described across multiple manuscripts, by different non-overlapping authors.  This makes citing the package a bit difficult, so here is a guide.

- If you are using MINFI in a publication, please cite [@minfi].  This publication includes details on sex estimation using `getSex()`, quality control using `getQC()`, quantile normalization using `preprocessQuantile()`, bump hunting using `bumphunter()` and block finding using `blockFinder()`.
- If you are using MINFI to analyze EPIC or 27k arrays, please cite [@minfiEPIC]. The publication includes details on `convertArray()` and `combineArrays()`, extending NOOB to work in single-sample mode as well as using `estimateCellCounts()` with reference data from the 450k array to estimate cell type composition for EPIC data.
- If you are using `preprocessQuantile()`, it would be considerate to also cite [@Touleimat:2012], since this publication describes a method essentially identical to `preprocessQuantile()`.
- If you are using `bumphunter()`, you should also cite the original bump hunter publication [@bumphunter].
- If you are using SWAN normalization as implemented in `preprocessSWAN()` please cite [@SWAN].
- If you are using noob background correction as implemented in `preprocessNoob()`, please cite [@noob].
- If you are using functional normalization as implemented in `preprocessFunnorm()`, please cite [@funnorm]. The default in `preprocessFunnorm()` is to do noob background correction. If this is used, please also cite [@noob].
- If you are estimating A/B compartments as implemented in `compartments()` and `extractAB()`, please cite [@compartments].

If you're using Bibtex, you can get the citations in this format by

```{r citation,eval=FALSE}
toBibtex(citation("minfi"))
```

## Terminology

The literature is often a bit unspecific wrt. the terminology for the DNA methylation microarrays.

For the 450k microarray, each sample is measured on a single array, in two different color channels (red and green).  Each array measures roughly 450,000 CpG positions.  Each CpG is associated with two measurements: a methylated measurement and an "un"-methylated measurement.  These two values can be measured in one of two ways: using a "Type I" design or a "Type II design".  CpGs measured using a Type I design are measured using a single color, with two different probes in the same color channel providing the methylated and the unmethylated measurements.  CpGs measured using a Type II design are measured using a single probe, and two different colors provide the methylated and the unmethylated measurements.  Practically, this implies that on this array there is **not** a one-to-one correspondence between probes and CpG positions.  We have therefore tried to be precise about this and we refer to a "locus" (or "CpG") when we refer to a single-base genomic locus, and we differentiate this from a "probe".  The previous generation 27k methylation array uses only the Type I design, and the EPIC arrays uses both Type I and Type II.

Differences in DNA methylation between samples can either be at a single CpG which is called a differentially methylated position (DMP), or at a regional level which is called a differentially methylated region (DMR).

Physically, each sample is measured on a single "array".  For the 450k design, there are 12 arrays on a single physical "slide" (organized in a 6 by 2 grid).  Slides are organized into "plates" containing at most 8 slides (96 arrays).  The EPIC array has 8 arrays per slide and 64 arrays per plate.

## Dependencies

This document has the following dependencies

```{r dependencies, warning=FALSE, message=FALSE}
library(minfi)
library(minfiData)
```



# minfi classes

The MINFI package is designed to be very flexible for methods developers.  This flexibility comes at a cost for users; they need to understand a few different data classes:

- `RGChannelSet` : raw data from the IDAT files; this data is organized at the probe (not CpG locus) level.  This data has two channels: Red and Green.
- `MethylSet` : data organized by the CpG locus level, but not mapped to a genome.  This data has two channels: Meth (methylated) and Unmeth (unmethylated).
- `RatioSet` : data organized by the CpG locus level, but not mapped to a genome. The data has at least one of two channels: Beta and/or M (logratio of Beta). It may optionally include a CN channel (copy number).
- `GenomicMethylSet` : like a `MethylSet`, but mapped to a genome.
- `GenomicRatioSet` : like a `RatioSet`, but mapped to the genome.

A class hierarchy is as follows

![The class hierarchy of minfi](./minfiClasses.pdf)

To make this more clear, let us look at the example data called `RGsetEx` from the `r Biocpkg("minfiData")` package. This is the the number of features and and classes as we move through the class hierarchy (code not run):

```{r RGsetEx}
RGsetEx
## RGsetEx: RGChannelSet, 622,399 features
MsetEx <- preprocessRaw(RGsetEx)
## MsetEx: MethylSet, 485,512 features
GMsetEx <- mapToGenome(MsetEx)
## GMsetEx: GenomicMethylSet, 485,512 features
```

Note how the number of features changes.  In the `RGChannelSet` a feature is a probe (which is different from a CpG locus, see the Terminology section).  In the `MethylSet` each feature is now a methylation locus, and it has fewer features because some loci are measured using two probes.  Finally, the `GenomicMethylSet` has the same size as the `MethylSet`, but it could in principle be smaller in case the annotation you use says that some probes do not map to the genome (in this case hg19).

Finally we can convert to a `RatioSet` by `ratioConvert()`.  The two functions `ratioConvert()` and `mapToGenome()` commute, as shown by the class hierarchy diagram above.  Many preprocessing functions goes through several steps in the diagram, for example if the function needs to know the genomic location of the probes (several preprocessing functions handle probes measured on the sex chromosomes in a different way).

# Reading data

This package supports analysis of IDAT files, containing the summarized bead information.

In our experience, most labs use a "Sample Sheet" CSV file to describe the layout of the experiment.  This is based on a sample sheet file provided by Illumina.  Our pipeline assumes the existence of such a file(s), but it is relatively easy to create such a file using for example Excel, if it is not available.

We use an example dataset with 6 samples, spread across two slides.  First we obtain the system path to the IDAT files; this requires a bit since the data comes from an installed package

```{r baseDir}
baseDir <- system.file("extdata", package = "minfiData")
list.files(baseDir)
```

This shows the typical layout of 450k data: each "slide" (containing 12 arrays, see Termiology) is stored in a separate directory, with a numeric name.  The top level directory contains the sample sheet file. Inside the slide directories we find the IDAT files (and possible a number of JPG images or other files):

```{r baseDir2}
list.files(file.path(baseDir, "5723646052"))
```

The files for each array has another numeric number and consists of a Red and a Grn (Green) IDAT
file.  Note that for this example data, each slide contains only 3 arrays and not 12.  This was done
because of file size limitations and because we only need 6 arrays to illustrate the package's
functionality. 

First we read the sample sheet.  We provide a convenience function for reading in this file `read.metharray.sheet()`.  This function has a couple of attractive bells and whistles.  Let us look at the output

```{r sheet}
targets <- read.metharray.sheet(baseDir)
targets
```

First the output: this is just a `data.frame`.  It contains a column `Basename` that describes the location of the IDAT file corresponding to the sample, as well as two columns `Array` and `Slide`.  In the sample sheet provided by Illumina, these two columns are named `Sentrix_Position` and `Sentrix_ID`, but we rename them.  We provide more detail on the use of this function below.  The `Basename` column tend to be too large for display, here it is simplified relative to `baseDir`:

```{r BasenameColumn>}
sub(baseDir, "", targets$Basename)
```

(This is just for display purposes).

With this `data.frame`, it is easy to read in the data

```{r readingTargets}
RGset <- read.metharray.exp(targets = targets)
```

Let us look at the associated pheno data, which is really just the information contained in the targets object above.

```{r pData}
RGset
pd <- pData(RGset)
pd[,1:4]
```

The `read.metharray.exp()` function also makes it possible to read in an entire directory or directory tree (with `recursive` set to `TRUE`) by using the function just with the argument `base` and `targets=NULL`, like
```{r read2}
RGset2 <- read.metharray.exp(file.path(baseDir, "5723646052"))
RGset3 <- read.metharray.exp(baseDir, recursive = TRUE)
```

## Advanced notes on Reading Data

The only important column in sheet `data.frame` used in the `targets` argument for the `read.metharray.exp()` function is a column names `Basename`.  Typically, such an object would also have columns named `Array`, `Slide`, and (optionally) `Plate`.

We used sheet data files build on top of the Sample Sheet data file provided by Illumina.  This is a CSV file, with a header.  In this case we assume that the phenotype data starts after a line beginning with `[Data]` (or that there is no header present).

It is also easy to read a sample sheet manually, using the function `read.csv()`.  Here, we know that we want to skip the first 7 lines of the file.
```{r sampleSheet2}
targets2 <- read.csv(file.path(baseDir, "SampleSheet.csv"), 
                     stringsAsFactors = FALSE, skip = 7)
targets2
```

We now need to populate a `Basename` column.  On possible approach is the following
```{r Basename}
targets2$Basename <- file.path(baseDir, targets2$Sentrix_ID, 
                               paste0(targets2$Sentrix_ID, 
                                      targets2$Sentrix_Position))
```

Finally, MINFI contains a file-based parser: `read.metharray()`.  The return object represents the red and the green channel measurements of the samples.  A useful function that we get from the package `r Biocpkg("Biobase")` is `combine()` that combines ("adds") two sets of samples. This allows the user to manually build up an `RGChannelSet`.

# Manifest / annotation

## What everyone needs to know

For a methylation array, we have two types of annotation packages: "manifest" packages which contains the array design and "annotation" packages which contains information about where the methylation loci are located on the genome, which genomic features they map to and possible whether they overlap any known SNPs.

You can see which packages are being used by

```{r annotation}
annotation(RGsetEx)
```

## Advanced discussion

This discussion is intended for package developers or users who want to understand the internals of MINFI.

A set of 450k data files will initially be read into an `RGChannelSet`, representing the raw intensities as two matrices: one being the green channel and one being the red channel.  This is a class which is very similar to an `ExpressionSet` or an `NChannelSet`. The `RGChannelSet` is, together with a `IlluminaMethylationManifest` object, preprocessed into a `MethylSet`.  The `IlluminaMethylationManifest` object contains the array design, and describes how probes and color channels are paired together to measure the methylation level at a specific CpG.  The object also contains information about control probes (also known as QC probes).  The `MethylSet` contains normalized data and essentially consists of two matrices containing the methylated and the unmethylated evidence for each CpG.  Only the `RGChannelSet` contains information about the control probes.

The process described in the previous paragraph is very similar to the paradigm for analyzing Affymetrix expression arrays using the `r Biocpkg("affy")` package (an `AffyBatch` is preprocessed into an `ExpressionSet` using array design information stored in a CDF environment (package)).






# Quality control

- use shinyMethyl
- minfiQC
- getSex checks
- mds plots
- plot conversion probes

# Preprocessing

Preprocessing in MINFI is done by a series of functions with names like `preprocessXX`.  Different functions has different classes as output.

Currently, we have

- `preprocessRaw` : No processing.
- `preprocessIllumina` : Illumina preprocessing, as performed by Genome Studio (reverse engineered by us).
- `preprocessSWAN` : SWAN normalization, described in [@SWAN].
- `preprocessQuantile` : Quantile normalization (adapted to DNA methylation arrays), described in [@Touleimat:2012, @minfi]
- `preprocessNoob` : Noob preprocessing, described in [@noob].
- `preprocessFunnorm` : Functional normalization as described in [@funnorm].

FIXME: discuss literature

# SNPs and other issues

- SNPs
- cross reactive probes
- remove bad probes
- detection P
- Gap Hunting

# Identifying differentially methylated regions

- DMP finding
- Bump hunting
- Block finding
- plot DMRs with annotation. 

# Correcting for cell type heterogenity

- estimateCellTypes
- reference packages

# Other stuff

- Horvath age estimation
- other packages


# Sessioninfo

```{r sessionInfo, results='asis', echo=FALSE}
toLatex(sessionInfo())
```

# References

%\VignetteIndexEntry{Old Minfi Guide}
%\VignetteDepends{minfi}
%\VignetteDepends{minfiData}
%\VignettePackage{minfi}
\documentclass[12pt]{article}
<<options,echo=FALSE,results=hide>>=
options(width=70)
@ 
\SweaveOpts{eps=FALSE,echo=TRUE}
\usepackage{times}
\usepackage{color,hyperref}
\usepackage{fullpage}
\usepackage[numbers]{natbib}
\definecolor{darkblue}{rgb}{0.0,0.0,0.75}
\hypersetup{colorlinks,breaklinks,
            linkcolor=darkblue,urlcolor=darkblue,
            anchorcolor=darkblue,citecolor=darkblue}
\usepackage{parskip}
          
\newcommand{\Rcode}[1]{{\texttt{#1}}}
\newcommand{\Rpackage}[1]{\textsf{#1}}
\newcommand{\software}[1]{\textsf{#1}}
\newcommand{\R}{\software{R}}

\title{The minfi User's Guide\\
  Analyzing Illumina 450k Methylation Arrays\\
OLD -- New version is being created}
\author{Kasper D.\ Hansen \and Martin J. Aryee}
\date{Modified: October 9, 2011.  Compiled: \today}

\begin{document}
\maketitle
\setcounter{secnumdepth}{1} 

This vignette is being modified and re-written.

\section{Introduction}

The \Rpackage{minfi} package provides tools for analyzing Illumina's Methylation arrays, with a
special focus on the new 450k array for humans.  At the moment Illumina's 27k methylation arrays are
not supported.

The tasks addressed in this package include preprocessing, QC assessments, identification of
interesting methylation loci and plotting functionality.  Analyzing these types of arrays is ongoing
research in ours and others groups.  In general, the analysis of 450k data is not straightforward
and we anticipate many advances in this area in the near future.

The input data to this package are IDAT files, representing two different color channels prior to
normalization.  It is possible to use Genome Studio files together with the data structures
contained in this package, but in general Genome Studio files are already normalized and we do not
recommend this.

If you are using \Rpackage{minfi} in a publication, please cite \citet{minfi}.  The SWAN
normalization method is described in \citep{Maksimovic:2012}.



\subsubsection{Chip design and terminology}

The 450k array has a complicated design.  What follows is a quick overview.

Each sample is measured on a single array, in two different color channels (red and green).  Each
array measures roughly 450,000 CpG positions.  Each CpG is associated with two measurements: a
methylated measurement and an ``un''-methylated measurement.  These two values can be measured in
one of two ways: using a ``Type I'' design or a ``Type II design''.  CpGs measured using a Type I
design are measured using a single color, with two different probes in the same color channel
providing the methylated and the unmethylated measurements.  CpGs measured using a Type II design
are measured using a single probe, and two different colors provide the methylated and the
unmethylated measurements.  Practically, this implies that on this array there is \emph{not} a
one-to-one correspondence between probes and CpG positions.  We have therefore tried to be precise
about this and we refer to a ``methylation position'' (or ``CpG'') when we refer to a single-base
genomic locus.  The previous generation 27k methlation array uses only the Type I design.

In this package we refer to differentially methylated positions (DMPs) by which we mean a single
genomic position that has a different methylation level in two different groups of samples (or
conditions).  This is different from differentially methylated regions (DMRs) which imply more that
more than one methylation positions are different between conditions.

Physically, each sample is measured on a single ``array''.  There are 12 arrays on a single physical
``slide'' (organized in a 6 by 2 grid).  Slides are organized into ``plates'' containing at most 8
slides (96 arrays).


\subsubsection{Workflow and R data classes}

A set of 450k data files will initially be read into an \Rcode{RGChannelSet}, representing the raw
intensities as two matrices: one being the green channel and one being the red channel.  This is a
class which is very similar to an \Rcode{ExpressionSet} or an \Rcode{NChannelSet}.

The \Rcode{RGChannelSet} is, together with a \Rcode{IlluminaMethylationManifest} object,
preprocessed into a \Rcode{MethylSet}.  The \Rcode{IlluminaMethylationManifest} object contains the
array design, and describes how probes and color channels are paired together to measure the
methylation level at a specific CpG.  The object also contains information about control probes
(also known as QC probes).  The \Rcode{MethylSet} contains normalized data and essentially consists
of two matrices containing the methylated and the unmethylated evidence for each CpG.  Only the
\Rcode{RGChannelSet} contains information about the control probes.

The process described in the previous paragraph is very similar to the paradigm for analyzing
Affymetrix expression arrays using the \Rpackage{affy} package (an \Rcode{AffyBatch} is preprocessed
into an \Rcode{ExpressionSet} using array design information stored in a CDF environment (package)).

A \Rcode{MethylSet} is the starting point for any post-normalization analysis, such as searching for
DMPs or DMRs.



\subsubsection{Getting Started}

<<load,results=hide>>=
require(minfi)
require(minfiData)
@ 

\section{Reading Data}

This package supports analysis of IDAT files, containing the summarized bead information.

In our experience, most labs use a ``Sample Sheet'' CSV file to describe the layout of the
experiment.  This is based on a sample sheet file provided by Illumina.  Our pipeline assumes the
existence of such a file(s), but it is relatively easy to create such a file using for example Excel,
if it is not available.

We use an example dataset with 6 samples, spread across two slides.  First we obtain the system path
to the IDAT files; this requires a bit since the data comes from an installed package

<<baseDir>>=
baseDir <- system.file("extdata", package = "minfiData")
list.files(baseDir)
@ 
This shows the typical layout of 450k data: each ``slide'' (containing 12 arrays) is stored in a
separate directory, with a numeric name.  The top level directory contains the sample sheet file.
Inside the slide directories we find the IDAT files (and possible a number of JPG images or other
files):
<<baseDir>>=
list.files(file.path(baseDir, "5723646052"))
@ 
The files for each array has another numeric number and consists of a Red and a Grn (Green) IDAT
file.  Note that for this example data, each slide contains only 3 arrays and not 12.  This was done
because of file size limitations and because we only need 6 arrays to illustrate the package's
functionality. 

First we read the sample sheet.  We provide a convenience function for reading in this file
\Rcode{read.metharray.sheet}.  This function has a couple of attractive bells and whistles.  Let us look
at the output
<<sampleSheet>>=
targets <- read.metharray.sheet(baseDir)
targets
@ 
First the output: this is just a \Rcode{data.frame}.  It contains a column \Rcode{Basename} that
describes the location of the IDAT file corresponding to the sample, as well as two columns
\Rcode{Array} and \Rcode{Slide}.  In the sample sheet provided by Illumina, these two columns are
named \Rcode{Sentrix\_Position} and \Rcode{Sentrix\_ID}, but we rename them.  We provide more detail
on the use of this function below.  The \Rcode{Basename} column tend to be too large for display,
here it is simplified relative to \Rcode{baseDir}:
<<BasenameColumn>>=
sub(baseDir, "", targets$Basename)
@ 
(This is just for display purposes).

With this \Rcode{data.frame}, it is easy to read in the data
<<paths>>= 
RGset <- read.metharray.exp(targets = targets)
@ 
Let us look at the associated pheno data, which is really just the information contained in the
targets object above.
<<pData>>=
RGset
pd <- pData(RGset)
pd[,1:4]
@ 

The \Rcode{read.metharray.exp} also makes it possible to read in an entire directory or directory tree
(with \Rcode{recursive} set to \Rcode{TRUE}) by using the function just with the argument \Rcode{base}
and \Rcode{targets=NULL}, like
<<read2>>=
RGset2 = read.metharray.exp(file.path(baseDir, "5723646052"))
RGset3 = read.metharray.exp(baseDir, recursive = TRUE)
@ 

\subsubsection{Advanced notes on Reading Data}

The only important column in sheet \Rcode{data.frame} used in the \Rcode{targets} argument for the
\Rcode{read.metharray.exp} function is a column names \Rcode{Basename}.  Typically, such an object would
also have columns named \Rcode{Array}, \Rcode{Slide}, and (optionally) \Rcode{Plate}.

We used sheet data files build on top of the Sample Sheet data file provided by Illumina.  This is a
CSV file, with a header.  In this case we assume that the phenotype data starts after a line
beginning with \Rcode{[Data]} (or that there is no header present).

It is also easy to read a sample sheet ``manually'', using the function \Rcode{read.csv}.  Here, we
know that we want to skip the first 7 lines of the file.
<<sampleSheet2>>=
targets2 <- read.csv(file.path(baseDir, "SampleSheet.csv"), 
                     stringsAsFactors = FALSE, skip = 7)
targets2
@ 
We now need to populate a \Rcode{Basename} column.  On possible approach is the following
<<Basename>>=
targets2$Basename <- file.path(baseDir, targets2$Sentrix_ID, 
                               paste0(targets2$Sentrix_ID, 
                                      targets2$Sentrix_Position))
@ 

Finally, \Rpackage{minfi} contains a file-based parser: \Rcode{read.metharray}.  The return object
represents the red and the green channel measurements of the samples.  A useful function that we get
from the package \Rpackage{Biobase} is \Rcode{combine} that combines (``adds'') two sets of samples.
This allows the user to manually build up an \Rcode{RGChannelSet}.

\section{Quality Control}

\Rcode{minfi} provides several plots that can be useful for identifying samples with data quality
problems.  These functions can display summaries of signal from the array (e.g. density plots) as
well as the values of several types of control probes included on the array. Our understanding of
the expected sample behavior in the QC plots is still evolving and will improve as the number of
available samples from the array increases. A good rule of thumb is to be wary of samples whose
behavior deviates from that of others in the same or similar experiments.

The wrapper function \Rcode{qcReport} function can be used to produce a PDF QC report of the most
common plots.  If provided, the optional sample name and group options will be used to label and
color plots. Samples within a group are assigned the same color. The sample group option can also be
used as a very cursory way to check for batch effects (e.g. by setting it to a processing day
variable.)

<<qcReport-quick,eval=FALSE>>=
qcReport(RGset, sampNames = pd$Sample_Name, 
         sampGroups = pd$Sample_Group, pdf = "qcReport.pdf")
@

The components of the QC report can also be customized and produced individually as detailed below.

\subsubsection{Density plots}

The \Rcode{densityPlot} function produces density plots of the methylation Beta values for all
samples, typically colored by sample group. While the density plots in Figure \ref{fig:densityPlot}
are useful for identifying deviant samples, it is not easy to identify the specific problem
sample. If there is a concern about outlier samples, a useful follow-up is the ``bean'' plot (Figure
\ref{fig:densityBeanPlot}) that shows each sample in its own section. While the shape of the 
distribution for ``good'' samples will differ from experiment to experiment, many conditions 
have methylation profiles characterized by  
two modes - one with close to 0\% methylation, and a second at close to 100\% methylation.

\begin{figure} 
\begin{center}
<<qcReport-density,fig=TRUE>>=
densityPlot(RGset, sampGroups = pd$Sample_Group, 
            main = "Beta", xlab = "Beta")
@
\end{center} 
\caption{Beta density plots}
\label{fig:densityPlot}
\end{figure}


\begin{figure} 
\begin{center}
<<qcReport-bean,fig=TRUE>>=
par(oma=c(2,10,1,1))
densityBeanPlot(RGset, sampGroups = pd$Sample_Group, 
                sampNames = pd$Sample_Name)
@
\end{center} 
\caption{Beta beanplots} 
\label{fig:densityBeanPlot}
\end{figure}

\subsubsection{Control probe plots}

The \Rcode{controlStripPlot} function allows plotting of individual control probe types (Figure \ref{fig:controlStripPlot}).
The following control probes are available on the array:
\begin{verbatim}
  BISULFITE CONVERSION I    12
  BISULFITE CONVERSION II    4
  EXTENSION                  4
  HYBRIDIZATION              3
  NEGATIVE                 614
  NON-POLYMORPHIC            4
  NORM_A                    32
  NORM_C                    61
  NORM_G                    32
  NORM_T                    61
  SPECIFICITY I             12
  SPECIFICITY II             3
  STAINING                   6
  TARGET REMOVAL             2
\end{verbatim}

\begin{figure} 
\begin{center}
<<qcReport-stripplot,fig=TRUE>>=
controlStripPlot(RGset, controls="BISULFITE CONVERSION II", 
                 sampNames = pd$Sample_Name)
@
\end{center} 
\caption{Beta stripplot} 
\label{fig:controlStripPlot}
\end{figure}


\section{Preprocessing (normalization)}

Preprocessing (normalization) takes as input a \Rcode{RGChannelSet} and returns a \Rcode{MethylSet}.

A number of preprocessing options are available (and we are working on more methods).  Each set of
methods are implemented as a function \Rcode{preprocessXXX} with \Rcode{XXX} being the name of the
method.  Each method may have a number of tuning parameters.

``Raw'' preprocessing  means simply converting the Red and the Green channel into a Methylated and
Unmethylated signal

<<Msetraw>>=
MSet.raw <- preprocessRaw(RGset)
@ 

We have also implemented preprocessing choices as available in Genome Studio.  These choices follow
the description provided in the Illumina documentation and has been validated by comparing the
output of Genome Studio to the output of these algorithms, and this shows the two approaches to be
roughly equivalent (for a precise statement, see the manual pages).

Genome studio allows for background subtraction (also called background normalization) as well as
something they term control normalization.  Both of these are optional and turning both of them off
is equivalent to raw preprocessing (\Rcode{preprocessRaw}).

<<allMsets>>=
MSet.norm <- preprocessIllumina(RGset, bg.correct = TRUE,
                                 normalize = "controls", reference = 2)
@ 

The \Rcode{reference = 2} selects which array to use as ``reference'' which is an arbitrary array
(we are not sure how Genome Studio makes its choice of reference).

\subsubsection{Operating on a MethylSet}

Once a \Rcode{MethylSet} has been generated, we have a various ways of getting access to the
methylation data.  The most basic functions are \Rcode{getMeth} and \Rcode{getUnmeth}, which returns
unlogged methylation channels.  The function \Rcode{getBeta} gets ``beta''-values which are values
between 0 and 1 with 1 interpreted as very high methylation.  If \Rcode{type = "Illumina"} (not the
default) these are computed using Illumina's formula
\begin{displaymath}
  \beta = \frac{M}{M + U + 100}
\end{displaymath}
Finally, we have the ``M-values'' (not to be confused with the methylation channel obtained by
\Rcode{getMeth}).   M-values are perhaps an unfortunate terminology, but it
seems to be standard in the methylation array world.  These are computed as
$\textrm{logit}(\beta)$ and are obtained by \Rcode{getM}. 


<<MSet>>=
getMeth(MSet.raw)[1:4,1:3]
getUnmeth(MSet.raw)[1:4,1:3]
getBeta(MSet.raw, type = "Illumina")[1:4,1:3]
getM(MSet.raw)[1:4,1:3]
@ 

\subsubsection{MDS plots}

After preprocessing the raw data to obtain methylation estimates, Multi-dimensional scaling (MDS)
plots provide a quick way to get a first sense of the relationship between samples. They are similar
to the more familiar PCA plots and display a two-dimensional approximation of sample-to-sample
Euclidean distance. Note that while the plot visualizes the distance in epigenomic profiles between samples, 
the absolute positions of the points is not meaningful. One often expects to see greater between-group 
than within-group distances
(although this clearly depends on the particular experiment).  The most variable locations are used
when calculating sample distances, with the number specified by the \Rcode{numPositions} option. Adding
sample labels to the MDS plot is a useful way of identifying outliers (figure \ref{fig:MDS}) that 
behave differently from their peers.
\begin{figure} 
\begin{center}
<<qcReport-mdsplot2,fig=TRUE>>=
mdsPlot(MSet.norm, numPositions = 1000, sampGroups = pd$Sample_Group, 
	sampNames = pd$Sample_Name)
@
\end{center} 
\caption{Multi-dimensional scaling plot}
\label{fig:MDS}
\end{figure}



\subsubsection{The validation of \Rcode{preprocessIllumina}}

By validation we mean ``yielding output that is equivalent to Genome Studio''.

Illumina offers two steps: control normalization and background subtraction (normalization).  Using
output from Genome Studio we are certain that the control normalization step is validated, with the
following caveat: control normalization requires the selection of one array among the 12 arrays on a
chip as a reference array.  It is currently unclear how Genome Studio selects the reference; if you
know the reference array we can recreate Genome Studio exactly.  Background subtraction
(normalization) is almost correct: for 18 out of 24 arrays we see exact equivalence and for the
remaining 6 out of 24 arrays we only see small discrepancies (a per-array max difference of 1-4 for
unlogged intensities).  A script for doing this is in \texttt{scripts/GenomeStudio.R}.

\subsubsection{Subset-quantile within array normalisation (SWAN)}

SWAN (subset-quantile within array normalisation) is a new normalization method for Illumina 450k
arrays.  What follows is a brief description of the methodology (written by the authors of SWAN):

Technical differences have been demonstrated to exist between the Type I and Type II assay designs
within a single 450K array\citep{Bibikova:2011,Dedeurwaerder:2011}. Using the SWAN method
substantially reduces the technical variability between the assay designs whilst maintaining the
important biological differences.  The SWAN method makes the assumption that the number of CpGs
within the 50bp probe sequence reflects the underlying biology of the region being
interrogated. Hence, the overall distribution of intensities of probes with the same number of CpGs
in the probe body should be the same regardless of design type. The method then uses a subset
quantile normalization approach to adjust the intensities of each array \citep{Maksimovic:2012}.
SWAN takes a \Rcode{MethylSet} as input. This can be generated by either \Rcode{preprocessRaw} or
\Rcode{preprocessIllumina}.  Calling the function without specifying a \Rcode{MethylSet} uses
\Rcode{preprocessRaw}.  It should be noted that, in order to create the normalization subset, SWAN
randomly selects Infinium I and II probes that have one, two and three underlying CpGs; as such, we
recommend setting a seed (using \Rcode{set.seed})before using \Rcode{preprocessSWAN} to ensure that
the normalized intensities will be identical, if the normalization is repeated.

<<preprocessSwan>>=
Mset.swan <- preprocessSWAN(RGsetEx, MsetEx)
@ 

The technical differences between Infinium I and II assay designs can result in aberrant beta value
distributions (Figure~\ref{fig:plotBetaTypes}, panel ``Raw''). Using SWAN corrects for the
technical differences between the Infinium I and II assay designs and produces a smoother overall
beta value distribution (Figure~\ref{fig:plotBetaTypes}, panel ``SWAN'').

\begin{figure}
  \centering
<<plotBetaType,fig=TRUE,width=8,height=4>>=
par(mfrow=c(1,2))
plotBetasByType(MsetEx[,1], main = "Raw")
plotBetasByType(Mset.swan[,1], main = "SWAN")
@ 
\caption{The effect of normalizing using SWAN.}\label{fig:plotBetaTypes}
\end{figure}

\section{Finding differentially methylated positions (DMPs)}

We are now ready to use the normalized data to identify DMPs, defined as CpG positions where the 
methylation level correlates with a phenotype of interest. The phenotype may be categorical (e.g. cancer vs.
normal) or continuous (e.g. blood pressure).

We will create a 20,000 CpG subset of our dataset to speed up the demo:

<<subset-mset>>=
mset <- MSet.norm[1:20000,]
@

\subsubsection{Categorical phenotypes}

The \Rcode{dmpFinder} function uses an F-test to identify positions that are differentially
methylated between (two or more) groups. Tests are performed on logit transformed Beta values as
recommended in Pan et al.  Care should be taken if you have zeroes in either the Meth or the Unmeth
matrix.  One possibility is to threshold the beta values, so they are always in the interval
$[\epsilon, 1-\epsilon]$.  We call $\epsilon$ the betaThreshold

Here we find the differences between GroupA and GroupB.

<<dmpFinder-categorical>>=
table(pd$Sample_Group)
M <- getM(mset, type = "beta", betaThreshold = 0.001)
dmp <- dmpFinder(M, pheno=pd$Sample_Group, type="categorical")
head(dmp)
@

\Rcode{dmpFinder} returns a table of CpG positions sorted by differential methylation p-value.

We can use the \Rcode{plotCpG} function to plot methylation levels at individual positions:
<<plot-dmps-categorical,fig=TRUE>>=
cpgs <- rownames(dmp)[1:4]
par(mfrow=c(2,2))
plotCpg(mset, cpg=cpgs, pheno=pd$Sample_Group)
@


\subsubsection{Continuous phenotypes}
We can also identify DMPs where the mean methylation level varies with a continuous covariate 
using linear regression. Since the sample dataset does not contain any continuous phenotypes 
we will simulate one for demonstration purposes:

<<set-seed, echo=FALSE>>=
set.seed(123)
@
<<sim-pheno>>=
continuousPheno <- rnorm(nrow(pd))
@

We now search for DMPs associated with this phenotype. 
<<dmpFinder-continuous>>=
dmp <- dmpFinder(mset, pheno=continuousPheno, type="continuous")
dmp[1:3,]
@

The \Rcode{beta} column gives the change in mean phenotype for each unit increase of methylation.
We can filter the DMP list to exclude positions with a small effect size:
<<filter-dmp>>=
dmp <- subset(dmp, abs(beta)>1)
@

The \Rcode{plotCpg} function can be used to visualise these continuous DMPs:
<<plot-dmps-continuous,fig=TRUE>>=
cpgs <- rownames(dmp)[1:4]
par(mfrow=c(2,2))
plotCpg(mset, cpg=cpgs, type="continuous",
        pheno=continuousPheno, xlab="Phenotype 1")
@




\section{Advanced: The manifest object}

In order to preprocess the data we need a ``manifest'' object.  This object is similar to the union
of a CDF environment and a probe package (and may be restructured).  Essentially it describes what
probes are on the array and how they are matched together.

\emph{The manifest object only depends on the array design.  It is not related to annotating the
  CpGs measured by the array.}

The internal structure of the manifest object should not be of concern to users. However, it may be
useful to know something about the array design.  First we have a look at the object:

<<manifest>>=
IlluminaHumanMethylation450kmanifest
head(getProbeInfo(IlluminaHumanMethylation450kmanifest, type = "I"), n = 3)
head(getProbeInfo(IlluminaHumanMethylation450kmanifest, type = "II"), n = 3)
head(getProbeInfo(IlluminaHumanMethylation450kmanifest, type = "Control"), n = 3)
@

The 450k array has a rather special design.  It is a two-color array, so each array will have an
associated Green signal and a Red signal.

On the 450k array, a CpG may be measured by a ``type I'' or ``type II'' design.  The literature
often uses the term``type I/II probes'' which we believe is unfortunate (see next paragraph).

Each CpG has an associated methylated and un-methylated signal.  If the CpG is of ``type I'', the
methylation and un-methylation signal are originating from two different probes (physical location
on the array).  There is one set of ``type I'' CpGs where the signal comes from the Green channel
for both probes (and the Red channel measures nothing) and another set where the signal comes from
the Red channel.  If the CpG is of ``type II'', a single probe (physical location) is being used to
measure the methylated/un-methylated signal and the methylated signal is always measured in the
Green channel.

This is reflected in the manifest object seen above: ``type I'' CpGs have ``AddressA'', ``AddressB''
(this is a link to the physical location on the array) as well as ``ProbeSeqA'' and ``ProbeSeqB''.
They also have a ``Col'' indicator (which channel is the methylated signal coming from).  In
contrast ``type II'' CpGs have a single ``Address'', one ``ProbeSeq'' and no color information.

Because CpGs of ``type I'' are measured using two different physical probes, we dislike calling the
probes ``type I/II'' and instead attaches the type to the CpG itself.

Note that Illumina uses a special ``cgXXX'' name for the CpGs.  There is actually a meaning to this,
not unlike the meaning associated with ``rsXXX'' numbers for SNPs.  Essentially the XXX is a hash of
the bases surrounding the CpG, making the cgXXX numbers independent of genome version.  Illumina has
a technical note describing this.

\section{SessionInfo}

<<sessionInfo,results=tex,echo=FALSE>>=
toLatex(sessionInfo())
@ 

\nocite{*}
\bibliographystyle{unsrturl}
\bibliography{minfi}
\end{document}

% Local Variables:
% eval: (add-hook 'LaTeX-mode-hook '(lambda () (if (string= (buffer-name) "minfi.Rnw") (setq fill-column 100))))
% LocalWords: LocalWords methylation CpG CDF ExpressionSet Affymetrix preprocessed preprocessing unlogged
% LocalWords: methylated Methylated unmethylated Unmethylated Grn IDAT dataset Illumina CSV
% End:

% \VignetteIndexEntry{minfi User's Guide}
% \VignettePackage{minfi}
% \VignetteDepends{BiocStyle}
% \VignetteEngine{knitr::knitr}

\documentclass{article}

<<style, eval=TRUE, echo=FALSE, results="asis">>=
BiocStyle::latex()
@ 

\title{The minfi user's guide}
\author{Kasper D.\ Hansen \and Jean-Phillipe Fortin}
\date{Modified: October 20, 2013.  Compiled: \today}

\newcommand{\minfi}{\Biocpkg{minfi}}

\begin{document}

\maketitle

\tableofcontents

\section{Introduction}

The \minfi\ package provides tools for analyzing Illumina's Methylation arrays, with a
special focus on the new 450k array for humans.  At the moment Illumina's 27k methylation arrays are
partially supported; some of the functionality may not work (or may not make sense) for the older
27k array, other functionality works fine.

The tasks addressed in this package include preprocessing, QC assessments, identification of
interesting methylation loci and plotting functionality.  Analyzing these types of arrays is ongoing
research in ours and others groups.  In general, the analysis of 450k data is not straightforward
and we anticipate many advances in this area in the near future.

The input data to this package are IDAT files, representing two different color channels prior to
normalization.  It is possible to use Genome Studio files together with the data structures
contained in this package, but in general Genome Studio files are already normalized and we do not
recommend this.

If you are using \minfi\ in a publication, please cite \cite{minfi}.  The SWAN normalization method
is described in \cite{Maksimovic:2012}.  The functional normalization method is described in
\cite{funnorm}.  The code for parsing Illumina IDAT files is from the \Biocpkg{illuminaio} package
described in \cite{illuminaio}.



\subsubsection*{Chip design and terminology}

The 450k array has a complicated design.  What follows is a quick overview.

Each sample is measured on a single array, in two different color channels (red and green).  Each
array measures roughly 450,000 CpG positions.  Each CpG is associated with two measurements: a
methylated measurement and an ``un''-methylated measurement.  These two values can be measured in
one of two ways: using a ``Type I'' design or a ``Type II design''.  CpGs measured using a Type I
design are measured using a single color, with two different probes in the same color channel
providing the methylated and the unmethylated measurements.  CpGs measured using a Type II design
are measured using a single probe, and two different colors provide the methylated and the
unmethylated measurements.  Practically, this implies that on this array there is \emph{not} a
one-to-one correspondence between probes and CpG positions.  We have therefore tried to be precise
about this and we refer to a ``methylation position'' (or ``CpG'') when we refer to a single-base
genomic locus.  The previous generation 27k methlation array uses only the Type I design.

In this package we refer to differentially methylated positions (DMPs) by which we mean a single
genomic position that has a different methylation level in two different groups of samples (or
conditions).  This is different from differentially methylated regions (DMRs) which imply more that
more than one methylation positions are different between conditions.

Physically, each sample is measured on a single ``array''.  There are 12 arrays on a single physical
``slide'' (organized in a 6 by 2 grid).  Slides are organized into ``plates'' containing at most 8
slides (96 arrays).


\subsubsection*{Workflow and R data classes}

A set of 450k data files will initially be read into an \Rclass{RGChannelSet}, representing the raw
intensities as two matrices: one being the green channel and one being the red channel.  This is a
class which is very similar to an \Rclass{ExpressionSet} or an \Rclass{NChannelSet}.

The \Rclass{RGChannelSet} is, together with a \Rclass{IlluminaMethylationManifest} object,
preprocessed into a \Rclass{MethylSet}.  The \Rclass{IlluminaMethylationManifest} object contains the
array design, and describes how probes and color channels are paired together to measure the
methylation level at a specific CpG.  The object also contains information about control probes
(also known as QC probes).  The \Rclass{MethylSet} contains normalized data and essentially consists
of two matrices containing the methylated and the unmethylated evidence for each CpG.  Only the
\Rclass{RGChannelSet} contains information about the control probes.

The process described in the previous paragraph is very similar to the paradigm for analyzing
Affymetrix expression arrays using the \Biocpkg{affy} package (an \Rclass{AffyBatch} is preprocessed
into an \Rclass{ExpressionSet} using array design information stored in a CDF environment
(package)).

A \Rclass{MethylSet} is the starting point for any post-normalization analysis, such as searching for
DMPs or DMRs.


\subsubsection*{Getting Started}

<<load,results="hide",warning=FALSE,message=FALSE>>=
require(minfi)
require(minfiData)
@ 


\section{Reading data}

This package supports analysis of IDAT files, containing the summarized bead information.

In our experience, most labs use a ``Sample Sheet'' CSV file to describe the layout of the
experiment.  This is based on a sample sheet file provided by Illumina.  Our pipeline assumes the
existence of such a file(s), but it is relatively easy to create such a file using for example Excel,
if it is not available.

Examples we have seen, either store all samples in a single directory, or has each slide in a
separate directory.

We use an example dataset with 6 samples, spread across two slides.  First we obtain the system path
to the IDAT files; this requires a bit since the data comes from an installed package

<<baseDir>>=
baseDir <- system.file("extdata", package = "minfiData")
list.files(baseDir)
@ 

This shows the typical layout of 450k data: each ``slide'' (containing 12 arrays) is stored in a
separate directory, with a numeric name.  The top level directory contains the sample sheet file.
Inside the slide directories we find the IDAT files (and possible a number of JPG images or other
files):

<<baseDir2>>=
list.files(file.path(baseDir, "5723646052"))
@ 

The files for each array has another numeric number and consists of a Red and a Grn (Green) IDAT
file.  Note that for this example data, each slide contains only 3 arrays and not 12.  This was done
because of file size limitations and because we only need 6 arrays to illustrate the package's
functionality. 

First we read the sample sheet.  We provide a convenience function for reading in this file
\Rfunction{read.metharray.sheet}.  This function has a couple of attractive bells and whistles.  Let us look
at the output

<<sampleSheet>>=
targets <- read.metharray.sheet(baseDir)
targets
@ 

First the output: this is just a \Rclass{data.frame}.  It contains a column \Rcode{Basename} that
describes the location of the IDAT file corresponding to the sample, as well as two columns
\Rcode{Array} and \Rcode{Slide}.  In the sample sheet provided by Illumina, these two columns are
named \Rcode{Sentrix\_Position} and \Rcode{Sentrix\_ID}, but we rename them.  We provide more detail
on the use of this function below.  The \Rcode{Basename} column tend to be too large for display,
here it is simplified relative to \Rcode{baseDir}:

<<BasenameColumn>>=
sub(baseDir, "", targets$Basename)
@ 
 
(This is just for display purposes). 

With this \Rcode{data.frame}, it is easy to read in the data

<<paths>>= 
RGset <- read.metharray.exp(base = baseDir, targets = targets)
@ 

Let us look at the associated pheno data, which is really just the information contained in the
targets object above.

<<pData>>=
RGset
pd <- pData(RGset)
pd[,1:4]
@ 

The \Rcode{read.metharray.exp} also makes it possible to read in an entire directory or directory tree
(with \Rcode{recursive} set to \Rcode{TRUE}) by using the function just with the argument
\Rcode{base} and \Rcode{targets=NULL}, like

<<read2>>=
RGset2 = read.metharray.exp(file.path(baseDir, "5723646052"))
RGset3 = read.metharray.exp(baseDir, recursive = TRUE)
@ 

\subsubsection*{Advanced notes on Reading Data}

The only important column in sheet \Rcode{data.frame} used in the \Rcode{targets} argument for the
\Rfunction{read.metharray.exp} function is a column named \Rcode{Basename}.  Typically, such an object would
also have columns named \Rcode{Array}, \Rcode{Slide}, and (optionally) \Rcode{Plate}.

We used sheet data files build on top of the Sample Sheet data file provided by Illumina.  This is a
CSV file, with a header.  In this case we assume that the phenotype data starts after a line
beginning with \Rcode{[Data]} (or that there is no header present).

It is also easy to read a sample sheet ``manually'', using the function \Rfunction{read.csv}.  Here, we
know that we want to skip the first 7 lines of the file.

<<sampleSheet2>>=
targets2 <- read.csv(file.path(baseDir, "SampleSheet.csv"), 
                     stringsAsFactors = FALSE, skip = 7)
targets2
@ 

We now need to populate a \Rcode{Basename} column.  On possible approach is the following

<<Basename>>=
targets2$Basename <- file.path(baseDir, targets2$Sentrix_ID, 
                               paste0(targets2$Sentrix_ID, 
                                      targets2$Sentrix_Position))
@ 

Finally, \minfi\ contains a file-based parser: \Rcode{read.metharray}.  The return object represents the
red and the green channel measurements of the samples.  A useful function that we get from the
package \Biocpkg{Biobase} is \Rcode{combine} that combines (``adds'') two sets of samples.  This
allows the user to manually build up an \Rclass{RGChannelSet}.

\section{Data classes in minfi}

\minfi\ contains several classes for representing 450k data, which can be confusing at first.  The
objective of the different classes is to allow flexibility for developers in how to represent and
transform the data.  All of the data classes consists of a data part (possibly large) as well as
additional phenotype information and book keep variables (small).

The basic input data class is \Rclass{RGChannelSet} which represents the Red and the Green color
channel measurements of the entire array.  The data measurements are (unsurprisingly) stored in two
matrices \Robject{Red} and \Robject{Green} with dimensions equal to the number of probes on the
array times the number of samples.

The other four data classes are \Rclass{MethylSet}, \Rclass{RatioSet}, \Rclass{GenomicMethylSet} and
\Rclass{GenomicRatioSet}.  The classes are very similar, at a superficual level.  Transformations
between the four classes are not guranteed to be invertible (two-way), for reasons explained below.

The natural end-point of a normalization strategy is, in our opinion, a \Rclass{GenomicRatioSet}
where a user will have obtained normalized methylation percentages (beta values or M-values) and
where the loci have been mapped to genomic coordinates.  The reason for the existence of four data
classes is to allow flexible normalization strategies.

The two classes with ``Genomic'' in front represents data which has been associated with genomic
coordinates (typically using the function \Rfunction{mapToGenome}).  This association is potentially
a non-invertible mapping since it drops loci which cannot be mapped.  Whether or not a loci is
mappable depends on the genome build that is being used.

The ``MethylSet'' classes contain the data as a \Robject{Meth} and \Robject{Unmeth} matrices,
representing the two methylation channels.  The ``RatioSet'' classes contains the data as either a
\Robject{Beta} or a \Robject{M} matrix (or both) with optionally a \Robject{CN} matrix of copy
number estimates.  We allow the presence of either or both of \Robject{Beta} and \Robject{M} in case
the two matrices are normalized separately.  

\begin{figure}
  \centering
  \includegraphics[width=.5\textwidth]{minfiClasses.pdf}
  \caption{The relationship between the various data classes in minfi.}
  \label{minfiClasses}
\end{figure}

\section{Quality control}

Short, but with pointers to shiny methyl of course.

\section{Dealing with SNPs}

Short, but important.

\section{Normalization}

Overview of normalization.

\section{Bump hunting}

How to do a bump hunting analysis.

\section{Using annotation}

\section{Other stuff}




\section*{Session info}

Here is the output of \Rfunction{sessionInfo} on the system on which
this document was compiled:
<<sessionInfo, results="asis", echo=FALSE, eval=TRUE>>=
toLatex(sessionInfo())
@

\nocite{*}
\bibliography{minfi}

\end{document}

\name{RGChannelSet-class}
\Rdversion{1.1}
\docType{class}
\alias{RGChannelSet-class}
\alias{RGChannelSetExtended-class}
\alias{RGChannelSet}
\alias{RGChannelSetExtended}
%% methods
\alias{show,RGChannelSet-method}
\alias{annotation,RGChannelSet-method}
\alias{annotation<-,RGChannelSet,ANY-method}
\alias{combine,RGChannelSet,RGChannelSet-method}
\alias{getBeta,RGChannelSet-method}
\alias{getManifest,RGChannelSet-method}
\alias{coerce,RGChannelSetExtended,RGChannelSet-method}
%% eSet-methods
\alias{pData,RGChannelSet-method}
\alias{pData<-,RGChannelSet,DataFrame-method}
\alias{sampleNames,RGChannelSet-method}
\alias{sampleNames<-,RGChannelSet,ANY-method}
\alias{featureNames,RGChannelSet-method}
\alias{featureNames<-,RGChannelSet-method}
%% functions
\alias{getGreen}
\alias{getRed}
\alias{getOOB}
\alias{getSnpBeta}
\alias{getNBeads}

\title{Class \code{"RGChannelSet"}}

\description{
  These classes represents raw (unprocessed) data from a two color
  micro array; specifically an Illumina methylation array.
}

\usage{
## Constructors

RGChannelSet(Green = new("matrix"), Red = new("matrix"),
             annotation = "", ...)

RGChannelSetExtended(Green = new("matrix"), Red = new("matrix"),
                    GreenSD = new("matrix"), RedSD = new("matrix"),
                    NBeads = new("matrix"), annotation = "", ...)

## Accessors

\S4method{annotation}{RGChannelSet}(object)
\S4method{pData}{RGChannelSet}(object)
\S4method{sampleNames}{RGChannelSet}(object)
\S4method{featureNames}{RGChannelSet}(object)
\S4method{getBeta}{RGChannelSet}(object, \dots)
getGreen(object)
getRed(object)
getNBeads(object)
\S4method{getManifest}{RGChannelSet}(object)

## Convenience functions
getOOB(object)
getSnpBeta(object)
}

\value{
  An object of class \code{RGChannelSet} or \code{RGChannelSetExtended}
  for the constructors.
}

\arguments{
  \item{object}{An \code{RGChannelSet} (or \code{RGChannelSetExtended}).}
  \item{Green}{A matrix of Green channel values (between zero and infinity)
    with each row being a methylation loci and each column a sample.}
  \item{Red}{See the \code{Green} argument, but for the Green channel.}
  \item{GreenSD}{See the \code{Green} argument, but for standard
    deviations of the Green channel summaries.}  
  \item{RedSD}{See the \code{Green}, but for standard deviations of the
    Red channel summaries.} 
  \item{NBeads}{See the \code{Green} argument, but contains the number
    of beads used to summarize the Green and Red channels.}
  \item{annotation}{An annotation string, optional.}
  \item{\dots}{For the constructor(s), additional arguments to be passed
    to \code{SummarizedExperiment}; of particular interest are
    \code{colData}, \code{rowData} and \code{metadata}.  For
    \code{getBeta} these values gets passed onto \code{getBeta}.} 
}

\section{Constructors}{
  Instances are constructed using the
  \code{RGChannelSet} or \code{RGChannelSetExtended} functions with the
  arguments outlined above.

  \code{as(object, "RGChannelSet")} coerces a
  \code{RGChannelSetExtended} object into a \code{RGChannelSet}.
}

\section{Accessors}{
  \describe{
    \item{\code{getGreen}:}{Gets the Green channel as a matrix.}
    \item{\code{getRed}:}{Gets the Red channel as a matrix.}
    \item{\code{getNBeads}:}{Gets the number of beads as a matrix, this
      requires an \code{RGChannelSetExtended}.}
    \item{\code{getManifest}:}{Gets the manifest object itself
      associated with the array type}
  }
}

\section{Convenience functions}{
  \describe{
    \item{\code{getOOB}:}{Retrives the so-called \dQuote{out-of-band}
      (OOB) probes.  These are the measurements of Type I probes in the
      \dQuote{wrong} color channel.  Return value is a list with two
      matrices, named \code{Red} and \code{Grn}.}
    \item{\code{getSnpBeta}:}{Retrives the measurements of the 65 SNP
      probes located on the array.  These SNP probes are intended to be used
      for sample tracking and sample mixups.  The return value is a
      matrix of beta values.  Each SNP probe ought to have values clustered
      around 3 distinct values corresponding to homo-, and
      hetero-zygotes.}
    \item{\code{combine}:}{Combines two different \code{RGChannelSet},
      eventually using the \code{combine} method for \code{eSet}.}
  }
}

\section{Tips}{
  The class inherits a number of useful methods from \code{SummarizedExperiment}.  In earliers versions of minfi, this class inherited
  from \code{eSet}, and we have kept of number of methods related to this, for example \code{pData}. 
  
  The best way to access phenotype data and sample names are \code{colData} and \code{colnames}.
  
  Amongst the useful methods are

  \describe{
    \item{\code{dim}, \code{nrow}, \code{ncol}}{The dimension (number of
    probes by number of samples) of the experiment.}
    \item{\code{colData}, \code{colnames}, \code{pData}, \code{sampleNames}}{Phenotype information and
    sample names.}
    \item{\code{rownames}, \code{featureNames}}{This is the addresses (probe identifiers)
    of the array.}
  }
}
  
\author{Kasper Daniel Hansen \email{khansen@jhsph.edu}}

\seealso{
  See \code{\linkS4class{SummarizedExperiment}} for the basic class that is used as a
  building block for \code{"RGChannelSet(Extended)"}.  See
  \code{\link{IlluminaMethylationManifest}} for a class representing the
  design of the array. 
}

\examples{
showClass("RGChannelSet")
}

\name{getSex}
\alias{getSex}
\alias{addSex}
\alias{plotSex}
\title{
  Estimating sample sex based on methylation data
}
\description{
  Estimates samples sex based on methylation data.
}
\usage{
getSex(object = NULL, cutoff = -2)
addSex(object, sex = NULL)
plotSex(object, id = NULL)
}
\arguments{
  \item{object}{An object of class \code{[Genomic]MethylSet}.}
  \item{cutoff}{What should the difference in log2 copynumber be between
  males and females.}
  \item{sex}{An optional character vector of sex (with values \code{M}
    and \code{F}).}
  \item{id}{Text used as plotting symbols in the \code{plotSex}
    function. Used for sample identification on the plot.}
}
\details{
  Estimation of sex is based on the median values of measurements on the
  X and Y chromosomes respectively. If \code{yMed} - \code{xMed} is
  less than \code{cutoff} we predict a female, otherwise male.
}
\value{
  For \code{getSex}, a \code{DataFrame} with columns \code{predictedSex}
  (a character with values \code{M} and \code{F}), \code{xMed} and
  \code{yMed}, which are the chip-wide medians of measurements on the two
  sex chromosomes.

  For \code{addSex}, an object of the same type as \code{object} but
  with the output of \code{getSex(object)} added to the pheno data.

  For \code{plotSex}, a plot of \code{xMed} vs. \code{yMed}, which are the
  chip-wide medians of measurements on the two sex chromosomes, coloured by
  \code{predictedSex}.

}
\author{
  Rafael A. Irizarry, Kasper D. Hansen, Peter F. Hickey
}
\examples{
if(require(minfiData)) {
  GMsetEx <- mapToGenome(MsetEx)
  estSex <- getSex(GMsetEx)
  GMsetEx <- addSex(GMsetEx, sex = estSex)
}
}

\name{getGenomicRatioSetFromGEO}
\alias{getGenomicRatioSetFromGEO}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Reading Illumina methylation array data from GEO.
}
\description{
Reading Illumina methylation array data from GEO.
}
\usage{
getGenomicRatioSetFromGEO(GSE = NULL, path = NULL, array = "IlluminaHumanMethylation450k",
                          annotation = .default.450k.annotation, what = c("Beta", "M"),
                          mergeManifest = FALSE, i = 1)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{GSE}{
The GSE ID of the dataset to be downloaded from GEO.
}
  \item{path}{
If data already downloaded, the path with soft files. Either \code{GSE} or \code{path}
are required.
}
  \item{array}{
Array name.
}
  \item{annotation}{
The feature annotation to be used. This includes the location of
features thus depends on genome build.
}
  \item{what}{
Are \code{Beta} or \code{M} values being downloaded.  
}
\item{mergeManifest}{
Should the Manifest be merged to the final object.
}
  \item{i}{
If the GEO download results in more than one dataset, it pickes entry \code{i}.
}
}
\details{
This function downloads data from GEO using
\code{\link[GEOquery]{getGEO}} from the \pkg{GEOquery} package. It then
returns a \linkS4class{GenomicRatioSet} object. Note that the rs probes
(used for genotyping) are dropped.
}
\value{
A \linkS4class{GenomicRatioSet} object.
}
\author{
  Tim Triche Jr. and 
  Rafael A. Irizarry\email{rafa@jimmy.harvard.edu}.
}
\seealso{
  If the data is already in memor you can use 
  \code{\link{makeGenomicRatioSetFromMatrix}}
}
\examples{
\dontrun{
mset=getGenomicRatioSetFromGEO("GSE42752")
}
}

\name{getQC}
\alias{getQC}
\alias{addQC}
\alias{plotQC}
\title{
  Estimate sample-specific quality control (QC) for methylation data
}
\description{
  Estimate sample-specific quality control (QC) for methylation data.
}
\usage{
getQC(object)
addQC(object, qc)
plotQC(qc, badSampleCutoff = 10.5)
}
\arguments{
  \item{object}{An object of class \code{[Genomic]MethylSet}.}
  \item{qc}{An object as produced by \code{getQC}.}
  \item{badSampleCutoff}{The cutoff for identifying a bad sample.}
}
\value{
  For \code{getQC}, a \code{DataFrame} with two columns: \code{mMed} and
  \code{uMed} which are the chipwide medians of the Meth and Unmeth
  channels.

  For \code{addQC}, essentially \code{object} supplied to the function,
  but with two new columns added to the pheno data slot: \code{uMed} and
  \code{mMed}.
}
\author{
  Rafael A. Irizarry and Kasper D. Hansen
}
\seealso{
  \code{\link{minfiQC}} for an all-in-one function.
}
\examples{
if(require(minfiData)){
  qc <- getQC(MsetEx)
  MsetEx <- addQC(MsetEx, qc = qc)
  ## plotQC(qc)
}
}

\name{IlluminaMethylationAnnotation-class}
\Rdversion{1.1}
\docType{class}
\alias{IlluminaMethylationAnnotation-class}
\alias{IlluminaMethylationAnnotation}
\alias{show,IlluminaMethylationAnnotation-method}
\alias{getManifest,IlluminaMethylationAnnotation-method}

\title{Class \code{IlluminaMethylationAnnotation}}

\description{
  This is a class for representing annotation associated with an
  Illumina methylation microarray.  Annotation is transient in the sense
  that it may change over time, wheres the information stored in the
  \code{IlluminaMethylationManifest} class only depends on the array design.
}

\usage{
## Constructor
IlluminaMethylationAnnotation(objectNames, annotation = "",
                              defaults = "", packageName = "")

## Data extraction
\S4method{getManifest}{IlluminaMethylationAnnotation}(object)
}

\arguments{
  \item{object}{An object of class
    \code{IlluminaMethylationAnnotation}.}
  \item{annotation}{An annotation \code{character}.}
  \item{defaults}{A vector of default choices for
    \code{getAnnotation(what = "everything")}.}
  \item{objectNames}{a \code{character} with object names used in the
    package.}
  \item{packageName}{The name of the package this object will be
    contained in.}
}

\value{
  An object of class \code{IlluminaMethylationAnnotation}.
}

\section{Utilitues}{
  In the following code, \code{object} is a \code{IlluminaMethylationAnnotation}.

  \describe{
    \item{\code{getManifest(object)}}{Get the manifest object associated
      with the array.}
  }
}

\author{
  Kasper Daniel Hansen \email{khansen@jhsph.edu}.
}
\seealso{
  \code{\linkS4class{IlluminaMethylationManifest}}
}
\keyword{classes}

\name{dmpFinder}
\alias{dmpFinder}
\title{
Find differentially methylated positions
}
\description{
Identify CpGs where methylation is associated with a continuous or categorical phenotype.
}
\usage{
dmpFinder(dat, pheno, type = c("categorical", "continuous"),
    qCutoff = 1, shrinkVar = FALSE)
}
\arguments{
  \item{dat}{ A \code{MethylSet} or a \code{matrix}.}
  \item{pheno}{ The phenotype to be tested for association with methylation. }
  \item{type}{ Is the phenotype '\sQuote{continuous} or \sQuote{categorical}?  }
  \item{qCutoff}{ DMPs with an FDR q-value greater than this will not be returned.  }
  \item{shrinkVar}{ Should variance shrinkage be used? See details. }
}
\details{

  This function tests each genomic position for association between methylation and
  a phenotype. Continuous phenotypes are tested with linear regression,
  while an F-test is used for categorical phenotypes.

  Variance shrinkage (\code{shrinkVar=TRUE}) is recommended when sample
  sizes are small (<10). The sample variances are squeezed
  by computing empirical Bayes posterior means using the \pkg{limma}
  package.

}
\value{
  A table with one row per CpG.
}
\author{
  Martin Aryee \email{aryee@jhu.edu}.
}
\seealso{
  \code{\link[limma]{squeezeVar}} and the \pkg{limma} package in general.
}
\examples{
if (require(minfiData)) {

grp <- pData(MsetEx)$Sample_Group
MsetExSmall <- MsetEx[1:1e4,] # To speed up the example
M <- getM(MsetExSmall, type = "beta", betaThreshold = 0.001)
dmp <- dmpFinder(M, pheno=grp, type="categorical")
sum(dmp$qval < 0.05, na.rm=TRUE)
head(dmp)

}
}

\name{plotBetasByType}
\alias{plotBetasByType}
\title{
  Plot the overall distribution of beta values and the distributions of
  the Infinium I and II probe types.
}
\description{
  Plot the overall density distribution of beta values and the density
  distributions of the Infinium I and II probe types.
}
\usage{
plotBetasByType(data, probeTypes = NULL, legendPos = "top",
                colors = c("black", "red", "blue"),
                main = "", lwd = 3, cex.legend = 1)
}
\arguments{
  \item{data}{A \code{MethylSet} or a \code{matrix} or a \code{vector}.
    We either use the \code{getBeta} function to get Beta values (in the
    first case) or we assume the matrix or vector contains Beta values.}
  \item{probeTypes}{If \code{data} is a \code{MethylSet} this argument
    is not needed.  Otherwise, a \code{data.frame} with a column 'Name'
    containing probe IDs and a column 'Type' containing their
    corresponding assay design type.}
  \item{legendPos}{The x and y co-ordinates to be used to position the
    legend. They can be specified by keyword or in any way which is
    accepted by \code{\link[grDevices]{xy.coords}}. See
    \code{\link[graphics]{legend}} for details.}
  \item{colors}{Colors to be used for the different beta value density
    distributions. Must be a vector of length 3.}
  \item{main}{Plot title.}
  \item{lwd}{The line width to be used for the different beta value
    density distributions.}
  \item{cex.legend}{The character expansion factor for the legend text.}
}
\details{

  The density distribution of the beta values for a single sample is
  plotted. The density distributions of the Infinium I and II probes are
  then plotted individually, showing how they contribute to the overall
  distribution. This is useful for visualising how using
  \code{\link{preprocessSWAN}} affects the data.

}
\value{
  No return value. Plot is produced as a side-effect.
}

\author{
  Jovana Maksimovic \email{jovana.maksimovic@mcri.edu.au}.
}

\seealso{
  \code{\link{densityPlot}}, \code{\link{densityBeanPlot}},
  \code{\link[graphics]{par}}, \code{\link[graphics]{legend}}
}
\examples{
\dontrun{
if (require(minfiData)) {
  Mset.swan <- preprocessSWAN(RGsetEx, MsetEx)
  par(mfrow=c(1,2))
  plotBetasByType(MsetEx[,1], main="Raw")
  plotBetasByType(Mset.swan[,1], main="SWAN")
}
}
}

\name{compartments}
\alias{compartments}
\alias{createCorMatrix}
\alias{extractAB}
\title{
  Estimates A/B compartments from Illumina methylation arrays
}
\description{
  Estimates A/B compartments as revealed by Hi-C by computing the first
  eigenvector on a binned probe correlation matrix.
}
\usage{
compartments(object, resolution=100*1000, what = "OpenSea", chr="chr22",
                  method = c("pearson", "spearman"), keep=TRUE)
}
\arguments{
  \item{object}{An object of class (Genomic)MethylSet or (Genomic)RatioSet}
  \item{resolution}{An integer specifying the binning resolution}
  \item{what}{Which subset of probes should be used?}
  \item{chr}{The chromosome to be analyzed.}
  \item{method}{Method of correlation.}
  \item{keep}{Should the correlation matrix be stored or not?}
}
\details{

This function extracts A/B compartments from Illumina methylation microarrays. Analysis of Hi-C data has shown that the genome can be divided into two compartments (A/B compartments) that are cell-type specific and are associated with open and closed chromatin respectively. The approximately 170,000 open sea probes on the 450k array can be used to estimate these compartments by computing the first eigenvector on a binned correlation matrix. The binning resolution can be specified by \code{resolution}, and by default is set to a 100 kb. We do not recommend higher resolutions because of the low-resolution probe design of the 450k array.
}

\value{an object of class \code{GRanges} containing the correlation matrix, the compartment eigenvector and the compartment labels (A or B) as metadata.}

\references{
  JP Fortin and KD Hansen.
  \emph{Reconstructing A/B compartments as revealed by Hi-C using long-range correlations in epigenetic data}.
  bioRxiv (2015).
  doi:\href{http://dx.doi.org/10.1101/019000}{10.1101/019000}.
}

\author{
  Jean-Philippe Fortin \email{jfortin@jhsph.edu},
  Kasper D. Hansen \email{kasperdanielhansen@gmail.com}
}
\examples{
if (require(minfiData)) {
  GMset <- mapToGenome(MsetEx)
  ## compartments at 1MB resolution; we recommend 100kb.
  comps <- compartments(GMset, res = 10^6)
}
}

\name{readGEORawFile}
\alias{readGEORawFile}
\title{
  Read in Unmethylated and Methylated signals from a GEO raw file.
}
\description{
  Read in Unmethylated and Methylated signals from a GEO raw file.
}
\usage{
readGEORawFile(filename, sep = ",", Uname = "Unmethylated signal",
               Mname = "Methylated signal", row.names = 1, pData = NULL,
               array = "IlluminaHumanMethylation450k",
               annotation = .default.450k.annotation, mergeManifest = FALSE,
               showProgress = TRUE, ...)
}

\arguments{
  \item{filename}{
    The name of the file to be read from.
  }
  \item{sep}{
    The field separator character.  Values on each line of the
    file are separated by this character.
  }
  \item{Uname}{
    A string that uniquely identifies the columns containing the
    unmethylated signals.
  }
  \item{Mname}{
    A string that uniquely identifies the columns containing the
    methylated signals.}
  \item{row.names}{
    The column containing the feature (CpG) IDs.
  }
  \item{pData}{
    A \code{DataFrame} or \code{data.frame} describing the samples represented by the columns of
    \code{mat}. If the rownames of the \code{pData} don't match the colnames
    of \code{mat} these colnames will be changed. If \code{pData} is not
    supplied, a minimal \code{DataFrame} is created.
  }
  \item{array}{
    Array name.
  }
  \item{annotation}{
    The feature annotation to be used. This includes the location of
    features thus depends on genome build.
  }
  \item{mergeManifest}{
    Should the Manifest be merged to the final object.
  }
  \item{showProgress}{
    TRUE displays progress on the console. It is
    produced in fread's C code.
  }
  \item{...}{
    Additional arguments passed to \code{data.table::\link[data.table]{fread}()}.
  }
}
\details{
 450K experiments uploaded to GEO typically include a raw data file as
 part of the supplementary materials. Unfortunately there does not
 appear to be a standard format. This function provides enough
 flexibility to read these files. Note that you will likely need to
 change the \code{sep}, \code{Uname}, and \code{Mname} arguments and
 make sure the first column includes the feature (CpG) IDs. You can use the
 \code{\link{readLines}} function to decipher how to set these
 arguments.

 Note that the function uses the \code{\link[data.table]{fread}}
function in the \pkg{data.table} package to read the data. To install
\pkg{data.table} type \code{install.packages("data.table")}. We use this
  package because the files too large for \code{read.table}.
}
\value{
  A \linkS4class{GenomicMethylSet} object.

}
\author{
  Rafael A. Irizarry\email{rafa@jimmy.harvard.edu}.
}
\seealso{
  \code{\link{getGenomicRatioSetFromGEO}}
}
\examples{
\dontrun{
library(GEOquery)
getGEOSuppFiles("GSE29290")
gunzip("GSE29290/GSE29290_Matrix_Signal.txt.gz")
# NOTE: This particular example file uses a comma as the decimal separator
#       (e.g., 0,00 instead of 0.00). We replace all such instances using the
#       command line tool 'sed' before reading in the modified file.
cmd <- paste0("sed s/,/\\./g GSE29290/GSE29290_Matrix_Signal.txt > ",
              "GSE29290/GSE29290_Matrix_Signal_mod.txt")
system(cmd)
gmset <- readGEORawFile(filename = "GSE29290/GSE29290_Matrix_Signal_mod.txt",
                        Uname = "Signal_A",
                        Mname = "Signal_B",
                        sep = "\t")
}
}

\name{readTCGA}
\alias{readTCGA}
\title{
Read in tab deliminited file in the TCGA format
}
\description{
Read in tab deliminited file in the TCGA format
}
\usage{
readTCGA(filename, sep = "\t", keyName = "Composite Element REF", Betaname = "Beta_value",
         pData = NULL, array = "IlluminaHumanMethylation450k",
         annotation = .default.450k.annotation, mergeManifest = FALSE,
         showProgress = TRUE)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{filename}{
The name of the file to be read from.
}
  \item{sep}{
 The field separator character.  Values on each line of the
          file are separated by this character.  
}
  \item{keyName}{
The column name of the field containing the feature IDs.
}
  \item{Betaname}{
The character string contained all column names of the beta value fields.
}
  \item{pData}{
A \code{DataFrame} or \code{data.frame} describing the samples represented by the columns of
\code{mat}. If the rownames of the \code{pData} don't match the colnames
of \code{mat} these colnames will be changed. If \code{pData} is not
supplied, a minimal \code{DataFrame} is created.
}
  \item{array}{
Array name.
}
  \item{annotation}{
The feature annotation to be used. This includes the location of
features thus depends on genome build.
}
  \item{mergeManifest}{
Should the Manifest be merged to the final object.
}
  \item{showProgress}{
    TRUE displays progress on the console. It is
    produced in fread's C code.
  }
}
\details{
  This function is a wrapper for
  \code{\link{makeGenomicRatioSetFromMatrix}}. It assumes a very specific
  format, used by TCGA, and then uses the \code{\link[data.table]{fread}}
  function in the \pkg{data.table} package to read the data. To install
  \pkg{data.table} type \code{install.packages("data.table")}. We use this
  package because the files too large for \code{read.table}.
    
  Currently, an example of a file that this function reads is here:
  \url{http://gdac.broadinstitute.org/runs/stddata__2014_10_17/data/UCEC/20141017/gdac.broadinstitute.org_UCEC.Merge_methylation__humanmethylation450__jhu_usc_edu__Level_3__within_bioassay_data_set_function__data.Level_3.2014101700.0.0.tar.gz}. Note it is a 8.1 GB archive.
}
\value{
  A \linkS4class{GenomicRatioSet} object.
}
\author{
  Rafael A. Irizarry\email{rafa@jimmy.harvard.edu}.
}
\seealso{
  \code{\link{makeGenomicRatioSetFromMatrix}}
}
\examples{
\dontrun{
    filename <- "example.txt" ##file must be in the specicif TCGA format
    readTCGA(filename)
}
}

\name{ratioConvert-methods}
\docType{methods}
\alias{ratioConvert}
\alias{ratioConvert-methods}
\alias{ratioConvert,MethylSet-method}
\alias{ratioConvert,GenomicMethylSet-method}
\title{Converting methylation signals to ratios (Beta or M-values)}
\description{
  Converting methylation data from methylation and unmethylation
  channels, to ratios (Beta and M-values).
}

\usage{

\S4method{ratioConvert}{MethylSet}(object, what = c("beta", "M", "both"), keepCN = TRUE, \dots)
\S4method{ratioConvert}{GenomicMethylSet}(object, what = c("beta", "M", "both"), keepCN = TRUE, \dots)
}

\arguments{
  \item{object}{Either a \code{MethylSet}, or a \code{GenomicRatioSet}.}
  \item{what}{Which ratios should be computed and stored?}
  \item{keepCN}{A logical, should copy number values be computed and stored in the
    object?}
  \item{\dots}{Passed to \code{getBeta}, \code{getM} methods.}
}

\value{
  An object of class \code{RatioSet} or \code{GenomicRatioSet}.
}

\author{Kasper Daniel Hansen \email{khansen@jhsph.edu}}

\seealso{

  \code{\link{RatioSet}} or code{\link{GenomicRatioSet}} for the output
  object and \code{\link{MethylSet}} or code{\link{GenomicMethylSet}}
  for the input object.

}

\keyword{methods}
\examples{
if (require(minfiData)) {
  ## MsetEx.sub is a small subset of MsetEx;
  ## only used for computational speed.
  RsetEx.sub <- ratioConvert(MsetEx.sub, keepCN = TRUE)
  }
}

\name{logit2}
\alias{logit2}
\alias{ilogit2}
\title{
  logit in base 2.
}
\description{
  Utility functions for computing logit and inverse logit in base 2.
}
\usage{
logit2(x)
ilogit2(x)
}
\arguments{
  \item{x}{A numeric vector.}
}
\value{A numeric vector.}
\author{
  Kasper Daniel Hansen \email{khansen@jhsph.edu}.
}
\examples{
logit2(c(0.25, 0.5, 0.75))
}

\name{MethylSet-class}
\Rdversion{1.1}
\docType{class}
\alias{MethylSet-class}
\alias{MethylSet}
%% generics
\alias{getMeth}
\alias{getUnmeth}
\alias{getBeta}
\alias{getM}
\alias{preprocessMethod}
%%methods
\alias{show,MethylSet-method}
\alias{preprocessMethod,MethylSet-method}
\alias{annotation,MethylSet-method}
\alias{annotation<-,MethylSet,ANY-method}
\alias{getMeth,MethylSet-method}
\alias{getUnmeth,MethylSet-method}
\alias{getBeta,MethylSet-method}
\alias{getM,MethylSet-method}
\alias{getCN,MethylSet-method}
\alias{getManifest,MethylSet-method}
\alias{combine,MethylSet,MethylSet-method}
\alias{dropMethylationLoci}
%% eSet-methods
\alias{pData,MethylSet-method}
\alias{pData<-,MethylSet,DataFrame-method}
\alias{sampleNames,MethylSet-method}
\alias{sampleNames<-,MethylSet,ANY-method}
\alias{featureNames,MethylSet-method}
\alias{featureNames<-,MethylSet-method}

\title{MethylSet instances}

\description{
  This class holds preprocessed data for Illumina methylation
  microarrays.
}

\usage{
## Constructor

MethylSet(Meth = new("matrix"), Unmeth = new("matrix"),
          annotation = "", preprocessMethod = "", ...)

## Data extraction / Accessors

\S4method{getMeth}{MethylSet}(object)
\S4method{getUnmeth}{MethylSet}(object)
\S4method{getBeta}{MethylSet}(object, type = "", offset = 0, betaThreshold = 0)
\S4method{getM}{MethylSet}(object, type = "", \dots)
\S4method{getCN}{MethylSet}(object, \dots)
\S4method{getManifest}{MethylSet}(object)
\S4method{preprocessMethod}{MethylSet}(object)
\S4method{annotation}{MethylSet}(object)
\S4method{pData}{MethylSet}(object)
\S4method{sampleNames}{MethylSet}(object)
\S4method{featureNames}{MethylSet}(object)

## Utilities
dropMethylationLoci(object, dropRS = TRUE, dropCH = TRUE)
}

\arguments{
  \item{object}{A \code{MethylSet}.}
  \item{Meth}{A matrix of methylation values (between zero and infinity)
    with each row being a methylation loci and each column a sample.}
  \item{Unmeth}{See the \code{Meth} argument.}
  \item{annotation}{An annotation string, optional.}
  \item{preprocessMethod}{A \code{character}, optional.}
  \item{type}{How are the values calculated?  For \code{getBeta} setting
    \code{type="Illumina"} sets \code{offset=100} as per Genome Studio.
    For \code{getM} setting \code{type=""} computes M-values as the
    logarithm of \code{Meth}/\code{Unmeth}, otherwise it is computed as
    the logit of \code{getBeta(object)}.}
  \item{offset}{Offset in the beta ratio, see detail.}
  \item{betaThreshold}{Constrains the beta values to be in the inverval
    betwen \code{betaThreshold} and 1-\code{betaThreshold}.}  
  \item{dropRS}{Should SNP probes be dropped?}
  \item{dropCH}{Should CH probes be dropped}
  \item{\dots}{For the constructor, additional arguments to be passed to
    \code{SummarizedExperiment}; of particular interest are
    \code{colData}, \code{rowData} and \code{metadata}.
    For \code{getM} these values gets passed onto
    \code{getBeta}.} 
  
}

\value{
  An object of class \code{MethylSet} for the constructor.
}

\details{
  This class inherits from \code{eSet}.  Essentially the class is a representation of a
  \code{Meth} matrix and a \code{Unmeth} matrix linked to a \code{pData} data frame.

  In addition, an annotation and a preprocessMethod slot is present.  The annotation slot describes
  the type of array and also which annotation package to use.  The preprocessMethod slot describes
  the kind of preprocessing that resulted in this dataset.

  A \code{MethylSet} stores \code{meth} and \code{Unmeth}.  From these it is easy to compute Beta
  values, defined as
  \deqn{\beta = \frac{\textrm{Meth}}{\textrm{Meth} + \textrm{Unmeth} + \textrm{offset}}}{%
    Beta = Meth / (Meth + Unmeth + offset)}
  The offset is chosen to avoid dividing with small values.  Illumina uses a default of 100.
  M-values (an unfortunate bad name) are defined as
  \deqn{M = \textrm{logit}(\beta) = \log(\textrm{Meth} / \textrm{Unmeth})}{%
    M = logit(Beta) = log( Meth / Unmeth )}
  This formula has problems if either Meth or Unmeth is zero.  For this reason, we can use
  \code{betaThreshold} to make sure Beta is neither 0 nor 1, before taken the logit.  What makes
  sense for the \code{offset} and \code{betaThreshold} depends crucially on how the data was
  preprocessed.  Do not expect the default values to be particular good.
}

\section{Constructor}{
Instances are constructed using the \code{MethylSet} function with the
arguments outlined above.
}

\section{Accessors}{
  In the following code, \code{object} is a \code{MethylSet}.

  \describe{
    \item{\code{getMeth(object)}, \code{getUnmeth(object)}}{Get the
      Meth or the Unmeth matrix }
    \item{\code{getBeta(object)}}{Get Beta, see details.}
    \item{\code{getM(object)}}{get M-values, see details.}
    \item{\code{getCN(object)}}{get copy number values which are defined
      as the sum of the methylation and unmethylation channel.}
    \item{\code{getManifest(object)}}{get the manifest associated with
      the object.}
    \item{\code{preprocessMethod(object)}}{Get the preprocess method \code{character}.}
      }
}

\section{Utilities}{
  In the following code, \code{object} is a \code{MethylSet}.
  
  \describe{
    \item{\code{dropMethylationLoci}(object)}{A unifed interface to removing methylation loci.  You
      can drop SNP probes (probes that measure SNPs, not probes containing SNPs) or CH probes (non-CpG
      methylation).}
    \item{\code{combine}:}{Combines two different \code{MethylSet},
      eventually using the \code{combine} method for \code{eSet}.}
  }
}

\author{Kasper Daniel Hansen \email{khansen@jhsph.edu}}

\seealso{
  \code{\linkS4class{eSet}} for the basic class structure.
  Objects of this class are typically created from an
  \code{\link{RGChannelSet}} using \code{\link{preprocessRaw}} or
  another preprocessing function.
}

\examples{
  showClass("MethylSet")
}

\name{preprocessFunnorm}
\alias{preprocessFunnorm}
\title{
  Functional normalization for Illumina 450k arrays
}
\description{
  Functional normalization (FunNorm) is a between-array normalization
  method for the Illumina Infinium HumanMethylation450 platform. It
  removes unwanted variation by regressing out variability explained by
  the control probes present on the array.
}
\usage{
preprocessFunnorm(rgSet, nPCs=2, sex = NULL, bgCorr = TRUE,
                  dyeCorr = TRUE, keepCN = TRUE, ratioConvert = TRUE,
                  verbose = TRUE)
}
\arguments{
  \item{rgSet}{An object of class \code{RGChannelSet}.}
  \item{nPCs}{Number of principal components from the control probes PCA}
  \item{sex}{An optional numeric vector containing the sex of the
    samples.}
  \item{bgCorr}{Should the NOOB background correction be done, prior to
    functional normalization (see \code{preprocessNoob})}
  \item{dyeCorr}{Should dye normalization be done as part of the NOOB
    background correction (see \code{preprocessNoob})?}
  \item{keepCN}{Should copy number estimates be kept around? Setting to
    \code{FALSE} will decrease the size of the output object significantly.}
  \item{ratioConvert}{Should we run \code{ratioConvert}, ie. should the
    output be a \code{GenomicRatioSet} or should it be kept as a
    \code{GenomicMethylSet}; the latter is for experts.}
  \item{verbose}{Should the function be verbose?}
}
\details{

This function implements functional normalization preprocessing for
Illumina methylation microarrays. Functional normalization extends the
idea of quantile normalization by adjusting for known covariates
measuring unwanted variation. For the 450k array, the first k principal
components of the internal control probes matrix play the role of the
covariates adjusting for technical variation. The number k of principal
components can be set by the argument \code{nPCs}. By default
\code{nPCs} is set to 2, and have been shown to perform consistently
well across different datasets.  This parameter should only be modified
by expert users.  The normalization procedure is applied to the Meth and
Unmeth intensities separately, and to type I and type II signals
separately. For the probes on the X and Y chromosomes we normalize males
and females separately using the gender information provided in the
\code{sex} argument. For the Y chromosome, standard quantile
normalization is used due to the small number of probes, which results
in instability for functional normalization. If sex is unspecified
(\code{NULL}), a guess is made using by the \code{getSex} function using
copy number information. Note that this algorithm does not rely on any
assumption and therefore can be be applicable for cases where global
changes are expected such as in cancer-normal comparisons or tissue
differences.

}

\value{an object of class \code{GenomicRatioSet}, unless
  \code{ratioConvert=FALSE} in which case an object of class
  \code{GenomicMethylSet}.}

\references{
  JP Fortin, A Labbe, M Lemire, BW Zanke, TJ Hudson, EJ Fertig, CMT
  Greenwood and KD Hansen.
  \emph{Functional normalization of 450k methylation array data improves
    replication in large cancer studies}.
  (2014) Genome Biology (2014) 15:503.
  doi:\href{http://www.dx.doi.org/10.1186/s13059-014-0503-2}{10.1186/s13059-014-0503-2}.
}

\author{
  Jean-Philippe Fortin \email{jfortin@jhsph.edu},
  Kasper D. Hansen \email{khansen@jhsph.edu}.
}
\seealso{
  \code{\linkS4class{RGChannelSet}}
  as well as \code{\linkS4class{IlluminaMethylationManifest}} for the
  basic classes involved in these functions.
  \code{\link{preprocessRaw}} and \code{\link{preprocessQuantile}} are
  other preprocessing functions. Background correction may be done using
  \code{\link{preprocessNoob}}.
}
\examples{
if (require(minfiData)) {
  ## RGsetEx.sub is a small subset of RGsetEx;
  ## only used for computational speed.
  Mset.sub.funnorm <- preprocessFunnorm(RGsetEx.sub)
}
}

\name{gaphunter}
\alias{gaphunter}
\title{
  Find gap signals in 450k data
}
\description{
  This function finds probes in the Illumina 450k Array for which calculated beta values
  cluster into distinct groups separated by a defined threshold. It identifies, for these
  \sQuote{gaps signals} the number of groups, the size of these groups, and the samples
  in each group.
}
\usage{
  gaphunter(object, threshold=0.05, keepOutliers=FALSE,
            outCutoff=0.01, verbose=TRUE)
}

\arguments{
  \item{object}{An object of class (Genomic)RatioSet, (Genomic)MethylSet, or matrix.
    If one of the first two, code{\link{getBeta}} is used to calculate beta values. If a
    matrix, must be one of beta values.}
  \item{threshold}{The difference in consecutive, ordered beta values
    that defines the presence of a gap signal. Defaults to 5 percent.}
  \item{keepOutliers}{Should outlier-driven gap signals be kept in the
    results? Defaults to \code{FALSE}}
  \item{outCutoff}{Value used to identify gap signals driven by outliers.
    Defined as the percentage of the total sample size; the sum of
    samples in all groups except the largest must exceed this number of
    samples in order for the probe to still be considered a gap signal.
    Defaults to 1 percent.}
 \item{verbose}{logical value. If \code{TRUE}, it writes some messages
   indicating progress. If \code{FALSE} nothing should be printed.}
}
\details{

  The function can calculate a beta matrix or utilize a user-supplied matrix of
  beta values.

  The function will idenfity probes with a gap in a beta signal greater than
  or equal to the defined \code{threshold}. These probes constitue an additional,
  dataset-specific subset of probes that merit special consideration due to their tendency
  to be driven by an underlying SNP or other genetic variant. In this manner, these probes
  can serve as surrogates for underlying genetic signal locally and/or in a broader
  (i.e. haplotype) context. Please see our upcoming manuscript for a detailed description
  of the utility of these probes.

  Outlier-driven gap signals are those in which the sum of the smaller group(s) does
  not exceed a certain percentage of the sample size, defined by the argument outCutoff.
}
\value{
  A list with three values,
  \item{proberesults}{A data frame listing, for each identified gap signal, the number
    of groups and the size of each group.}
  \item{sampleresults}{a matrix of dimemsions probes (rows) by samples (columns). Individuals
    are assigned numbers based onthe groups into which they cluster. Lower number groups
    indicate lower mean methylation values for the group. For example, individuals coded as
    \sQuote{1} will have a lower mean methylation value than those individuals coded as
    \sQuote{2}.}
  \item{algorithm}{A list detailing the arguments supplied to the function.}
}
\references{
  SV Andrews, C Ladd-Acosta, AP Feinberg, KD Hansen, MD Fallin.
  \emph{\sQuote{Gap hunting} to characterize clustered probe signals in Illumina methylation array data.}
  Epigenetics & Chromatin (2016) 9:56.
  doi:\href{http://dx.doi.org/10.1186/s13072-016-0107-z}{10.1186/s13072-016-0107-z}.
}
\author{
  Shan V. Andrews \email{sandre17@jhu.edu}.
}
\examples{
if(require(minfiData)) {
  gapres <- gaphunter(MsetEx.sub, threshold=0.3, keepOutliers=TRUE)
  #Note: the threshold argument is increased from the default value in this small example
  #dataset with 6 people to avoid the reporting of a large amount of probes as gap signals.
  #In a typical EWAS setting with hundreds of samples, the default arguments should be
  #sufficient.
}
}

\name{combineArrays}
\alias{combineArrays}
\alias{combineArrays,RGChannelSet,RGChannelSet-method}
\alias{combineArrays,MethylSet,MethylSet-method}
\alias{combineArrays,GenomicMethylSet,GenomicMethylSet-method}
\alias{combineArrays,RatioSet,RatioSet-method}
\alias{combineArrays,GenomicRatioSet,GenomicRatioSet-method}
\title{
A method for combining different types of methylation arrays into a
virtual array.
}
\description{
  A method for combining different types of methylation arrays into a
virtual array. The three generations of Illumina methylation arrays are
supported: the 27k, the 450k and the EPIC arrays.  Specifically, the
450k array and the EPIC array share many probes in common.  This
function combines data from the two different array types and outputs a
data object of the user-specified type.  Essentially, this new object
will be like (for example) an EPIC array with many probes missing.
}
\usage{
\S4method{combineArrays}{RGChannelSet,RGChannelSet}(object1, object2,
                  outType = c("IlluminaHumanMethylation450k",
                              "IlluminaHumanMethylationEPIC"),
                  verbose = TRUE)
\S4method{combineArrays}{MethylSet,MethylSet}(object1, object2,
                  outType = c("IlluminaHumanMethylation450k",
                              "IlluminaHumanMethylationEPIC",
                              "IlluminaHumanMethylation27k"),
                  verbose = TRUE)
\S4method{combineArrays}{RatioSet,RatioSet}(object1, object2,
                  outType = c("IlluminaHumanMethylation450k",
                              "IlluminaHumanMethylationEPIC",
                              "IlluminaHumanMethylation27k"),
                  verbose = TRUE)
\S4method{combineArrays}{GenomicMethylSet,GenomicMethylSet}(object1, object2,
                  outType = c("IlluminaHumanMethylation450k",
                              "IlluminaHumanMethylationEPIC",
                              "IlluminaHumanMethylation27k"),
                  verbose = TRUE)
\S4method{combineArrays}{GenomicRatioSet,GenomicRatioSet}(object1, object2,
                  outType = c("IlluminaHumanMethylation450k",
                              "IlluminaHumanMethylationEPIC",
                              "IlluminaHumanMethylation27k"),
                  verbose = TRUE)
}
\arguments{
  \item{object1}{The first object.}
  \item{object2}{The second object.}
  \item{outType}{The array type of the output.}
  \item{verbose}{Should the function be verbose?}
}
\details{
  FIXME: describe the RCChannelSet combination.
}
\value{
  The output object has the same class as the two input objects, that is
  either an \code{RGChannelSet}, a \code{MethylSet}, a \code{RatioSet}, a
  \code{GemomicMethylSet} or a \code{GenomicRatioSet}, with the type of
  the array given by the \code{outType} argument.
}
\author{
  Jean-Philippe Fortin and Kasper D. Hansen.
}
\examples{
if(require(minfiData) && require(minfiDataEPIC)) {
  data(RGsetEx.sub)
  data(RGsetEPIC)
  rgSet <- combineArrays(RGsetEPIC, RGsetEx.sub)
  rgSet
}
}

\name{preprocessIllumina}
\alias{preprocessIllumina}
\alias{bgcorrect.illumina}
\alias{normalize.illumina.control}
\title{
  Perform preprocessing as Genome Studio.
}
\description{
  These functions implements preprocessing for Illumina methylation
  microarrays as used in Genome Studio, the standard software provided
  by Illumina.
}
\usage{
preprocessIllumina(rgSet, bg.correct = TRUE, normalize = c("controls", "no"),
    reference = 1)
bgcorrect.illumina(rgSet)
normalize.illumina.control(rgSet, reference = 1)
}
\arguments{
  \item{rgSet}{An object of class \code{RGChannelSet}.}
  \item{bg.correct}{logical, should background correction be performed?}
  \item{normalize}{logical, should (control) normalization be performed?}
  \item{reference}{for control normalization, which array is the reference?}
}
\details{
  We have reverse engineered the preprocessing methods from Genome
  Studio, based on the documentation.

  The current implementation of control normalization is equal to what
  Genome Studio provides (this statement is based on comparing Genome
  Studio output to the output of this function), with the following
  caveat: this kind of normalization requires the selection of a
  reference array.  It is unclear how Genome Studio selects the
  reference array, but we allow for the manual specification of this
  parameter.

  The current implementation of background correction is roughly equal
  to Genome Studio.  Based on examining the output of 24 arrays, we are
  able to exactly recreate 18 out of the 24.  The remaining 6 arrays had
  a max discrepancy in the Red and/or Green channel of 1-4 (this is on
  the unlogged intensity scale, so 4 is very small).

  A script for doing this comparison may be found in the \code{scripts}
  directory (although it is of limited use without the data files).
}
\value{
  \code{preprocessIllumina} returns a \code{MethylSet}, while
  \code{bgcorrect.illumina} and \code{normalize.illumina.control} both
  return a \code{RGChannelSet} with corrected color channels.
}
\author{
  Kasper Daniel Hansen \email{khansen@jhsph.edu}.
}
\seealso{
  \code{\linkS4class{RGChannelSet}} and \code{\linkS4class{MethylSet}}
  as well as \code{\linkS4class{IlluminaMethylationManifest}} for the
  basic classes involved in these functions.
  \code{\link{preprocessRaw}} is another basic preprocessing function.
}
\examples{
if (require(minfiData)) {

dat <- preprocessIllumina(RGsetEx, bg.correct=FALSE, normalize="controls")
slot(name="preprocessMethod", dat)[1]

}
}

\name{makeGenomicRatioSetFromMatrix}
\alias{makeGenomicRatioSetFromMatrix}
\title{
Make a GenomicRatioSet from a matrix
}
\description{
Make a GenomicRatioSet from a matrix.
}
\usage{
makeGenomicRatioSetFromMatrix(mat, rownames = NULL, pData = NULL,
                              array = "IlluminaHumanMethylation450k",
                              annotation = .default.450k.annotation,
                              mergeManifest = FALSE, what = c("Beta", "M"))
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{mat}{
The matrix that will be converted.
}
  \item{rownames}{
The feature IDs associated with the rows of \code{mat} that will be used
to match to the IlluminaHumanMethylation450k feature IDs.
}
  \item{pData}{
A \code{DataFrame} or \code{data.frame} describing the samples represented by the columns of
\code{mat}. If the rownames of the \code{pData} don't match the colnames
of \code{mat} these colnames will be changed. If \code{pData} is not
supplied, a minimal \code{DataFrame} is created.
}
  \item{array}{
Array name.
}
  \item{annotation}{
The feature annotation to be used. This includes the location of
features thus depends on genome build.
}
\item{mergeManifest}{
Should the Manifest be merged to the final object.
}
\item{what}{
Are \code{Beta} or \code{M} values being downloaded.
}
}
\details{
Many 450K data is provided as csv files. This function permits you to
convert a matrix of values into the class that is used by functions such
as \code{bumphunter} and \code{blockFinder}. The rownames of \code{mat}
are used to match the 450K array features. Alternatively the rownames
can be supplied directly through \code{rownames}.
}
\value{
  A \linkS4class{GenomicRatioSet} object.
}
\author{
  Rafael A. Irizarry\email{rafa@jimmy.harvard.edu}.
}
\seealso{
  \code{\link{getGenomicRatioSetFromGEO}} is similar but reads data from GEO.
}
\examples{
mat <- matrix(10,5,2)
rownames(mat) <- c( "cg13869341", "cg14008030","cg12045430", "cg20826792","cg00381604")
grset <- makeGenomicRatioSetFromMatrix(mat)
}

\name{read.metharray.sheet}
\alias{read.metharray.sheet}
\title{
  Reading an Illumina methylation sample sheet
}
\description{
  Reading an Illumina methylation sample sheet, containing pheno-data
  information for the samples in an experiment.
}
\usage{
read.metharray.sheet(base, pattern = "csv$", ignore.case = TRUE,
    recursive = TRUE, verbose = TRUE)
}
\arguments{
  \item{base}{The base directory from which the search is started.}
  \item{pattern}{What pattern is used to identify a sample sheet file,
    see \code{list.files}}
  \item{ignore.case}{Should the file search be case sensitive?}
  \item{recursive}{Should the file search be recursive, see \code{list.files}?}
  \item{verbose}{Should the function be verbose?}
}
\details{
  This function search the directory \code{base} (possibly including
  subdirectories depending on the argument \code{recursive} for
  \dQuote{sample sheet} files (see below).  These files are identified
  solely on the base of their filename given by the arguments
  \code{pattern} and \code{ignore.case} (note the use of a dollarsign to
  mean end of file name).

  In case multiple sheet files are found, they are all read and the
  return object will contain the concatenation of the files.
  
  A sample sheet file is essentially a CSV (comma-separated) file
  containing one line per sample, with a number of columns describing
  pheno-data or other important information about the sample.  The file
  may contain a header, in which case it is assumed that all lines up to
  and including a line starting with \code{\[Data\]} should be dropped.
  This is modelled after a sample sheet file Illumina provides.  It is
  also very similar to the \code{targets} file made used by the popular
  \code{limma} package (see the extensive package vignette).

  An attempt at guessing the file path to the IDAT files represented in
  the sheet is made.  This should be doublechecked and might need to
  manually changed.

  The type of methylation array is guess by looking at the number of
  probes in the IDAT files.
}
\value{
  A \code{data.frame} containing the columns of all the sample sheets.
  As described in details, a column named \code{Sentrix_Position} is
  renamed to \code{Array} and \code{Sentrix_ID} is renamed to
  \code{Slide}.  In addition the \code{data.frame} will contain a column
  named \code{Basename}.
}
\author{
  Kasper Daniel Hansen\email{khansen@jhsph.edu}.
}
\seealso{
  \code{\link{read.metharray.exp}} and \code{\link{read.metharray}} for functions
  reading IDAT files.  \code{\link[base]{list.files}} for help on the
  arguments \code{recursive} and \code{ignore.case}.
}
\examples{
if(require(minfiData)) {

baseDir <- system.file("extdata", package = "minfiData")
sheet <- read.metharray.sheet(baseDir)

}
}

\name{preprocessQuantile}
\alias{preprocessQuantile}
\title{
  Stratified quantile normalization for an Illumina methylation array.
}

\description{
  Stratified quantile normalization for Illumina amethylation arrays.

  This function implements stratified quantile normalization preprocessing for
  Illumina methylation microarrays. Probes are stratified by region (CpG island,
  shore, etc.)

}
\usage{
preprocessQuantile(object, fixOutliers = TRUE, removeBadSamples = FALSE,
                   badSampleCutoff = 10.5, quantileNormalize = TRUE,
                   stratified = TRUE, mergeManifest = FALSE, sex = NULL,
                   verbose = TRUE)
}
\arguments{
  \item{object}{An object of class \code{RGChannelSet} or
    \code{[Genomic]MethylSet}.}
  \item{fixOutliers}{Should low outlier Meth and Unmeth signals be
    fixed?}
  \item{removeBadSamples}{Should bad samples be removed?}
  \item{badSampleCutoff}{Samples with median Meth and Umneth signals
    below this cutoff will be labelled \sQuote{bad}.}
  \item{quantileNormalize}{Should quantile normalization be performed?}
  \item{stratified}{Should quantile normalization be performed within
    genomic region strata (e.g. CpG island, shore, etc.)?}
  \item{mergeManifest}{Should the information in the associated manifest
    package be merged into the output object?}
  \item{sex}{Gender}
  \item{verbose}{Should the function be verbose?}
}
\details{

This function implements stratified quantile normalization preprocessing
for Illumina methylation microarrays. If \code{removeBadSamples} is
\code{TRUE} we calculate the median Meth and median Unmeth signal for
each sample, and remove those samples where their average falls below
\code{badSampleCutoff}. The normalization procedure is applied to the
Meth and Unmeth intensities separately. The distribution of type I and
type II signals is forced to be the same by first quantile normalizing
the type II probes across samples and then interpolating a reference
distribution to which we normalize the type I probes. Since probe types
and probe regions are confounded and we know that DNAm distributions
vary across regions we stratify the probes by region before applying
this interpolation. For the probes on the X and Y chromosomes we
normalize males and females separately using the gender information
provided in the \code{sex} argument. If gender is unspecified
(\code{NULL}), a guess is made using by the \code{getSex} function using
copy number information. Background correction is not used, but very
small intensities close to zero are thresholded using the
\code{fixMethOutlier}. Note that this algorithm relies on the
assumptions necessary for quantile normalization to be applicable and
thus is not recommended for cases where global changes are expected such
as in cancer-normal comparisons.

Note that this normalization procedure is essentially similar to one
previously presented (Touleimat and Tost, 2012), but has been
independently re-implemented due to the present lack of a released,
supported version.

}
\value{
  a \code{GenomicRatioSet}
}

\note{
  A bug in the function was found to affect the Beta values of type I
  probes, when \code{stratified=TRUE} (default). This is fixed in minfi
  version 1.19.7 and 1.18.4 and greater.
}

\references{
  N Touleimat and J Tost.
  \emph{Complete pipeline for Infinium Human
    Methylation 450K BeadChip data processing using subset quantile
    normalization for accurate DNA methylation estimation.}
  Epigenomics (2012) 4:325-341.
}

\author{
  Rafael A. Irizarry
}

\seealso{
  \code{\link{getSex}}, \code{\link{minfiQC}},
  \code{\link{fixMethOutliers}} for functions used as part of
  \code{preprocessQuantile}.
  }

\examples{
if (require(minfiData)) {
  # NOTE: RGsetEx.sub is a small subset of RGsetEx; only used for computational
  #       speed
  GMset.sub.quantile <- preprocessQuantile(RGsetEx.sub)
}
\dontrun{
if(require(minfiData)) {
  GMset <- preprocessQuantile(RGsetEx)
}
}
}

\name{subsetByLoci}
\alias{subsetByLoci}
\title{
  Subset an RGChannelset by CpG loci.
}
\description{
  Subset an RGChannelSet by CpG loci.
}
\usage{
subsetByLoci(rgSet, includeLoci = NULL, excludeLoci = NULL,
             keepControls = TRUE, keepSnps = TRUE)
}
\arguments{
  \item{rgSet}{An object of class \code{RGChannelSet} (or
    \code{RGChannelSetExtended}).}
  \item{includeLoci}{A character vector of CpG identifiers which should
    be kept.}
  \item{excludeLoci}{A character vector of CpG identifiers which should
    be excluded.}
  \item{keepControls}{Should control probes be kept?}
  \item{keepSnps}{Should SNP probes be kept?}
}
\details{
  This task is non-trivial because an \code{RGChannelSet} is indexed by
  probe position on the array, not by loci name.
}
\value{
  An object of class \code{RGChannelSet}, which some probes removed.
}
\examples{
if(require(minfiData)) {
  loci <- c("cg00050873", "cg00212031", "cg00213748", "cg00214611")
  subsetByLoci(RGsetEx.sub, includeLoci = loci)
  subsetByLoci(RGsetEx.sub, excludeLoci = loci)
}
}

\name{RatioSet-class}
\Rdversion{1.1}
\docType{class}
\alias{RatioSet-class}
\alias{RatioSet}
%%methods
\alias{show,RatioSet-method}
\alias{preprocessMethod,RatioSet-method}
\alias{annotation,RatioSet-method}
\alias{annotation<-,RatioSet,ANY-method}
\alias{getBeta,RatioSet-method}
\alias{getM,RatioSet-method}
\alias{getCN,RatioSet-method}
\alias{combine,RatioSet,RatioSet-method}
%% eSet-methods
\alias{pData,RatioSet-method}
\alias{pData<-,RatioSet,DataFrame-method}
\alias{sampleNames,RatioSet-method}
\alias{sampleNames<-,RatioSet,ANY-method}
\alias{featureNames,RatioSet-method}
\alias{featureNames<-,RatioSet-method}

\title{RatioSet instances}

\description{
  This class holds preprocessed data for Illumina methylation
  microarrays.
}

\usage{
## Constructor

RatioSet(Beta = NULL, M = NULL, CN = NULL,
        annotation = "", preprocessMethod = "", ...)

## Data extraction / Accessors

\S4method{getBeta}{RatioSet}(object)
\S4method{getM}{RatioSet}(object)
\S4method{getCN}{RatioSet}(object)
\S4method{preprocessMethod}{RatioSet}(object)
\S4method{annotation}{RatioSet}(object)
\S4method{pData}{RatioSet}(object)
\S4method{sampleNames}{RatioSet}(object)
\S4method{featureNames}{RatioSet}(object)
}

\arguments{
  \item{object}{A \code{RatioSet}.}
  \item{Beta}{A matrix of beta values (between zero and one)
    with each row being a methylation loci and each column a sample.}
  \item{M}{A matrix of log-ratios (between minus infinity and infinity)
    with each row being a methylation loci and each column a sample.}
  \item{CN}{An optional matrix of copy number estimates
    with each row being a methylation loci and each column a sample.}
    \item{annotation}{An annotation string, optional.}
  \item{preprocessMethod}{A \code{character}, optional.}
  \item{\dots}{For the constructor, additional arguments to be passed to
    \code{SummarizedExperiment}; of particular interest are
    \code{colData}, \code{rowData} and \code{metadata}.
    For \code{getM} these values gets passed onto
    \code{getBeta}.} 
}

\value{
  An object of class \code{RatioSet} for the constructor.
}

\details{
  This class inherits from \code{eSet}.  Essentially the class
  is a representation of a \code{Beta} matrix and/or a \code{M} matrix
  and optionally a \code{CN} (copy number) matrix linked to a
  \code{pData} data frame.

  In addition, an annotation and a preprocessMethod slot is present.  The annotation slot describes
  the type of array and also which annotation package to use.  The preprocessMethod slot describes
  the kind of preprocessing that resulted in this dataset.

  For a \code{RatioSet}, M-values are defined as \code{logit2} of the
  Beta-values if the M-values are not present in the object.  Similarly,
  if only M-values are present in the object, Beta-values are
  \code{ilogit2} of the M-values.
}

\section{Constructor}{
Instances are constructed using the \code{RatioSet} function with the
arguments outlined above.
}

\section{Accessors}{
  In the following code, \code{object} is a \code{RatioSet}.

  \describe{
    \item{\code{getBeta(object)}, \code{getM(object)}, \code{CN(object)}}{Get the
      Beta, M or CN matrix.}
    \item{\code{getManifest(object)}}{get the manifest associated with
      the object.}
    \item{\code{preprocessMethod(object)}}{Get the preprocess method \code{character}.}
      }
}


\section{Utilities}{
  In the following code, \code{object} is a \code{RatioSet}.
  
  \describe{
    \item{\code{combine}:}{Combines two different \code{RatioSet},
      eventually using the \code{combine} method for \code{eSet}.}
  }
}

\author{Kasper Daniel Hansen \email{khansen@jhsph.edu}}

\seealso{
  \code{\linkS4class{eSet}} for the basic class structure.
  Objects of this class are typically created from an
  \code{\link{MethylSet}} using \code{\link{ratioConvert}}.
}

\examples{
  showClass("RatioSet")
}

\name{mdsPlot}
\alias{mdsPlot}
\title{
Multi-dimensional scaling plots giving an overview of similarities and differences between samples.
}
\description{
Multi-dimensional scaling (MDS) plots showing a 2-d projection of distances between samples.
}
\usage{
mdsPlot(dat, numPositions = 1000, sampNames = NULL, sampGroups = NULL, xlim, ylim,
    pch = 1, pal = brewer.pal(8, "Dark2"), legendPos = "bottomleft",
    legendNCol, main = NULL)
}
\arguments{
  \item{dat}{An \code{RGChannelSet}, a \code{MethylSet} or a
    \code{matrix}.  We either use the \code{getBeta} function to get
    Beta values (for the first two) or we assume the matrix contains
    Beta values.}
  \item{numPositions}{Use the \code{numPositions} genomic positions with the most methylation variability when calculating distance between samples.}
  \item{sampNames}{Optional sample names. See details.}
  \item{sampGroups}{Optional sample group labels. See details.}
  \item{xlim}{x-axis limits.}
  \item{ylim}{y-axis limits.}
  \item{pch}{Point type. See \code{\link[graphics]{par}} for details.}
  \item{pal}{Color palette.}
  \item{legendPos}{The legend position. See
    \code{\link[graphics]{legend}} for details.}
  \item{legendNCol}{The number of columns in the legend. See
    \code{\link[graphics]{legend}} for details.}
  \item{main}{Plot title.}
}
\details{

  Euclidean distance is calculated between samples using the
  \code{numPositions} most variable CpG positions. These distances are then
  projected into a 2-d plane using classical multidimensional scaling
  transformation.

}
\value{
No return value. Plots are produced as a side-effect.
}
\references{
  I Borg, P Groenen.
  \emph{Modern Multidimensional Scaling: theory and applications (2nd
    ed.)}
  New York: Springer-Verlag (2005) pp. 207-212. ISBN 0387948457.

  \url{http://en.wikipedia.org/wiki/Multidimensional_scaling}
}
\author{
Martin Aryee \email{aryee@jhu.edu}.
}

\seealso{
\code{\link{qcReport}}, \code{\link{controlStripPlot}},
\code{\link{densityPlot}}, \code{\link{densityBeanPlot}},
\code{\link[graphics]{par}}, \code{\link[graphics]{legend}}
}
\examples{
if (require(minfiData)) {

names <- pData(MsetEx)$Sample_Name
groups <- pData(MsetEx)$Sample_Group
mdsPlot(MsetEx, sampNames=names, sampGroups=groups)

}
}

\name{getMethSignal}
\alias{getMethSignal}
\title{
  Various utilities
}
\description{
  Utility functions operating on objects from the minfi package.
}
\usage{
getMethSignal(object, what = c("Beta", "M"), \dots)
}
\arguments{
  \item{object}{An object from the minfi package supporting either
    \code{getBeta} or \code{getM}.}
  \item{what}{Which signal is returned.}
  \item{\dots}{Passed to the method described by argument \code{what}.}
}
\value{A matrix.}
\author{
  Kasper Daniel Hansen \email{khansen@jhsph.edu}.
}
\examples{
if(require(minfiData)) {
  head(getMethSignal(MsetEx, what = "Beta"))
}
}

\name{IlluminaMethylationManifest-class}
\Rdversion{1.1}
\docType{class}
\alias{IlluminaMethylationManifest-class}
\alias{IlluminaMethylationManifest}
\alias{show,IlluminaMethylationManifest-method}
\alias{getManifest,IlluminaMethylationManifest-method}
\alias{getManifest,character-method}
\alias{getManifest}
\alias{getProbeInfo}
\alias{getManifestInfo}
\alias{getControlAddress}

\title{Class \code{"IlluminaMethylationManifest"}}

\description{
  This is a class for representing an Illumina methylation microarray
  design, ie. the physical location and the probe sequences.  This
  information should be independent of genome build and annotation.
}

\usage{
## Constructor

IlluminaMethylationManifest(TypeI = DataFrame(),
                            TypeII = DataFrame(),
                            TypeControl = DataFrame(),
                            TypeSnpI = DataFrame(),
                            TypeSnpII = DataFrame(),
                            annotation = "")

## Data extraction
\S4method{getManifest}{IlluminaMethylationManifest}(object)
\S4method{getManifest}{character}(object)
getProbeInfo(object, type = c("I", "II", "Control",
                              "I-Green", "I-Red", "SnpI", "SnpII"))
getManifestInfo(object, type = c("nLoci", "locusNames"))
getControlAddress(object, controlType = c("NORM_A", "NORM_C",
                                          "NORM_G", "NORM_T"),
                  asList = FALSE)
}

\arguments{
  \item{object}{Either an object of class
    \code{IlluminaMethylationManifest} or class \code{character} for
  \code{getManifest}.  For \code{getProbeInfo}, \code{getManifestInfo}
  and \code{getControlAddress} an object of either class
  \code{RGChannelSet}, \code{IlluminaMethylationManifest}.}
  \item{TypeI}{A \code{DataFrame} of type I probes.}
  \item{TypeII}{A \code{DataFrame} of type II probes.}
  \item{TypeControl}{A \code{DataFrame} of control probes.}
  \item{TypeSnpI}{A \code{DataFrame} of SNP type I probes.}
  \item{TypeSnpII}{A \code{DataFrame} of SNP type II probes.}
  \item{annotation}{An annotation \code{character}.}
  \item{type}{A single character describing what kind of information
    should be returned.  For \code{getProbeInfo} it represents the following
    subtypes of probes on the array: Type I, Type II, Controls as well
    as Type I (methylation measured in the Green channel) and Type II
    (methylation measured in the Red channel).  For
    \code{getManifestInfo} it represents either the number of
    methylation loci (approx. number of CpGs) on the array or the locus names.}
  \item{controlType}{A character vector of control types.}
  \item{asList}{If \code{TRUE} the return object is a list with one
  component for each \code{controlType}.}
}

\value{
  An object of class \code{IlluminaMethylationManifest} for the constructor.
}

\section{Details}{
  The \code{data} slot contains the following objects: \code{TypeI},
  \code{TypeII} and \code{TypeControl} which are all of class
  \code{data.frame}, describing the array design.

  Methylation loci of type I are measured using two different probes, in
  either the red or the green channel.  The columns \code{AddressA},
  \code{AddresB} describes the physical location of the two probes on
  the array (with \code{ProbeSeqA}, \code{ProbeSeqB} giving the probe
  sequences), and the column \code{Color} describes which color channel
  is used.

  Methylation loci of type II are measured using a single probe, but
  with two different color channels.  The methylation signal is always
  measured in the green channel.
}

\section{Utilities}{
  In the following code, \code{object} is a \code{IlluminaMethylationManifest}.

  \describe{
    \item{\code{getManifest(object)}}{Get the manifest object.}
    \item{\code{getProbeInfo(object)}}{Returns a \code{DataFrame}
    giving the type I, type II or control probes.  It is also possible
    to get the type I probes measured in the Green or Red channel.  This
    function ensures that the return object only contains probes which
    are part of the input object.  In case of a \code{RGChannelSet} and
    type I probes, both addresses needs to be in the object.}
   \item{\code{getManifestInfo(object)}}{Get some information about the
    manifest object (the chip design).}
   \item{\code{getControlAddress(object)}}{Get the control addresses for
    control probes of a certain type.}
   \item{\code{getControlTypes(object)}}{Returns the types and the
    numbers of control probes of each type.}
  }
}

\author{
  Kasper Daniel Hansen \email{khansen@jhsph.edu}.
}

\seealso{
  \code{\link{IlluminaMethylationAnnotation}} for annotation
  information for the array (information depending on a specific genome
  build).
}
\examples{
if(require(IlluminaHumanMethylation450kmanifest)) {

show(IlluminaHumanMethylation450kmanifest)
head(getProbeInfo(IlluminaHumanMethylation450kmanifest, type = "I"))
head(IlluminaHumanMethylation450kmanifest@data$TypeI)
head(IlluminaHumanMethylation450kmanifest@data$TypeII)
head(IlluminaHumanMethylation450kmanifest@data$TypeControl)

}
}
\keyword{classes}

\name{preprocessRaw}
\alias{preprocessRaw}
\title{
  Creation of a MethylSet without normalization
}
\description{
  Converts the Red/Green channel for an Illumina methylation array into
  methylation signal, without using any normalization.
}
\usage{
preprocessRaw(rgSet)
}
\arguments{
  \item{rgSet}{An object of class \code{RGChannelSet}.}
}
\details{
  This function takes the Red and the Green channel of an Illumina
  methylation array, together with its associated manifest object and
  converts it into a \code{MethylSet} containing the methylated and
  unmethylated signal.
}
\value{
  An object of class \code{MethylSet}
}
\author{
  Kasper Daniel Hansen\email{khansen@jhsph.edu}.
}
\seealso{
  \code{\linkS4class{RGChannelSet}} and \code{\linkS4class{MethylSet}}
  as well as \code{\linkS4class{IlluminaMethylationManifest}}.
}
\examples{
if (require(minfiData)) {

dat <- preprocessRaw(RGsetEx)
slot(name="preprocessMethod", dat)[1]

}
}

\name{mapToGenome-methods}
\docType{methods}
\alias{mapToGenome}
\alias{mapToGenome-methods}
\alias{mapToGenome,MethylSet-method}
\alias{mapToGenome,RatioSet-method}
\alias{mapToGenome,RGChannelSet-method}
\title{Mapping methylation data to the genome}
\description{
  Mapping Ilumina methylation array data to the genome using an
  annotation package.  Depending on the genome, not all methylation loci
  may have a genomic position.
}

\usage{

\S4method{mapToGenome}{MethylSet}(object, mergeManifest = FALSE)
\S4method{mapToGenome}{MethylSet}(object, mergeManifest = FALSE)
\S4method{mapToGenome}{RGChannelSet}(object, \dots)
}

\arguments{
  \item{object}{Either a \code{MethylSet}, a \code{RGChannelSet} or a
    \code{RatioSet}.}
  \item{mergeManifest}{Should the information in the associated manifest
    package be merged into the location \code{GRanges}?}
  \item{\dots}{Passed to the method for \code{MethylSet}.}
}

\value{
  An object of class \code{GenomicMethylSet} or \code{GenomicRatioSet}.
}
\details{
  FIXME: details on the MethylSet method.

  The \code{RGChannelSet} method of this function is a convenience
  function: the \code{RGChannelSet} is first transformed into a
  \code{MethylSet} using \code{preprocessRaw}.  The resulting
  \code{MethylSet} is then mapped directly to the genome.

  This function silently drops loci which cannot be mapped to a genomic
  position, based on the associated annotation package.
}

\author{Kasper Daniel Hansen \email{khansen@jhsph.edu}}

\seealso{
  \code{\link{GenomicMethylSet}} for the output object and
  \code{\link{MethylSet}} for the input object.  Also,
  \code{\link{getLocations}} obtains the genomic locations for a given object.
}

\keyword{methods}
\examples{
if (require(minfiData)) {
  ## MsetEx.sub is a small subset of MsetEx;
  ## only used for computational speed.
  GMsetEx.sub <- mapToGenome(MsetEx.sub)
}
}

\name{qcReport}
\alias{qcReport}
\title{
  QC report for Illumina Infinium Human Methylation 450k arrays
}
\description{
  Produces a PDF QC report for Illumina Infinium Human Methylation 450k
  arrays, useful for identifying failed samples.
}
\usage{
qcReport(rgSet, sampNames = NULL, sampGroups = NULL, pdf = "qcReport.pdf",
    maxSamplesPerPage = 24, controls = c("BISULFITE CONVERSION I",
    "BISULFITE CONVERSION II", "EXTENSION", "HYBRIDIZATION",
    "NON-POLYMORPHIC", "SPECIFICITY I", "SPECIFICITY II", "TARGET REMOVAL"))
}
\arguments{
  \item{rgSet}{An object of class \code{RGChannelSet}.}
  \item{sampNames}{Sample names to be used for labels.}
  \item{sampGroups}{Sample groups to be used for labels.}
  \item{pdf}{Path and name of the PDF output file.}
  \item{maxSamplesPerPage}{Maximum number of samples to plot per page in
    those sections that plot each sample separately.}
  \item{controls}{The control probe types to include in the report.}
}
\details{

  This function produces a QC report as a PDF file. It is a useful first
  step after reading in a new dataset to get an overview of quality and
  to flag potentially problematic samples.

}
\value{
No return value. A PDF is produced as a side-effect.
}
\author{
Martin Aryee \email{aryee@jhu.edu}.
}

\seealso{
\code{\link{mdsPlot}}, \code{\link{controlStripPlot}}, \code{\link{densityPlot}}, \code{\link{densityBeanPlot}}
}
\examples{
if (require(minfiData)) {

names <- pData(RGsetEx)$Sample_Name
groups <- pData(RGsetEx)$Sample_Group

\dontrun{
qcReport(RGsetEx, sampNames=names, sampGroups=groups, pdf="qcReport.pdf")
}

}
}

\name{cpgCollapse}
\alias{cpgCollapse}
\title{
  Collapse methylation values of adjacent CpGs into a summary value.
}

\description{

  This function groups adjacent loci into clusters with a specified
maximum gap between CpGs in the cluster, and a specified maximum cluster
width.  The loci within each cluster are summarized resulting in a
single methylation estimate per cluster.

}

\usage{
cpgCollapse(object, what = c("Beta", "M"), maxGap = 500,
            blockMaxGap = 2.5 * 10^5, maxClusterWidth = 1500,
            dataSummary = colMeans, na.rm = FALSE,
            returnBlockInfo = TRUE, islandAnno = NULL, verbose = TRUE,
            \dots)
}

\arguments{
  \item{object}{An object of class \code{[Genomic]MethylSet} or \code{[Genomic]RatioSet}.}
  \item{what}{Should operation be performed on the M-scale or
  Beta-scale?}
  \item{maxGap}{Maximum gap between CpGs in a cluster}
  \item{blockMaxGap}{Maximum block gap}
  \item{maxClusterWidth}{Maximum cluster width}
  \item{dataSummary}{Function used to summarize methylation across CpGs
    in the cluster.}
  \item{na.rm}{Should NAs be removed when summarizing? Passed on to the
    dataSummary function.}
  \item{returnBlockInfo}{Should the block annotation table be returned
    in addition to the block table?}
  \item{islandAnno}{Which Island annotation should be used.  \code{NULL}
    indicates the default.  This argument is only useful if the
    annotatio object contains more than one island annotation.}
  \item{verbose}{Should the function be verbose?}
  \item{\dots}{Passed on to getMethSignal and getCN. Can be used to
    specify } 
}
\details{

  This function is used as the first step of block-finding. It groups
  adjacent loci into clusters with a default maximum gap of 500bp and a
  maximum cluster width of 1,500bp.  The loci within each cluster are
  then summarized (using the mean by default) resulting in a single
  methylation estimate per cluster. Cluster estimates from open-sea
  probes are used in block-finding.

}

\value{

If returnBlockInfo is \code{FALSE}: a GenomicRatioSet of collapsed CpG clusters.

If returnBlockInfo is \code{TRUE}:
\item{object }{A GenomicRatioSet of collapsed CpG clusters}
\item{blockInfo }{A cluster annotation data frame}
}

\author{
Rafael Irizarry
}

\seealso{
  \code{\link{blockFinder}}
}

\name{minfi-defunct}
\alias{minfi-defunct}
\alias{read.450k}
\alias{read.450k.sheet}
\alias{read.450k.exp}
\title{Defunct functions in package \sQuote{minfi}}

\description{
  These functions are provided now defunct in \sQuote{minfi}.
}

\details{
  The following functions are now defunct (not working anymore); use the
  replacement indicated below: 
  \itemize{
    \item{\code{read.450k}: Use \code{\link{read.metharray}}}
    \item{\code{read.450k.sheet}: Use \code{\link{read.metharray.sheet}}}
    \item{\code{read.450k.exp}: Use \code{\link{read.metharray.exp}}}
  }
}

\seealso{
  \code{\link{Defunct}}.
}

\name{blockFinder}
\alias{blockFinder}
\title{
  Finds blocks of methylation differences for Illumina methylation arrays
}
\description{
  Finds blocks (large scale regions) of methylation differences for
  Illumina methylation arrays
}

\usage{
blockFinder(object, design, coef = 2, what = c("Beta", "M"),
                        cluster = NULL, cutoff = NULL,
                        pickCutoff = FALSE, pickCutoffQ = 0.99,
                        nullMethod = c("permutation","bootstrap"),
                        smooth = TRUE, smoothFunction = locfitByCluster,
                        B = ncol(permutations), permutations = NULL,
                        verbose = TRUE, bpSpan = 2.5*10^5,\dots)
}

\arguments{
  \item{object}{An object of class GenomicRatioSet.}
  \item{design}{Design matrix with rows representing samples and columns
    representing covariates. Regression is applied to each row of mat.}
  \item{coef}{An integer denoting the column of the design matrix
    containing the covariate of interest. The hunt for bumps will be
    only be done for the estimate of this coefficient.}
  \item{what}{Should blockfinding be performed on M-values or Beta
    values?}
  \item{cluster}{The clusters of locations that are to be analyzed
    together.  In the case of microarrays, the clusters are many times
    supplied by the manufacturer. If not available the function
    \code{\link{clusterMaker}} can be used to cluster nearby locations.}
 \item{cutoff}{A numeric value. Values of the estimate of the genomic
    profile above the cutoff or below the negative of the cutoff will be
    used as candidate regions.  It is possible to give two separate
    values (upper and lower bounds).  If one value is given, the lower
    bound is minus the value.}
  \item{pickCutoff}{Should a cutoff be picked automatically?}
  \item{pickCutoffQ}{The quantile used for picking the cutoff using the
    permutation distribution.}
  \item{nullMethod}{
    Method used to generate null candidate regions, must be one of \sQuote{bootstrap} or
    \sQuote{permutation} (defaults to \sQuote{permutation}). However, if covariates in addition to the 
    outcome of interest are included in the design matrix
    (ncol(design)>2), the \sQuote{permutation} approach is not
    recommended. See vignette and original paper for more information.}
  \item{smooth}{
    A logical value. If TRUE the estimated profile will be smoothed with the
    smoother defined by \code{smoothFunction}}
  \item{smoothFunction}{
    A function to be used for smoothing the estimate of the genomic
    profile. Two functions are provided by the package: \code{loessByCluster}
    and \code{runmedByCluster}.}
   \item{B}{An integer denoting the number of resamples to use when computing
   null distributions. This defaults to 0. If \code{permutations} is
  supplied that defines the number of permutations/bootstraps and \code{B} is
  ignored.}
\item{permutations}{ is a matrix with columns providing indexes to be used to
  scramble the data and create a null distribution. If this matrix is not supplied and \code{B}>0 then
  these indexes created using the function \code{sample}. }
  \item{verbose}{Should the function be verbose?}
  \item{bpSpan}{Smoothing span. Note that this defaults to a large value
  becuase we are searching for large scale changes.}
  \item{\dots}{further arguments sent to \code{bumphunterEngine}.}
}

\details{

The approximately 170,000 open sea probes on the 450k can be used to detect 
long-range changes in methylation status. These large scale changes that can range up
to several Mb have typically been identified only through whole-genome bisulfite 
sequencing. \code{blockFinder} groups the averaged methylation values in open-sea 
probe clusters (See \link{cpgCollapse}) into large regions in which the \link{bumphunter}
procedure is applied with a large (250KB+) smoothing window.

Note that estimating the precise boundaries of these blocks are constrained by the 
resolution of the array.
}

\value{
  FIXME
}

\seealso{
  \code{\link{cpgCollapse}}, and \code{\link{bumphunter}}
}

\name{minfi-package}
\alias{minfi-package}
\alias{minfi}
\docType{package}
\title{
  Analyze Illumina's methylation arrays
}
\description{
  Tools for analyzing and visualizing Illumina methylation array data.
  There is special focus on the 450k array; the 27k array is not
  supported at the moment.
}
\details{
  The package contains a (hopefully) useful vignette; this vignette
  contains a lengthy description of the package content and
  capabilities.
}
\keyword{ package }

\name{fixMethOutliers}
\alias{fixMethOutliers}
\title{
  Fix methylation outliers
}
\description{
  Methylation outliers (loci with very extreme values of the Meth or
  Unmeth channel) are identified and fixed (see details).
}
\usage{
fixMethOutliers(object, K = -3, verbose = FALSE)
}
\arguments{
  \item{object}{An object of class \code{[Genomic]MethylSet}.}
  \item{K}{The number of standard deviations away from the median when
    defining the outlier cutoff, see details.}
  \item{verbose}{Should the function be verbose?}
}
\details{
  This function fixes outlying methylation calls in the Meth channel and
  Unmeth channel separately.

  Unlike other types of arrays, all loci on a methylation array ought to
  measure something (apart from loci on the Y chromosome in a female
  sample).  An outlier is a loci with a very low value in one of the two
  methylation channels.  Typically, relatively few loci ought to be outliers.

  An outlier is defined in a sample and methylation channel specific
  way.  First the (sample, methylation channel) values are
  \code{log2(x+0.5)} transformed and then the median and mad of these
  values are computed.  An outlier is then defined to be any value less
  than the median plus \code{K} times the mad, and these outlier values
  are thresholded at the cutoff (on the original scale).
}
\value{
  An object of the same class as \code{object} where outlier values in
  the methylation channels have been thresholded.
}
\author{
  Rafael A. Irizarry and Kasper D. Hansen
}
\seealso{
  \code{\link{minfiQC}}
}
\examples{
if(require(minfiData)) {
  MsetEx <- fixMethOutliers(MsetEx)
}
}

\name{controlStripPlot}
\alias{controlStripPlot}
\title{
Plot control probe signals.
}
\description{
Strip plots are produced for each control probe type specified.
}
\usage{
controlStripPlot(rgSet, controls = c("BISULFITE CONVERSION I",
    "BISULFITE CONVERSION II"), sampNames = NULL, xlim = c(5, 17))
}
\arguments{
  \item{rgSet}{An \code{RGChannelSet}.}
  \item{controls}{A vector of control probe types to plot.}
  \item{sampNames}{Sample names to be used for labels.}
  \item{xlim}{x-axis limits.}
}
\details{
This function produces the control probe signal plot component of the QC report.
}
\value{
No return value. Plots are produced as a side-effect.
}
\author{
Martin Aryee \email{aryee@jhu.edu}.
}
\seealso{
\code{\link{qcReport}}, \code{\link{mdsPlot}}, \code{\link{densityPlot}}, \code{\link{densityBeanPlot}}
}
\examples{
if (require(minfiData)) {

names <- pData(RGsetEx)$Sample_Name
controlStripPlot(RGsetEx, controls=c("BISULFITE CONVERSION I"), sampNames=names)

}
}

\name{getAnnotation}
\alias{getAnnotation}
\alias{getAnnotationObject}
\alias{getLocations}
\alias{getSnpInfo}
\alias{addSnpInfo}
\alias{dropLociWithSnps}
\alias{getIslandStatus}
\alias{getProbeType}
\title{
  Accessing annotation for Illumina methylation objects
}
\description{
  These functions access provided annotation for various Illumina
  methylation objects.
}
\usage{
getAnnotation(object, what = "everything", lociNames = NULL,
              orderByLocation = FALSE, dropNonMapping = FALSE)

getLocations(object, mergeManifest = FALSE,
             orderByLocation = FALSE, lociNames = NULL)

getAnnotationObject(object)

getSnpInfo(object, snpAnno = NULL)
addSnpInfo(object, snpAnno = NULL)

dropLociWithSnps(object, snps = c("CpG", "SBE"), maf = 0, snpAnno = NULL)

getProbeType(object, withColor = FALSE)

getIslandStatus(object, islandAnno = NULL)
}

\arguments{
  \item{object}{A \code{minfi} object.}
  \item{what}{Which annotation objects should be returned?}
  \item{lociNames}{Restrict the return values to these loci.}
  \item{orderByLocation}{Should the return object be ordered according
    to genomic location.}
  \item{dropNonMapping}{Should loci that do not have a genomic location
    associated with it (by being marked as \code{unmapped} or
    \code{multi}) be dropped from the return object.}
  \item{mergeManifest}{Should the manifest be merged into the return
    object?}
  \item{snpAnno}{The snp annotation you want to use; \code{NULL}
    signifies picking the default.}
  \item{withColor}{Should the return object have the type I probe color
    labelled?}
  \item{snps}{The type of SNPs used.}
  \item{maf}{Minor allelle fraction.}
  \item{islandAnno}{Like \code{snpAnno}, but for islands.}
}

\details{

  \code{getAnnotation} returns requested annotation as a
  \code{DataFrame}, with each row corresponding to a methylation loci.
  If \code{object} is of class \code{IlluminaHumanAnnotation} no
  specific ordering of the return object is imposed.  If, on the other
  hand, the class of \code{object} imposes some natural order on the
  return object (ie. if the object is of class
  \code{[Genomic](Methyl|Ratio)Set}), this order is kept in the return
  object.  Note that \code{RGChannelSet} does not impose a specific
  ordering on the methylation loci.

  \code{getAnnotationObject} returns the annotation object, as opposed
  to the annotation the object contains.  This is useful for printing
  and examining the contents of the object.

  \code{getLocations} is a convenience function which returns
  \code{Locations} as a \code{GRanges} and which furthermore drops
  unmapped loci.  A user should not need to call this function, instead
  \code{mapToGenome} should be used to get genomic coordinates and
  \code{granges} to return those coordinates.

  \code{getSnpInfo} is a conevnience function which gets a SNP DataFrame
  containing information on which probes contains SNPs
  where. \code{addSnpInfo} adds this information to the \code{rowRanges}
  or \code{granges} of the object.  \code{dropLociWithSnps} is a
  convenience function for removing loci with SNPs based on their MAF.

  To see which options are available for \code{what}, simply print the
  annotation object, possibly using \code{getAnnotationObject}.
}

\value{

  For \code{getAnnotation}, a \code{DataFrame} with the requested
  information.

  For \code{getAnnotationObject}, a \code{IlluminaMethylationAnnotation}
  object.

  For \code{getLocations}, a \code{GRanges} with the locations.

  For \code{getProbeType} and \code{getIslandStatus}, a \code{character}
  vector with the requested information.

  For \code{getSnpInfo}, a \code{DataFrame} with the requested
  information.  For \code{addSnpInfo}, an object of the same class as
  \code{object} but with the SNP information added to the metadata columns
  of the granges of the object.

  For \code{dropLociWithSnps} an object of the same kind as the input,
  possibly with fewer loci.
}

\author{
Kasper Daniel Hansen\email{khansen@jhsph.edu}
}
\seealso{
\code{\linkS4class{IlluminaMethylationAnnotation}} for the basic class,
\code{\link{mapToGenome}} for a better alternative (for users) to
\code{getLocations}.
}
\examples{
if(require(minfiData)) {
  table(getIslandStatus(MsetEx))
  getAnnotation(MsetEx, what = "Manifest")
}
}

\name{GenomicRatioSet-class}
\Rdversion{1.1}
\docType{class}
\alias{GenomicRatioSet-class}
\alias{GenomicRatioSet}
\alias{getCN}
%%methods
\alias{show,GenomicRatioSet-method}
\alias{annotation,GenomicRatioSet-method}
\alias{annotation<-,GenomicRatioSet,ANY-method}
\alias{preprocessMethod,GenomicRatioSet-method}
\alias{getBeta,GenomicRatioSet-method}
\alias{getM,GenomicRatioSet-method}
\alias{getCN,GenomicRatioSet-method}
\alias{mapToGenome,GenomicRatioSet-method}
\alias{combine,GenomicRatioSet,GenomicRatioSet-method}
%% eSet-methods
\alias{pData,GenomicRatioSet-method}
\alias{pData<-,GenomicRatioSet,DataFrame-method}
\alias{sampleNames,GenomicRatioSet-method}
\alias{sampleNames<-,GenomicRatioSet,ANY-method}
\alias{featureNames,GenomicRatioSet-method}
\alias{featureNames<-,GenomicRatioSet-method}



\title{GenomicRatioSet instances}

\description{
  This class holds preprocessed data for Illumina methylation
  microarrays, mapped to a genomic location.
}

\usage{
## Constructor

GenomicRatioSet(gr = GRanges(), Beta = NULL, M = NULL,
                CN = NULL, annotation = "",
                preprocessMethod = "", ...)

## Data extraction / Accessors

\S4method{getBeta}{GenomicRatioSet}(object)
\S4method{getM}{GenomicRatioSet}(object)
\S4method{getCN}{GenomicRatioSet}(object)
\S4method{pData}{GenomicRatioSet}(object)
\S4method{sampleNames}{GenomicRatioSet}(object)
\S4method{featureNames}{GenomicRatioSet}(object)
\S4method{annotation}{GenomicRatioSet}(object)
\S4method{preprocessMethod}{GenomicRatioSet}(object)
\S4method{mapToGenome}{GenomicRatioSet}(object, \dots)
}

\arguments{
  \item{object}{A \code{GenomicRatioSet}.}
  \item{gr}{A \code{GRanges} object.}
  \item{Beta}{A matrix of beta values (optional, see details).}
  \item{M}{A matrix of M values (optional, see details).}
  \item{CN}{A matrix of copy number values.}
  \item{annotation}{An annotation character string.}
  \item{preprocessMethod}{A preprocess method character string.}
  \item{\dots}{For the constructor, additional arguments to be passed to
    \code{SummarizedExperiment}; of particular interest are
    \code{colData} and \code{metadata}. For \code{mapToGenome}, this is
    ignored.} 
}

\value{
  An object of class \code{GenomicRatioSet} for the constructor.
}

\details{
  This class holds \code{M} or \code{Beta} values (or both) together
  with associated genomic coordinates.  It is not possible to get
  \code{Meth} or \code{Unmeth} values from this object.  The intention
  is to use this kind of object as an analysis end point.

  In case one of \code{M} or \code{Beta} is missing, the other is
  computed on the fly.  For example, M is computed from Beta as the
  logit (base 2) of the Beta values.
}

\section{Constructor}{
Instances are constructed using the \code{GenomicRatioSet} function with the
arguments outlined above.
}

\section{Accessors}{
  A number of useful accessors are inherited from
  \code{RangedSummarizedExperiment}.
  
  In the following code, \code{object} is a \code{GenomicRatioSet}.
  
  \describe{
    \item{\code{getBeta(object)}}{Get Beta, see details.}
    \item{\code{getM(object)}}{get M-values, see details.}
    \item{\code{getCN(object)}}{get copy number, see details.}
    \item{\code{getManifest(object)}}{get the manifest associated with
      the object.}
    \item{\code{sampleNames(object)}, \code{featureNames(object)}}{Get
      the sampleNames (colnames) or the featureNames (rownames).}
    \item{\code{preprocessMethod(object)},
      \code{annotation(object)}}{Get the preprocess method or annotation
      \code{character}.}
  }
}

\section{Utilities}{
  \describe{
    \item{\code{mapToGenome(object)}}{ Since \code{object} is already
      mapped to the genome, this method simply returns \code{object}
      unchanged.}
    \item{\code{combine}:}{Combines two different \code{GenomicRatioSet},
      eventually using the \code{cbind} method for \code{SummarizedExperiment}.}
  }
}

\author{Kasper Daniel Hansen \email{khansen@jhsph.edu}}

\seealso{
  \code{\link[SummarizedExperiment]{RangedSummarizedExperiment}} in the
  \pkg{SummarizedExperiment} package for the basic class structure.
}

\examples{
  showClass("GenomicRatioSet")
}

\name{densityBeanPlot}
\alias{densityBeanPlot}
\title{
  Density bean plots of methylation Beta values.
}
\description{
  Density \sQuote{bean} plots of methylation Beta values, primarily for QC.
}
\usage{
densityBeanPlot(dat, sampGroups = NULL, sampNames = NULL, main = NULL,
    pal = brewer.pal(8, "Dark2"), numPositions = 10000)
}
\arguments{
  \item{dat}{An \code{RGChannelSet}, a \code{MethylSet} or a
    \code{matrix}.  We either use the \code{getBeta} function to get
    Beta values (for the first two) or we assume the matrix contains
    Beta values.}
  \item{sampGroups}{Optional sample group labels. See details.}
  \item{sampNames}{Optional sample names. See details.}
  \item{main}{Plot title.}
  \item{pal}{Color palette.}
  \item{numPositions}{The density calculation uses \code{numPositions}
    randomly selected CpG positions.  If \code{NULL} use all positions.}
}
\details{
  This function produces the density bean plot component of the QC
  report. If \code{sampGroups} is specified, group-specific colors will
  be used. For speed reasons the plots are produced using a random
  subset of CpG positions. The number of positions used is specified by the
  \code{numPositions} option.
}
\value{
No return value. Plots are produced as a side-effect.

}
\references{
  P Kampstra.
  \emph{Beanplot: A boxplot alternative for visual comparison of
    distributions.}
  Journal of Statistical Software 28, (2008). \url{http://www.jstatsoft.org/v28/c01}
}
\author{
Martin Aryee \email{aryee@jhu.edu}.
}

\seealso{
\code{\link{qcReport}}, \code{\link{mdsPlot}}, \code{\link{controlStripPlot}}, \code{\link{densityPlot}}
}
\examples{
if (require(minfiData)) {

names <- pData(RGsetEx)$Sample_Name
groups <- pData(RGsetEx)$Sample_Group
par(mar=c(5,6,4,2))
densityBeanPlot(RGsetEx, sampNames=names, sampGroups=groups)

}
}

\name{estimateCellCounts}
\alias{estimateCellCounts}
\title{
Cell Proportion Estimation
}
\description{
Estimates the relative proportion of pure cell types within a sample. For example,
given peripheral blood samples, this function will return the relative proportions
of lymphocytes, monocytes, B-cells, and neutrophils.
}
\usage{
estimateCellCounts(rgSet, compositeCellType = "Blood",
                   processMethod = "auto", probeSelect = "auto",
                   cellTypes = c("CD8T","CD4T", "NK","Bcell","Mono","Gran"),
                   referencePlatform = c("IlluminaHumanMethylation450k",
                                         "IlluminaHumanMethylationEPIC",
                                         "IlluminaHumanMethylation27k"),
                   returnAll = FALSE, meanPlot = FALSE, verbose = TRUE, \dots)
}
\arguments{
  \item{rgSet}{The input \code{RGChannelSet} for the procedure.}
  \item{compositeCellType}{Which composite cell type is being deconvoluted. Should
    be one of "Blood", "CordBlood", or "DLPFC". See details. }
  \item{processMethod}{How should the user and reference data be processed together?
    Default input "auto" will use \code{preprocessQuantile} for
    Blood and DLPFC and \code{preprocessNoob} otherwise, in line
    with the existing literature.  Set it to the name of a
    preprocessing function as a character if you want to override
    it, like \code{"preprocessFunnorm"}.}
  \item{probeSelect}{How should probes be selected to distinguish cell types? Options include
    "both", which selects an equal number (50) of probes (with F-stat p-value < 1E-8) with the
    greatest magnitude of effect from the hyper- and hypo-methylated sides, and "any", which
    selects the 100 probes (with F-stat p-value < 1E-8) with the greatest magnitude of difference
    regardless of direction of effect. Default input "auto" will use "any" for cord blood and
    "both" otherwise, in line with previous versions of this function and/or our recommendations.
    Please see the references for more details.}
  \item{cellTypes}{Which cell types, from the reference object, should be
    we use for the deconvolution? See details. }
  \item{referencePlatform}{The platform for the reference dataset; if
    the input \code{rgSet} belongs to another platform, it will be
    converted using \code{\link{convertArray}}.}
  \item{returnAll}{Should the composition table and the normalized user
    supplied data be return?}
  \item{verbose}{Should the function be verbose?}
  \item{meanPlot}{
    Whether to plots the average DNA methylation across the cell-type discrimating probes within
    the mixed and sorted samples.
  }
  \item{\dots}{Passed to \code{preprocessQuantile}.}
}
\details{

This is an implementaion of the Houseman et al (2012) regression calibration approachalgorithm
to the Illumina 450k microarray for deconvoluting heterogeneous tissue sources like blood.
For example, this function will take an \code{RGChannelSet} from a DNA methylation (DNAm)
study of blood, and return the relative proportions of CD4+ and CD8+ T-cells, natural
killer cells, monocytes,  granulocytes, and b-cells in each sample.

The function currently supports cell composition estimation for blood, cord blood, and
the frontal cortex, through \code{compositeCellType} values of "Blood", "CordBlood", and
"DLPFC", respectively. Packages containing the appropriate reference data should be installed
before running the function for the first time ("FlowSorted.Blood.450k", "FlowSorted.DLPFC.450k",
"FlowSorted.CordBlood.450k"). Each tissue supports the estimation of different cell types, delimited
via the \code{cellTypes} argument. For blood, these are "Bcell", "CD4T", "CD8T", "Eos", "Gran",
"Mono", "Neu", and "NK" (though the default value for \code{cellTypes} is often sufficient).
For cord blood, these are "Bcell", "CD4T", "CD8T", "Gran", "Mono", "Neu", and "nRBC". For frontal
cortex, these are "NeuN_neg" and "NeuN_pos". See documentation of individual reference packages for
more details.

The \code{meanPlot} should be used to check for large batch effects in the data,
reducing the confidence placed in the composition estimates. This plot
depicts the average DNA methylation across the cell-type discrimating probes
in both the provided and sorted data. The means from the provided
heterogeneous samples should be within the range of the sorted samples.
If the sample means fall outside the range of the sorted means,
the cell type estimates will inflated to the closest cell type. Note that we
quantile normalize the sorted data with the provided data to reduce these
batch effects.

}
\value{
  Matrix of composition estimates across all samples and cell types.

  If \code{returnAll=TRUE} a list of a count matrix (see previous
  paragraph), a composition table and the normalized user data in form
  of a GenomicMethylSet.
}

\references{
  EA Houseman, WP Accomando, DC Koestler, BC Christensen, CJ Marsit, HH
  Nelson, JK Wiencke and KT Kelsey.
  \emph{DNA methylation arrays as surrogate measures of cell mixture distribution.}
  BMC bioinformatics (2012) 13:86.
  doi:\href{http://www.dx.doi.org10.1186/1471-2105-13-86}{10.1186/1471-2105-13-86}.

  AE Jaffe and RA Irizarry. \emph{Accounting for cellular
    heterogeneity is critical in epigenome-wide association studies.}
  Genome Biology (2014) 15:R31.
  doi:\href{http://www.dx.doi.org/10.1186/gb-2014-15-2-r31}{10.1186/gb-2014-15-2-r31}.

  KM Bakulski, JI Feinberg, SV Andrews, J Yang, S Brown, S McKenney,
  F Witter, J Walston, AP Feinberg, and MD Fallin.
  \emph{DNA methylation of cord blood cell types: Applications for mixed cell birth studies.}
  Epigenetics (2016) 11:5.
  doi:\href{http://dx.doi.org/10.1080/15592294.2016.1161875}{10.1080/15592294.2016.1161875}.
}

\author{
Andrew E. Jaffe, Shan V. Andrews, E. Andres Houseman
}

\seealso{\code{\link{preprocessQuantile}} and \code{\link{convertArray}}.}

\examples{
\dontrun{
if(require(FlowSorted.Blood.450k)) {
  wh.WBC <- which(FlowSorted.Blood.450k$CellType == "WBC")
  wh.PBMC <- which(FlowSorted.Blood.450k$CellType == "PBMC")
  RGset <- FlowSorted.Blood.450k[, c(wh.WBC, wh.PBMC)]
  ## The following line is purely to work around an issue with repeated
  ## sampleNames and Biobase::combine()
  sampleNames(RGset) <- paste(RGset$CellType,
    c(seq(along = wh.WBC), seq(along = wh.PBMC)), sep = "_")
  counts <- estimateCellCounts(RGset, meanPlot = FALSE)
  round(counts, 2)
}
}
}

\name{minfi-deprecated}
\alias{minfi-deprecated}
\title{Deprecated functions in package \sQuote{minfi}}

\description{
  These functions are provided for compatibility with older versions
  of \sQuote{minfi} only, and will be defunct at the next release.
}

\details{
  No functions are currently deprecated.
  
  The following functions are deprecated and will be made defunct; use
  the replacement indicated below:
  \itemize{
    \item{\code{read.450k}: \code{\link{read.metharray}}}
    \item{\code{read.450k.sheet}: \code{\link{read.metharray.sheet}}}
    \item{\code{read.450k.exp}: \code{\link{read.metharray.exp}}}
  }
}

\name{plotCpg}
\alias{plotCpg}
\title{
Plot methylation values at an single genomic position
}
\description{
Plot single-position (single CpG) methylation values as a function of a categorical or continuous phenotype
}
\usage{
plotCpg(dat, cpg, pheno, type = c("categorical", "continuous"),
    measure = c("beta", "M"), ylim = NULL, ylab = NULL, xlab = "",
    fitLine = TRUE, mainPrefix = NULL, mainSuffix = NULL)
}
\arguments{
    \item{dat}{An \code{RGChannelSet}, a \code{MethylSet} or a
    \code{matrix}.  We either use the \code{getBeta} (or \code{getM} for
    \code{measure="M"}) function to get Beta values (or M-values) (for
    the first two) or we assume the matrix contains Beta values (or
    M-values).}
  \item{cpg}{A character vector of the CpG position identifiers to be plotted.}
  \item{pheno}{A vector of phenotype values.}
  \item{type}{Is the phenotype categorical or continuous?}
  \item{measure}{Should Beta values or log-ratios (M) be  plotted?}
  \item{ylim}{y-axis limits.}
  \item{ylab}{y-axis label.}
  \item{xlab}{x-axis label.}
  \item{fitLine}{Fit a least-squares best fit line when using a continuous phenotype.}
  \item{mainPrefix}{Text to prepend to the CpG name in the plot main title.}
  \item{mainSuffix}{Text to append to the CpG name in the plot main title.}
}
\details{
  This function plots methylation values (Betas or log-ratios) at
  individual CpG loci as a function of a phenotype.

}
\value{
No return value. Plots are produced as a side-effect.
}

\author{
Martin Aryee \email{aryee@jhu.edu}.
}

\examples{
if (require(minfiData)) {

grp <- pData(MsetEx)$Sample_Group
cpgs <- c("cg00050873", "cg00212031", "cg26684946", "cg00128718")
par(mfrow=c(2,2))
plotCpg(MsetEx, cpg=cpgs, pheno=grp, type="categorical")

}
}

\name{read.metharray.exp}
\alias{read.metharray.exp}
\title{
  Reads an entire metharray experiment using a sample sheet
}
\description{
  Reads an entire methylation array experiment using a sample sheet or (optionally)
  a target like data.frame.
}
\usage{
read.metharray.exp(base = NULL, targets = NULL, extended = FALSE,
    recursive = FALSE, verbose = FALSE, force = FALSE)
}
\arguments{
  \item{base}{The base directory.}
  \item{targets}{A targets \code{data.frame}, see details}
  \item{extended}{Should the output of the function be a
    \code{"RGChannelSetExtended"} (default is \code{"RGChannelSet"}).}
  \item{recursive}{Should the search be recursive (see details)}
  \item{verbose}{Should the function be verbose?}
  \item{force}{Should reading different size IDAT files be forced? See
    the documentation for \code{\link{read.metharray}}}
}
\details{
  If the \code{targets} argument is \code{NULL}, the function finds all
  two-color IDAT files in the directory given by \code{base}.  If
  \code{recursive} is \code{TRUE}, the function searches \code{base} and
  all subdirectories.  A two-color IDAT files are pair of files with
  names ending in \code{_Red.idat} or \code{_Grn.idat}.

  If the \code{targets} argument is not \code{NULL} it is assumed it has
  a columned named \code{Basename}, and this is assumed to be pointing
  to the base name of a two color IDAT file, ie. a name that can be made
  into a real IDAT file by appending either \code{_Red.idat} or
  \code{_Grn.idat}.

  The type of methylation array is guess by looking at the number of
  probes in the IDAT files.
}
\value{
  An object of class \code{"RGChannelSet"} or
  \code{"RGChannelSetExtended"}.
}
\author{
  Kasper Daniel Hansen \email{khansen@jhsph.edu}.
}
\seealso{
  \code{\link{read.metharray}} for the workhorse function,
  \code{\link{read.metharray.sheet}} for reading a sample sheet and
  \code{\linkS4class{RGChannelSet}} for the output class.
}
\examples{
if(require(minfiData)) {

baseDir <- system.file("extdata", package = "minfiData")
RGset <- read.metharray.exp(file.path(baseDir, "5723646052"))

}
}

\name{preprocessSWAN}
\alias{preprocessSWAN}

\title{Subset-quantile Within Array Normalisation for Illumina Infinium
  HumanMethylation450 BeadChips}

\description{Subset-quantile Within Array Normalisation (SWAN) is a
  within array normalisation method for the Illumina Infinium
  HumanMethylation450 platform. It allows Infinium I and II type probes
  on a single array to be normalized together.}

\usage{
  preprocessSWAN(rgSet, mSet = NULL, verbose = FALSE)
}

\arguments{
  \item{rgSet}{An object of class \code{RGChannelSet}.}
  \item{mSet}{An optional object of class \code{MethylSet}.  If set to
    \code{NULL} preprocessSwan uses \code{preprocessRaw} on the
    \code{rgSet} argument.  In case \code{mSet} is supplied, make sure it
    is the result of preprocessing the \code{rgSet} argument.}
  \item{verbose}{Should the function be verbose?}
}

\details{The SWAN method has two parts. First, an average quantile
distribution is created using a subset of probes defined to be
biologically similar based on the number of CpGs underlying the probe
body. This is achieved by randomly selecting N Infinium I and II probes
that have 1, 2 and 3 underlying CpGs, where N is the minimum number of
probes in the 6 sets of Infinium I and II probes with 1, 2 or 3 probe
body CpGs. If no probes have previously been filtered out e.g. sex
chromosome probes, etc. N=11,303. This results in a pool of 3N Infinium
I and 3N Infinium II probes. The subset for each probe type is then
sorted by increasing intensity. The value of each of the 3N pairs of
observations is subsequently assigned to be the mean intensity of the
two probe types for that row or \dQuote{quantile}. This is the standard
quantile procedure. The intensities of the remaining probes are then
separately adjusted for each probe type using linear interpolation
between the subset probes.}

\note{
  SWAN uses a random subset of probes to do the between array
  normalization.  In order to achive reproducible results, the seed
  needs to be set using \code{set.seed}.
}

\value{an object of class \code{MethylSet}}

\references{
  J Maksimovic, L Gordon and A Oshlack (2012).  \emph{SWAN: Subset
  quantile Within-Array Normalization for Illumina Infinium
  HumanMethylation450 BeadChips}.  Genome Biology 13, R44.
}

\author{Jovana Maksimovic\email{jovana.maksimovic@mcri.edu.au}}

\seealso{\code{\linkS4class{RGChannelSet}} and
\code{\linkS4class{MethylSet}} as well as
\code{\linkS4class{IlluminaMethylationManifest}}.}

\examples{
if (require(minfiData)) {
  ## RGsetEx.sub is a small subset of RGsetEx;
  ## only used for computational speed.
  MsetEx.sub.swan <- preprocessSWAN(RGsetEx.sub)
}
\dontrun{
if (require(minfiData)) {
  dat <- preprocessRaw(RGsetEx)
  preprocessMethod(dat)
  datSwan <- preprocessSWAN(RGsetEx, mSet = dat)
  datIlmn <- preprocessIllumina(RGsetEx)
  preprocessMethod(datIlmn)
  datIlmnSwan <- preprocessSWAN(RGsetEx, mSet = datIlmn)
}
}
}

\name{preprocessNoob}
\alias{preprocessNoob}
\title{
  The Noob/ssNoob preprocessing method for Infinium methylation microarrays.
}
\description{
  Noob (normal-exponential out-of-band) is a background correction method
  with dye-bias normalization for Illumina Infinium methylation arrays.
}
\usage{
preprocessNoob(rgSet, offset = 15, dyeCorr = TRUE, verbose = FALSE,
               dyeMethod=c("single", "reference"))
}
\arguments{
  \item{rgSet}{An object of class \code{RGChannelSet}.}
  \item{offset}{An offset for the normexp background correction.}
  \item{dyeCorr}{Should dye correction be done?}
  \item{verbose}{Should the function be verbose?}
  \item{dyeMethod}{How should dye bias correction be done: use a single sample approach (ssNoob), or a reference array?}
}
\value{
  An object of class \code{MethylSet}.
}
\references{
  TJ Triche, DJ Weisenberger, D Van Den Berg, PW Laird and KD Siegmund
  \emph{Low-level processing of Illumina Infinium DNA Methylation
    BeadArrays}.
  Nucleic Acids Res (2013) 41, e90.
  doi:\href{http://www.dx.doi.org/10.1093/nar/gkt090}{10.1093/nar/gkt090}.
}
\author{
  Tim Triche, Jr.
}
\seealso{
  \code{\linkS4class{RGChannelSet}}
  as well as \code{\linkS4class{IlluminaMethylationManifest}} for the
  basic classes involved in these functions.
  \code{\link{preprocessRaw}} and \code{\link{preprocessQuantile}} are other preprocessing functions.
}
\examples{
if (require(minfiData)) {
  ## RGsetEx.sub is a small subset of RGsetEx;
  ## only used for computational speed.
  MsetEx.sub.noob <- preprocessNoob(RGsetEx.sub)
}
\dontrun{
if (require(minfiData)) {
  dyeMethods <- c(ssNoob="single", refNoob="reference")
  GRsets <- lapply(dyeMethods,
                   function(m) preprocessNoob(RGsetEx, dyeMethod=m))
  all.equal(getBeta(GRsets$refNoob), getBeta(GRsets$ssNoob)) # TRUE
}
}
}

\name{convertArray}
\alias{convertArray}
\alias{convertArray,RGChannelSet-method}
\alias{convertArray,MethylSet-method}
\alias{convertArray,GenomicMethylSet-method}
\alias{convertArray,RatioSet-method}
\alias{convertArray,GenomicRatioSet-method}
\title{
A method for converting a type of methylation arrays into a
virtual array of another type.
}
\description{
  A method for converting a type of methylation array into a array of
another type. The three generations of Illumina methylation arrays are
supported: the 27k, the 450k and the EPIC arrays.  Specifically, the
450k array and the EPIC array share many probes in common. For
\code{RGChannelSet}, this function will convert an EPIC array into a
450k array (or vice-versa) by dropping probes that differ between the
two arrays. Because most of the probes on the 27k array have a different
chemistry than the 450k and EPIC probes, converting an 27k
\code{RGChannelSet} into another array is not supported.  Each array can
be converted into another array at the CpG site level, that is any
\code{MethylSet} and \code{RatioSet} (or \code{GenomicMethylSet} and
\code{GenomicRatioSet}) can be converted to a 27k, 450k or EPIC array.
The output array is specified by the \code{outType} argument.
}
\usage{
\S4method{convertArray}{RGChannelSet}(object,
                  outType = c("IlluminaHumanMethylation450k",
                              "IlluminaHumanMethylationEPIC"),
                  verbose = TRUE)
\S4method{convertArray}{MethylSet}(object,
                  outType = c("IlluminaHumanMethylation450k",
                              "IlluminaHumanMethylationEPIC",
                              "IlluminaHumanMethylation27k"),
                  verbose = TRUE)
\S4method{convertArray}{RatioSet}(object,
                  outType = c("IlluminaHumanMethylation450k",
                              "IlluminaHumanMethylationEPIC",
                              "IlluminaHumanMethylation27k"),
                  verbose = TRUE)
\S4method{convertArray}{GenomicMethylSet}(object,
                  outType = c("IlluminaHumanMethylation450k",
                              "IlluminaHumanMethylationEPIC",
                              "IlluminaHumanMethylation27k"),
                  verbose = TRUE)
\S4method{convertArray}{GenomicRatioSet}(object,
                  outType = c("IlluminaHumanMethylation450k",
                              "IlluminaHumanMethylationEPIC",
                              "IlluminaHumanMethylation27k"),
                  verbose = TRUE)
}
\arguments{
  \item{object}{The input object.}
  \item{outType}{The array type of the output.}
  \item{verbose}{Should the function be verbose?}
}
\details{
  FIXME: describe the RGChannelSet conversion.
}
\value{
  The output object has the same class as the input object, that is
  either an \code{RGChannelSet}, a \code{MethylSet}, a \code{RatioSet}, a
  \code{GemomicMethylSet} or a \code{GenomicRatioSet}, with the type of
  the array given by the \code{outType} argument.
  }
\author{
  Jean-Philippe Fortin and Kasper D. Hansen.
}
\examples{
if(require(minfiData)) {
  data(RGsetEx.sub)
  rgSet <- convertArray(RGsetEx.sub, outType = "IlluminaHumanMethylationEPIC")
  rgSet
}
}

\name{detectionP}
\alias{detectionP}
\title{
  Detection p-values for all probed genomic positions.
}
\description{
  This function identifies failed positions defined as both the methylated and unmethylated channel reporting background signal levels.
}
\usage{
detectionP(rgSet, type = "m+u")
}
\arguments{
  \item{rgSet}{An \code{RGChannelSet}.}
  \item{type}{How to calculate p-values. Only \code{m+u} is currently implemented (See details).}
}
\details{

  A detection p-value is returned for every genomic position in every sample. Small
  p-values indicate a good position. Positions with non-significant p-values
  (typically >0.01) should not be trusted.

  The \code{m+u} method compares the total DNA signal (Methylated +
  Unmethylated) for each position to the background signal level. The
  background is estimated using negative control positions, assuming a
  normal distribution. Calculations are performed on the original
  (non-log) scale.

  This function is different from the detection routine in Genome Studio.
}
\value{
A matrix with detection p-values.
}
\author{
Martin Aryee \email{aryee@jhu.edu}.
}
\examples{
if (require(minfiData)) {
  detP <- detectionP(RGsetEx.sub)
  failed <- detP>0.01
  colMeans(failed) # Fraction of failed positions per sample
  sum(rowMeans(failed)>0.5) # How many positions failed in >50% of samples?
}
}

\name{GenomicMethylSet-class}
\Rdversion{1.1}
\docType{class}
\alias{GenomicMethylSet-class}
\alias{GenomicMethylSet}
%%methods
\alias{show,GenomicMethylSet-method}
\alias{annotation,GenomicMethylSet-method}
\alias{annotation<-,GenomicMethylSet,ANY-method}
\alias{preprocessMethod,GenomicMethylSet-method}
\alias{getMeth,GenomicMethylSet-method}
\alias{getUnmeth,GenomicMethylSet-method}
\alias{getBeta,GenomicMethylSet-method}
\alias{getM,GenomicMethylSet-method}
\alias{getCN,GenomicMethylSet-method}
\alias{mapToGenome,GenomicMethylSet-method}
\alias{combine,GenomicMethylSet,GenomicMethylSet-method}
%% eSet-methods
\alias{pData,GenomicMethylSet-method}
\alias{pData<-,GenomicMethylSet,DataFrame-method}
\alias{sampleNames,GenomicMethylSet-method}
\alias{sampleNames<-,GenomicMethylSet,ANY-method}
\alias{featureNames,GenomicMethylSet-method}
\alias{featureNames<-,GenomicMethylSet-method}

\title{GenomicMethylSet instances}

\description{
  This class holds preprocessed data for Illumina methylation
  microarrays, mapped to a genomic location.
}

\usage{
## Constructor

GenomicMethylSet(gr = GRanges(), Meth = new("matrix"),
                 Unmeth = new("matrix"), annotation = "",
                 preprocessMethod = "", ...)

## Data extraction / Accessors

\S4method{getMeth}{GenomicMethylSet}(object)
\S4method{getUnmeth}{GenomicMethylSet}(object)
\S4method{getBeta}{GenomicMethylSet}(object, type = "", offset = 0, betaThreshold = 0)
\S4method{getM}{GenomicMethylSet}(object, type = "", \dots)
\S4method{getCN}{GenomicMethylSet}(object, \dots)
\S4method{pData}{GenomicMethylSet}(object)
\S4method{sampleNames}{GenomicMethylSet}(object)
\S4method{featureNames}{GenomicMethylSet}(object)
\S4method{annotation}{GenomicMethylSet}(object)
\S4method{preprocessMethod}{GenomicMethylSet}(object)
\S4method{mapToGenome}{GenomicMethylSet}(object, \dots)
}

\arguments{
  \item{object}{A \code{GenomicMethylSet}.}
  \item{gr}{A \code{GRanges} object.}
  \item{Meth}{A matrix of methylation values (between zero and infinity)
    with each row being a methylation loci and each column a sample.}
  \item{Unmeth}{See the \code{Meth} argument.}
  \item{annotation}{An annotation character string.}
  \item{preprocessMethod}{A preprocess method character string.}
  \item{type}{How are the values calculated?  For \code{getBeta} setting
  \code{type="Illumina"} sets \code{offset=100} as per Genome Studio.
  For \code{getM} setting \code{type=""} computes M-values as the
  logarithm of \code{Meth}/\code{Unmeth}, otherwise it is computed as
  the logit of \code{getBeta(object)}.}
  \item{offset}{Offset in the beta ratio, see detail.}
  \item{betaThreshold}{Constrains the beta values to be in the inverval
    betwen \code{betaThreshold} and 1-\code{betaThreshold}.}  
  \item{\dots}{For the constructor, additional arguments to be passed to
    \code{SummarizedExperiment}; of particular interest are
    \code{colData} and \code{metadata}. For \code{getM} these values
    gets passed onto \code{getBeta}. For \code{mapToGenome}, this is
    ignored.}
}

\value{
  An object of class \code{GenomicMethylSet} for the constructor.
}

\details{
  For a detailed discussion of \code{getBeta} and \code{getM} see the
  deatils section of \code{\link{MethylSet}}.
}

\section{Constructor}{
Instances are constructed using the \code{GenomicMethylSet} function with the
arguments outlined above.
}

\section{Accessors}{
  A number of useful accessors are inherited from
  \code{RangedSummarizedExperiment}.
  
  In the following code, \code{object} is a \code{GenomicMethylSet}.
  
  \describe{
    \item{\code{getMeth(object)}, \code{getUnmeth(object)}}{Get the
      Meth or Unmeth matrix.}
    \item{\code{getBeta(object)}}{Get Beta, see details.}
    \item{\code{getM(object)}}{get M-values, see details.}
    \item{\code{getCN(object)}}{get copy number values which are defined
      as the sum of the methylation and unmethylation channel.}
    \item{\code{getManifest(object)}}{get the manifest associated with
      the object.}
    \item{\code{sampleNames(object)}, \code{featureNames(object)}}{Get
      the sampleNames (colnames) or the featureNames (rownames).}
    \item{\code{preprocessMethod(object)},
      \code{annotation(object)}}{Get the preprocess method or annotation
      \code{character}.}
  }
}

\section{Utilities}{
  \describe{
    \item{\code{mapToGenome(object)}}{ Since \code{object} is already
      mapped to the genome, this method simply returns \code{object}
      unchanged.}
    \item{\code{combine}:}{Combines two different \code{GenomicMethylSet},
      eventually using the \code{cbind} method for \code{SummarizedExperiment}.}
  }
}


\author{Kasper Daniel Hansen \email{khansen@jhsph.edu}}

\seealso{
  \code{\link[SummarizedExperiment]{RangedSummarizedExperiment}} in the
  \pkg{SummarizedExperiment} package for the basic class structure.
  Objects of this class are typically created by using the function
  \code{\link{mapToGenome}} on a \code{\link{MethylSet}}.}

\examples{
  showClass("GenomicMethylSet")
}

\name{minfiQC}
\alias{minfiQC}
\title{
  easy one-step QC of methylation object
}
\description{
  This function combines a number of functions into a simple to use, one
  step QC step/
}
\usage{
minfiQC(object, fixOutliers = TRUE, verbose = FALSE)
}
\arguments{
  \item{object}{An object of class \code{[Genomic]MethylSet}.}
  \item{fixOutliers}{Should the function fix outlying observations
    (using \code{fixMethOutliers}) before running QC?}
  \item{verbose}{Should the function be verbose?}
}
\details{
  A number of functions are run sequentially on the \code{object}.

  First outlier values are thresholded using \code{fixMethOutliers}.
  Then qc is performed using \code{getQC} and then sample specific sex
  is estimated using \code{getSex}.
}
\value{
  A list with two values,
  \item{object}{The \code{object} processed by \code{fixMethOutliers}
    and with a column \code{predictedSex} added to the pheno data.}
  \item{qc}{A \code{DataFrame} with columns from the output of
  \code{getQC} and \code{getSex}}
}
\author{
  Kasper D. Hansen
}
\seealso{
  \code{\link{getSex}}, \code{\link{getQC}}, \code{\link{fixMethOutliers}}
}
\examples{
if(require(minfiData)) {
  out <- minfiQC(MsetEx)
  ## plotQC(out$qc)
  ## plotSex(out$sex)
}
}

\name{densityPlot}
\alias{densityPlot}
\title{
Density plots of methylation Beta values.
}
\description{
Density plots of methylation Beta values, primarily for QC.
}
\usage{
densityPlot(dat, sampGroups = NULL, main = "", xlab = "Beta",
    pal = brewer.pal(8, "Dark2"), xlim, ylim, add = TRUE, legend = TRUE,
    \dots)
}
\arguments{
  \item{dat}{An \code{RGChannelSet}, a \code{MethylSet} or a
    \code{matrix}.  We either use the \code{getBeta} function to get
    Beta values (for the first two) or we assume the matrix contains
    Beta values.}
  \item{sampGroups}{Optional sample group labels. See details.}
  \item{main}{Plot title.}
  \item{xlab}{x-axis label.}
  \item{pal}{Color palette.}
  \item{xlim}{x-axis limits.}
  \item{ylim}{y-axis limits.}
  \item{add}{Start a new plot?}
  \item{legend}{Plot legend.}
  \item{\dots}{Additional options to be passed to the plot command.}
}
\details{
This function produces the density plot component of the QC report. If \code{sampGroups} is specified, group-specific colors will be used.
}
\value{
No return value. Plots are produced as a side-effect.
}
\author{
Martin Aryee \email{aryee@jhu.edu}.
}
\seealso{
\code{\link{qcReport}}, \code{\link{mdsPlot}}, \code{\link{controlStripPlot}}, \code{\link{densityBeanPlot}}
}
\examples{
if (require(minfiData)) {

groups <- pData(RGsetEx)$Sample_Group
densityPlot(RGsetEx, sampGroups=groups)

}
}

\name{read.metharray}
\alias{read.metharray}
\title{
  Parsing IDAT files from Illumina methylation arrays.
}
\description{
  Parsing IDAT files from Illumina methylation arrays.
}
\usage{
read.metharray(basenames, extended = FALSE, verbose = FALSE, force = FALSE)
}
\arguments{
  \item{basenames}{The basenames or filenames of the IDAT files.  By
    basenames we mean the filename without the ending \code{_Grn.idat}
    or \code{_Red.idat} (such that each sample occur once).  By
    filenames we mean filenames including \code{_Grn.idat}
    or \code{_Red.idat} (but only one of the colors)}
  \item{extended}{Should a \code{RGChannelSet} or a
    \code{RGChannelSetExtended} be returned. }
  \item{verbose}{Should the function be verbose?}
  \item{force}{Should reading different size IDAT files be forced? See
    Details.}
}
\details{
  The type of methylation array is guess by looking at the number of
  probes in the IDAT files.

  We have seen IDAT files from the same array, but with different number
  of probes in the wild.  Specifically this is the case for early access
  EPIC arrays which have fewer probes than final release EPIC arrays.
  It is possible to combine IDAT files from the same inferred array, but
  with different number of probes, into the same \code{RGChannelSet} by
  setting \code{force=TRUE}.  The output object will have the same
  number of probes as the smallest array being parsed; effectively
  removing probes which could have been analyzed.
}
\value{
  An object of class \code{RGChannelSet} or
  \code{RGChannelSetExtended}. 
}
\author{
  Kasper Daniel Hansen\email{khansen@jhsph.edu}.
}
\seealso{
  \code{\link{read.metharray.exp}} for a convenience function for
  reading an experiment, \code{\link{read.metharray.sheet}} for
  reading a sample sheet and \code{\linkS4class{RGChannelSet}} for the
  output class. 
}
\examples{
if(require(minfiData)) {

baseDir <- system.file("extdata", package = "minfiData")
RGset1 <- read.metharray(file.path(baseDir, "5723646052", "5723646052_R02C02"))

}
}

\name{DelayedArray_utils.Rd}

\alias{type,HDF5RealizationSink-method}
\alias{type,arrayRealizationSink-method}
\alias{type,RleRealizationSink-method}

\alias{dimnames,arrayRealizationSink-method}

\title{
Stubs for internal functions
}

\description{Stubs for internal functions}

%\usage{
%dimnames(x)
%type(x)
%}
%\arguments{
%  \item{x}{A RealizationSink (e.g.,
%  arrayRealizationSink, HDF5RealizationSink, RleRealizationSink) instance}
%}

\keyword{internal}

\name{bumphunter-methods}
\docType{methods}
\alias{bumphunter,GenomicRatioSet-method}

\title{Methods for function \code{bumphunter} in Package \pkg{minfi}}

\description{
  Estimate regions for which a genomic profile deviates from its baseline
  value. Originally implemented to detect differentially methylated
  genomic regions between two populations, but can be applied to any CpG-level
  coefficient of interest.

}
\usage{
\S4method{bumphunter}{GenomicRatioSet}(object, design, cluster=NULL,
          coef=2,  cutoff=NULL, pickCutoff=FALSE, pickCutoffQ=0.99,
          maxGap=500,  nullMethod=c("permutation","bootstrap"),
          smooth=FALSE, smoothFunction=locfitByCluster,
          useWeights=FALSE,   B=ncol(permutations), permutations=NULL,
          verbose=TRUE, type = c("Beta","M"), \dots)
}

\arguments{
  \item{object}{An object of class GenomicRatioSet.}
  \item{design}{Design matrix with rows representing samples and columns
    representing covariates. Regression is applied to each row of mat.}
  \item{cluster}{The clusters of locations that are to be analyzed
    together.  In the case of microarrays, the clusters are many times
    supplied by the manufacturer. If not available the function
    \code{\link{clusterMaker}} can be used to cluster nearby locations.}
  \item{coef}{An integer denoting the column of the design matrix
    containing the covariate of interest. The hunt for bumps will be
    only be done for the estimate of this coefficient.}
  \item{cutoff}{A numeric value. Values of the estimate of the genomic
    profile above the cutoff or below the negative of the cutoff will be
    used as candidate regions.  It is possible to give two separate
    values (upper and lower bounds).  If one value is given, the lower
    bound is minus the value.}
  \item{pickCutoff}{Should bumphunter attempt to pick a cutoff using the
    permutation distribution?}
  \item{pickCutoffQ}{The quantile used for picking the cutoff using the
    permutation distribution.}
  \item{maxGap}{ If cluster is not provided this maximum location gap
    will be used to define cluster via the  \code{\link{clusterMaker}}
    function.}
  \item{nullMethod}{
    Method used to generate null candidate regions, must be one of \sQuote{boots
      trap} or
    \sQuote{permutation} (defaults to \sQuote{permutation}).
    However, if covariates in addition to the
    outcome of interest are included in the design matrix
    (ncol(design)>2), the \sQuote{permutation} approach is not
    recommended. See vignette and original paper for more information.}
  \item{smooth}{
    A logical value. If TRUE the estimated profile will be smoothed with the
    smoother defined by \code{smoothFunction}}
  \item{smoothFunction}{
    A function to be used for smoothing the estimate of the genomic
    profile. Two functions are provided by the package: \code{loessByCluster}
    and \code{runmedByCluster}.}
  \item{useWeights}{
    A logical value. If \code{TRUE} then the standard errors of the
   point-wise estimates of the profile function will be used as weights
   in the loess smoother \code{loessByCluster}. If the
   \code{runmedByCluster} smoother is used this argument is ignored.}
\item{B}{An integer denoting the number of resamples to use when computing
   null distributions. This defaults to 0. If \code{permutations} is
  supplied that defines the number of permutations/bootstraps and \code{B} is
  ignored.}
\item{permutations}{ is a matrix with columns providing indexes to be used to
  scramble the data and create a null distribution when
  \code{nullMethod} is set to  permutations. If the bootstrap approach is used this argument is
  ignored. If this matrix is not supplied and \code{B}>0 then
  these indexes are created using the function \code{sample}. }
\item{verbose}{
    logical value. If \code{TRUE}, it writes out some messages
    indicating progress. If \code{FALSE} nothing should be printed.
  }
  \item{type}{
    Should bumphunting be performed on M-values ("M") or Beta values
    ("Beta")?}
  \item{\dots}{
    further arguments to be passed to the smoother functions.
}
}
\details{
  See help file for \code{\link[bumphunter]{bumphunter}} method in the
  \code{bumphunter} package for for details.
}

\value{
An object of class \code{bumps} with the following components:
\item{tab}{The table with candidate regions and annotation for these.}
\item{coef}{The single loci coefficients.}
\item{fitted}{The estimated genomic profile used to determine the
  regions.}
\item{pvaluesMarginal}{marginal p-value for each genomic location.}
\item{null}{The null distribution.}
\item{algorithm}{details on the algorithm.}
}

\references{
AE Jaffe, P Murakami, H Lee, JT Leek, MD Fallin, AP Feinberg, and RA
Irizarry.
\emph{Bump hunting to identify differentially methylated regions in
  epigenetic epidemiology studies.}
International Journal of Epidemiology (2012) 41(1):200-209.
doi:\href{http://www.dx.doi.org/10.1093/ije/dyr238}{10.1093/ije/dyr238}
}

\seealso{\code{\link[bumphunter]{bumphunter}}}

\author{
  Rafael A. Irizarry, Martin J. Aryee and Kasper D. Hansen
}


\examples{
if(require(minfiData)) {
  gmSet <- preprocessQuantile(MsetEx)
  design <- model.matrix(~ gmSet$status)
  bumps <- bumphunter(gmSet, design = design, B = 0,
                      type = "Beta", cutoff = 0.25)
}
}
